/*
checkaport.c: clean up and check security in AUTHDIR
*/

/* WARNING! checkaport should not be setuid auth, but it's probably */
/* running as either auth or root! All privileged programs are */
/* dangerous! Check them carefully! */

static char checkaportauthor[] =
"checkaport was written by Daniel J. Bernstein.\n\
Internet address: brnstnd@acf10.nyu.edu.\n";

static char checkaportversion[] = 
"checkaport version 1.0, April 23, 1990.\n\
Copyright (c) 1990, Daniel J. Bernstein.\n\
All rights reserved.\n";

static char checkaportcopyright[] =
"checkaport version 1.0, April 23, 1990.\n\
Copyright (c) 1990, Daniel J. Bernstein.\n\
All rights reserved.\n\
\n\
Until January 1, 1995, you are granted the following rights: A. To make\n\
copies of this work in original form, so long as (1) the copies are exact\n\
and complete; (2) the copies include the copyright notice, this paragraph,\n\
and the disclaimer of warranty in their entirety. B. To distribute this\n\
work, or copies made under the provisions above, so long as (1) this is\n\
the original work and not a derivative form; (2) you do not charge a fee\n\
for copying or for distribution; (3) you ensure that the distributed form\n\
includes the copyright notice, this paragraph, and the disclaimer of\n\
warranty in their entirety. These rights are temporary and revocable upon\n\
written, oral, or other notice by Daniel J. Bernstein. These rights are\n\
automatically revoked on January 1, 1995. This copyright notice shall be\n\
governed by the laws of the state of New York.\n\
\n\
If you have questions about checkaport or about this copyright notice,\n\
or if you would like additional rights beyond those granted above,\n\
please feel free to contact the author at brnstnd@acf10.nyu.edu\n\
on the Internet.\n";

static char checkaportwarranty[] =
"To the extent permitted by applicable law, Daniel J. Bernstein disclaims\n\
all warranties, explicit or implied, including but not limited to the\n\
implied warranties of merchantability and fitness for a particular purpose.\n\
Daniel J. Bernstein is not and shall not be liable for any damages,\n\
incidental or consequential, arising from the use of this program, even\n\
if you inform him of the possibility of such damages. This disclaimer\n\
shall be governed by the laws of the state of New York.\n\
\n\
In other words, use this program at your own risk.\n\
\n\
If you have questions about checkaport or about this disclaimer of warranty,\n\
please feel free to contact the author at brnstnd@acf10.nyu.edu\n\
on the Internet.\n";

static char checkaportusage[] =
"Usage: checkaport [ -ACHUVW ] \n\
Help:  checkaport -H\n";

static char checkaporthelp[] =
"checkaport goes through the auth directory, cleaning up and reporting\n\
any inconsistencies or security violations.\n\
\n\
checkaport -A: print authorship notice\n\
checkaport -C: print copyright notice\n\
checkaport -H: print this notice\n\
checkaport -U: print short usage summary\n\
checkaport -V: print version number\n\
checkaport -W: print disclaimer of warranty\n\
\n\
checkaport: start checking\n\
\n\
If you have questions about or suggestions for checkaport, please feel free\n\
to contact the author, Daniel J. Bernstein, at brnstnd@acf10.nyu.edu\n\
on the Internet.\n";

#include <sys/types.h>
#include <sys/dir.h>
#include <sys/file.h>
#ifdef BSD
#include <limits.h>
#endif
#include <stdio.h>
extern int getopt();
extern char *optarg; /* these should be in getopt.h! */
extern int optind;
#include "djberr.h"

#ifndef AUTHDIR
#define AUTHDIR "/usr/etc/auth"
#endif

main(argc,argv,envp)
int argc;
char *argv[];
char *envp[];
{
 int opt;
 int uid = getuid();
 int euid = geteuid();
 DIR *dirp;
 struct direct *dp;
 int fd;
 char buf[32]; /* same length as in attachport */
 int r;
 char username[10];
 char usertemp[10];
 int pid;
 int pidchild;
 unsigned short port;
 unsigned long in;
 unsigned short rport;
 char pfn[sizeof(AUTHDIR) + 30];

 /* WARNING! We are probably running as auth or root! */

 while ((opt = getopt(argc,argv,"ACHUVW")) != EOF)
   switch(opt)
    {
     case 'A': (void) err(checkaportauthor); exit(1);
     case 'C': (void) err(checkaportcopyright); exit(1);
     case 'H': (void) err(checkaporthelp); exit(1);
     case 'U': (void) err(checkaportusage); exit(1);
     case 'V': (void) err(checkaportversion); exit(1);
     case 'W': (void) err(checkaportwarranty); exit(1);
     case '?': (void) err(checkaportusage); exit(1);
    }
 argv += optind, argc -= optind;

 if ((chdir(AUTHDIR) == -1) || chdir("tcp") == -1)
  {
   perrn2("checkaport: fatal: cannot change to directory %s/tcp",AUTHDIR);
   exit(2);
  }

 if (!(dirp = opendir(".")))
  {
   perrn2("checkaport: fatal: cannot read directory %s/tcp",AUTHDIR);
   exit(3);
  }

 while (dp = readdir(dirp))
  {
   if ((dp->d_name[0] == '.')
     &&((dp->d_name[1] == '\0')
      ||((dp->d_name[1] == '.')
       &&(dp->d_name[2] == '\0'))))
     continue;
   switch(dp->d_name[0])
    {
     case 'l':
       if (sscanf(dp->d_name,"lock.%hd",&port) < 1)
	{
	 errn2("checkaport: warning: UFO %s",dp->d_name);
	}
       else if (((fd = open(dp->d_name,O_RDONLY)) == -1)
	       &&(flock(fd,LOCK_EX) == -1))
	{
	 perrn2("checkaport: warning: cannot open %s",dp->d_name);
	}
       else if ((r = read(fd,buf,31)) <= 0)
	{
	 errn2("checkaport: warning: cannot read %s, deleting",dp->d_name);
	 (void) close(fd);
	 (void) unlink(dp->d_name);
	}
       else
	{
	 buf[r] = '\0';
	 (void) close(fd);
	 if (buf[0] == '!')
	  {
	   errn2("checkaport: warning: corrupted lock file %s",dp->d_name);
	   errn2("checkaport: contents of lock file: %s",buf);
	   errn("checkaport: this may be an attempted security breach!");
	  }
	 else if (sscanf(buf,"%d-%8s",&pid,username) < 2)
	   if (sscanf(buf,"%d:%D.%hd %8s",&pid,&in,&rport,username) < 4)
	    {
	     errn2("checkaport: warning: UFO contents %s",dp->d_name);
	    }
	   else /* authtcp lock file */
	    {
	     sprintf(buf,"%D.%u.%u",in,port,rport);
	     if (((fd = open(buf,O_RDONLY)) == -1)
	       &&(flock(fd,LOCK_EX) == -1))
	      {
	       perrn2("checkaport: warning: cannot open %s",buf);
	       errn2("checkaport: came from lock %s",dp->d_name);
	      }
	     else
	      {
	       if ((r = read(fd,usertemp,8)) <= 0)
		 r = 0;
	       usertemp[r] = '\0';
	       if (r == 0 || strcmp(usertemp,username))
		{
		 errn("checkaport: warning: security breach!");
		 errn2("checkaport: lock file %s",dp->d_name);
		 errn2("checkaport: ILR %s",buf);
		 errn2("checkaport: lock file user %s",username);
		 errn2("checkaport: ILR user %s",usertemp);
		 errn("checkaport: deleting both files");
		 (void) unlink(dp->d_name);
		 (void) unlink(buf);
		}
	       (void) close(fd);
	      }
	    }
	 if (kill(pid,0) == -1)
	  {
	   errn2("checkaport: warning: no pid for lock file %s, deleting",
		 dp->d_name);
	   (void) unlink(dp->d_name);
	  }
	}
       break;
     case 'p':
       if (sscanf(dp->d_name,"ps.%d.%d",&pid,&pidchild) < 2)
	{
	 errn2("checkaport: warning: UFO %s",dp->d_name);
	}
       else if (((fd = open(dp->d_name,O_RDONLY)) == -1)
	       &&(flock(fd,LOCK_EX) == -1))
	{
	 perrn2("checkaport: warning: cannot open %s",dp->d_name);
	}
       else if ((r = read(fd,pfn,sizeof(pfn) - 1)) <= 0)
	{
	 errn2("checkaport: warning: cannot read %s",dp->d_name);
	 (void) close(fd);
	}
       else
	{
	 pfn[r] = '\0';
	 (void) close(fd);
	 if (strncmp(pfn,AUTHDIR,strlen(AUTHDIR)))
	  {
	   errn2("checkaport: warning: ps %s out of directory, deleting",
		 dp->d_name);
	   (void) unlink(dp->d_name);
	  }
	 else if (((fd = open(pfn,O_RDONLY)) == -1)
	        &&(flock(fd,LOCK_EX) == -1))
	  {
	   perrn2("checkaport: warning: cannot open ilr %s",pfn);
	   errn2("checkaport: came from ps %s, deleting",dp->d_name);
	   (void) unlink(dp->d_name);
	  }
	 else
	  {
	   /* XXX: Can we do more tests here? */
	   (void) close(fd);
	   if (kill(pid,0) == -1)
	    {
	     errn2("checkaport: warning: no pid for ps file %s, deleting",buf);
	     (void) unlink(buf);
	    }
	   else if (kill(pidchild,0) == -1)
	    {
	     errn2("checkaport: warning: no pid for ps file %s, deleting",buf);
	     (void) unlink(buf);
	    }
	  }
	}
       break;
     default:
       if (sscanf(dp->d_name,"%D.%hd.%hd",&in,&port,&rport) < 3)
	{
	 errn2("checkaport: warning: UFO %s",dp->d_name);
	}
       else if (((fd = open(dp->d_name,O_RDONLY)) == -1)
	       &&(flock(fd,LOCK_EX) == -1))
	{
	 perrn2("checkaport: warning: cannot open %s",dp->d_name);
	}
       else if ((r = read(fd,usertemp,8)) <= 0)
	{
	 errn2("checkaport: warning: cannot read %s, deleting",dp->d_name);
	 (void) close(fd);
	 (void) unlink(dp->d_name);
	}
       else
	{
	 usertemp[r] = '\0';
	 (void) close(fd);
	 sprintf(buf,"lock.%u",port);
	 if (((fd = open(buf,O_RDONLY)) == -1)
	       &&(flock(fd,LOCK_EX) == -1))
	  {
	   perrn2("checkaport: warning: cannot open lock %s",buf);
	   errn2("checkaport: came from ilr %s, deleting",dp->d_name);
	   (void) unlink(dp->d_name);
	  }
	 if ((r = read(fd,buf,31)) <= 0)
	  {
	   errn2("checkaport: weird: cannot read lock.%u",port);
	  }
	 else
	  {
	   buf[r] = '\0';
	   (void) close(fd);
	   if ((sscanf(buf,"%d-%8s",&pid,username) < 2)
	     &&(sscanf(buf,"%d:%D.%hd %8s",&pid,&in,&rport,username) < 4))
	    {
	     errn2("checkaport: warning: corrupted lock file lock.%u",port);
	     errn2("checkaport: came from ilr %s, deleting",dp->d_name);
	     errn("checkaport: this may be an attempted security breach!");
	     (void) unlink(dp->d_name);
	    }
	   if (r == 0 || strcmp(usertemp,username))
	    {
	     errn2("checkaport: warning: corrupted lock file lock.%u",port);
	     errn2("checkaport: came from ilr %s, deleting",dp->d_name);
	     errn2("checkaport: ilr user %s",usertemp);
	     errn2("checkaport: lock file user %s",username);
	     errn("checkaport: this may be an attempted security breach!");
	     (void) unlink(dp->d_name);
	    }
	  }
	}
       break;
    }
  }

 (void) closedir(dirp);

 exit(0);
}
