/*
killaport.c: kill an attachport
*/

/* WARNING! killaport is probably setuid auth! */
/* All setuid programs are dangerous! Check them carefully! */

static char killaportauthor[] =
"killaport was written by Daniel J. Bernstein.\n\
Internet address: brnstnd@acf10.nyu.edu.\n";

static char killaportversion[] = 
"killaport version 1.0, April 23, 1990.\n\
Copyright (c) 1990, Daniel J. Bernstein.\n\
All rights reserved.\n";

static char killaportcopyright[] =
"killaport version 1.0, April 23, 1990.\n\
Copyright (c) 1990, Daniel J. Bernstein.\n\
All rights reserved.\n\
\n\
Until January 1, 1995, you are granted the following rights: A. To make\n\
copies of this work in original form, so long as (1) the copies are exact\n\
and complete; (2) the copies include the copyright notice, this paragraph,\n\
and the disclaimer of warranty in their entirety. B. To distribute this\n\
work, or copies made under the provisions above, so long as (1) this is\n\
the original work and not a derivative form; (2) you do not charge a fee\n\
for copying or for distribution; (3) you ensure that the distributed form\n\
includes the copyright notice, this paragraph, and the disclaimer of\n\
warranty in their entirety. These rights are temporary and revocable upon\n\
written, oral, or other notice by Daniel J. Bernstein. These rights are\n\
automatically revoked on January 1, 1995. This copyright notice shall be\n\
governed by the laws of the state of New York.\n\
\n\
If you have questions about killaport or about this copyright notice,\n\
or if you would like additional rights beyond those granted above,\n\
please feel free to contact the author at brnstnd@acf10.nyu.edu\n\
on the Internet.\n";

static char killaportwarranty[] =
"To the extent permitted by applicable law, Daniel J. Bernstein disclaims\n\
all warranties, explicit or implied, including but not limited to the\n\
implied warranties of merchantability and fitness for a particular purpose.\n\
Daniel J. Bernstein is not and shall not be liable for any damages,\n\
incidental or consequential, arising from the use of this program, even\n\
if you inform him of the possibility of such damages. This disclaimer\n\
shall be governed by the laws of the state of New York.\n\
\n\
In other words, use this program at your own risk.\n\
\n\
If you have questions about killaport or about this disclaimer of warranty,\n\
please feel free to contact the author at brnstnd@acf10.nyu.edu\n\
on the Internet.\n";

static char killaportusage[] =
"Usage: killaport [ -ACHUVW ] port\n\
Help:  killaport -H\n";

static char killaporthelp[] =
"killaport kills an attachport you're running.\n\
\n\
killaport -A: print authorship notice\n\
killaport -C: print copyright notice\n\
killaport -H: print this notice\n\
killaport -U: print short usage summary\n\
killaport -V: print version number\n\
killaport -W: print disclaimer of warranty\n\
\n\
killaport port: kill, kill, kill!\n\
\n\
If you have questions about or suggestions for killaport, please feel free\n\
to contact the author, Daniel J. Bernstein, at brnstnd@acf10.nyu.edu\n\
on the Internet.\n";

#include <sys/types.h>
#include <netinet/in.h>
#include <sys/file.h>
#ifdef BSD
#include <limits.h>
#endif
#include <stdio.h>
#include <netdb.h>
#include <signal.h>
#include <errno.h>
extern int errno;
#include <pwd.h>
extern int getopt();
extern char *optarg; /* these should be in getopt.h! */
extern int optind;
#include "djberr.h"

#ifndef AUTHDIR
#define AUTHDIR "/usr/etc/auth"
#endif

main(argc,argv,envp)
int argc;
char *argv[];
char *envp[];
{
 int opt;
 int uid = getuid();
 int euid = geteuid();
 int fd;
 char buf[32]; /* same length as in attachport */
 int r;
 struct passwd *pw;
 char username[10];
 int pid;
 unsigned short port;
 struct servent *se;

 /* WARNING! We are probably running setuid auth! */

 while ((opt = getopt(argc,argv,"ACHUVW")) != EOF)
   switch(opt)
    {
     case 'A': (void) err(killaportauthor); exit(1);
     case 'C': (void) err(killaportcopyright); exit(1);
     case 'H': (void) err(killaporthelp); exit(1);
     case 'U': (void) err(killaportusage); exit(1);
     case 'V': (void) err(killaportversion); exit(1);
     case 'W': (void) err(killaportwarranty); exit(1);
     case '?': (void) err(killaportusage); exit(1);
    }
 argv += optind, argc -= optind;

 if (!*argv)
  {
   (void) err(killaportusage);
   exit(1);
  }
 if (sscanf(*argv,"%hd",&port) < 1)
   if (!(se = getservbyname(*argv,"tcp")))
    {
     (void) err(killaportusage);
     exit(1);
    }
   else
     port = ntohs(se->s_port); /* inconsistency alert! s_port is int! */

 if (!(pw = getpwuid(uid)))
  {
   (void) errn("killaport: fatal: who are you?");
   exit(4);
  }

 if ((chdir(AUTHDIR) == -1) || chdir("tcp") == -1)
  {
   perrn2("killaport: fatal: cannot change to directory %s/tcp",AUTHDIR);
   exit(2);
  }

 (void) sprintf(buf,"lock.%u",port);

 if (((fd = open(buf,O_RDONLY)) == -1) || (flock(fd,LOCK_EX) == -1))
  {
   perrn2("killaport: fatal: cannot open lock file lock.%u",port);
   exit(5);
  }
 if ((r = read(fd,buf,31)) <= 0)
  {
   errn2("killaport: fatal: cannot read lock file lock.%u",port);
   exit(6);
  }
 buf[r] = '\0';
 if (buf[0] == '!')
  {
   errn2("killaport: fatal: corrupted lock file lock.%u",port);
   exit(7);
  }
 if (sscanf(buf,"%d-%s",&pid,username) < 2)
  {
   errn2("killaport: fatal: port %u not owned by an attachport",port);
   exit(8);
  }
 if (uid && strcmp(username,pw->pw_name))
  {
   (void)
   errn2("killaport: fatal: cannot kill port %u: Permission denied",port);
   exit(9);
  }
 if (kill(pid,SIGTERM) == -1)
   switch(errno)
    {
     case ESRCH:
       (void) errn2("killaport: fatal: port %u in trauma",port);
       exit(10);
     default:
       (void) errn2("killaport: weird: can't kill port %u attachport?",port);
       exit(11);
    }
 (void) errn2("killaport: port %u killed,\
 will be free after last connection drops",port);

 (void) close(fd);
 exit(0);
}
