/*
listaport.c: list current attachports
*/

/* WARNING! listaport is probably setuid auth! */
/* All setuid programs are dangerous! Check them carefully! */

static char listaportauthor[] =
"listaport was written by Daniel J. Bernstein.\n\
Internet address: brnstnd@acf10.nyu.edu.\n";

static char listaportversion[] = 
"listaport version 1.0, April 23, 1990.\n\
Copyright (c) 1990, Daniel J. Bernstein.\n\
All rights reserved.\n";

static char listaportcopyright[] =
"listaport version 1.0, April 23, 1990.\n\
Copyright (c) 1990, Daniel J. Bernstein.\n\
All rights reserved.\n\
\n\
Until January 1, 1995, you are granted the following rights: A. To make\n\
copies of this work in original form, so long as (1) the copies are exact\n\
and complete; (2) the copies include the copyright notice, this paragraph,\n\
and the disclaimer of warranty in their entirety. B. To distribute this\n\
work, or copies made under the provisions above, so long as (1) this is\n\
the original work and not a derivative form; (2) you do not charge a fee\n\
for copying or for distribution; (3) you ensure that the distributed form\n\
includes the copyright notice, this paragraph, and the disclaimer of\n\
warranty in their entirety. These rights are temporary and revocable upon\n\
written, oral, or other notice by Daniel J. Bernstein. These rights are\n\
automatically revoked on January 1, 1995. This copyright notice shall be\n\
governed by the laws of the state of New York.\n\
\n\
If you have questions about listaport or about this copyright notice,\n\
or if you would like additional rights beyond those granted above,\n\
please feel free to contact the author at brnstnd@acf10.nyu.edu\n\
on the Internet.\n";

static char listaportwarranty[] =
"To the extent permitted by applicable law, Daniel J. Bernstein disclaims\n\
all warranties, explicit or implied, including but not limited to the\n\
implied warranties of merchantability and fitness for a particular purpose.\n\
Daniel J. Bernstein is not and shall not be liable for any damages,\n\
incidental or consequential, arising from the use of this program, even\n\
if you inform him of the possibility of such damages. This disclaimer\n\
shall be governed by the laws of the state of New York.\n\
\n\
In other words, use this program at your own risk.\n\
\n\
If you have questions about listaport or about this disclaimer of warranty,\n\
please feel free to contact the author at brnstnd@acf10.nyu.edu\n\
on the Internet.\n";

static char listaportusage[] =
"Usage: listaport [ -ACHUVW ] \n\
Help:  listaport -H\n";

static char listaporthelp[] =
"listaport lists all attachports you're running.\n\
\n\
listaport -A: print authorship notice\n\
listaport -C: print copyright notice\n\
listaport -H: print this notice\n\
listaport -U: print short usage summary\n\
listaport -V: print version number\n\
listaport -W: print disclaimer of warranty\n\
\n\
listaport: list ports\n\
\n\
If you have questions about or suggestions for listaport, please feel free\n\
to contact the author, Daniel J. Bernstein, at brnstnd@acf10.nyu.edu\n\
on the Internet.\n";

#include <sys/types.h>
#include <sys/dir.h>
#include <sys/file.h>
#ifdef BSD
#include <limits.h>
#endif
#include <stdio.h>
#include <pwd.h>
extern int getopt();
extern char *optarg; /* these should be in getopt.h! */
extern int optind;
#include "djberr.h"

#ifndef AUTHDIR
#define AUTHDIR "/usr/etc/auth"
#endif

main(argc,argv,envp)
int argc;
char *argv[];
char *envp[];
{
 int opt;
 int uid = getuid();
 int euid = geteuid();
 DIR *dirp;
 struct direct *dp;
 int fd;
 char buf[32]; /* same length as in attachport */
 int r;
 struct passwd *pw;
 char username[10];
 int pid;
 unsigned short port;

 /* WARNING! We are probably running setuid auth! */

 while ((opt = getopt(argc,argv,"ACHUVW")) != EOF)
   switch(opt)
    {
     case 'A': (void) err(listaportauthor); exit(1);
     case 'C': (void) err(listaportcopyright); exit(1);
     case 'H': (void) err(listaporthelp); exit(1);
     case 'U': (void) err(listaportusage); exit(1);
     case 'V': (void) err(listaportversion); exit(1);
     case 'W': (void) err(listaportwarranty); exit(1);
     case '?': (void) err(listaportusage); exit(1);
    }
 argv += optind, argc -= optind;

 if (!(pw = getpwuid(uid)))
  {
   (void) errn("listaport: fatal: who are you?");
   exit(4);
  }

 if ((chdir(AUTHDIR) == -1) || chdir("tcp") == -1)
  {
   perrn2("listaport: fatal: cannot change to directory %s/tcp",AUTHDIR);
   exit(2);
  }

 if (!(dirp = opendir(".")))
  {
   perrn2("listaport: fatal: cannot read directory %s/tcp",AUTHDIR);
   exit(3);
  }

 while (dp = readdir(dirp))
  {
   if (sscanf(dp->d_name,"lock.%hd",&port) < 1)
     continue; /* fine, skip the file */
   if (((fd = open(dp->d_name,O_RDONLY)) == -1) || (flock(fd,LOCK_EX)) == -1)
    { perrn2("listaport: warning: cannot open lock file %s",dp->d_name); }
   if ((r = read(fd,buf,31)) <= 0)
    { perrn2("listaport: warning: cannot read lock file %s",dp->d_name); }
   (void) close(fd);
   buf[r] = '\0';
   if (buf[0] == '!')
    { perrn2("listaport: warning: corrupted lock file %s",dp->d_name); }
   if (sscanf(buf,"%d-%s",&pid,username) < 2)
     continue; /* fine, we just don't recognize this lock file's format */
   if (!uid || !strcmp(username,pw->pw_name))
     printf("user %s port %u pid %d\n",username,port,pid);
  }

 (void) closedir(dirp);

 exit(0);
}
