.TH attachport 1
.SH NAME
attachport \- attach a server program to a TCP port
.SH SYNTAX
attachport
[
\fB\-ACHUVW\fI
] [
\fB\-01rRvxX\fI
] [
\fB\-p\fIportnum
]
program
[
arg ...
] &
.SH DESCRIPTION
.I attachport
attaches a server to a TCP port.
The port may be specified by
.I portnum,
or one will be assigned;
the server is specified by
.I program.
.I attachport
will pass any further arguments through to the server program.
.PP
.I portnum
can be fully specified as a port number,
or given as a service name.
.PP
Any incoming request on the TCP port
invokes a new copy of the server,
which runs with all input directed from
and output directed to the port.
All environment variables and similar process data
are taken directly from the original invocation of
.I attachport;
however,
.I attachport
will attempt to dissociate itself from its controlling terminal.
.PP
.I attachport
works with
the Authentication Server,
.I authd(8):
the connecting host can verify the identity of the server.
.PP
Options
.B ACHUVW
print the authorship notice,
copyright notice,
help notice,
short usage summary,
version number,
and warranty information respectively.
.PP
.I attachport
has several flags:
.TP 12
\fB\-p\fIportnum
Attempt to attach the server to port
.I portnum.
The port number is restricted on most machines to a number between
0 and 65535; any larger number will be mapped down into this range.
Ports 0 through 1023 are generally restricted to root processes.
Ports above 50000 are a good choice for user (nonstandard) servers.
If you do not specify a port,
.I attachport
will assign one.
.TP
\fB\-v\fI
Verbose:
.I attachport
will report success.
.TP
\fB\-1\fI
Write the port number to standard output as an integer followed by
a newline.
This is the non-user-friendly version of
\fB\-v.
.TP
\fB\-0\fI
Every ten seconds, check that the (otherwise unused) standard input
file still has links. If it doesn't,
.I attachport
will slowly wither away,
dying as soon as its children have exited and cleaned up.
.I attachport
will refuse further connections on the same port in the meantime.
This is appropriate for a shell script;
use
.I killaport(1)
to kill
.I attachport
manually.
.TP
\fB\-x\fI
Locally authenticate each connection (default).
.TP
\fB\-X\fI
Do not locally authenticate connections.
.TP
\fB\-r\fI
Attempt to determine the identity of the other end of
each connection
through the remote Authentication Server (default).
.I attachport
will place the identity
into environment variable
REMOTE, with the form
.I user@in.et.ad.dr
where
.I user
is a string giving the user name
and
.I in.et.ad.dr
is a numerical Internet address.
If the other end is not authenticated,
.I user
will be blank.
Note that
.I user
may contain @ signs;
REMOTE should be parsed from
right to left.
.I attachport
also sets environment variable
PROTO to the string
.I TCP.
.TP
\fB\-R\fI
Do not remotely authenticate.
This is the default if the local port is 113 (auth):
two machines caught in an Authentication Server loop
would rapidly be brought to their knees,
though the loop would end after user
.I auth
hit its process limit on either machine.
.PP
The server could use
.I getpeername(2)
to find out who it's talking to,
but it's faster, simpler, and more portable
to use PROTO and REMOTE.
Future versions of
.I attachport,
versions for other network interfaces,
and versions for completely different communications systems
will all provide the same PROTO/REMOTE interface.
Of course, if PROTO is not TCP, REMOTE will probably
have a different format.
.PP
Note that the insertion of carriage returns after newlines
is normally done automatically by the terminal driver.
You must be careful to output carriage returns before newlines
for your program to talk correctly with a low-level communications server.
.PP
You should make sure to run 
.I attachport
in the background.
Also, you should make sure that its diagnostic output goes
either into a log file or into /dev/null.
.SH DIAGNOSTICS
.TP 1.5i
\fIwarning: can't find my own Internet number?!\fB
Your host is not listed in your hosts table?!
.TP
.I cannot setreuid
This should never happen.
.TP
.I cannot create socket
Something went wrong in creating the communications socket.
.TP
.I cannot bind: Address already in use
Some other process is attached to that port.
(Note that even if that process goes away,
it may be a few minutes before the port is free again.)
.TP
.I cannot bind: some other problem
You can't bind the communications socket to the port for some
other reason. Do you have permission for ports 0 through 1023?
.TP
.I cannot listen
Something went wrong in setting up the socket to listen for
incoming calls.
.TP
.I warning: cannot open /dev/tty
.I attachport
will not be able to dissociate itself from your terminal.
You should get this warning if you invoke
.I attachport
without a controlling terminal in the first place.
.TP
.I warning: cannot dissociate /dev/tty
.I attachport
will not be able to dissociate itself from your terminal.
.TP
.I cannot confirm connection
.I attachport
is unable to access
TCP status information for the connection.
This shouldn't happen;
let your system administrator
and travel agent know if it does.
.TP
.I cannot allocate environment
There's so little memory available that
.I attachport
is unable to find space for
the REMOTE and PROTO environment variables.
If REMOTE and PROTO are in the environment
when
.I attachport
starts,
this can't happen,
and
.I attachport
will run just a tiny bit faster.
(On the other hand, most other programs
will run a tiny bit slower.)
.TP
.I cannot get socket name
This shouldn't happen.
.TP
.I local port locked
.I attachport
is set up incorrectly.
.TP
.I attached to port nnnnn
.I attachport
has successfully attached itself to a port,
and you specified
.B\-v.
.TP
.I cannot execute
.I attachport
is unable to start the server.
.TP
.I local port locked
.I attachport
is set up incorrectly.
.TP
.I warning: cannot unlink authentication entry
This should never happen;
if it does, report the problem to your system administrator
and make sure the entry is removed.
.TP
.I warning: cannot authenticate
.I attachport
is set up incorrectly.
.SH MACHINES
.I attachport
has been tested on a VAX 8700 running ULTRIX 2.0,
a Sun 3 running SunOS,
a Sun 4 running SunOS,
a Convex C-210 running Convex UNIX,
an Astronautics ZS-2 running ZSUnix,
and several other machines.
.SH FILES
None.
.SH BUGS
None known.
.SH RESTRICTIONS
.I attachport
does not let you easily specify a pipeline or other sequence
of commands; you must first set up an executable shell script
that does the desired combination of actions.
.PP
.I attachport
deals correctly but rigidly with multihomed hosts.
.PP
.I attachport
does not support
.I syslog.
.PP
Because of failures in the signal handling semantics,
.I attachport
may fail to notice a child exiting if too many children
exit at once. Normally this doesn't matter, but if
.I attachport
is withering away under
.B\-1
or
.I killaport,
it may stick around waiting for nothing.
.I attachport
checks for this condition and corrects it periodically.
.PP
If
.I program
forks and exits,
passing the connection to its children,
it will lose authentication.
This is consistent with other network programs and
with common sense: it's impossible to define
``the'' user on the other side of the connection if
several processes, all with different uids, have that
connection open.
.SH VERSION
attachport version 4.1, dated April 18, 1990.
.SH AUTHOR
Copyright 1990, Daniel J. Bernstein.
.SH "SEE ALSO"
authtcp(1),
killaport(1),
telnet(1),
authuser(3),
getservbyname(3),
authd(8),
inetd(8)
