.TH authtcp 1
.SH NAME
authtcp \- create a locally authenticated TCP connection
.SH SYNTAX
authtcp
[
\fB\-d\fIn
] [
\fB\-p\fIlocalport
] [
\fB\-rRxXv\fI
] [
\fB\-ACHUVW\fI
]
\fIinetaddr\fB
\fItcpport\fB
\fIprogram\fB
[
\fIarg ...\fB
]
.SH DESCRIPTION
.I authtcp
creates a TCP connection to Internet host
\fIinetaddr\fB
at port
\fItcpport\fB,
then runs
\fIprogram\fB
with the specified
arguments.
The local TCP port may be given as
\fIlocalport\fB,
or
.I authtcp
will assign one.
.PP
Until
.I program
exits,
.I authd(8)
will report that the user owns the TCP connection.
Hence the other end of the connection can
verify the identity of this end,
modulo the lack of Internet security.
.PP
.I inetaddr
can be fully specified as a dotted Internet address,
or given as a domain name.
.I tcpport
can be fully specified as a decimal port number,
or given as a service name.
.PP
.I authtcp
makes the connection available
to 
.I program
as a socket in file descriptor 6,
leaving all other file descriptors intact.
.PP
Options
.B ACHUVW
print the authorship notice,
copyright notice,
help notice,
short usage summary,
version number,
and warranty information respectively.
.PP
.I authtcp
has several flags:
.TP 12
\fB\-d\fIn
Provide the connection in file descriptor
\fIn\fB,
rather than the default descriptor, 6.
If 
\fIn\fB
is not specified,
.I authtcp
will open the connection in the first available file descriptor
and pass that number in place of the first argument to
.I program
that contains solely an equals sign (if there is one).
.TP
\fB\-p\fIlocalport
Attempt to use
.I localport
as the local TCP port number.
This may fail if that port number is out of
range (usually 1\-65535),
or if another process is using that port
(or has used it very recently).
Ports 1 through 1023 are generally reserved for root processes,
and ports above 50000 are generally reserved for user servers.
If you specify
.B \-p0
(default),
.I authtcp
will assign a number.
Several instances of
.B \-p
defer to the last.
.TP
\fB\-X\fI
Do not attempt to locally authenticate the connection;
just set up the connection and run
.I program.
.TP
\fB\-x\fI
Locally authenticate the connection (default).
.TP
\fB\-r\fI
Attempt to determine the identity of the other end of
this connection
through the remote Authentication Server (default).
.I authtcp
will place the identity
into environment variable
REMOTE, with the form
.I user@in.et.ad.dr
where
.I user
is a string giving the user name
and
.I in.et.ad.dr
is a numerical Internet address.
If the other end is not authenticated,
.I user
will be blank.
Note that
.I user
may contain @ signs;
REMOTE should be parsed from
right to left.
.I authtcp
also sets environment variable
PROTO to the string
.I TCP.
.TP
\fB\-R\fI
Do not remotely authenticate.
.TP
\fB\-v\fI
Verbose: Print a message when the connection is established.
Also, report unusual termination of
.I program.
.PP
If
.I program
terminates normally,
.I authtcp
will terminate with the same exit code.
Otherwise it will terminate with exit code 1.
.PP
.SH DIAGNOSTICS
.TP
.I do not understand inetaddr
You probably specified a domain name address
that
.I authtcp
can't decode.
.TP
.I cannot execute
.I authtcp
is unable to execute
.I program.
.TP
.I cannot unlink authentication entry
This should never happen; if it does,
report the problem to your system administrator and
make sure the entry is removed.
.TP
.I cannot bind local port
You probably specified a protected or out-of-range
port with
.I\-p.
If you didn't specify
.I\-p
and this message appears without a number,
all TCP ports are in use.
Report this to your system administrator.
.TP
.I cannot confirm connection
.I authtcp
is unable to access
TCP status information for the connection.
This shouldn't happen;
let your system administrator
and travel agent know if it does.
.TP
.I cannot allocate environment
There's so little memory available that
.I authtcp
is unable to find space for
the REMOTE and PROTO environment variables.
If REMOTE and PROTO are in the environment
when
.I authtcp
starts,
this can't happen,
and
.I authtcp
will run just a tiny bit faster.
(On the other hand, most other programs
will run a tiny bit slower.)
.TP
.I cannot connect
Self-explanatory.
.TP
.I connected to
Self-explanatory.
.TP
.I killed by signal
Self-explanatory.
.TP
.I cannot authenticate
.I authtcp
is not set up correctly.
.TP
.I local port locked
.I authtcp
is not set up correctly.
.TP
.I cannot setreuid
This should never happen.
.TP
.I cannot create socket
This shouldn't happen, unless you have too many files open.
.TP
.I cannot use file descriptor
This should never happen.
.TP
.I cannot get socket name
This shouldn't happen.
.SH MACHINES
.I authtcp
has been tested
on an Astronautics ZS-2
running ZSUnix,
a Sun 3 running SunOS,
a Sun 4 running SunOS,
a Convex C-210 running Convex UNIX,
and several other machines.
.SH FILES
/usr/etc/auth/tcp/*
.SH BUGS
None known.
.SH RESTRICTIONS
If
.I program
closes the connection long before exiting,
another user can with a little effort make the same connection
and pretend to be the user running
.I authtcp.
Hence
.I program
should exit soon after
closing the connection.
(Within several seconds is usually good enough.)
A slightly more subtle
security problem is that a program may set up a connection
under
.I authtcp,
break the connection without exiting,
and wait for a victim program to make the same connection.
If the attacker chose the correct local port number,
there are two possibilities: Either the victim uses
the authentication mechanism and will fail to connect,
or the victim does not understand the mechanism and will
be misrepresented by
.I authd.
As of version 1.5,
.I authtcp
closes both of these holes,
by keeping the connection open until
.I program
exits.
Hence
.I program
must not depend upon the connection being closed
before it exits.
.PP
.I authtcp's
most important function is to create a locally authenticated
connection; remote authentication is useful but can be
performed by
.I program.
It is sometimes difficult to explain that the
.I auth
in
.I authtcp
stands for local, not remote, authentication.
.PP
If
.I gethostbyname(3)
doesn't understand the Domain Name Server,
.I authtcp
won't either.
.PP
.I authtcp
should try all the possible addresses returned
by
.I gethostbyname(3);
it only tries the first.
.PP
If
.I program
passes the connection to another
process and exits,
authentication will be lost.
.SH VERSION
authtcp version 2.1, dated April 18, 1990.
.SH AUTHOR
Copyright 1990, Daniel J. Bernstein.
.SH "SEE ALSO"
attachport(1),
authuser(3),
tcp(4),
authd(8),
gethostbyname(3),
getservbyname(3)
