/*
 * $Header: lex.l,v 1.1 89/03/14 20:59:10 cthuang Exp $
 *
 * C function prototype generator
 * Lexical analyzer specification for flex
 */

WLF             ([ \t\n\f]*)
LETTER		[a-zA-Z_]
DIGIT		[0-9]
ID		{LETTER}({LETTER}|{DIGIT})*
SKIPQUOTED	(\"([^"\n]|\\\")*\"|\'.\'|\\.)

%{
static int curly = 0;		/* number of curly brace nesting levels */
int line_num = 1;
%}

%x INIT1 INIT2 CURLY COMMENT
%%
\n		{ ++line_num; }

#[ \t]*[0-9]+.*$	{ sscanf(yytext, "# %d ", &line_num); --line_num; }
#.*$			;

"("		return '(';
")"		return ')';
"*"		return '*';
","		return ',';
";"		return ';';
"..."		return ELLIPSIS;

auto		return AUTO;
extern		return EXTERN;
register	return REGISTER;
static		return STATIC;
typedef		return TYPEDEF;
char		return CHAR;
double		return DOUBLE;
float		return FLOAT;
int		return INT;
void		return VOID;
long		return LONG;
short		return SHORT;
signed		return SIGNED;
unsigned	return UNSIGNED;
enum		return ENUM;
struct		return STRUCT;
union		return UNION;
const		return CONST;
volatile	return VOLATILE;

\[[^\]]*\]	{
		    strcpy(yylval.text, yytext);
		    return BRACKETS;
		}

{ID}		{
		    strcpy(yylval.text, yytext);
		    if (is_typedef_name(yytext))
			return TYPEDEF_NAME;
		    else
			return IDENTIFIER;
		}

"="{WLF}"{"		{ curly = 1; BEGIN INIT1; }
<INIT1>\n		{ ++line_num; }
<INIT1>"{"		++curly;
<INIT1>"}"		{
			    if (--curly == 0) {
				BEGIN 0;
				return INITIALIZER;
			    }
			}
<INIT1>{SKIPQUOTED}|.	;

"="			{ BEGIN INIT2; }
<INIT2>\n		{ ++line_num; }
<INIT2>({SKIPQUOTED}|[^,;])+[,;]	{
			    yyless(yyleng-1);
			    BEGIN 0;
			    return INITIALIZER;
			}

"{"			{ curly=1; BEGIN CURLY; }
<CURLY>\n		{ ++line_num; }
<CURLY>"{"		{ ++curly; }
<CURLY>"}"		{
			    --curly;
			    if (curly == 0) {
				BEGIN 0;
				return BRACES;
			    }
			}
<CURLY>{SKIPQUOTED}|.	;

"/*"		BEGIN COMMENT;
<COMMENT>\n	{ ++line_num; }
<COMMENT>"*/"	BEGIN 0;
<COMMENT>.	;

[ \t]+		;
.		print_error("syntax error");
