#
PROGRAM MAIN(INP,OUTPUT,HEX,MNEMFILE);
(*
 *  a6809 - mc6809 cross-assembler.
 *
 * Copyright : Jack Jansen en Hans Pronk, H.T.S."A", 1982.
 * History :
 * Jack Jansen, 10-10-83 , V1.0 PRIME :
 *      FCC verbeterd, string werd niet gelezen (a6809.pars)
 *      ORG aan begin pass 2 (a6809.main)
 *      R mode file gemaakt, programmanaam veranderd in MAIN.
 *      Errors detected op de terminal (a6809.main)
 *      Parity strippen in strings (a6809.exec)
 *      Octale getallen (a6809.inpt)
 *      ESC-L voor de hex file (a6809.main)
 *      Filenamen goed inlezen (a6809.main)
 * Jack Jansen, 11-10-83, V1.0 UNIX :
 *      Versie UNIX gemaakt.
 *      Upper/Lower case mapping.
 * Jack, 28-feb-84 :
 *      NEXTCH checkte niet of er >= 80 chars waren ingelezen.
 * Hans Pronk, 16-11-84 , V1.1 Unix :
 *      fatal error ( eof ) verbeterd (a6809.main)
 *      direct page initialiseerd nu goed
 *      start adress voor auto start geimplementeerd.
 *      PC is gelijk aan PCR ( a6809.exec )
 *      op0 geen error op commentaar ( a6809.pars )
 *      MAKEOPER modulair gemaakt ( en gotoes weggewerkt ) ( a6809.pars )
 *
 *)

 (* Define ONE of the following constants : *)
#define UNIX      (* For a UNIX version *)
 (* #define PRIME    (* For a PRIME version *)
(*
                A6809 CONSTANT DEFINITIONS.
                ====== ======== ============
*)
CONST
#ifdef PRIME
    VERSION = 'A6809 V1.1 PR1ME';
    MNEMNAM = 'HTSAME>ETC>A6809.MNEMONICS';
#else
    VERSION = 'A6809 V1.1 UNIX ';
#endif
    FILENAMELENGTH = 32;
    NOFNAME = '                                ';
    MAXMNEM  = 160;
    STRLEN  = 6;                (* LENGTH OF IDENTIFIERS *)
    MAXERR  = 3;                (* # ERRORS PER LINE     *)
    MAXCODE = 5;                (* # CODES PER LINE         *)
    HBMAX   = 30;                (* SIZE OF HEX BUFFER         *)
    LINESPP = 55;                (* LISTING LINES/PAGE    *)
    LINLEN  = 80;                       (* CHARS/LINE *)
    LEGEID   = '      ';        (* GEEN IDENTIFIER         *)
(*
                A6809 TYPE DEFINITIONS.
                ====== ==== ============
*)
TYPE
    STRING = PACKED ARRAY[ 1 .. STRLEN ] OF CHAR;
#ifdef PRIME
    FILENAME = PACKED ARRAY[ 1 .. FILENAMELENGTH ] OF CHAR;
#endif

    VARSTRING = ^VARSRECORD;
    VARSRECORD = RECORD
        INHOUD : CHAR;
        NEXT   : VARSTRING;
    END;

    IDRECORD = ^IDENTRY;
    IDENTRY = RECORD
        WAARDE,DEFLIN : INTEGER;
    END;

    ARGTYPE = ( ARGIND,ARGNUM,ARGREG,ARGSTR,ARGIMM,ARGOPT );

    OPTYPE = ( OPNAM, OPFCB, OPFCC, OPRMB, OPEQU, OPSDP, OPEND,
                OPOPT, OP0, OP1B, OP1W, OPEMT, OPREL, OPREG, OPSTK );
    OPCSET = SET OF OPTYPE;

    REGISTER = ( REGX,REGY,REGU,REGS,REGPC,REGD,PCREG,
                      REGA,REGB,REGCC,REGDP,NOREG );
    REGSET    = SET OF REGISTER;

    SYMBOL = ( NAMSY,NUMSY,SPACESY,EOFSY,ADDSY,MINSY,MULSY,DIVSY,
               MODSY,ANDSY,ORSY,LBRACKSY,RBRACKSY,LESSY,GREATERSY,
               LPARSY,RPARSY,IMMSY,COMMASY,DOTSY,EOLNSY,ERRORSY);

    MNEMRECORD = RECORD
        NAME : STRING;
        OPT  : OPTYPE;
        OPC  : INTEGER;
    END;

    OPLIST = ^OPRECORD;
    OPRECORD = RECORD
        NEXT : OPLIST;
        CASE ARGTP : ARGTYPE OF
        ARGIND : ( AILIST : OPLIST );        (* [ ...... ]           *)
        ARGNUM : ( ANVAL  : INTEGER;        (* NUM, <NUM, >NUM *)
                   ANFORC ,
                   ANLONG : BOOLEAN );
        ARGREG : ( ARREG  : REGISTER;        (* REGISTER NAME   *)
                   ARINC  : -2 .. 2 );  (* # OF INC/DEC    *)
        ARGSTR : ( ASTEXT : VARSTRING );(* OTHER STRINGS   *)
        ARGIMM : ( AIVAL  : INTEGER );        (* #<EXPRESSION>   *)
        ARGOPT : ( AOOPT : STRING );       (* STRING FOR OPT *)
    END;

    STMT = ^STMTRECORD;
    STMTRECORD = RECORD
        LEBEL : STRING;
        OPCODE : INTEGER;
        OPT : OPTYPE;
        OPERANDS : OPLIST;
    END;

    TREE = ^TREELEAF;
    TREELEAF = RECORD
        LLINK,RLINK : TREE;                (* LINKER/RECHTER ZOON *)
        NAME : STRING;                        (* IDENTIFIER NODE     *)
        DATA : IDRECORD;                (* DATA IN DEZE NODE   *)
    END;
(*
                A6809 GLOBAL VAR DEFINITIONS.
                ====== ====== === ============
*)
VAR
        I : INTEGER;
        INP,
        HEX    : TEXT;                             (* HEX OUTPUT FILE      *)
        MNEMFILE : FILE OF MNEMRECORD;
#ifdef PRIME
        INPNAME,                       (* INPUT FILE NAME *)
        OUTNAME,                       (* OUTPUT FILE NAME *)
        HEXNAME : FILENAME;           (* AND HEXFILE NAME *)
#endif
        C      : CHAR;                             (* INGELEZEN CHARACTER  *)
        SY     : SYMBOL;                     (* INGELEZEN TERMINAL   *)
        SYNAM  : STRING;                     (* INGELEZEN IDENTIFIER *)
        SYNUM  : INTEGER;                     (* INGELEZEN GETAL             *)
        SYCHAR : ARRAY[CHAR] OF SYMBOL;      (* MAP CHAR->SYMBOLTYPE *)
        REGNAME: ARRAY[REGISTER] OF STRING;    (*NAMES OF REGISTERS *)
        MNEMTAB : ARRAY[1..MAXMNEM] OF MNEMRECORD;  (* MNEMONIC TABLE*)
        TITLE  : VARSTRING;                     (* PAGE HEADER          *)
        ROOT : TREE;                              (* FIRST IDENTIFIER    *)
        ST   : STMT;                              (* STATEMENT           *)
        COMMENT,                             (* TRUE IF COMMENTLINE  *)
        DEBUG,                              (* DEBUGGING ON *)
        OPTLIST,                             (* TRUE IF LISTING WTD  *)
        OPTBIN,                                     (* TRUE IF BINARY WANTED*)
        OPTSYM,                                     (* TRUE IF SYMTABLE WTD *)
        PASS2,                                     (* TRUE ALS IN PASS 2   *)
        INITIALIZING,                         (* TRUE ALS IN INITIALISATIE*)
        STOPPED: BOOLEAN;                     (* TRUE ALS 'END'       *)
        LOCCNTR,                             (* LOCATION COUNTER     *)
        OLOCCNTR,                             (* OLD LOC. COUNTER     *)
        CODELOC,                             (* HEXBUF LOCATION      *)
        CODELIN,                             (* INDEX IN 'CODES'     *)
        CODECNTR,                             (* INDEX IN 'HEXBUF'    *)
        LINCNTR,                             (* LINE COUNTER             *)
        PAGCNTR,                             (* PAGE COUNTER             *)
        CHRCNTR,                             (* CHARPOS COUNTER      *)
        ERRLIN ,                             (* # ERRORS IN LINE     *)
        DIRPAG ,                             (* SETDP VARIABLE       *)
        STARTADR ,                           (* ADRESS FOR AUTOSTART *)
        MNEMLEN,                         (* LENGTH OF MNEMTAB*)
        ERRCNTR: INTEGER;                     (* ERROR COUNTER             *)
        ASSOPC,                              (* PSUEDO-OPERATIONS    *)
        PROOPC : OPCSET;                     (* REAL OPERATIONS      *)
        INXREG,                              (* INDEX REGISTERS      *)
        ACCREG : REGSET;                     (* ACCU OFFSET REGS.    *)
        LINE   : PACKED ARRAY[1..LINLEN] OF CHAR;
                                             (* LINE FOR LISTING     *)
        ERRORS : PACKED ARRAY[1..MAXERR] OF CHAR;
                                             (* ERROR CHARACTERS     *)
        CODES  : ARRAY[1..MAXCODE] OF INTEGER;(*LISTING BINARY CODES *)
        HEXBUF : ARRAY[1..HBMAX] OF INTEGER; (* HEXFILE BUFFER       *)

(*
                A6809 PROCEDURE/FUNCTION HEADERS.
                ====== ========= ======== ========
DE ROUTINES STAAN OP DE VOLGENDE FILES :

A6809.SYMB
                GETNAM
                NEWNAM

A6809.INPT
                NEXTCH
                INSYMBOL
                        INNAM
                        INNUM
                ISINIT
A6809.OUTP
                LISTLINE
                PRINTHEX
                OUTHEX
                FLUSHEX
A6809.PARS
                MAKEOPER
                        MAKEXPR
                                MAK1NUM
                MAKESTMT

A6809.EXEC
                DOINIT
                DOSTMT
                        REMTITLE
                        REGNYB
                        REGBIT
                        MKLEBEL
                        REMSTMT
                                REMOPLIST
                        DOOPER

*)
{ ***************************************

PROCEDURE NEXTCH; EXTERN;
(##* LEES VOLGENDE KARAKTER, EN STOP DAT IN 'C'.                        *##)


PROCEDURE INSYMBOL; EXTERN;
(##*   LEES EEN SYMBOL EN ZET GOLBALE VAR'S SY,SYNUM,SYNAM. *##)


PROCEDURE ISINIT; EXTERN;
(##* ISINIT INITIALISEERT VOOR INSYMBOL.                                *##)


FUNCTION MAKEOPER : OPLIST; EXTERN;
(##*  LEEST EEN LIJST MET OPERANDEN, EN RETURNT EEN POINTER NAAR *##)
(##*  HET RESULTAAT                                               *##)



FUNCTION MAKESTMT : STMT; EXTERN;
(##*  LEEST (MBV MAKEOPER) EEN REGEL, EN RETURNT EEN POINTER NAAR *##)
(##*  HET RESULTAAT                                                *##)


PROCEDURE DOSTMT(S : STMT); EXTERN;
(##*  DOSTMT VOERT STATEMENTS UIT. *##)


FUNCTION NEWNAM(NAME : STRING; DATA : IDRECORD) : BOOLEAN; EXTERN;
(##*  NEWNAM ZET NAAM 'NAME' MET DATA 'DATA' IN DE SYMBOLTABLE. *##)
(##*  ER WORDT 'TRUE' GERETURNED ALS 'DATA' NIET GELIJK IS AAN  *##)
(##*  EEN EVENTUELE VORIGE 'DATA'.                              *##)


FUNCTION GETNAM(NAME : STRING) : IDRECORD; EXTERN;
(##*  GETNAM RETURNT DE DATA BEHORENDE BIJ 'NAME', EN 'NIL' ALS  *##)
(##*  'NAME' NIET GEVONDEN WORDT.                                       *##)


PROCEDURE OUTHEX(VAL,LEN : INTEGER); EXTERN;
(##* OUTHEX OUTPUT 'LEN' BYTES VANUIT VAL NAAR DE LISTING EN NAAR *##)
(##* DE HEX FILE.                                                        *##)


PROCEDURE FLUSHEX; EXTERN;
(##* FLUSHEX SCHRIJFT DE BUFFER 'HEXBUF' NAAR DE 'HEX' FILE.        *##)


PROCEDURE ERROR(C : CHAR); EXTERN;
(##* GEEFT ERRORMELDING 'C'.                                        *##)


FUNCTION FIND(MNEM : STRING;VAR OPC : INTEGER; VAR TP : OPCTYP);
                                                EXTERN;
(##* FIND ZOEKT MNEMONICS OP EN RETURNT 'OPC' EN 'TP'.                *##)


PROCEDURE LISTLINE; EXTERN;
(##* LISTLINE LIST 1 REGEL, EN ZORGT VOOR PAGINERING,ETC.                *##)

********************************** }
(*  FORWARD DEFINITIONS *)

PROCEDURE ERROR( C : CHAR ) ; FORWARD;

PROCEDURE FLUSHEX (LASTBLOK : BOOLEAN ); FORWARD;

PROCEDURE PRINTHEX( VAR F : TEXT; NUM,SIZ : INTEGER);FORWARD;

(*  EXTERN DEFINITIONS *)
#ifdef PRIME
FUNCTION IAND(I,J : INTEGER) : INTEGER; EXTERN;  (* BINARY AND *)

FUNCTION IOR(I,J : INTEGER) : INTEGER; EXTERN; (* BINARY OR *)

#else
FUNCTION IAND(I,J : INTEGER) : INTEGER;
BEGIN
    ERROR('?');
    IAND := 0;
END;

FUNCTION IOR(I,J : INTEGER) : INTEGER;
BEGIN
    ERROR('?');
    IOR := 0;
END;

#endif

#include "symb.inc"
#include "inpt.inc"
#include "outp.inc"
#include "pars.inc"
#include "exec.inc"

#ifdef PRIME
PROCEDURE INFNAM(VAR NM : FILENAME);
(* INFNAM LEEST EEN FILENAME VAN DE TERMINAL  *)
(* VAR I : INTEGER; *)
BEGIN
    WHILE (INPUT^ = ' ') AND NOT EOLN(INPUT) DO GET(INPUT);
(*    FOR I := 1 TO FILENAMELENGTH DO   *)
(*      IF EOLN(INPUT) THEN NM[I] := ' ' ELSE READ(INPUT,NM[I]); *)
    READ(INPUT,NM);
END (* INFNAM *);

PROCEDURE READOPT;
(* VAR I : INTEGER; *)
BEGIN
    READLN;
    WHILE ( INPUT^ = ' ') AND NOT EOLN DO
      GET(INPUT);
(*  FOR I := 1 TO STRLEN DO     *)
(*      IF INPUT^ IN ['A' .. 'Z'] THEN READ(SYNAM[I]) ELSE SYNAM[I] := ' '; *)
READ(SYNAM);
FOR I := 1 TO STRLEN DO
  IF SYNAM[I] IN ['a'..'z'] THEN
    SYNAM[I] := CHR(ORD(SYNAM[I])+ORD('A')-ORD('a'));
END (* READOPT *);
#endif


BEGIN (* OF MAIN PROGRAM *)
    CHRCNTR := 0;
#ifdef PRIME
    WRITELN(OUTPUT,'[',VERSION,']');
    WRITE(OUTPUT,'Input file - ');
    INFNAM(INPNAME);
    WRITE(OUTPUT,'Listing file - ');
    READLN;
    INFNAM(OUTNAME);
    WRITE(OUTPUT,'Hex file - ');
    READLN;
    INFNAM(HEXNAME);
    OPTBIN := HEXNAME <> NOFNAME;
    OPTLIST:= OUTNAME <> NOFNAME;
    DEBUG := FALSE;
    INITIALIZING := TRUE;
    REPEAT
        WRITE('Option - ');
        READOPT;
        IF SYNAM <> LEGEID THEN OPTION(SYNAM);
    UNTIL SYNAM = LEGEID;
    IF INPNAME <> NOFNAME THEN RESET(INP,INPNAME);
    IF HEXNAME <> NOFNAME THEN REWRITE(HEX,HEXNAME)
    ELSE IF OPTBIN THEN REWRITE(HEX,'HEX.6809');
    IF OUTNAME <> NOFNAME THEN REWRITE(OUTPUT,OUTNAME);
%CHECKS OFF;
    IF OUTNAME <> NOFNAME THEN WRITELN(CHR(1),CHR(1));
%CHECKS ON;
#else
    RESET(HEX);
    READ(HEX,I);
    OPTBIN := I <> 0;
    READ(HEX,I);
    OPTLIST := I <> 0;
    READ(HEX,I);
    DEBUG := I <> 0;
    READ(HEX,I);
    OPTSYM := I <> 0;
    REWRITE(HEX);
#endif
    IF OPTBIN THEN
      WRITELN(HEX,CHR(27),'L'); (* ESC-L, labbus load sequence *)
    INITIALIZING := FALSE;
    ROOT := NIL;
    PASS2 := FALSE;
    TITLE := NIL;
#ifdef PRIME
    IF INPNAME <> NOFNAME THEN RESET(INP,INPNAME) ELSE RESET(INP);
#else
    RESET(INP);
#endif
    NEXTCH;                        (* LEES EERSTE CHAR *)
    MNEMINIT;                        (* INIT MNEMONICTABLE *)
    ISINIT;                        (* INSYMBOL INIT. *)
    DOINIT;                        (* DOSTMT INIT.   *)
(**************                PASS 1                 *************)
    LOCCNTR := 0;
    OLOCCNTR := 0;
    LINCNTR := 0;
    PAGCNTR := 0;
    ERRCNTR := 0;
    CODELIN := 0;
    DIRPAG := 0;
    CODELOC := 0;
    STARTADR := 0;
    STOPPED := FALSE;
    WHILE NOT STOPPED AND NOT EOF(INP) DO BEGIN
        OLOCCNTR := LOCCNTR;
        COMMENT := FALSE;
        LINCNTR := LINCNTR+1;
        ST := MAKESTMT;                (* LEES STATEMENT *)
        ERRORS := '   ';
        ERRLIN := 0;
        IF NOT COMMENT THEN
            DOSTMT(ST);         (* VOER STATEMENT UIT *)
        IF DEBUG THEN LISTLINE;
        CHRCNTR := 0;
        CODELIN := 0;
        IF NOT STOPPED AND NOT EOF(INP) THEN
        BEGIN
           READLN(INP);
           NEXTCH;
        END;
    END;
    OLOCCNTR := 0;
    FLUSHEX(FALSE);
(**************         PASS 2                *************)
    PASS2 := TRUE;
    STOPPED := FALSE;
    LOCCNTR := 0;
    OLOCCNTR := 0;
    LINCNTR := 0;
    CODELIN := 0;
    PAGCNTR := 0;
    ERRCNTR := 0;
    CHRCNTR := 0;
    CODELOC := 0;
    STARTADR:=0;
    DIRPAG := 0;
#ifdef PRIME
    IF INPNAME <> NOFNAME THEN RESET(INP,INPNAME) ELSE RESET(INP);
#else
    RESET(INP);
#endif
    NEXTCH;
    WHILE NOT STOPPED AND NOT EOF(INP) DO
    BEGIN
        COMMENT := FALSE;
        LINCNTR := LINCNTR+1;
        OLOCCNTR := LOCCNTR;
        ERRLIN := 0;
        ERRORS := '   ';
        ERRLIN := 0;
        ST := MAKESTMT;
        IF NOT COMMENT THEN
            DOSTMT(ST);
        IF OPTLIST OR (ERRLIN > 0) THEN LISTLINE;
        CHRCNTR := 0;
        IF NOT STOPPED AND NOT EOF(INP) THEN
        BEGIN
           READLN(INP);
           NEXTCH;
        END;
    END;
    IF NOT STOPPED THEN (* EOF WITHOUT END PSEUDO OP *)
	BEGIN
	LINCNTR := LINCNTR +1;
	ERRCNTR := ERRCNTR +1;
	WRITELN('E  ',LINCNTR:5,'     ****  NO END STATEMENT ***** ');
	END;
    IF OPTSYM THEN SYMTABLE;
    WRITELN('Errors detected : ',ERRCNTR:1);
#ifdef PRIME
    REWRITE(OUTPUT,'@TTY');
    WRITELN('Errors detected : ',ERRCNTR:1);
#endif
    FLUSHEX(FALSE);
    FLUSHEX(TRUE);
END.
