/* diffc.c - provide diff -c output on non-berkeley systems
**
** cc -o diffc diffc.c
**
**  or for V7 type systems
**
** cc -o diffc -DV7 diffc.c
*/

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef V7
#define strchr index
#include <string.h>		/* BEZERKELY 4.2 is strings.h */
#else
#include <string.h>
#endif

#define WIDTH 3
#define SEPERATE "***************"
#define max(_a,_b) (((_a)>(_b))?(_a):(_b))

FILE *diff, *fd1, *fd2;
int line1 = 1, line2 = 1;
long off1, off2;
char buffer[BUFSIZ];
int width = WIDTH;

char *ctime();
FILE *popen();

main(argc, argv)
int argc;
char **argv;
    {
    int i;
    struct stat statbuf;

    strcpy(buffer, "diff");
    /* skip over minus args - they must be for diff */
    for (i = 1; i < argc; ++i)
	{
	if (argv[i][0] != '-')
	    break;
	else if (!strncmp(argv[i], "-c", 2))
	    {
	    width = atoi(argv[i]+2);
	    if (width < 1 || width > 20)
		width = WIDTH;
	    }
	else
	    {
	    strcat(buffer, " ");
	    strcat(buffer, argv[i]);
	    }
	}
    if (argc - i != 2)
	{
	fprintf(stderr, "Usage: %s [-minusargs] file1 file2\n", argv[0]);
	exit(1);
	}
    strcat(buffer, " ");
    strcat(buffer, argv[i]);
    strcat(buffer, " ");
    strcat(buffer, argv[i+1]);

    fd1 = fopen(argv[i], "r");
    fd2 = fopen(argv[i+1], "r");

    if (!fd1 || !fd2)
	{
	fprintf(stderr, "%s: File not accessible\n", argv[0]);
	exit(1);
	}

    if ((diff = popen(buffer, "r")) == NULL)
	{
	fprintf(stderr, "%s: Cant find diff\n", argv[0]);
	exit(1);
	}

    stat(argv[i], &statbuf);
    printf("*** %s\t%s", argv[i], ctime(&(statbuf.st_mtime)));
    stat(argv[i+1], &statbuf);
    printf("--- %s\t%s", argv[i+1], ctime(&(statbuf.st_mtime)));

    dodiffs();
    }

dodiffs()
    {
    char line[BUFSIZ];
    char type, disp;
    int from1, to1, from2, to2;
    int dfrom1, dto1, dfrom2, dto2;
    int i;
    int skip;

    while (fgets(line, sizeof(line), diff))
	{
	parse(line, &from1, &from2, &to1, &to2, &type);

	dfrom1 = max(from1 - width, 1);
	dfrom2 = max(from2 - width, 1);

	line1 = toline(fd1, line1, dfrom1);
	line2 = toline(fd2, line2, dfrom2);

	off1 = ftell(fd1);
	off2 = ftell(fd2);

	dto1 = checkline(fd1, line1, to1 + width);
	dto2 = checkline(fd2, line2, to2 + width);

	/* checkline moved the position, so seek back to last point */
	fseek(fd1, off1, 0);
	fseek(fd2, off2, 0);

	printf("*** %d,%d\n", dfrom1, dto1);
	for (i = dfrom1; i <= dto1; ++i)
	    {
	    if (!fgets(buffer, sizeof(buffer), fd1))
		break;
	    disp = ' ';
	    if (i >= from1 && i <= to1)
		disp = (type == 'c') ? '!': '-';
	    printf("%c %s", disp, buffer);
	    }
	printf("\n--- %d,%d -----\n", dfrom2, dto2);
	for (i = dfrom2; i <= dto2; ++i)
	    {
	    if (!fgets(buffer, sizeof(buffer), fd2))
		break;
	    disp = ' ';
	    if (i >= from2 && i <= to2)
		disp = (type == 'c') ? '!': '+';
	    printf("%c %s", disp, buffer);
	    }

	/* seek back because next diff may overlap */
	fseek(fd1, off1, 0);
	fseek(fd2, off2, 0);

	/* calculate the number of display lines generated by diff */
	skip = to1 - from1 + to2 - from2 + 2;
	if (type == 'c')
	    skip++;		/* add one for the seperator */
	toline(diff, 0, skip);
	}
    }

parse(line, from1, from2, to1, to2, type)
char *line, *type;
int *from1, *from2, *to1, *to2;
    {
    char *part2, *ptr;

    /* expect a line range type range */
    if(!isdigit(line[0]))
	{
	fprintf(stderr, "Unknown junk from diff: %s\n", line);
	exit(2);
	}
    if ((part2 = strchr(line, 'a')) || (part2 = strchr(line, 'c')) || (part2 = strchr(line, 'd')))
	{
	*type = *part2;
	*part2++ = '\0';
	printf("%s\n", SEPERATE);
	}
    else
	{
	fprintf(stderr, "Unknown junk from diff: %s\n", line);
	exit(2);
	}
/* file 1 */
    *from1 = atoi(line);
    if (ptr = strchr(line, ','))
	*to1 = atoi(++ptr);
    else
	*to1 = *from1;
/* file 2 */
    *from2 = atoi(part2);
    if (ptr = strchr(part2, ','))
	*to2 = atoi(++ptr);
    else
	*to2 = *from2;

    if (*type == 'a')
	++*from1;

    if (*type == 'd')
	++*from2;
    }


toline(fd, lineis, linenum)
FILE *fd;
int lineis, linenum;
    {
    /* note: toline can seek to one more line than there actually is (EOF) */
    while (lineis < linenum)
	{
	if (fgets(buffer, sizeof(buffer), fd))
	    ++lineis;
	else
	    break;
	}
    return lineis;
    }

checkline(fd, lineis, linenum)
FILE *fd;
int lineis, linenum;
    {
    lineis = toline(fd, lineis, linenum);
    if (!fgets(buffer, sizeof(buffer), fd))
	--lineis;
    return lineis;
    }
