.TH UniDesk 3 "March 14, 1988"
.SH NAME
UniDesk \- Simplified textual desktop user interface package
.SH SYNOPSIS
.I cc [options] [files] UniDesk.o -ltermcap [libraries]
.SH DESCRIPTION
.I UniDesk
is a library package which implements a desktop user interface.  It may
be thought of as "a level above curses(3)".  It consists of two distinct
parts: ADD (Ascii Desktop Driver) and SDI (Simplified Desktop Interface).
As a minimum, the programmer need only be concerned with a set of 10 SDI
calls, although the advanced user wishing to customize the desktop
environment may call other routines as well, provided the file "unidesk.h"
is included in the user's source.

From the end-user's point of view, the UniDesk user interface as defined
by the SDI is as follows:  The input device for characters remains the
keyboard.  An analog input device, such as a Logitech Serial PC-Mouse or
bitpad, is searched for.  If not found, then the keyboard arrow keys, if
any, plus the keys h/j/k/l and turbo H/J/K/L allow cursor movement,
with the space bar serving as the button.  Objects must be frontmost
in order to interact with the user; clicking on any part of an object
or program output to it will make it so.  The pull-down menus work in
the usual fashion; clicking again on the bodies of windows or icons will
invoke an appropriate action by the applications program; clicking on
the names of windows or icons will allow them to be moved to a new
location; and clicking near the lower right corner of a window allows
it to be resized.  To quit the application, a standard "Quit" option is
always installed in a standard "UniDesk" menu in the upper left corner.

If a directory entitled ".unideskrc" is present in the user's home
directory, then the names of any files therein will be
installed in a standard "UniDesk" menu at the upper-left corner.  When
one of these names is selected, UniDesk attempts to execute them by
using system(3) to invoke sh(1) on the given file.

Signals such as stopjob and interrupt are handled gracefully; almost all
I/O (see bugs section below) is done directly to the user's terminal so
that standard file descriptors may be closed if desired.

In order to use UniDesk from the applications programmer's point of view,
the following description should suffice:

.nf
For the calls below:
	int (*DefProc) ()	Pointer to user's procedure
	char *Name		Name of object
	char *Item		Item name for menus
	char *String		Input or output string for windows
	int MaxLength		Maximum allowable input for windows

The 10 self-explanatory calls are:
	SDIEasyReady ()
	SDIEasyGo (DefProc)
	SDIEasyOpenWindow (Name,DefProc)
	SDIEasyCloseWindow (DefProc)
	SDIEasyOpenMenu (Name,Item,DefProc)
	SDIEasyCloseMenu (DefProc)
	SDIEasyOpenIcon (Name,Image,DefProc)
	SDIEasyCloseIcon (DefProc)
	SDIEasyOutput (DefProc,String)
	SDIEasyInput (DefProc,String,MaxLength)

.fi
The input and output routines work with windows only; the DefProc parameter
to SDIEasyGo is a pseudo-concurrency feature:  it is executed periodically
even if the end-user is doing nothing.  The constant NULL may be passed
for any of these DefProcs if desired.

This is only a cursory description; for further documentation, the user
is presently directed to the documented source code.

.SH FILES
.nf
UniDesk.o			Object file to link with
unidesk.h			Header file necessary only for advanced use
~/.unideskrc		Directory containing desk accessories
/tmp/unidesk?		Housekeeping file used during initialization
sdi.hi.c			Source: Smart init; Desktop handling
sdi.lo.c			Source: Event handling
add.hi.c			Source: Object handling
add.med.c			Source: Object drawing
add.inp.c			Source: Analog input
add.lo.c			Source: Low-level output; recordkeeping
.fi
.SH SEE ALSO
window(1), termcap(3), UniDesk source code
.SH AUTHOR
Ben Koning (408/738-1763) (ben@apple.com)
.SH BUGS
The auto-screen-location routine in SDI causes icons to eventually be
drawn off the desktop.

The SDI text input routine is blocking and filedes 0 is still used from
keyboard input (rather than its own filedes directly to the /dev/tty?,
as is already done with the output).  These quirks must be fixed before
UniDesk is truly ready for full multiprogramming.
