
.TL
UNIDESK USER'S MANUAL
.TL
UniDesk: An Undergraduate Thesis in Computer Science
.AU
Ben Koning [556498717 408/738-1763 ben@apple.com]
.AI
University of California, Santa Cruz
.AB
A single-process device-independent textual desktop user interface
library package which is reasonably flexible yet extremely easy to
use for applications programmers, such as students in an intermediate
programming class, to use from relatively slow remote terminals posessing
a bare minimum of features, is described.
This package is uses \fItermcap\fP to achieve terminal independence, and
supports several analog input devices, if available at runtime.  It is
compatible without any
modification with the 4.2BSD, 4.3BSD, and XENIX System V dialects of
UNIX.  This
library of routines, entitled UniDesk, is implemented successfully and
completely in the C programming language, and is
concisely defined and described.
.AE
.bp
.LP
.nr PS 10
.DA





.TL
INTRODUCTION
.PP
UniDesk is a desktop user interface which works on any ASCII terminal on
any UNIX system, provided the terminal has a valid \fItermcap\fP data
base entry.
After compilation, UniDesk consists of a single useful file entitled
\fIUniDesk.o\fP.  One may view UniDesk as "a layer on top of \fIcurses\fP".
It is important to realize that UniDesk does \fBnot\fP fork off any processes
(except when executing a desk accessroy; see below);
it is simply a toolbox of routines to implement to "look" of a desktop.
However, it is quite possible to write a shell which \fIdoes\fP fork off a
new process for each new window, and thus one could have desktop full of
windows acting as viewports into many different processes' outputs.


.TL
RUNNING A PROGRAM WHICH USES UNIDESK
.PP
When a program which has been compiled with UniDesk begins executing and
turns control over to UniDesk, it first checks to see if the output device
meets some minimum requirements dealing with cursor addressability and
screen size.  Next, the best possible locator device is started up, using
the following hierarchy: Logitech PC Mouse > GTCO DigiPad 5 > Summagraphics
Bitpad One > terminal keyboard.  Next, memory resources are allocated, and
finally, a standard desktop background pattern is drawn.
If any errors occur during this startup sequence, then the program's execution
is terminated with an error message.
.PP
If startup occurs smoothly, then control will bounce back and forth between
UniDesk and the application program during normal execution.  Barrning an
unusual event (described below), however, the only clean path out of UniDesk
control is through the standard "Quit" option in the standard "UniDesk" menu
(see below).


.TL
USING THE DESKTOP ENVIRONMENT
.PP
If an analog locator device such as a mouse is available, then cursor movement
is obvious.  Use the locator device to move the cursor over the desired region,
and then \fIselect\fP it by pressing a button.  No distinction is made between
different buttons if the locator device has more than one.  If the locator
device is the keyboard, then the \fBk\fP, \fBj\fP, \fBh\fP, \fBl\fP, keys may be
used to move up, down, left, and right, respecitively, while the space bar
serves as the "button".  If the terminal has
specialized arrow keys, then those are supported also.  In addition, the
capitalized versions of the above (\fBK\fP, \fBJ\fP, \fBH\fP, \fBL\fP) are
turbo cursor moves.
.PP
Selecting an object at any visible part will cause it to become the frontmost
object on the desktop.  All objects (the term \fIobject\fP refers to either
a window, menu, or icon) have two parts which are distinguished: the
\fIname region\fP and the \fIcontent region\fP.  If an object is frontmost and
its content region is selected, then the procedure in the program's code which
has been attached to that object (see below) is activated, causing some
program-specific action.  This is called a \fIselection event\fP.  Menus work
somewhat differently; selecting a menu title produces a "pull-down" menu, of
which an item may be selected (or not selected, by selecting outside of the
pull-down menu). 
.PP
If the name region is
selected, then the \fIdrag frame\fP is superimposed over the object, and the
locator device will then move the drag frame freely around the desktop.  When
the locator button goes down again, the object is redrawn in its new
location.  This is called a \fImove event\fP.
.PP
If the area near the lower right-hand corner of a window is selected, then
the drag frame is superimposed on the window frame just as in a move event; but
this time the locator controls the new size of the window.  The window's
contents are reinterpreted in terms of its new aspect ratio, and the window's
next-character position is moved to the lower left-hand corner.  This is called
a \fIresize event\fP.
.PP
During all of these events, the cursor image changes its shape to reflect the
current operation in progress.


.TL
THE STANDARD UNIDESK MENU
.PP
Unless explicitly closed by the application program, a standard "UniDesk" menu
will always exist as the first menu in the upper left-hand corner of the
desktop.  The first item in this menu is called the \fIstandard Quit option\fP.
Selecting this option will result in a clean exit from UniDesk, returning
control to the application, which will then usually initiate shutdown routines
of its own.
.PP
If the user has a directory at her root level entitled " ~/.unideskrc", then
the rest of the items in the standard menu will contain the names of the files
in that directory, provided there aren't too many files (in which case not all
appear), and provided the names aren't too long (in which case UniDesk
terminates abnormally).  If one of these names is selected, then its full
pathname is passed as a command line to the standard shell \fIsh\fP(1) after
the screen is cleared.  When the selected command has finished execution,
UniDesk waits for the press of any key before resuming normal operation.


.TL
ABNORMAL EVENTS
.PP
If the interrupt, quit, or software termination signals are recieved by
UniDesk, usually as a result of \fB^C\fP, \fB^\\\fP, or
\fIkill\fP(1) activity, UniDesk
will terminate the program that it is serving gracefully with an error
message.  For systems with job control, the stop-job and continue-job
signals, usually caused by typing \fB^Z\fP and \fBfg\fP, prompt behavior
just like \fIvi\fP(1), clearing and redrawing the screen as appropriate.


.TL
GETTING STARTED WITH PROGRAMMING UNDER UNIDESK
.PP
UniDesk is divided into two main parts:  The ASCII-Terminal Desktop Driver
(\fBADD\fP) implements the "overlapping windows" model of the desktop out
of character- and elementary \fItermcap\fP terminal-control operations, as
well as input device support.  The direct use of ADD is quite challending
and requires the programmer to include the source file "unidesk.h" while
maintaining an intricate awareness of ADD data structures.  However, UniDesk
can also present a friendlier side.
.PP
The Simplified Desktop Interface (\fBSDI\fP) is that part of UniDesk which
goes out of its way to be easy to use, from applications programmer's point
of view.  Its use is outlined in detail below.
.PP
The identifiers of all UniDesk calls, globals, and defines are guaranteed
to be prefixed by
either "ADD" or "SDI", in order to prevent collisions with user identifiers.
In addition, rather strict naming conventions govern the naming of all
UniDesk identifiers, as a quick look at the call names below will indicate.
.PP
UniDesk's SDI achieves its simplictiy by the use of two powerful concepts:
The definition procedure, or \fIdefproc\fP, which is a procedure in the
user's code which may be called by UniDesk at certain times; and the concept
of a \fIdesktop manager\fP routine in Unidesk which handles all
desktop-related activity (such as object movement, window resizing,
and menu selection) except those object selection events which actually require
the calling program's attention.  Thus, the concept of a desktop is somewhat
abstracted.
.PP
The following box and model procedure headers describe exactly how to
use the UniDesk SDI routines:



.nf
************* THE BIG TEN: EASY DESKTOP INTERFACE CALLS *****************
                                                                        
   These are the ONLY calls which the *casual* UniDesk user will ever   
   have to make.  They take care of everything, including all event     
   handling and the location of objects.  The names of all of these     
   calls is of the form "SDIEasy___".  In order to use UniDesk with     
   these calls, do the following:                                       
                                                                        
   (0) The program which uses UniDesk must be compiled as follows:      
       cc [options] [files] UniDesk.o [libraries] -ltermcap             
                                                                        
   (1) Call SDIEasyReady.  Everything is initialized and standard       
       objects are drawn on the desktop.  A standard "UniDesk" menu     
       with desk accessories from the user's ".unideskrc" directory     
       and a standard "Quit" option is installed, over a standard       
       desktop pattern.  This routine must be called first; after       
       calling this routine, any of the other "SDIEasy___" routines     
       may be called.                                                   
                                                                        
   (2) Call any number of "SDIEasy___" routines to setup the desktop    
       before event handling starts, if it is so desired.  These        
       calls open and close objects and may be called to close stan-    
       dard objects, if desired.  They may also be called during        
       execution of the program as appropriate.  They associate         
       user-procedures with the objects opened, so that they may be     
       called if the appropriate object is selected.  Duplicate pro-    
       cedure associations are not allowed; that is, every object on    
       the desktop must have its own associated procedure.              
                                                                        
   (3) Call SDIEasyGo, to initiate totally automatic event handling.    
       SDIEasyGo takes total control, and should be the last call be-   
       fore the user's own shutdown code, if any.  The user's program   
       will regain control again (hopefully only momentarily) when an   
       object is selected.  If desired, a procedure given by the user   
       is executed every few times this routine does its main loop.     
       Both the latter user procedure and the user procedure that is    
       called when an object is selected should return as soon as       
       possible.  SDIEasyGo returns only if the standard.  "Quit"       
       option in the standard "UniDesk" menu is selected.               
                                                                        
   (4) During user program execution, call SDIEasyOutput to output      
       strings to windows, and call SDIEasyInput to get strings from    
       windows.                                                         

   (5) SDIEasyOpenIcon expects an \fIicon string\fP; an icon string is
       a null-terminated string with an integer specifying the X dimension,
       the Y dimension, and the data in left-to-right top-to-bottom
       raster-scanned format.  These three fields are separated by commas.
       The tab character in the data ("\\t") specifies a transparent space
       with respect to the background.
                                                                        
    These routines are designed to be quite robust against bogus pa-    
    rameters.  There are, in all, ten easy-to-use routines which        
    completely control the desktop user interface:                      
                                                                        
	HOUSEKEEPING CONTROL:
		SDIEasyReady		First call; draws desktop
		SDIEasyGo		Takes control until Quit event

	WINDOW MANIPUATION:
		SDIEasyOpenWindow	Create new window on desktop
		SDIEasyCloseWindow	Remove window from desktop
		SDIEasyInput			Get a string from window
		SDIEasyOutput		Print string to window

	MENU MANIPULATION:
		SDIEasyOpenMenu		Create new menu/menu item
		SDIEasyCloseMenu		Remove menu/menu item

	ICON MANIPULATION:
		SDIEasyOpenIcon		Create new icon on desktop
		SDIEasyCloseIcon		Remove icon from desktop
                                                                        
*****************************************************************************  


SDIEasyReady ()

   This routine initializes everything.  If it fails, it terminates program
   execution.  It sets up signal-catching functions so interrupts, etc.,
   will hopefully be handled cleanly.  This routine draws a clean desktop,
   and installs a standard menu with desk accessories and a "Quit" option.
   It must be the first call made, or else bad things will happen.  WARNING:
   Desk Accessory titles must be less than ~80 characters, else crash.
   USAGE: SDIEasyReady ()   


SDIEasyGo (ConcurrentDefProc)

int	(*ConcurrentDefProc) ();      Input: User proc to run concurrently   

   This routine takes control of the calling process by handling desktop
   events.  If the user selects an object, then the procedure in the
   user's code that has been bound to it by an "SDIEasy___" routine (see
   below) is called.  If the standard "Quit" option from the standard
   menu is selected, then everything will be shut down cleanly, and this
   routine will just return (it won't terminate the calling process).  If
   a desk accessory is selected, then it is handled as well.  Every few
   times through the loop, the given user procedure is called, unless
   the NULL procedure is given.  WARNING: If the user program is to execute
   on a slow (ie, < 9600 baud) terminal, it is NOT recommended that the
   given user procedure perform desktop input or outut, as it can be
   quite a nuisance; the best use of this facility is for computation only.
   USAGE: SDIEasyGo (NULL) or SDIEasyGo (realtimeproc)   


SDIEasyOpenWindow (DefProc,Name)

int	(*DefProc) ();		   Input: Associated procedure   
char	*Name;			   Input: Window title string    

   This routine opens a window with the given name and binds the given
   procedure in the user's code to it.  All details pertaining to initial
   location on the desktop, etc., are taken care of.  The given name need
   not be unique, but the given procedure must be.  In other words, each
   window must have its own procedure in the caller's code.  Like the
   window input/output routines (below), the cursor is hidden first.
   USAGE: SDIEasyOpenWindow ("Window Title",main)   


SDIEasyCloseWindow (DefProc)

int	(*DefProc) ();		   Input: Associated procedure   

   The window that was bound to the given procedure is closed by this
   routine, with all housekeeping taken care of.  Like the window
   input/output routines (below), the cursor is hidden first.
   USAGE: SDIEasyCloseWindow (main)   


SDIEasyOpenMenu (DefProc,TitleName,ItemName)

int	(*DefProc) ();		   Input: Associated procedure   
char	*TitleName;		   Input: Menu title name        
char	*ItemName;		   Input: Menu item name         

   This routine handles the opening of a new menu and new menu items
   (The menu title is actually implemented on the screen as an icon
   with a null image string).  The given menu item associated with the
   given menu title is bound to the user's procedure.  If the given menu
   title exists, then the given item is added to that menu.  Otherwise,
   the given (ie., new) menu title is added to the menu bar.  All details
   pertaining to initial location on the desktop, etc., are taken care of.
   The given name(s) need not be unique, but the given procedure must be.
   Like the window input/output routines (below), the cursor is hidden first.
   USAGE: SDIEasyOpenMenu (main,"Menu 1","Item 1")   


SDIEasyCloseMenu (DefProc)

int	(*DefProc) ();		   Input: Associated procedure   

   This routine handles the closing of the menu item bound to the
   given procedure.  If the menu title associated with the closed
   item then has no more items in it, then it too is removed from
   the desktop.  Like the window input/output routines (below), the
   cursor is hidden first.
   USAGE: SDIEasyCloseMenu (main)   


SDIEasyOpenIcon (DefProc,Name,Image)

int	(*DefProc) ();		   Input: Associated procedure           
char	*Name;			   Input: Icon title                     
char	*Image;			   Input: IconString-format icon image   

   This routine opens an icon with the given name and binds the given
   procedure in the user's code to it.  All details pertaining to initial
   location on the desktop, etc., are taken care of.  The given name need
   not be unique, but the given procedure must be.  Like the window 
   input/output routines (below), the cursor is hidden first.
   USAGE: SDIEasyOpenIcon (main,"Icon Title","3,3,+-+|i|+-+")   


SDIEasyCloseIcon (DefProc)

int	(*DefProc) ();		   Input: Associated procedure   

   The icon that was bound to the given procedure is closed by this
   routine, with all housekeeping taken care of.  Like the window
   input/output routines (below), the cursor is hidden first.
   USAGE: SDIEasyCloseIcon (main)   


SDIEasyInput (DefProc,String,Limit)

int	(*DefProc);		   Input:  Associated procedure   
char	*String;		   Output: String typed by user   
int	Limit;			   Input:  Maximum # characters   

   This routine gets a string of input from the window bound to the given
   procedure.  The routine will return unsuccessfully if the given pro-
   cedure does not belong to a window.  If the window is not frontmost,
   then it is first made frontmost before input occurs.  The "Limit" para-
   meter specifies the maximum number of characters that may be entered;
   this will usually be the length of the caller's string character-array.
   The cursor is hidden before input occurs, and is kept hidden until the
   cursor is tracked again (usually immediately).
   USAGE: SDIEasyInput (main,MyString,80)   


SDIEasyOutput (DefProc,String)

int	(*DefProc) ();		   Input: Associated procedure   
char	*String;		   Input: String to print        

   This routine prints the given string to the window bound to the given
   procedure.  The routine will return unsuccessfully if the given pro-
   cedure does not belong to a window.  If the window is not frontmost,
   then it is first made frontmost before output occurs.  The cursor is
   hidden before output occurs, and is kept hidden until the cursor is
   tracked again (usually immediately).
   USAGE: SDIEasyOutput (main,"Hello, World!!\\n\\n\\n")   
.fi



.TL
TECHNICAL DESCRIPTION: DATA STRUCTURES
.PP
We now prepare for a more technical look at UniDesk.  A definition of the basic
data structures is in order.
.PP
The screen contents are maintained by a parallel array whose contents always
match the screen contents between UniDesk calls.  This array is a global
variable.  There are many other global variables to keep track of such things
as the current screen size, file descriptors, etc.  All globals are declared
as \fBexternal\fP in the file "unidesk.h".
.PP
Each object is represented by an \fIObjectRecord\fP.  When referring to
objects between ADD calls, the ObjectRecord is always used.  Each ObjectRecord
contains information fields describing object type, size, location, visibility
status, cursor location, pointers to user-supplied data such as DefProcs and
object titles, etc.  All ObjectRecords are kept in a singly linked list.
For the purposes of menus, menu items are represented by \fIItemRecord\fPs,
which are similar to object records but contain less information and have
their list-head in the ObjectRecord of the menu to which they belong.


.TL
TECHNICAL DESCRIPTION: FLOW OF CONTROL
.PP
The internal execution flow of control in UniDesk is logically divided
between the six source files, which have the following major features
and provide ever increasing levels of abstraction:

.nf
	\(bu"sdi.hi.c"  :  Implements easy-to-use DefProc'ed interface
			   (User program-level abstraction
	\(bu"sdi.lo.c"  :  Cursor tracking, text input, positioning, menu bar
			   (Event-level abstraction)
	\(bu"add.hi.c"  :  Initialization, Object-independent open/close
			   (Object-level abstraction)
	\(bu"add.med.c" :  Window/Menu/Icon image drawing
			   (Image-level abstraction)
	\(bu"add.inp.c" :  Locator input device drivers, dragframe/cursor
			   (Locator-device independence)
	\(bu"add.lo.c"  :  ADT List Op, Clip Output, Scroll, Block save/restore
			   (Overlapping blocks on any terminal; ObjectRecords)
.fi

.PP
In order to understand the flow of control between the different routines,
we will follow an example:  Let's assume that the user just called
SDIEasyOpenMenu with the proper parameters.  What happens during this call
is typical of the way the hierarchy works.
.PP
SDIEasyOpenMenu first checks to see if its given DefProc hasn't been used
yet.  It also checks to see if there is a menu title with the given name.
When it is ready to actually show the menu, it calls ADDOpenObject, a
high-level ADD routine, to create
a new ObjectRecord.  ADDOpenObject, in turn, calls ADDListInsertObject, a
low-level ADD routine.  Then, SDIEasyOpenMenu, a high-level SDI routine,
calls SDIOpenShowMenuTitle, a low-level SDI routine, which draws a menu
title without the body on the screen by actually drawing a null icon.
So it in turn calls ADDOpenObject again, and then calls ADDShowObject, which
calls the low-level ADDScreenBlockSave routine to save the screen contents,
and then calls the medium-level ADDDrawNewIcon to draw the stuff on the
screen.  Of course, ADDDrawNewIcon calls low-level routines such as
ADDScreenOutput and ADDFlushChar.  This kind of delegation of duties is
quite common; most SDI calls go two to four levels deep.
.PP
The boxes and model procedure headers below go into more intricate detail;
they appear in top-down fashion, with the higher-level routines listed first.



.nf
***************** EASY DESKTOP CALLS SUPPORT ROUTINES *******************
   These routines are used by the easy-to-use desktop interface rou-    
   tines below.  They are for internal use only.                        
*****************************************************************************  


SDI_shutdown_ ()

   This routine is for internal use only.  It is an emergency shutdown
   which is called when a signal is caught, etc.   


SDI_daselected_ (String)

char	*String;

   This routine is for internal use only.  It handles the execution of a
   desk accessory.  WARNING: The stdin handling is somewhat of a kludge;
   it works for now (ADD V1.0) but may require changing later.   


SDI_quitselected_ ()

   This routine is for internal use only.  It is a dummy placeholder
   routine for identifying when the standard quit option is selected.   


SDI_stopjob_ ()

   This procedure is called when a stopjob signal has been issued under
   BSD unix. It sets the keyboard back to normal and then sends the stopjob
   signal to ourselves.  Again, this works only with ADD V1.0.   


SDI_contjob_ ()

   This procedure is called when a continuejob signal has been issued under
   BSD unix.  It restarts our keyboard handler and resets our signal
   handling vector.  Again, this works only with ADD V1.0.   


******************** SDI LO-LEVEL DESKTOP ROUTINES **********************
   These routines are intended for internal SDI use only.  They im-   
   plement some basic Apple Macintosh-like behavior out of the ADD    
   primitives, such as the implementation of a "menu bar", stacked    
   windows, etc.  These routines are used extensively by the high-    
   level SDI routines.                                                
****************************************************************************  


SDIWindowInput (Object,String,Limit)

ADDObjectRecord	**Object;	   Input:  ObjectRecordPtr of window    
char		*String;	   Output: Complete string as entered   
int		Limit;		   Input:  Maximum number of charctrs   

   This routine represents a first attempt to get input from inside a
   window.  Unfortunately, it blocks until a carriage return is recieved.
   WARNING: It is assumed that the tty driver is currently in CBREAK mode
   with ~ECHO, as it will be if the locator startup routine was called.
   The carriage return is not included within the output string.
   USAGE: SDIWindowInput (&FrontMostWindow,MyStrPtr)   


SDISetNewObjectPosition (Object)

ADDObjectRecord	**Object;

   This routine finds the "best" place to put the presumably just-created-
   but-still-not-visible object on the desktop.  It puts different objects
   at different places, and modifies the X1Y1X2Y2 fields of the object
   directly.  WARNING: The numerical constants hardcoded in this routine are
   rather subjective.  The current policy is to stack windows from near
   the UL corner, to build menus from UL corner to the right, and to
   build icons near the LL corner to the right and then up.  The assigned
   size for windows, for example, is 1/2 of screen size; they are
   spaced apart by 4 positions across and 2 positions down.
   USAGE: SDISetNewObjectPosition (&ObjectJustCreatedButStillHidden)   


SDIOpenShowMenuTitle (Object)

ADDObjectRecord	**Object;	   Input: Should be just-created menu   

   This routine opens and makes visible an icon with a null image
   string at the position of the menu title, as if adding a new
   menu to the "menu bar" on top of the screen.   


SDICloseHideMenuTitle (Object)

ADDObjectRecord	**Object;	   Input: Should be just-created menu   

   This routine hides and closes the icon with a null image string
   which was created by ADDOpenInstallMenuTitle, above.   


*********************** CURSOR TRACKING ROUTINES ************************
   These routines track the current locator device with their own      
   respective cursors, sometimes dragging with them the "drag frame"   
   to provide feedback on where the new position will be.  Most of     
   these routines block until something happens.  Most of these rou-   
   tines will return new coordinates with which to call ADD hi-level   
   movement and resizing routines.                                     
*****************************************************************************  


SDITrackCursor (UpDown)

   This routine tracks a standard crosshair cursor around the desktop,
   until the button goes up or down.  This routine is nonblocking; it returns
   immediately.  This routine will aways return with the cursor showing,
   unless the locator button went up or down.  However, it is not necessary
   to have the cursor showing before calling it for the first time.  This
   routine need not be the exclusive caller of ADDLocatorUpdate; that is
   why this routine maintains its own static variables for last position.
   However, it must have exclusive access to ADD/Show/Hide/Cursor.
   The parameter should be 1 if waiting for the button to go down; it
   should be 0 if waiting for the button to come up.  If the parameter is
   -1, then the only effect of calling this routine will be to make sure
   that the cursor is removed from the screen.  This should be done
   before outputting text to the frontmost window, for example.
   USAGE: if (SDITrackCursor (1)) { button went down... }   


SDITrackObjectMove (Object,NewX1,NewY1)

ADDObjectRecord	**Object;	   Input:  Object to move      
int		*NewX1;		   Output: New X1 coordinate   
int		*NewY1;		   Output: New Y1 coordinate   

   This routine tracks the movement of any object around the desktop,
   using a special cursor and the drag frame.  It blocks until the user has
   selected a new size, and then returns new X1 and Y1 coordinates, ready
   to be handed to ADDMoveObject.  The object must be visible and frontmost,
   and the locator position should be close to or on top of the object.
   USAGE: if (move_event) button = SDITrackObjectMove (&ObjectToMove,&nx,&ny)
          ADDMoveObject (&ObjectToMove,nx,ny)   


SDITrackWindowResize (Object,NewX2,NewY2)

ADDObjectRecord	**Object;	   Input:  Window to resize    
int		*NewX2;		   Output: New X2 coordinate   
int		*NewY2;		   Output: New Y2 coordinate   

   This routine tracks the resizing of a window.  It blocks until the
   resize is done, returning new X2 and Y2 coordinates ready to be
   handed to ADDWindowResize.  It uses a special cursor icon.  The
   given window must be frontmost and visible. 
   USAGE: if (resize_event) button = SDITrackWindowResize (&ObjWind,&nx,&ny)
          ADDWindowResize (&ObjWind,nx,ny)   


int	(*SDITrackMenuSelect (Object,OptionalName)) ()

ADDObjectRecord	**Object;	   Input:  Menu (title) that was selected   
char		**OptionalName;	   Output: Name string of item selected     

   This routine tracks the menu item selection procedure, blocking,
   drawing the menu body underneath the menu title, using a special
   cursor, and undrawing the menu body before returning.  During the
   selection process the menu item underneath the cursor at any time
   is highlighted.  If an item is selected, it is blinked a few times
   using (un)highlighting, in order to confirm the choice.  If the object
   isn't a menu, an error is returned.  If nothing is selected,
   then NULL is returned; else the approriate DefProc is returned.
   USAGE: menu_item_defproc = SDITrackMenuSelect (&MenuObjectRecord)   


int	(*SDIDesktopManager (OptionalName)) ()

char	**OptionalName;		   Optional output: Name of selection   

   This routine is usually nonblocking, normally returning immediately
   with NULL.  It should be called very frequently; the responsiveness of
   the user interface depends on it.  This routine tracks the cursor, and
   blocks while calling the tracking routines above to handle any movement
   or resizing of objects.  Thus, all aspects of (non- opening/closing)
   desktop management is taken care of.  If a selection event occurred, then
   the selected object/item's DefProc is returned instead of NULL.
   If this routine returns with a DefProc, then the cursor is removed before
   exiting; else it is left on the screen.  The decisions governing the
   calling of the tracking routines above are in the following hierarchy:

	(1) If hit in X1Y1X2Y2 or NX1X2Y then bring to front and hilite.
	(2) If hit in NX1X2Y then move.
	(3) If hit in box 25% big around X2Y2 then resize.

   USAGE: if (ProcPtr = SDIDesktopManager()) { Call the user's DefProc } or
          if (SDIDesktopManager (String)) { String now contains sel. name )   


*************************** GLOBAL VARIABLES ****************************
   These global variables must be maintained throughout all ADD calls.   
   They are all declared as "external" in the file "unidesk.h".          
*****************************************************************************  


   Text message strings:   

char	ADDCreditMessage [512] = "\n\
UniDesk -- (C) 1988 Ben Koning [556498717 ben@ucscf.UCSC.EDU 408/738-1763]\n\
University of California, Santa Cruz, CA, 95064\n\
1360 Flicker Way, Sunnyvale, CA, 94087-3407\n\
ADD (Ascii-Terminal Desktop Driver)    V1.0\n\
SDI (Simplified Desktop Interface)     V1.0\n\
";
char	SDISignalMessage  [80] = "UniDesk: Abnormal Termination.";
char	SDIStartupMessage [80] = "UniDesk: Insufficient Resources.";


   Desktop frame buffer:   

char	ADDScreenBuffer [ADDMAXX][ADDMAXY];
int	ADDScreenSizeX;
int	ADDScreenSizeY;


   Global variables pertaining to desktop object record management:   

int			ADDNextLayer;	    	     Next available layer   
struct ADDObjectRecord	*ADDObjectRecordListHead;    Head of linked list    


   Global variables pertaining to cursor/dragframe drawing:   

char		ADDCursorSave [ADDMAXX][ADDMAXY];      ScreenSave   
char		ADDDragFrameSave [4][4];	       ScreenSave   


   Global variables pertaining to locator input device handling:   

int		ADDLocatorX;		   X (0..ScreenSize) after update   
int		ADDLocatorY;		   Y (0..ScreenSize) after update   
int		ADDLocatorDevice;	   Currently selected lctr device   

int		ADDLocatorFDS;		   File descriptor for input dev    
struct sgttyb	ADDLocatorIFL;		   Original ioctl flag save         


   Global variables pertaining to TERMCAP usage:   

char	 ADDTcapBuff [1024];	   This is where tgetent() stores data   
char	 ADDTcapArea [1024];	   This is where our own tgetstr strings go   
char	*ADDTcapAreaPtr;	   Pointer into area array for next tgetent()   
char	*ADDTcapClear;		   Where in area array the cls string is   
char	*ADDTcapCursorMotion;	   Where in area array curs motion string is   
char	*ADDTcapKeyUp;		   Where in area array key-up string is   
char	*ADDTcapKeyDown;	   Where in area array key-down string is   
char	*ADDTcapKeyLeft;	   Where in area array key-left string is   
char	*ADDTcapKeyRight;	   Where in area array key-right string is   


   Global file stream pointer for output terminal that is not stdout:   

FILE	*ADDTtyPtr;		   Stream pointer of our file "/dev/tty?"   


   Globals to keep track of next positions for auto object placement:   

int	SDIWindowNextX;		   Next window position   
int	SDIWindowNextY;
int	SDIMenuNextX;		   Next menu position   
int	SDIMenuNextY;
int	SDIIconNextX;		   Next icon position   
int	SDIIconNextY;


************************* HOUSEKEEPING ROUTINES *************************  
   The startup routine below should be called before any other ADD         
   routine is called; it performs important initialization tasks.  Like-   
   wise, the shutdown routine below should be called before exiting a      
   process that has been using ADD; it deallocates resources used.  The    
   redraw routine should be called when the integrity of the terminal      
   screen is uncertain; ie, if the user types control-L.                   
*****************************************************************************  


ADDStartUp ()

   Main initialization routine for ADD.  Must be called before any other
   ADD routine is called.  This routine returns the ADD version ID number,
   or -1 if error.  If an error occurs, other ADD routines should not
   be called.   


ADDRedraw ()

   This routine is mainly provided as a total-screen refresh for use in
   case the integrity of the terminal screen has been violated.  Should
   be called on a control-L keypress, or after executing a "desk accessory"
   process, etc.   


ADDShutDown ()

   This routine reallocates dynamic resources.  Must be the last ADD
   routine called.   


*************** HIGH-LEVEL OBJECT MANIPULATION ROUTINES *****************

   These calls allow the ADD-user to dynamically create new objects,    
   manipulate them at a fairly high level of abstraction by adding      
   and removing items to them, etc., and finally to delete them.        
   While calling these operations, the integrity of the desktop is      
   fully maintainted.  The usage of these routines is as follows:       
                                                                        
    *  Call ADDOpenObject to identify the object-type and provide       
       necessary information.  At this time, a new ObjectRecord will    
       be set up in the global linked list and it will be filled        
       with default style characters, etc.                              
                                                                        
    *  Optionally, modify some of the fields in the new object's        
       ObjectRecord, by dereferencing the ObjectRecord pointer          
       provided by ADDOpenObject.                                       
                                                                        
    *  Optionally, call ADDOpenItem to add items, for example, if       
       the object is a menu.                                            
                                                                        
    *  Optionally, call ADDCloseItem to remove items, for example, if   
       the object is a menu.                                            
                                                                        
    *  Call ADDShowObject to actually cause the object to appear on     
       the desktop.  It will be the new topmost object, overlaying      
       whatever was fully or partially underneath.                      
                                                                        
    *  Optionally, call ADDHideObject to restore the desktop contents   
       underneath the object and to cause it to disappear.  Since       
       its ObjectRecord (with all its customizations and/or items)      
       still exists, however, calling ADDShowObject will cause it to    
       appear again (on top of everything else, thus bringing it        
       to the top).                                                     
                                                                        
    *  Call ADDCloseObject to remove the object from the desktop and    
       from the object list.  This works correctly whether or not       
       the object is currently visible on the desktop.                  
                                                                        
    While the object is open and visible on the desktop, routines in    
    next section can be called for movement, resizing, output, etc.     
                                                                        
*****************************************************************************  


ADDOpenObject (Name,DefProc,Type,Image,X1,Y1,X2,Y2,Object)

char		*Name;		   Input:  Title string of object (for all)   
int		(*DefProc) ();	   Input:  Assoc user-procedure   (for all)   
int		Type;		   Input:  Window/Menu/Icon/etc?  (for all)   
char		*Image;		   Input:  IconString-format str (for icon)   
int		X1,Y1;		   Input:  Upper-Left boundary    (for all)   
int		X2,Y2;		   Input:  Lower-Right boundary  (for wndw)   
ADDObjectRecord	**Object;	   Output: ObjectRecord of new object         

   This routine does the internal bookkeeping for creating a new object
   on the desktop.  It adds a new ObjectRecord to the global linked list,
   and stuffs it with the given information, plus default information
   for all types of objects and specific default information based on
   the type of object being opened.  The object is set up as invisible
   (since this routine doesn't actually draw it yet).  NOTE:  Some of the
   input parameters will not make sense for all objects, such as X2Y2 for an
   icon, for example, or the DefProc, in some cases.  If so, they should
   be given dummy values or null strings when calling this routine.
   The output ObjectRecord parameter retured by this routine may be dis-
   carded by the caller, if desired; the new ObjectRecord is now in the
   linked list and can be found by the ADD- linked list search operations.
   USAGE EXAMPLES: 
	ADDOpenObject ("Window Title",main,ADDTYPEWINDOW,"",5,1,50,6,&Obj)
	ADDOpenObject ("Menu Title",main,ADDTYPEMENU,"",5,1,0,0,&Obj)
	ADDOpenObject ("Icon Title,main,ADDTYPEICON,"2,2,#%%#",5,1,0,0,&Obj)   


ADDOpenItem (Object,Name,DefProc,Item)

ADDObjectRecord	**Object;	   Input:  Object to which item is added   
char		*Name;		   Input:  Title string of new item        
int		(*DefProc) ();	   Input:  Associated user-procedure       
ADDItemRecord	**Item;		   Output: Newly created ItemRecord        

   This routine attributes a new item to the given object.  The newly
   created ItemRecord is stuffed with the given- and default-
   information.  NOTE: If this is done to an item whose object is currently
   visible on the desktop, then the effects will not be seen until
   the item is hidden and then shown again.  This routine is normally
   used to add items to objects which are menus.  As with ADDOpenObject,
   not all of the parameters must be given nondummy values.
   USAGE: ADDOpenItem (&MyMenuObjectRecord,"Item Title 1",main,&ItemRecord)   


ADDShowObject (Object)

ADDObjectRecord	**Object;	   Input: Object to display   

   This routine actually displays the given object, which must be
   currently open.  The object drawn will be the new topmost object;
   hence a new layer number is assigned here.  This routine can be 
   used freely in conjunction with ADDHideObject, and can be called
   safely on *any* open object.   


ADDHideObject (Object)

ADDObjectRecord	**Object;	   Input: Object to undisplay   

   This routine actually un-displays the given object, restoring
   what was underneath the object when ADDShowObject was called on
   it.  This routine can be used freely in conjunction with
   ADDShowObject, can be called safely on *any* open object, and
   even works correctly if the given object is not frontmost.   


ADDCloseItem (Object,Item)

ADDObjectRecord	**Object;	   Input: Object from which to remove item   
ADDItemRecord	**Item;		   Input: Item to close                      

   This routine removes the given item from the given object.  NOTE: If this
   is done to an item whose object is currently visible on the desktop, then
   the effects will not be seen until the item is hidden and then shown again.
   This routine is normally used to remove items from objects which are menus.
   USAGE: ADDCloseItem (&MyMenuObjectRecord,&IllFatedItemRecord)   


ADDCloseObject (Object)

ADDObjectRecord	**Object;	   Input: Object to close   

   This does the internal bookkeeping for permanently removing an object
   from the desktop.  This routine may be called with the given object
   either as visible or hidden on the desktop; if it is visible, then it
   is hidden before ObjectRecord removal.   


*************** (VISIBLE)-OBJECT OUTPUT/ACTION ROUTINES *****************
   These routines perform vital miscellaneous and ergonomic functions   
   such as object movement, window resizing, and "rubber-banding".      
   A text-to-window output routine is provided, as well as some         
   global object manipulation calls.  Some of these routines should     
   be called only on objects that are both visible and frontmost.       
*****************************************************************************  


ADDBringObjectToFront (Object)

ADDObjectRecord	**Object;	   Input: Object to bring to front   

   This routine brings the given object to the front of the desktop,
   so that it can be manipulated by the routines below.  This routine
   should be called if the user selects an object, for example.   


ADDMoveObject (Object,NewX1,NewY1)

ADDObjectRecord	**Object;	   Input: Object to move to new location   
int		NewX1,NewY1;	   Input: New origin coordinates           

   This routine moves the given object to a new location on the desktop,
   provided that the object is currently frontmost and visible and that
   the object will fit in its new location.
   USAGE: ADDMoveObject (&ObjectRecord,15,6)   


ADDCloseAllObjects ()

   This routine simply closes every single object on the desktop.
   This routine should be used to start fresh without having to
   shut the ADD package down first, or, alternatively, before calling
   the shutdown routine if one desires to *watch* everything close.   


ADDWindowOutput (Object,String)

ADDObjectRecord	**Object;	   Input: Window to clip output to   
char		*String;	   Input: Output character string    

   This routine should be used for all text output to a window.  The given
   object should be a window that is currently topmost and visible.
   The given text string is clipped to within the X1Y1X2Y2 bounds given
   in the ObjectRecord, minus one character in all directions.  This is 
   done to preserve a window border.  The CXCY fields in the ObjectRecord
   are used and updated to maintain the current relative character positions
   within the window, and are expected to contain sane values upon entry.
   When the character position is at the bottom right limit and this routine 
   is called with a non-null string, the window contents are scrolled 
   up as necessary.  If the newline character ('\\n') is encountered, then
   the following characters will be forced to the next line in the window.
   Nonprintable (control) characters are currently output as spaces.
   This routine also updates the object's TextContent field.
   USAGE: ADDWindowOutput (&ObjectRecordOfWindow,"TextTextText")   


ADDWindowResize (Object,NewX2,NewY2)

ADDObjectRecord **Object;	   Input: Window object to resize    
int		NewX2,NewY2;	   Input: New coordinates for X2Y2   

   This routine allows the resizing of a window, provided it is visible
   and frontmost, by the specification of new X2Y2 coordinates.  These
   coordinates must be right/below the current X1Y1 coordinates, and must
   not go off the desktop.  NOTE: The window contents, which are saved in
   the object's TextContent field will be redrawn in rasterscanned format
   but will look funny due to the new aspect ratio; if the new window's 
   area is smaller, then some of the contents will be lost.
   USAGE: ADDWindowResize (&ObjRecOfWindow_0_0_55_15_Coords,56,16)   


******************* GENERALIZED LOCATOR INPUT ROUTINES ******************  
                                                                           
   These routines provide an easy-to-use, device-independent means of      
   gathering analog positional input from the user.  They exploit spec-    
   ialized hardware such as a mouse or a bitpad/tablet when available,     
   but can also automatically revert to a simple keyboard interface,       
   which is able to take advantage of special moving keys if available.    
   The globals ADDLocatorDevice, ADDLocatorX, and ADDLocatorY are set by   
   these routines.                                                         
                                                                           
   In order to support a NEW input device, do the following:  (1) Add      
   a new #define'd ID number in the file unidesk.h.  (2) Add a new case    
   for the device in ADDLocatorStartUp.  (3) Add a new case for the de-    
   vice in ADDLocatorShutDown.  (4) Add a new case in ADDLocatorUpdate.    
   (5) Add a new startup routine for the device.  (6) Add a new shutdown   
   routine for the device.  (7) Add a new status/coordinate update rou-    
   tine.  The latter three should be placed in the section immediately     
   following the next three routines.                                      
                                                                           
**************************************************************************  
**

ADDLocatorStartUp (DeviceRequest)

   Explicit Input Parameter:	Locator-ID of dev desired or 0 for autoscan   
   Implicit Output Parameter:	Global ADDLocatorDevice updated if success    
   Return Value:		-1 Error if locator could not be found        

   This routine attempts to initialize the locator user-input device whose
   ID number is given.  If successful, then the cursor coordinate globals
   ADDLocatorX and ADDLocatorY will be reset to somewhere on the desktop,
   and the global ADDLocatorDevice set to the ID of the new device.  If
   "0" is given for the device ID, then this routine will attempt to
   initialize devices with successively higher ID's until successful (thus
   if the system does not have a mouse, it will try to initialize for a
   bitpad/tablet, then for a keyboard, and so on).
   If the routine cannot find a default or requested input device, then it
   will return with an error, leaving the global ADDLocatorDevice, as
   well as the globals ADDLocatorX and ADDLocatorY, unchanged.
   This routine is called by ADDStartUp with a "0" parameter, so it should
   be called afterward only to select a nondefault input device.
   Warning: This routine may cause a few garbage characters to be output
   while initializing some devices; refresh the screen after calling.  Also,
   it is not advisable to startup several input devices without shutting
   them down first, although this is not prevented.
   USAGE EXAMPLES: ADDLocatorStartUp (0)
                   ADDLocatorStartUp (ADDLOCATORTABLET)   


ADDLocatorShutDown ()

   This routine attempts to shut down the current locator-input device,
   deallocating any resources used or returning them to a sane state
   (as is the case with the keyboard).  This routine should be called
   before attempting to start up another input device.  This routine
   is called by ADDShutDown.
   USAGE:  ADDLocatorShutDown ()   


ADDLocatorUpdate ()

   Implicit Input Parameter:	ADDLocatorDevice global for selecting device   
   Implicit Output Parameter:	ADDLocatorX global variable is set to nw pos   
   Implicit Output Parameter:	ADDLocatorY global variable is set to nw pos   
   Return Value (Output): 	Integer	containing the bits of the button(s)   

   This routine reads the locator device given by the ADDLocatorDevice
   global variable.  The globals ADDLocatorX and ADDLocatorY are updated
   with new absolute positions on the desktop scaled to the range
   0..ADDScreenSizeX and 0..ADDScreenSizeY, respectively.  The return
   value contains the state(s) of any buttons(s) on the locator in
   positive-logic bitwise format; hence the (usual) return value of zero
   indicates no button-push event occurred, and nonzero indicates otherwise.
   This routine returns immediately; it must be called repeatedly for
   cursor tracking, etc.
   USAGE: if (ADDLocatorUpdate ()) { Button down; check coordinates... }   


*************** LOW-LEVEL LOCATOR INPUT DEVICE DRIVERS: *****************
                                                                            
   These routines are for private use only, and are called by the three     
   locator-input interface routines above.  This is where the "startup",    
   "shutdown", and "update" device driver routines for a new locator de-    
   vice should go.  Each "startup" routine MUST return -1 on error, and     
   set any globals its "update" routine may need later on, such as          
   ADDLocatorFDS.  Each "update" routine MUST return -1 on error, and if    
   successful it MUST update ADDLocatorX and ADDLocatorY, scaling them to   
   the current screen dimensions, and it MUST return the status of its      
   button(s).  However, it should not make sure that the new coordinates    
   be within screen range; that is done elsewhere (it is done this way      
   for possible future reasons).  The "shutdown" routine should deallo-     
   cate/return to normal any resources allocated/altered while starting     
   up the device.                                                           
                                                                            
   Some of these device drivers are based on those written by Jim Murphy;   
   some of them were written by the author; all of them exhibit awful       
   coding style.  But they work :-).                                        
                                                                            
*****************************************************************************  


ADDLocator_mousestartup_ ()

   Device: Logitech serial PC-Mouse   


ADDLocator_mouseshutdown_ ()

   Device: Logitech serial PC-Mouse   


ADDLocator_mouseupdate_ ()

   Device: Logitech serial PC-Mouse   


ADDLocator_bitpadstartup_ ()

   Device: Summagraphics Bitpad One   


ADDLocator_bitpadshutdown_ ()

   Device: Summagraphics Bitpad One   


ADDLocator_bitpadupdate_ ()

   Device: Summagraphics Bitpad One   


ADDLocator_tabletstartup_ ()

   Device: GTCO DigiPad 5   


ADDLocator_tabletshutdown_ ()

   Device: GTCO DigiPad 5   


ADDLocator_tabletupdate_ ()

   Device: GTCO DigiPad 5   


ADDLocator_keyboardstartup_ ()

   Device: Stdin keyboard, using termcap   


ADDLocator_keyboardshutdown_ ()

   Device: Stdin keyboard, using termcap   


ADDLocator_keyboardupdate_ ()

   Device: Stdin keyboard, using termcap   


************* INPUT-EVENT - RELATED HILEVEL DRAWING ROUTINES ************  
   The following routines draw things on or about objects, to be used in   
   response to user input events in order to provide ergonomic feedback.   
   Some of these should only be done to frontmost objects; for example,    
   highlighting simply involves drawing characters to either side of the   
   "Name" (title) string given in the appropriate Item- or Object-         
   Record (This is why, by convention, all title strings are bordered by   
   spaces in the medium-level drawing routines).  WARNING: Although it     
   is safe to call these routines at any time, they should not be inter-   
   spersed between other hi-level calls.  This is because all of these     
   operations act on top of everything else without regard for the list    
   of objects.  Thus, if the cursor is currently on the screen, it         
   should be hidden before resizing a window underneath it, for example.   
*****************************************************************************  


ADDShowObjectHilite (Object)

ADDObjectRecord **Object;	   This is where we get necessary info   

   This routine uses information in the given ObjectRecord to change the
   highlighting status of an active image on the screen.  NOTE: It
   presently does not make any sense to (un)highlight a menu title. 
   USAGE: ADDShowObjectHilite (&ObjectRecord)   


ADDHideObjectHilite (Object)

ADDObjectRecord **Object;	   This is where we get necessary info   

   This routine uses information in the given ObjectRecord to change the
   highlighting status of an active image on the screen.  NOTE: It
   presently does not make any sense to (un)highlight a menu title. 
   USAGE: ADDHideObjectHilite (&ObjectRecord)   


ADDShowItemHilite (Item)

ADDItemRecord	**Item;		   This is where we get our info   

   This routine uses information in the given ItemRecord to change the
   highlighting status of an active item line on the screen.
   USAGE: ADDShowItemHilite (&ItemRecord)   


ADDHideItemHilite (Item)

ADDItemRecord	**Item;		   This is where we get our info   

   This routine uses information in the given ItemRecord to change the
   highlighting status of an active item line on the screen.
   USAGE: ADDHideItemHilite (&ItemRecord)   


ADDShowDragFrame (X1,Y1,X2,Y2)

   This routine draws a "drag frame" with the given coordinates on
   top of everything else.  Drag frames should be used to provide
   ergonomic feedback while moving or resizing objects.  The
   drag frame should be hidden before doing any other desktop 
   operations.  This routine does not do any clipping whatsoever.
   USAGE: ADDShowDragFrame (3,3,30,30)   


ADDHideDragFrame (X1,Y1,X2,Y2)

   This routine hides the drag frame drawn by an earlier call to
   ADDShowDragFrame.  This routine should be given the same parameters
   as were given to the ADDShowDragFrame call.
   USAGE: ADDHideDragFrame (3,3,30,30)   


ADDShowCursor (CursorImage,X,Y)

char	*CursorImage;	   Input: IconString that is the cursor to be drawn    
int 	X,Y;		   Input: Coordinates where center of cursor will be   

   This routine displays the given cursor image (which should be in
   IconString format) and draws it on top of everything else.  The
   cursor should be hidden before doing other desktop activity.
   The cursor is drawn such that the given X,Y coordinates will
   coincide with the *center* of the cursor image, clipping off the
   screen if necessary.  After drawing the cursor image, the terminal
   cursor is positioned to its center.
   USAGE: ADDShowCursor (0,0,ADDCURSORICONPOINT)   


ADDHideCursor (CursorImage,X,Y)

char	*CursorImage;	   Input: IconString that is the cursor to be erased   
int 	X,Y;		   Input: Coordinates where center of cursor is        

   This routine hides the cursor that was drawn by an earlier call
   to ADDShowCursor.  This routine must be given the same parameters
   as were given to the ADDShowCursor call.
   USAGE: ADDHideCursor (0,0,ADDCURSORICONPOINT)   


******************* MEDIUM-LEVEL SCREEN IMAGE DRAWING *******************
   These routines use the low-level ADD routines to draw icons, custom   
   text strings, etc., to the screen.  These are in turn used by the     
   other ADD calls in this file as they draw actual desktop objects to   
   the screen with their inherent style.                                 
*****************************************************************************  


ADDDrawCustomStrings (String,X1,Y1,X2,Y2)

char *String;			   Encoded CustomStrings   
int  X1,Y1,X2,Y2;		   Bounding rectangle      

   This routine draws the given CustomStrings-format text on the
   screen/buffer.  The format is assumed to be: Xrelative, Comma, 
   Yrelative, Comma, Text; Null to terminate or Comma followed 
   by another such sequence.  Coordinates in the CustomStrings string 
   must be in the range 0-100 to indicate a percentage between the
   upper left and the bottom right corners of the given bounding
   rectangle for the origin of each individual string.  In order to
   place a comma within a text sequence, use two commas in a row.
   CustomStrings are used to optionally customize desktop objects;
   for example, to add "scroll buttons" to windows, etc.
   USAGE: ADDDrawCustomStrings ("0,0,Hi,, there!,50,50,Center",0,0,19,19)   


ADDDrawIconToBufferOnly (String,X,Y,DimensionX,DimensionY,Inhibit)

char *String;			   Input:  Encoded icon string         
int  X,Y;			   Input:  Where on screen to put it   
int  *DimensionX,*DimensionY;	   Output: Dimensions of icon image    
int  Inhibit;			   Input:  1=Don't even draw buffer    

   This routine is used by several routines, some of which need to draw
   many icons to the buffer before updating the screen.  This routine
   does the most of the actual work for ADDDrawIcon.  If the Inhibit
   parameter is set, then only the size calculations are done and the
   screen buffer is left untouched.
   USAGE: ADDDrawIconToBufferOnly (str,x,y,&sizeX,&sizeY,0)   


ADDDrawIcon (String,X,Y)

char *String;
int  X,Y;

   This routine draws the given icon at the given coordinates, clipping to
   the screen boundaries.  The format of an icon-string is as follows:
   X and Y dimensions (in counting numbers; a 2x4 image has dimensions 2 and
   4, not 1 and 3), followed by the characters of the string in left-to-
   right, top-to-bottom raster-scanned format.  These are separated by
   commas.  Commas may appear in the icon-data string without an escape
   character.  The tab character ('\\t') denotes a transparent pixel.
   USAGE: ADDDrawIcon ("5,3,+===+|,,,|+===+",x,y)    


ADDDrawDesktop (String)

char *String;

   This routine draws a repeating pattern on the screen, by filling it
   with a repeating mosaic of icons decoded from the given icon string. 
   USAGE: ADDDrawDesktop ("2,2,:  :")   


************************* OBJECT DRAWING ROUTINES ***********************

   These routines actually draw the images of desktop objects to the       
   screen for when they are presented for the first time.  They use some   
   of the medium-level routines (above) as well as some of the lower-      
   level routines.  It is in these calls that the standard UniDesk         
   "look" is defined (other than in the modifiable characters in each      
   ObjectRecord).  These calls are used directly by the high-level ADD     
   routines which "open", "close", "move", etc., objects.                  
   All of the calls in this section use information in the given           
   ObjectRecords.  The horizontal, vertical, and corner characters in      
   the ObjectRecord are used, as are its bounding rectangle coordinates.   
   The inside area of the object is cleared to spaces, if appropriate.     
   For highlighting, all Name-strings are flanked by spaces.               
   These routines call ADDScreenBlockSave to save the screen area under-   
   neath before drawing.  In most cases, some of the given object's        
   coordinate data fields in the Object or Item -Records are set to        
   useful values (see especially ADDDrawNewMenu).  For example, the        
   name position fields (NX1,NX2,NY) and lower-right corner position       
   fields (X2,Y2) are set, because they are calculated by these calls.     
   Note, however, that the custom-strings are not drawn on the objects     
   by these routines; nor are layer numbers or object-types assigned.      
   This is done by the high-level routines.

*****************************************************************************  


ADDDrawNewWindow (Object)

ADDObjectRecord **Object;	   This is where we get necessary info   

   This routine uses information in the given ObjectRecord to draw a new
   image on the screen.  Please see the comment header for this section.
   The name position fields (NX1,NX2,NY) ARE SET by this routine.  The
   TextContent field in the given object record is drawn to the window
   and is thus assumed to contain sane values.  The screen area over-
   written by this routine is saved first.
   USAGE: ADDDrawNewWindow (&ObjectRecord)   


ADDDrawNewMenu (Object)

ADDObjectRecord **Object;	   This is where we get necessary info   

   This routine uses information in the given ObjectRecord to draw a new
   image on the screen.  Please see the comment header for this section.
   This routine creates a menu with all items in the ItemList, if any,
   including the menu title at top.  It reads the X1Y1 fields of the
   ObjectRecord to determine the origin, AND IT WRITES to the X2Y2 fields
   the lower-right corner coordinates of the menu box, since its size is
   usually unknown until calling this procedure.  IT ALSO WRITES to the
   X1X2Y fields in each ItemRecord, since the coordinates of the individual
   item strings is also unknown before calling this routine.  This routine
   also sets the NX1,NX2,Y fields.  The screen area overwritten by this
   routine is saved by it first.
   USAGE: ADDDrawNewMenu (&ObjectRecord)   


ADDDrawNewIcon (Object)

ADDObjectRecord **Object;	   This is where we get necessary info   

   This routine uses information in the given ObjectRecord to draw a new
   icon image on the screen.  Please see the comment header for this section.
   The IconString field in the given ObjectRecord is used and must be set up.
   The title is centered and drawn underneath the icon image, with one
   blank space to either side to allow highlighting.  This routine WRITES
   to the X2Y2 fields of the given ObjectRecord, setting them to the
   lower-right corner coordinates of the image.  It also WRITES to the
   NX1,NX2,NY fields, setting them to the bounds of the title string.
   The screen area overwritten by this routine is saved by it first.  If an
   empty string or NULL is passed as the icon image, then no icon is drawn
   and the title is put at the X1+1,Y1+2 coordinates.  This is useful
   for implementing a menu bar.
   USAGE: ADDDrawNewIcon (&ObjectRecord)   


********************** LOWEST-LEVEL OUTPUT ROUTINES: ********************
   These routines output text to the screen in a cartesian coordinate       
   environment, while always maintaining or using the global character      
   buffer, which is at all times identical to screen contents.  Only        
   these routines should be used for screen manipulation.                   
*****************************************************************************  


ADDPutChar (c)

char c;

   This routine does nothing more than make the putchar macro into a call. 
   This makes it easier to use the termcap "tput" call, and since all 
   low-level ADD calls use it, all ADD output goes here.

   WARNING:  Characters are output only when this routine is called many
   times or if the output buffer is flushed.  See ADDFlushChar, below.
   It is done this way to optimize performance.   


ADDFlushChar ()

   This routine should be called after performing a few calls to ADDPutChar,
   to insure that enqueued characters are output.  Typically, this routine
   should be called after loops containing ADDPutChar calls or just before
   exitting a routine which produces output.   


ADDScreenClear ()

   This routine quickly clears the entire screen and frame buffer to 
   all SPACES.  Should be used to initialize the frame buffer at first.   


ADDScreenRefresh (X1,Y1,X2,Y2)

   This routine redraws the screen from the frame buffer within the
   given bounding rectange, clipping if out of range.  
   USAGE:  ADDScreenRefresh (0,0,ADDScreenSizeX,ADDScreenSizeY)   


ADDScreenOutput (String,PosX,PosY)

char *String;
int  PosX,PosY;

   This routine draws the given string to buffer and screen starting with
   the first character at the given coordinate.  The string should only
   be one line; displaying multiple lines requires repetitive calling
   of this routine.  It clips to the terminal bounds, rather than wrapping 
   text around.  It is the caller's responsibility to insert numbers, etc.,
   into the string by using sprintf, for example.  Control characters are
   reduced to spaces.
   USAGE: ADDScreenOutput ("(Clipped) Hello, world!",-10,10)   


ADDScreenScrollUp (X1,Y1,X2,Y2)

   This routine takes the given region of text on the screen and scrolls
   it up one line.  The freed-up line is then cleared to spaces.  This
   routine is heavily optimized for minimum-screen-output performance:
   For every line, the contents are manipulated in memory first.  Then,
   only the lines which actually need updating are drawn to the terminal
   screen, and even those lines are right-truncated if possible.   


************* OBJECT/ITEM-RECORD LINKED LIST OPERATIONS *****************
   The following calls manipulate the global linked list of object      
   records, providing various general "abstract data type" insert,      
   delete, and search operations.                                       
*****************************************************************************  


ADDListInsertObject (Object)

ADDObjectRecord  **Object;    Output parameter   

   Allocates storage for an object record, inserts it into the list, and
   returns a pointer to the new record.  This routine should be called FIRST,
   and then the record can be filled with data by the caller. 
   USAGE: ADDListInsertObject (&var), where var is pointer to record.   


ADDListDeleteObject (Object)

ADDObjectRecord  **Object;    Input parameter   

   Removes the pointed-to object record from the linked list.  Warning:
   The object to be deleted must actually exist in the list!
   USAGE: ADDListDeleteObject (&var), where var is pointer to record.   


ADDListFindObjectTop (Object)

ADDObjectRecord  **Object; 	   Output: Topmost visible object if found   

   Searches for the TOPMOST object that is VISIBLE and returns a pointer to
   its ObjectRecord or NULL.
   USAGE: ADDListFindObjectTop (&var), where var is pointer to record.   


ADDListFindObjectByName (Object,NameString)

ADDObjectRecord	    **Object; 		   Output parameter   
char		    *NameString; 	   Input  parameter   

   Finds the FIRST object in the list with the given name or NULL. 
   USAGE: ADDListFindObjectByName (&var,str), where var is pointer to record.   


ADDListFindObjectByDefProc (Object,TheDefProc)

ADDObjectRecord	    **Object; 		   Output parameter   
int		    (*TheDefProc) (); 	   Input  parameter   

   Finds the FIRST object in the list with the given DefProc or NULL. 
   USAGE: ADDListFindObjectByDefProc (&Object,main)   


ADDListFindObjectByLoc (Object,x,y)

ADDObjectRecord     **Object; 		   Output parameter   
int		    x,y;		   Input parameters   

   Finds the TOPMOST, VISIBLE object in the list, IN WHICH the given
   coordinate point is CONTAINED; or NULL if not found.  This works
   even if the given point is inside the bounds of the object's "Name"
   field.  This routine is of great importance to event handling and
   screen area restore routines.
   USAGE: ADDListFindObjectByLoc (&ObjectIfFound,x,y)   


ADDListInsertItem (Object,Item)

ADDObjectRecord	**Object;	   Input:  Object that is to have this item   
ADDItemRecord	**Item;		   Output: The item ready to be filled data   

   This call inserts a new item in the list of items associated with the
   given object.  This routine should be called FIRST, and then the
   record can be filled with data by the caller.  Note that this routine
   inserts the new item at the END of the item list.  This is done so
   that menus, etc., turn out to look as expected.
   USAGE: ADDListInsertItem (&ExistingObjectRecord,&NewItemRecord)   


ADDListDeleteItem (Object,Item)

ADDObjectRecord	**Object;	   Input: ObjectRecord our item belongs to   
ADDItemRecord	**Item;		   Input: The ill-fated item to be deleted   

   Removes the pointed-to ItemRecord from its linked list.  Warning:  The
   item to be removed must actually exist in the list!
   USAGE: ADDListDeleteItem (&ObjectRecord,&ItemRecord)   


ADDListFindItemByName (Object,Item,NameString)

ADDObjectRecord	**Object;	   Input:  ObjectRecord with item list   
ADDItemRecord	**Item;		   Output: ItemRecord being searched     
char		*NameString;	   Input:  The name string               

   Finds the FIRST item in the list with the given name or NULL.
   USAGE: ADDListFindItemByName (&ObjectRecord,&TheItem,"SearchKey")   


ADDListFindItemByDefProc (Object,Item,TheDefProc)

ADDObjectRecord	**Object;		   Input:  ObjectRec with item list   
ADDItemRecord	**Item;			   Output: ItemRec being searched     
int		(*TheDefProc) ();	   Input:  The definition procedure   

   Finds the FIRST item in the list with the given DefProc or NULL.
   USAGE: ADDListFindItemByDefProc (&ObjectRecord,&TheItem,main)   


ADDListFindItemByLoc (Object,Item,X,Y)

ADDObjectRecord	**Object;	   Input:  The object record with item list   
ADDItemRecord	**Item;		   Output: The item record we want or NULL    
int		X,Y;		   Input:  Coordinates as search key          

   Finds the FIRST item in the list belonging to the given object 
   IN WHICH the given coordinate point is CONTAINED or NULL if not found.
   USAGE: ADDListFindItemByLoc (&ObjectRecord,&TheItem,5,5)   


************** OVERLAPPING SCREEN AREA SUPPORT ROUTINES: ****************
   These calls implement the "overlapping papers" model of the         
   "desktop", by saving or restoring patches of the screen about to    
   be, or previously obscured by, other display activity.              
*****************************************************************************  


ADDScreenBlockSave (Object)

ADDObjectRecord **Object;

   This routine saves the screen area inside the rectangle given inside the
   given ObjectRecord from the screen buffer into the buffer contained inside
   the given ObjectRecord.  Unlike ADDScreenBlockRestore, it does not pay any 
   attention to whatever layer number might be inside the ObjectRecord; so the
   object whose image is being saved should be topmost.  Usually, this routine
   will be called while drawing a new object, so that the new object's 
   ObjectRecord contains an image of the screen area it is obscuring, so that
   it may later be restored when the object is removed.  In addition to the
   given rectange, the object title line is also saved.
   USAGE:  ADDSaveScreenArea (&ObjectRecordPointer)   


ADDScreenBlockRestore (Object)

ADDObjectRecord **Object;

   This routine restores the screen area previously obscured by the
   given object after calling ADDScreenBlockSave (see description above). 
   This routine takes into account the object's layer number, so that it 
   works correctly even if the object being deleted is not currently topmost,
   and is possibly partially obscured.  Usually, this routine is called
   while either closing an object or resizing or moving an object.  In
   addition to the object contents, the screen area previously obscured
   by its title line is also restored.  WARNING: The object whose area
   underneath is being restored by this routine should currently be visible.
   And no, the author is *not* proud of the complexity of this routine.
   USAGE:  ADDRRestoreScreenArea (&ObjectRecordPointer)   


*********************** ADD DEFAULT DEFINITIONS *************************
   These define style and limitations of the ADD package.                    
***************************************************************************  


   Maximum allowable display screen size:   

#define 	ADDMAXX		132
#define		ADDMAXY		48


   Desktop object "Type" classifications defined so far:   

#define		ADDTYPEWINDOW		1
#define		ADDTYPEMENU		2
#define		ADDTYPEICON		3


   Locator device ordering; numerical order determines order of priority:   

#define		ADDLOCATORMOUSE		1
#define		ADDLOCATORBITPAD	2
#define		ADDLOCATORTABLET	3
#define		ADDLOCATORKEYBOARD	4


   Some critical characters:   

#define		ADDCOMMA	','	   Searched for in formatted strings   
#define		ADDSPACE	' '	   Drawn to clear windows, etc.        
#define		ADDKEYUP	'k'	   Nondedicated locator-up key         
#define		ADDKEYUPFAST	'K'	   Nondedicated locator-up key         
#define		ADDKEYDOWN	'j'	   Nondedicated locator-down key       
#define		ADDKEYDOWNFAST	'J'	   Nondedicated locator-down key       
#define		ADDKEYLEFT	'h'	   Nondedicated locator-left key       
#define		ADDKEYLEFTFAST	'H'	   Nondedicated locator-left key       
#define		ADDKEYRIGHT	'l'	   Nondedicated locator-right key      
#define		ADDKEYRIGHTFAST	'L'	   Nondedicated locator-right key      
#define		ADDKEYBUTTON	' '	   (Non)dedicated locator-button key   

#define		ADDTIMEOUT	350	   Locator-driver routine constant     
#define		ADDNUMFASTX	15	   For fast cursor-motion keys above   
#define		ADDNUMFASTY	5	   For fast cursor-motion keys above   


   Some "look" defaults:   

#define		ADDCHARHORIZONTAL		'-'
#define		ADDCHARVERTICAL			'|'
#define		ADDCHARCORNER			'+'

#define	     	ADDCHARHORIZONTALWINDOW		'='
#define	     	ADDCHARHORIZONTALMENU		'_'

#define		ADDCHARHILITELEFT		'>'
#define		ADDCHARHILITERIGHT		'<'

#define		ADDCHARDRAGFRAME		'#'

#define		ADDCURSORICONPOINT		"5,3,\t\t|\t\t--+--\t\t|\t\t"
#define		ADDCURSORICONSELECT		"5,3,\t\t|\t\t\t-X-\t\t\t|\t\t"
#define		ADDCURSORICONMOVE		"5,3,\t\t|\t\t<-+->\t\t|\t\t"
#define		ADDCURSORICONRESIZE		"5,3,\t\t|\t\t->+<-\t\t|\t\t"

#define		ADDDESKTOPICON 			"2,2,.  ."

#define		SDINUMBLINK			25
#define		SDIHITWINDOWRESIZEX		4
#define		SDIHITWINDOWRESIZEY		2

#define		SDITEMPPATH			"/tmp/unideskXXXXXX"
#define		SDIDAPATH			"$HOME/.unideskrc"

#define		SDISTDMENUNAME			"UniDesk"
#define		SDISTDQUITNAME			"Quit"

#define		SDIEASYGOLOOP			12


********************* ADD DATA STRUCTURE DEFINITIONS ********************
                                                                               
   The following data structures are used by ADD routines to communicate       
   information about desktop objects.  Most of the data fields declared        
   below are self-explanatory; however, some of the strings are understood     
   to have certain implicit formats.  Some of the implicit types of data       
   structures used by ADD are:                                                 
                                                                               
  	* IconString -- A string representing a two-dimensional picture made   
  			of characters.  Used by icon, cursor, background,      
  			etc., drawing routines.  Format: string, containing    
  			Xdimension, Comma, Ydimension, Comma, Data, Null.      
  			The data is in left-to-right, top-to-bottom raster-    
  			scanned format and may contain commas.  Dimensions     
  			are in numbers counted starting at one, not zero.      
                        Commas may freely appear in the data text.  If the     
                        tab character ('\\t') appears anywhere within the       
                        string, then that position will be transparent with    
                        respect to what was on the desktop before the icon     
                        was drawn.                                             
                                                                               
  	* CustomStrings -- A string used to place user-defined text in         
                           the vicinity of a desktop object, using relative    
                           coordinates in the range 0-100 which indicate a     
                           percentage displacement between the upper left      
                           lower right corners of object.  Format: string:     
                           Xcoordinate, Comma, Ycoordinate, Comma, Text;       
                           followed either by Null or another comma and        
                           another such x,y,text sequence.  Commas are         
                           allowed within text sequences like this: ",,".      
                                                                               
***************************************************************************  


typedef struct ADDObjectRecord
	{
		int	Dummy;

	For all objects:   
		char	*Name;				Title of object     
		int	(*DefProc) ();			Assoc user proc     
		int 	Hilite;				Highlighted?        

	For specific objects:   
		int 	Type;				Wnd/Menu/Icon ?!!   
		int	CX,CY;				Rel cur pos. wndw   
		char	TextContent[(ADDMAXX*ADDMAXY)];	Text in window      
		struct	ADDItemRecord	*ItemListHead;	Item list in menu   
		struct	ADDObjectRecord	*MenuTitleIcon;	Menu bar entry      
		char	*Image;				Image str in icon   

	For maintaining desktop integrity:   
		int	Layer;				For ID & overlap    
		int	Visible;			Visible?            
		int	X1,Y1,X2,Y2;			Current pos/size    
		char	ScreenSave [ADDMAXX][ADDMAXY];	Chars underneath    
		int	NX1,NX2,NY;			Pos of name strng   
		char	ScreenSaveName [ADDMAXX];	Chars underneath    

	For custom appearances:   
		char	*CustomStrings;			Customizing chars   
		char	CharHorizontal;			Horiz border char   
		char	CharVertical;			Verti border char   
		char	CharCorner;			Cornr border char   

		struct	ADDObjectRecord	*Link;		Next ObjectRecord   
	} ADDObjectRecord;


typedef struct ADDItemRecord
	{
		int	X1,X2,Y;			Current pos/size    

		char	*Name;				Item title          
		int	(*DefProc) ();			Assoc user proc     
		int	Hilite;				Highlighted?        

		struct	ADDItemRecord	*Link;		Next ItemRecord     
	} ADDItemRecord;
.fi




.TL
APPENDIX 1: ALPHABETICAL INDEX OF ROUTINES
.PP
.nf

	ADDBringObjectToFront (Object)
	ADDCloseAllObjects ()
	ADDCloseItem (Object,Item)
	ADDCloseObject (Object)
	ADDDrawCustomStrings (String,X1,Y1,X2,Y2)
	ADDDrawDesktop (String)
	ADDDrawIcon (String,X,Y)
	ADDDrawIconToBufferOnly (String,X,Y,DimensionX,DimensionY,Inhibit)
	ADDDrawNewIcon (Object)
	ADDDrawNewMenu (Object)
	ADDDrawNewWindow (Object)
	ADDFlushChar ()
	ADDHideCursor (CursorImage,X,Y)
	ADDHideDragFrame (X1,Y1,X2,Y2)
	ADDHideItemHilite (Item)
	ADDHideObject (Object)
	ADDHideObjectHilite (Object)
	ADDListDeleteItem (Object,Item)
	ADDListDeleteObject (Object)
	ADDListFindItemByDefProc (Object,Item,TheDefProc)
	ADDListFindItemByLoc (Object,Item,X,Y)
	ADDListFindItemByName (Object,Item,NameString)
	ADDListFindObjectByDefProc (Object,TheDefProc)
	ADDListFindObjectByLoc (Object,x,y)
	ADDListFindObjectByName (Object,NameString)
	ADDListFindObjectTop (Object)
	ADDListInsertItem (Object,Item)
	ADDListInsertObject (Object)
	ADDLocatorShutDown ()
	ADDLocatorStartUp (DeviceRequest)
	ADDLocatorUpdate ()
	ADDLocator_bitpadshutdown_ ()
	ADDLocator_bitpadstartup_ ()
	ADDLocator_bitpadupdate_ ()
	ADDLocator_keyboardshutdown_ ()
	ADDLocator_keyboardstartup_ ()
	ADDLocator_keyboardupdate_ ()
	ADDLocator_mouseshutdown_ ()
	ADDLocator_mousestartup_ ()
	ADDLocator_mouseupdate_ ()
	ADDLocator_tabletshutdown_ ()
	ADDLocator_tabletstartup_ ()
	ADDLocator_tabletupdate_ ()
	ADDMoveObject (Object,NewX1,NewY1)
	ADDOpenItem (Object,Name,DefProc,Item)
	ADDOpenObject (Name,DefProc,Type,Image,X1,Y1,X2,Y2,Object)
	ADDPutChar (c)
	ADDRedraw ()
	ADDScreenBlockRestore (Object)
	ADDScreenBlockSave (Object)
	ADDScreenClear ()
	ADDScreenOutput (String,PosX,PosY)
	ADDScreenRefresh (X1,Y1,X2,Y2)
	ADDScreenScrollUp (X1,Y1,X2,Y2)
	ADDShowCursor (CursorImage,X,Y)
	ADDShowDragFrame (X1,Y1,X2,Y2)
	ADDShowItemHilite (Item)
	ADDShowObject (Object)
	ADDShowObjectHilite (Object)
	ADDShutDown ()
	ADDStartUp ()
	ADDWindowOutput (Object,String)
	ADDWindowResize (Object,NewX2,NewY2)
	SDICloseHideMenuTitle (Object)
	(*SDIDesktopManager (OptionalName)) ()
	SDIEasyCloseIcon (DefProc)
	SDIEasyCloseMenu (DefProc)
	SDIEasyCloseWindow (DefProc)
	SDIEasyGo (ConcurrentDefProc)
	SDIEasyInput (DefProc,String,Limit)
	SDIEasyOpenIcon (DefProc,Name,Image)
	SDIEasyOpenMenu (DefProc,TitleName,ItemName)
	SDIEasyOpenWindow (DefProc,Name)
	SDIEasyOutput (DefProc,String)
	SDIEasyReady ()
	SDIOpenShowMenuTitle (Object)
	SDISetNewObjectPosition (Object)
	SDITrackCursor (UpDown)
	(*SDITrackMenuSelect (Object,OptionalName)) ()
	SDITrackObjectMove (Object,NewX1,NewY1)
	SDITrackWindowResize (Object,NewX2,NewY2)
	SDIWindowInput (Object,String,Limit)
	SDI_contjob_ ()
	SDI_daselected_ (String)
	SDI_quitselected_ ()
	SDI_shutdown_ ()
	SDI_stopjob_ ()
.fi


.TL
APPENDIX 2: THE IMPLEMENTATION OF A SHELL
.PP
In order to write a shell using UniDesk which would presumably allow
the user to have multiple processes running in different windows, two
minor improvements must first be made to the UniDesk package as it
is now.  Both modifications are related to input handling.
.PP
The first problem is that UniDesk currently gathers its keyboard input
from file descriptor 0, which is traditionally associated with the stream
file pointer \fIstdin\fP.  When forking off processes which ask for
standard input, the shell may want to close file descriptor 0 and
handle the input request itself (by setting F_ASYNC to catch the request
(?) and using SDIWindowInput to handle the meat of it).  Therefore, input
handling will have to be done in the same manner as output handling
is currently done; that is, directly to the keyboard's /dev/tty.
.PP
The second problem is that SDIWindowInput is not nonblocking.  It will
have to be modified to return immediately.  The ObjectRecord data
structure will probably have to include a few new fields to facilitate
context save/restore between different SDIWindowInput calls.
