/* CHK=0xAE13 */
/*+-------------------------------------------------------------------------
	afterlint.c -- process -Zg output

Use with zgcc:

	#!/bin/csh
	#zgcc -- build prototyping "LINT_ARGS"
	echo C sources in $1';' result file $2
	rm -f zg.t
	foreach i(`cat $1`)
	cc -DBUILDING_LINT_ARGS -Zg $argv[3-] $i > zg1.t
	sort zg1.t > zg2.t
	echo '/ *' $i '* /' >> zg.t    #remove spaces inside comment markers
	cat zg2.t >> zg.t
	end
	afterlint zg.t $2
	rm -f zg.t zg?.t
	echo Output in $2

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:23-wht@n4hgf---- 2.80 --- */
/*:03-25-1990-14:11-wht@n4hgf------ x2.70 ------- */
/*:07-13-1988-19:50-wht-creation */

#include <stdio.h>

#define ff fprintf

#ifdef BSD4
#include <sys/time.h>
#define strchr index
#define strrchr rindex
char *index();
char *rindex();
/**************************/
#else	/* assuming XENIX */
/**************************/
#include <time.h>
#include <sys/types.h>
#include <sys/timeb.h>
char *strchr();
char *strrchr();
#endif	/* system dependencies */

extern long time();
extern struct tm *localtime();

/*+-------------------------------------------------------------------------
	emit_editnote(fp)
--------------------------------------------------------------------------*/
void
emit_editnote(fp)
FILE	*fp;
{
	struct tm *ltime;
	long cur_time;
#ifdef XENIX
#ifdef MESSED_UP_DST
	struct timeb tp;
#endif
#endif

	cur_time = time((long *)0);

#ifdef XENIX
#ifdef MESSED_UP_DST
	ftime(&tp);				/* localtime() does not know about DST */
	if(tp.dstflag)			/* so we adjust with ftime, which does know */
		cur_time += 3600L;	/* problem in BSD too, but no ftime there */
#endif
#endif

	ltime = localtime(&cur_time);

	fputs("/*+:EDITS:*/\n",fp);

	ff(fp,"/*:%02d-%02d-%04d-%02d:%02d-afterlint-creation */\n",
	    ltime->tm_mon+1,ltime->tm_mday,ltime->tm_year + 1900,
	    ltime->tm_hour,ltime->tm_min);

}	/* end of emit_editnote */

/*+-------------------------------------------------------------------------
	main(argc,argv,envp)
--------------------------------------------------------------------------*/
main(argc,argv,envp)
int argc;
char	**argv;
char	**envp;
{
	register char *cptr;
	register FILE *fpin;
	register FILE *fpout;
	char	buf[256];
	char	*basename;

	if(argc < 2)
	{
		ff(stderr,"usage: afterlint <infile> [<outfile>]\n");
		ff(stderr,"if outfile not supplied, output is to stdout\n");
		exit(1);
	}

	if((fpin = fopen(argv[1],"r")) == NULL)
	{
		perror(argv[1]);
		exit(1);
	}

	if(argc > 2)
	{
		if((fpout = fopen(argv[2],"w")) == NULL)
		{
			perror(argv[2]);
			exit(1);
		}
		basename = argv[2];
	}
	else
		fpout = stdout;

	ff(fpout,
	    "/*+-----------------------------------------------------------------------\n");
	if(argc > 2)
		ff(fpout,"\t%s\n",basename);
	else
		ff(fpout,"\tfunction declarations\n",basename);
	ff(fpout,
	    "------------------------------------------------------------------------*/\n");

	emit_editnote(fpout);
	ff(fpout,"\n");
	ff(fpout,"#ifndef BUILDING_LINT_ARGS\n");
	ff(fpout,"#ifdef LINT_ARGS\n\n");

	while(fgets(buf,sizeof(buf),fpin) != NULL)
	{
		cptr = (strncmp(buf,"/*global*/  ",12)) ? buf : buf + 12;
		fputs(cptr,fpout);
	}

	ff(fpout,"\n#else\t\t/* compiler doesn't know about prototyping */\n\n");

	fclose(fpin);
	fpin = fopen(argv[1],"r");

	while(fgets(buf,sizeof(buf),fpin) != NULL)
	{
		cptr = (strncmp(buf,"/*global*/  ",12)) ? buf : buf + 12;
		if(strncmp(cptr,"int ",4))
		{
			if(strchr(cptr,'('))
			{
				while(*cptr != '(')
					fputc(*cptr++,fpout);
				fputs("();\n",fpout);
			}
			else
				fputs(cptr,fpout);
		}
	}

	ff(fpout,"\n#endif /* LINT_ARGS */\n");
	ff(fpout,"#endif /* BUILDING_LINT_ARGS */\n");
	ff(fpout,"\n/* end of %s */\n",
	    (argc > 2) ? basename : "function declarations");

	fclose(fpin);
	fclose(fpout);
	exit(0);
}	/* end of main */

/* vi: set tabstop=4 shiftwidth=4: */
