/* CHK=0x27D4 */
/*+-------------------------------------------------------------------------
	ecuchdir.c
	wht%n4hgf@gatech.edu

  Defined functions:
	cd_array_delete(arg,narg)
	cd_array_delete_usage()
	cd_array_init()
	cd_array_read(arg,narg)
	cd_array_save(arg,narg)
	change_directory(cdarg,arg_present_flag)
	expand_dirname(dirname)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:23-wht@n4hgf---- 2.80 --- */
/*:03-25-1990-14:11-wht@n4hgf------ x2.70 ------- */
/*:01-01-1990-16:22-wht-use VTYPE */
/*:11-23-1989-02:49-wht-UNIX defines an ARG_MAX -- use BLD_ARG_MAX */
/*:07-03-1989-22:57-wht------ x2.00 ----- */
/*:06-24-1989-16:52-wht-flush edits --- ecu 1.95 */

#include "ecu.h"
#include "ecukey.h"
#include <pwd.h>

#define CD_QUAN		44
#define CD_PATHLEN	130
char *cd_array[CD_QUAN];
uint cd_in_use = 0;

extern char curr_dir[];		/* current working directory */
extern int errno;

/*+-------------------------------------------------------------------------
	cd_array_read(arg,narg)
--------------------------------------------------------------------------*/
void
cd_array_read(arg,narg)
char **arg;
int narg;
{
char dirpath[CD_PATHLEN];
FILE *fpcd;
FILE *fopen();
register char *cptr;
char *skip_ld_break();

	get_home_dir(dirpath);
	strcat(dirpath,"/.ecu/dir");
	if((fpcd = fopen(dirpath,"r")) == (FILE *)0)
		return;		/* none found */

	for(cd_in_use = 0; cd_in_use < CD_QUAN; cd_in_use++)
	{
		if(fgets(dirpath,sizeof(dirpath),fpcd) == (char *)0)
			break;
		dirpath[strlen(dirpath) - 1] = 0;
		cptr = skip_ld_break(dirpath);
		if(strlen(cptr) == 0)
		{
			--cd_in_use;
			continue;
		}
		strcpy(cd_array[cd_in_use],cptr);
	}
	fclose(fpcd);
}	/* end of cd_array_read */

/*+-------------------------------------------------------------------------
	cd_array_save(arg,narg)
--------------------------------------------------------------------------*/
void cd_array_save(arg,narg)
char **arg;
int narg;
{
register icd;
char savepath[256];
FILE *fpcd;
FILE *fopen();

	get_home_dir(savepath);
	strcat(savepath,"/.ecu/dir");

	if(cd_in_use == 0)
	{
		ff(se,"No directory list to save in %s\r\n",savepath);
		return;
	}
	if((fpcd = fopen(savepath,"w")) == (FILE *)0)
	{
		ff(se,"%s could not be opened\r\n",savepath);
		return;
	}

	for(icd = 0; icd < cd_in_use; icd++)
		fprintf(fpcd,"%s\n",cd_array[icd]);
	fclose(fpcd);
	ff(se,"%d entries saved in %s\r\n",cd_in_use,savepath);

}	/* end of cd_array_save */

/*+-------------------------------------------------------------------------
	cd_array_delete_usage()
--------------------------------------------------------------------------*/
void
cd_array_delete_usage()
{
	ff(se,"usage: d[elete] <1st> [<last>]\r\n");
}	/* end of cd_array_delete_usage */

/*+-------------------------------------------------------------------------
	cd_array_delete(arg,narg)
--------------------------------------------------------------------------*/
cd_array_delete(arg,narg)
char **arg;
int narg;
{
uint first;		/* 1st to delete */
uint last;		/* last to delete */

	if((narg < 2) || (narg > 3))
	{
		cd_array_delete_usage();
		return(-1);
	}

	first = atoi(arg[1]) - 1;
	if(narg == 2)
		last = first;
	else
		last = atoi(arg[2]) - 1;

	if((first > (cd_in_use - 1)) || (last > (cd_in_use - 1)) || (last < first))
	{
		cd_array_delete_usage();
		return(-1);
	}

	if(last == (cd_in_use - 1))
	{
		cd_in_use = first;
	}
	else
	{
	int count_less = last - first + 1;
		last++;
		while(last != cd_in_use)
			strcpy(cd_array[first++],cd_array[last++]);
		cd_in_use -= count_less;
	}
	return(0);
}	/* end of cd_array_delete */

/*+-------------------------------------------------------------------------
	cd_array_init()
--------------------------------------------------------------------------*/
void
cd_array_init()
{
register itmp;

/*allocate change_directory stack */
	for(itmp = 0; itmp < CD_QUAN; itmp++)
	{
		if(!(cd_array[itmp] = malloc(CD_PATHLEN + 1)))
		{
			ff(se,"Not enough memory for cd stack\r\n");
			exit(1);
		}
		*cd_array[itmp] = 0;
	}
	(void)cd_array_read(cd_array,0);
}	/* end of cd_array_init */

/*+-------------------------------------------------------------------------
	expand_dirname(dirname) - convert dirnames with shell chars
--------------------------------------------------------------------------*/
expand_dirname(dirname,maxlen)
char *dirname;
int maxlen;
{
char s256[256];
char *expcmd;

	if(!find_shell_chars(dirname))
		return(0);

	sprintf(s256,"`ls -d %s`",dirname);
	if(expand_cmd_with_wildlist(s256,&expcmd))
	{
		ff(se,"No files match\r\n");
		return(-1);
	}
	strncpy(dirname,expcmd,maxlen);
	dirname[maxlen - 1] = 0;
	free(expcmd);
	if(strchr(dirname,' '))
	{
		fputs("Too many files:\r\n",se);
		fputs(dirname,se);
		fputs("\r\n",se);
		return(-1);
	}
	return(0);

}	/* end of expand_dirname */

/*+-------------------------------------------------------------------------
	change_directory(dirname,arg_present_flag)

  Change directory to 'dirname' if arg_present_flag is true,
  else if flag 0, ask for new directory name and change to it
  This procedure maintains the global variable 'curr_dir' that
  reflects the ecu transmitter and receiver process current
  working directory.
--------------------------------------------------------------------------*/
change_directory(cdarg,arg_present_flag)
char *cdarg;
int arg_present_flag;
{
register icd;
register itmp;
char s130[130];
#define BLD_ARG_MAX	5
char *arg[BLD_ARG_MAX];
int narg;
int longest;

	if(cd_in_use == 0)
		cd_array_read(arg,0);

	fputs("  ",se);

	if(arg_present_flag)		/* if there is an argument ... */
	{
		if(isdigit(*cdarg))		/* ... and first char is digit */
		{
			icd = atoi(cdarg) - 1;
			if(icd >= cd_in_use)
				goto DISPLAY_CD_ARRAY;
			strncpy(s130,cd_array[icd],sizeof(s130) - 1);
		}
		else
			strncpy(s130,cdarg,sizeof(s130) - 1);	/* literal dir spec */

		s130[sizeof(s130) - 1] = 0;
	}
	else		/* no arg to cd command */
	{
DISPLAY_CD_ARRAY:
		fputs("\r\n",se);
		longest = 0;
		for(icd = 0; icd < CD_QUAN/2; icd++)
		{
			if(icd >= cd_in_use)
				break;
			if(longest < (itmp = strlen(cd_array[icd])))
				longest = itmp;
		}
		longest += 4;
		if(longest < 36)
			longest += 4;
		for(icd = 0; icd < CD_QUAN/2; icd++)
		{
			if(icd >= cd_in_use)
				break;
			sprintf(s130,"%2d %s ",icd + 1,cd_array[icd]);
			fputs(s130,se);
			if(icd + CD_QUAN/2 >= cd_in_use)
				fputs("\r\n",se);
			else
			{
				itmp = longest - strlen(s130);
				while(itmp-- > 0)
					fputc(' ',se);
				sprintf(s130,"%2d %s\r\n",
					icd + 1 + CD_QUAN/2,cd_array[icd + CD_QUAN/2]);
				fputs(s130,se);
			
			}
		}
		fputs("current dir: ",se);
		stand_out();
		ff(se," %s ",curr_dir);
		stand_end();
		eeol();
		fputs("\r\n",se);

GET_NEW_DIR:
		fputs("New dir, <#>, %save, %read, %del, %xmitcd, <enter>:  ",se);
		ttygets(s130,sizeof(s130),1);
		if( (s130[0] == ESC) || (strlen(s130) == 0) )
		{
			ff(se,"no directory change\r\n");
			return(0);
		}
		else if(s130[0] == '%')
		{
			build_str_array(s130,arg,BLD_ARG_MAX,&narg);

			if(minunique("save",&s130[1],1))
			{
				cd_array_save(arg,narg);
				goto GET_NEW_DIR;
			}
			else if(minunique("read",&s130[1],1))
			{
				cd_array_read(arg,narg);
				goto DISPLAY_CD_ARRAY;
			}
			else if(minunique("delete",&s130[1],1))
			{
				if(cd_array_delete(arg,narg))
					goto GET_NEW_DIR;
				else
					goto DISPLAY_CD_ARRAY;
			}
			else if(minunique("xmitcd",&s130[1],1))
			{
				lputs("cd ");
				lputs(curr_dir);
				lputc('\r');
				return(0);
			}
			else
			{
				ff(se,"Invalid cd subcommand\r\n");
				goto GET_NEW_DIR;
			}
		}
		else if(icd = atoi(s130))
		{
			icd--;
			if(icd >= cd_in_use)
				goto GET_NEW_DIR;
			strncpy(s130,cd_array[icd],sizeof(s130) - 1);
			s130[sizeof(s130) - 1] = 0;
		}
	}
	if(expand_dirname(s130,sizeof(s130)))
		return(-1);
	if(chdir(s130) < 0)		/* now change to the new directory */
	{
		perror(s130);		/* print error if we get one */
		ff(se,"\r\n");
		return(-1);
	}
	get_curr_dir(curr_dir,256);

	fputs("\r\nconfirmed: ",se);
	stand_out();
	ff(se," %s ",curr_dir);
	stand_end();
	fputs("\r\n",se);
	fflush(se);

	for(icd = 0; icd < cd_in_use; icd++)
	{
		if(strcmp(curr_dir,cd_array[icd]) == 0)
			return(0);
	}
	if(cd_in_use == CD_QUAN)
	{
		for(icd = 1; icd < CD_QUAN; icd++)
		{
			strcpy(cd_array[icd - 1],cd_array[icd]);
		}
		strcpy(cd_array[CD_QUAN - 1],curr_dir);
	}
	else
		strcpy(cd_array[cd_in_use++],curr_dir);

	return(0);

}	/* end of change_directory */
/* end of ecuchdir.c */
/* vi: set tabstop=4 shiftwidth=4: */
