/* CHK=0xC850 */
#define USE_XON_XOFF
/*+-------------------------------------------------------------------------
	ecufinsert.c -- insert file onto comm line
	wht%n4hgf@gatech.edu

  Defined functions:
	expand_filename(fname,maxlen)
	file_insert_clear_xoff()
	file_insert_to_line(narg,arg)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:23-wht@n4hgf---- 2.80 --- */
/*:04-19-1990-03:07-wht@n4hgf-GCC run found unused vars -- rm them */
/*:04-14-1990-19:46-wht@n4hgf-add 20-msec paced mode */
/*:03-25-1990-14:11-wht@n4hgf------ x2.70 ------- */
/*:12-03-1989-20:40-wht-beef up user interface + add echo pacing */
/*:07-03-1989-22:57-wht------ x2.00 ----- */
/*:06-27-1989-20:32-wht-allow wildcard spec or squiggle dir ref */
/*:06-24-1989-16:52-wht-flush edits --- ecu 1.95 */

#include "ecu.h"
#include "ecukey.h"

extern int interrupt;
extern char kbdintr;		/* current input INTR */
extern ulong colors_current;
extern ulong colors_alert;
extern ulong colors_errors;

/*+-------------------------------------------------------------------------
	expand_filename(fname) - convert fnames with shell chars
--------------------------------------------------------------------------*/
expand_filename(fname,maxlen)
char *fname;
int maxlen;
{
char s256[256];
char *expcmd;

	if(!find_shell_chars(fname))
		return(0);

	sprintf(s256,"`ls %s`",fname);
	if(expand_cmd_with_wildlist(s256,&expcmd))
	{
		ff(se,"\r\nNo files match\r\n");
		return(-1);
	}
	strncpy(fname,expcmd,maxlen);
	fname[maxlen - 1] = 0;
	if(strchr(expcmd,' '))
	{
		fputs("\r\nToo many files:\r\n",se);
		fputs(expcmd,se);
		fputs("\r\n",se);
		free(expcmd);
		return(-1);
	}
	free(expcmd);
	return(0);

}	/* end of expand_filename */

/*+-------------------------------------------------------------------------
	file_insert_clear_xoff()
--------------------------------------------------------------------------*/
void
file_insert_clear_xoff()
{
#ifdef USE_XON_XOFF
ulong colors_at_entry = colors_current;

	lclear_xmtr_xoff();
#ifdef SAY_CLEARED_XOFF
	setcolor(colors_alert);
	fputs("--> local XOFF cleared\r",se);
	setcolor(colors_at_entry);
#endif
#endif
}	/* end of file_insert_clear_xoff */

/*+-------------------------------------------------------------------------
	file_insert_to_line(narg,arg)
--------------------------------------------------------------------------*/
file_insert_to_line(narg,arg)
int narg;
char **arg;
{
register rchar;
int old_ttymode = get_ttymode();
long total_chars = 0L;
long total_lines = 0L;
ulong colors_at_entry = colors_current;
long timeout_msecs;
FILE *fp;
char file_string[512];
char s512[512];
char xmit_mode;
#ifdef USE_XON_XOFF
int ixon;
int ixoff;
#endif
char *make_char_graphic();

	if(narg > 1)
		strncpy(s512,arg[1],sizeof(s512));	
	else
	{
		ff(se,"\r\n--> File to insert on comm line: ");
		ttygets(s512,sizeof(s512),1);
		if( (s512[0] == ESC) || (strlen(s512) == 0) )
		{
			ff(se,"--> no transmission\r\n");
			return(0);
		}
	}
	if(expand_filename(s512,sizeof(s512)))
		return(-1);
	if((fp = fopen(s512,"r")) == (FILE *)0)
	{
		ff(se,"--> ");
		perror(s512);			/* print error if we get one */
		ff(se,"\r\n");
		return(-1);
	}

	if(narg > 1)
		ff(se,"\r\n");

	if(narg > 2)
		xmit_mode = *arg[2];
	else
	{
ASK_OPTION:
		ff(se,"--> (S)ingle line at a time\r\n");
		ff(se,"    (E)cho pacing\r\n");
		ff(se,"    (F)ull speed transmission\r\n");
		ff(se,"    (P)aced transmission (20 msec/schar)\r\n");
		ff(se,"    (Q)uit (or ESC):             ");
		xmit_mode = ttygetc(0) & 0x7F;
		fputs(make_char_graphic(xmit_mode,0),se);
		fputs("\r\n",se);
	}

	kill_rcvr_process(SIGUSR1);

	switch(xmit_mode = to_lower(xmit_mode))
	{
		case 's':
			setcolor(colors_alert);
			fputs("--> press SPACE to continue or ESC/'s' to stop\r\n",se);
			setcolor(colors_at_entry);
			break;

		case 'e':
			/* fall through */

		case 'f':
		case 'p':
			setcolor(colors_alert);
			ff(se,"--> press %s to abort\r\n",make_char_graphic(kbdintr,0));
			setcolor(colors_at_entry);
			ttymode(2);
			break;

		case 'q':
		case ESC:
			setcolor(colors_alert);
			ff(se,"--> file insertion aborted\r\n");
			setcolor(colors_at_entry);
			fclose(fp);
			return(0);

		default:
			ring_bell();
			fputs("\r\n",se);
			goto ASK_OPTION;
	}

#ifdef USE_XON_XOFF
	lget_xon_xoff(&ixon,&ixoff);		/* get current line xon/xoff status */
	lxon_xoff(IXON);					/* set it for us */
#endif

	interrupt = 0;	/* reset SIGINT flag */
	while(fgets(file_string,sizeof(file_string),fp))
	{
	int xmit_len = strlen(file_string);
	int xmit_cr = xmit_len && (file_string[xmit_len - 1] == NL);

		if(xmit_cr)
		{
			xmit_len--;
			file_string[xmit_len] = 0;
		}
		total_chars += xmit_len;
		total_lines++;

/* some destinations, like BBS msg entry, take a blank line to mean
end of message, so do not send completely blank lines */
		if(!xmit_len && xmit_cr)
		{
			lputc(' ');
			xmit_len = 1;
		}
		else if(xmit_mode == 'p')
		{
		register char *cptr = file_string;
			while(*cptr)
			{
				lputc(*cptr++);
				nap(20L);
				while(rdchk(shm->Liofd))
				{
					rchar = lgetc_xmtr();
					process_xmtr_rcvd_char((char)rchar,1);
				}
			}
		}
		else
			lputs(file_string);
		if(xmit_cr)
		{
			if(xmit_mode == 'p')
				nap(20L);
			lputc('\r');
			xmit_len++;
		}

		if(interrupt)
			break;

		switch(xmit_mode)
		{
			case 's':
				while(1)
				{
					if(rdchk(0))
						break;
					rchar = lgetc_timeout(5 * 1000L);
					if(rchar < 0)
						file_insert_clear_xoff();
					else
						process_xmtr_rcvd_char((char)rchar,1);
					if(rchar == 0x0A)
						break;
				}
				rchar = to_lower(ttygetc(1));
				if((rchar == 's') || (rchar == ESC))
					goto INSERT_DONE;
				break;

			case 'e':
				timeout_msecs = 5 * 1000L;
				while(1)
				{
					if(interrupt)
						break;
					rchar = lgetc_timeout(timeout_msecs);
					if(rchar < 0)
					{
						if(!xmit_len)
							break;
						file_insert_clear_xoff();
						timeout_msecs = 1 * 1000L;
					}
					else
					{
						process_xmtr_rcvd_char((char)rchar,1);
						timeout_msecs = 100L;
						if(xmit_len)
							xmit_len--;
					}
					if(rchar == 0x0A)
						break;
				}
				break;

			case 'f':
			case 'p':
				while(rdchk(shm->Liofd))
				{
					rchar = lgetc_xmtr();
					process_xmtr_rcvd_char((char)rchar,1);
				}
				break;
		}
		if(interrupt)
			break;
	}

INSERT_DONE:

	if(interrupt)
	{
		setcolor(colors_error);
		ff(se,"--> Interrupted\r\n");
		setcolor(colors_at_entry);
	}

	fclose(fp);

	ttymode(old_ttymode);		/* restore old console mode */

	while(((rchar = lgetc_timeout(200L)) > 0) && !interrupt)
		process_xmtr_rcvd_char((char)rchar,1);

	interrupt = 0;	/* reset SIGINT flag */
	setcolor(colors_success);
	ff(se,"\r\n-->  done ... sent %ld lines, %ld characters\r\n",
				total_lines,total_chars);
	setcolor(colors_at_entry);
	lclear_xmtr_xoff();
#ifdef USE_XON_XOFF
	lxon_xoff(ixon | ixoff);	/* restore old line xon/xoff status */
#endif
	start_rcvr_process(1);
	return(0);

}	/* end of file_insert_to_line */

/* vi: set tabstop=4 shiftwidth=4: */
/* end of ecufinsert.c */
