/* CHK=0xE2AA */
/*+-------------------------------------------------------------------------
	ecufork.c -- ecu spawning ground
	wht%n4hgf@gatech.edu

  Defined functions:
	exec_cmd(cmdstr)
	expand_cmd_with_wildlist(cmd,expcmd)
	find_executable(progname)
	is_executable(progname)
	shell(shellcmd)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-08-1990-21:40-wht@n4hgf-exec_cmd "null cmd" */
/*:05-07-1990-21:23-wht@n4hgf---- 2.80 --- */
/*:04-19-1990-03:07-wht@n4hgf-GCC run found unused vars -- rm them */
/*:03-27-1990-15:18-wht@n4hgf-overhaul for/wait/SIGCLD handling */
/*:03-25-1990-14:11-wht@n4hgf------ x2.70 ------- */
/*:01-01-1990-16:22-wht-use VTYPE */
/*:11-23-1989-03:11-wht-conditional compile for M_XENIX or M_UNIX */
/*:07-03-1989-22:57-wht------ x2.00 ----- */
/*:06-24-1989-16:52-wht-flush edits --- ecu 1.95 */

#include "ecu.h"
#include "ecufork.h"

extern int rcvr_pid;
extern int last_child_wait_status;
extern int last_child_wait_pid;

/*+-----------------------------------------------------------------------
	shell(shellcmd)

  param 'shellcmd' is a shell command prefixed with either
  a '!', '$', '>' character.

  '!' causes the command to run as a normal subshell of a process.
  '$' causes the communications line to be stdin and stdout
      for the spawned shell
  '>' causes spawned shell to receive exactly sames files as ecu
------------------------------------------------------------------------*/
void
shell(shellcmd)
char *shellcmd;
{
register shellpid;
register itmp;
register char *cptr;
char s40[40];
int wait_status;

int rcvr_alive = (rcvr_pid > 0);
char *getenv();

	if(rcvr_alive)
		kill_rcvr_process(SIGUSR1);		/* stop receiver process gracefully */

	signal(SIGINT,SIG_IGN);
	signal(SIGTERM,SIG_IGN);
	signal(SIGUSR1,SIG_IGN);
	signal(SIGUSR2,SIG_IGN);
	signal(SIGCLD,SIG_DFL);

	if((shellpid = fork()) < 0)
	{
		ff(se,"Cannot fork\r\n");
		if(rcvr_alive)
			start_rcvr_process(1);
		xmtr_signals();
		return;
	}

	ttymode(0);  	     		/* set canonical tty mode */
	if(shellpid == 0)			/* we are the spawned (going to call shell) */
	{
		if(*shellcmd != '>')	/* '>' prefix means leave fd's alone! */
		{
			/* Hook-up our "standard output" to either the tty or
			 * the line as appropriate for '!' or '$' */
			close(TTYOUT);
			fcntl(((*shellcmd == '$') ? shm->Liofd : TTYERR),F_DUPFD,TTYOUT);
			if(*shellcmd == '$')
			{
				close(TTYIN);
				fcntl(shm->Liofd,F_DUPFD,TTYIN);
			}
			close(shm->Liofd);
		}

		child_signals();		/* signals for child */

		if(*shellcmd == '!')
		{
			cptr = getenv("SHELL");
			if(cptr == (char *)0)
				cptr = "/bin/csh";
		}
		else
			cptr = "/bin/sh";

		shellcmd++;
		child_signals();
		if(ulindex(cptr,"csh") > -1)
		{
			if(*shellcmd == '\0')
				execl(cptr,"csh",(char *)0);
			else
				execl(cptr,"csh","-c",shellcmd,(char *)0);
		}
		else
		{
			if(*shellcmd == '\0')
				execl(cptr,"sh",(char *)0);
			else
				execl(cptr,"sh","-c",shellcmd,(char *)0);
		}

		ff(se,"cannot execute %s\r\n",cptr);	/* should not get here */
		_exit(255);								/* end of spawned process */
	}	/* end of if child process */

#if defined(FORK_DEBUG)
	sprintf(s40,"DEBUG fork shell pid %d",shellpid);
	ecu_log_event(s40); /* shell */
#endif

	while(((itmp = wait(&wait_status)) != shellpid) && (itmp != -1))
		;
	last_child_wait_status = wait_status;
	last_child_wait_pid = shellpid;

	xmtr_signals();			/* restore standard xmtr signals */
	ttymode(1);				/* control tty back to raw mode */

/* any comm program will probably doodle with the line characteristics. */
/* we want to make sure they are restored to normal */
	lreset_ksr();			/* restore comm line params */

	if(rcvr_alive)
		start_rcvr_process(1);

}	/* end of shell */

/*+-------------------------------------------------------------------------
	is_executable(progname)
--------------------------------------------------------------------------*/
is_executable(progname)
char *progname;
{
struct stat ss;

	if(stat(progname,&ss) < 0)			/* if cannot stat, flunk */
		return(0);
	if((ss.st_mode & 0111) == 0)		/* if no --x--x--x, flunk */
		return(0);
	return(1);	/* whew, this OUGHT to work */

}	/* end of is_executable */

/*+-------------------------------------------------------------------------
	find_executable(progname)
PATH=':/usr/wht/bin:/bin:/usr/bin:/usr/wht/bin:/etc/tuckerware' len=56
--------------------------------------------------------------------------*/
char *
find_executable(progname)
char *progname;
{
register itmp;
static char *path_buf = (char *)0;
#define PATHNAME_QUAN 32
static char *path_name[PATHNAME_QUAN + 1];
static char rtn_path[256];
static int path_count = 0;
char *cptr;
char *getenv();
char *str_token();

	if(path_buf == (char *)0)
	{
		if((cptr = getenv("PATH")) == (char *)0)
			return(cptr);
		if(!(path_buf = malloc(strlen(cptr) + 1)))
			return((char *)0);
		strcpy(path_buf,cptr);
		path_name[PATHNAME_QUAN + 1] = (char *)0;
		cptr = path_buf;
		for(path_count = 0; path_count < PATHNAME_QUAN; path_count++)
		{
			if(*cptr == 0)
				break;
			path_name[path_count] = cptr;
			while((*cptr != ':') && (*cptr != 0))
				cptr++;
			if(*cptr == ':')
				*cptr++ = 0;
		}
	}	/* end of get and process path env variable */

/* look for executable */
	for(itmp = 0; itmp < path_count; itmp++)
	{
		if(*path_name[itmp] == 0)	/* if null path (./) */
			strcpy(rtn_path,"./");
		else
			sprintf(rtn_path,"%s/",path_name[itmp]);
		strcat(rtn_path,progname);
		if(is_executable(rtn_path))
			return(rtn_path);
	}
	return((char *)0);
}	/* end of find_executable */

/*+-------------------------------------------------------------------------
	exec_cmd(cmdstr) - execute an arbitrary program with arguments
kills rcvr process if alive and restarts it when done if was alive
--------------------------------------------------------------------------*/
exec_cmd(cmdstr)
char *cmdstr;
{
char *cmdpath;
#define MAX_EXEC_ARG 512
char *cmdargv[MAX_EXEC_ARG];
int itmp;
int execpid;
int rcvr_alive = (rcvr_pid > 0);
int old_ttymode = get_ttymode();
int wait_status = 0;
extern int errno;
char *strrchr();

#if defined(FORK_DEBUG)
char s80[80];
	strcpy(s80,"DEBUG exec ");
	strncat(s80,cmdstr,sizeof(s80)-12);
	s80[sizeof(s80)-12] = 0;
	ecu_log_event(s80);
#endif

	build_arg_array(cmdstr,cmdargv,MAX_EXEC_ARG,&itmp);
	if(itmp == MAX_EXEC_ARG)
	{
		ff(se,"Too many arguments to command\r\n");
		return(-1);
	}
	else if(!itmp)
	{
		ff(se,"null command\r\n");
		return(-1);
	}

	if(*cmdargv[0] == '/')
	{
		cmdpath = cmdargv[0];
		cmdargv[0] = strrchr(cmdargv[0],'/') + 1;
	}
	else
	{
		if((cmdpath = find_executable(cmdargv[0])) == (char *)0)
		{
			ff(se,"Cannot find %s\r\n",cmdargv[0]);
			return(-1);
		}
	}

	if(rcvr_alive)
		kill_rcvr_process(SIGUSR1);		/* stop receiver process gracefully */

/* this code executed by the father (forking) process */
/* wait on death of child (morbid in life, but ok here) */

	signal(SIGINT,SIG_IGN);
	signal(SIGTERM,SIG_IGN);
	signal(SIGUSR1,SIG_IGN);
	signal(SIGUSR2,SIG_IGN);
	signal(SIGCLD,SIG_DFL);

	if((execpid = fork()) < 0)
	{
		ff(se,"Cannot fork\r\n");
		if(rcvr_alive)
			start_rcvr_process(1);
		xmtr_signals();
		return(-1);
	}

	if(execpid == 0)			/* we are the spawned (going to call exec) */
	{
		ttymode(0);  	     	/* set canonical tty mode */
		child_signals();
		execv(cmdpath,cmdargv);
		perror(cmdpath);
		exit(255);				/* end of spawned process */
	}	/* end of if child process */

	wait_status = 0;
	while(((itmp = wait(&wait_status)) != execpid) && (itmp != -1))
		;
	last_child_wait_status = wait_status;
	last_child_wait_pid = execpid;

/* resume our normally scheduled program */
	lreset_ksr();				/* restore comm line params */
	ttymode(old_ttymode);		/* control tty back to original */
	if(rcvr_alive)
		start_rcvr_process(1);
	xmtr_signals();
	return(0);

}	/* end of exec_cmd */

/*+-------------------------------------------------------------------------
	expand_cmd_with_wildlist(cmd,&expcmd)
if return -1, error, expcmd has error message (static message)
if return 0, cmd has been expanded, expcmd must be free()'d when done
--------------------------------------------------------------------------*/
expand_cmd_with_wildlist(cmd,expcmd)
char *cmd;
char **expcmd;
{
register char *cptr;
#define P_READ 0
#define P_WRITE 1
int pipe_pid;
int stdout_pipe[2];
int stderr_pipe[2];
int count;
int itmp;
int wait_status;
int rcvr_alive = (rcvr_pid > 0);
char *echo_cmd;
static char s132[132];
extern int errno;
extern char *sys_errlist[];
char *strchr();
static char *pipe_err_msg = "system error: no pipe";
static char *mem_err_msg = "system error: no memory";

	if(strchr(cmd,'<') || strchr(cmd,'>') || strchr(cmd,'&'))
	{
		*expcmd = "illegal characters: '<', '>' or '&'";
		return(-1);
	}

	if(pipe(stdout_pipe) < 0)
	{
		*expcmd = pipe_err_msg;
		return(-1);
	}
	if(pipe(stderr_pipe) < 0)
	{
		close(stdout_pipe[P_READ]);
		close(stdout_pipe[P_WRITE]);
		*expcmd = pipe_err_msg;
		return(-1);
	}
	if(!(echo_cmd = malloc(strlen(cmd) + 10)))
	{
		close(stdout_pipe[P_READ]);
		close(stdout_pipe[P_WRITE]);
		close(stderr_pipe[P_READ]);
		close(stderr_pipe[P_WRITE]);
		*expcmd = mem_err_msg;
		return(-1);
	}

	strcpy(echo_cmd,"echo ");
	strcat(echo_cmd,cmd);


	if(rcvr_alive)
		kill_rcvr_process(SIGUSR1);		/* stop receiver process gracefully */

	signal(SIGINT,SIG_IGN);
	signal(SIGTERM,SIG_IGN);
	signal(SIGUSR1,SIG_IGN);
	signal(SIGUSR2,SIG_IGN);
	signal(SIGCLD,SIG_DFL);

	if((pipe_pid = fork()) == 0)
	{
	int null = open("/dev/null",O_WRONLY,0);

		close(stdout_pipe[P_READ]);
		close(TTYOUT);
		dup(stdout_pipe[P_WRITE]);
		close(stdout_pipe[P_WRITE]);
		close(TTYERR);
		dup(stderr_pipe[P_WRITE]);
		close(stderr_pipe[P_WRITE]);
		close(null);
		child_signals();
		execl("/bin/csh","csh","-e","-f","-c",echo_cmd,(char *)0);
		_exit(255);
	}
#if defined(FORK_DEBUG)
	sprintf(s132,"DEBUG expand pid %d",pipe_pid);
	ecu_log_event(s132);		/* expand_cmd_with_wildlist */
#endif

	free(echo_cmd);

	close(stdout_pipe[P_WRITE]);
	close(stderr_pipe[P_WRITE]);
	if(pipe_pid == -1)
	{
		close(stdout_pipe[P_READ]);
		close(stderr_pipe[P_READ]);
		*expcmd = "could not fork";
		if(rcvr_alive)
			start_rcvr_process(0);
		xmtr_signals();
		return(-1);
	}

	if(!(*expcmd = malloc(5120)))
	{
		close(stdout_pipe[P_READ]);
		close(stderr_pipe[P_READ]);
		kill(pipe_pid,SIGKILL);
		*expcmd = mem_err_msg;
		if(rcvr_alive)
			start_rcvr_process(0);
		return(-1);
	}

	cptr = *expcmd;
	while((count = read(stdout_pipe[P_READ],cptr,64)) != 0)
	{
		if(count < 0)
		{
			if(errno == EINTR)
			{
				errno = 0;
				continue;
			}
			free(*expcmd);
			kill(pipe_pid,SIGKILL);
			close(stdout_pipe[P_READ]);
			close(stderr_pipe[P_READ]);
			*expcmd = "error reading wild list expansion";
			if(rcvr_alive)
				start_rcvr_process(0);
			return(-1);
		}
		cptr += count;
		*cptr = 0;
		if(*(cptr - 1) == '\n')
		{
			*(cptr - 1) = 0;
			break;
		}
	}
	close(stdout_pipe[P_READ]);
	if(!strlen(*expcmd))
	{
		free(*expcmd);
		count = read(stderr_pipe[P_READ],s132,sizeof(s132) - 1);
		if(count < 0)
			strcpy(s132,sys_errlist[errno]);
		else
			s132[count] = 0;
		if(s132[count - 1] == '\n')
			s132[count - 1] = 0;
		close(stderr_pipe[P_READ]);
		if(strncmp(s132,"echo: ",6))
			*expcmd = s132;
		else
			*expcmd = s132 + 6;
		if(rcvr_alive)
			start_rcvr_process(0);
		return(-1);
	}

	wait_status = 0;
	while(((itmp = wait(&wait_status)) != pipe_pid) && (itmp != -1))
		;

	xmtr_signals();
	if(wait_status)
	{
		free(*expcmd);
		count = read(stderr_pipe[P_READ],s132,sizeof(s132) - 1);
		if(count < 0)
			strcpy(s132,sys_errlist[errno]);
		else
			s132[count] = 0;
		if(s132[count - 1] == '\n')
			s132[count - 1] = 0;
		close(stderr_pipe[P_READ]);
		if(strncmp(s132,"echo: ",6))
			*expcmd = s132;
		else
			*expcmd = s132 + 6;
		if(rcvr_alive)
			start_rcvr_process(0);
		return(-1);
	}
	close(stderr_pipe[P_READ]);

	if(rcvr_alive)
		start_rcvr_process(0);
	return(0);
}	/* end of expand_cmd_with_wildlist */
/* vi: set tabstop=4 shiftwidth=4: */
