/* CHK=0xE625 */
/*+-------------------------------------------------------------------------
	ecuicmaux.c -- ecuicmd.c auxiliary functions
	wht%n4hgf@gatech.edu

  Defined functions:
	cmd_escape_str()
	display_ascii_names()
	nlin_nlout_control(token,narg,arg)
	icmd_conversion(token,narg,arg)
	icmd_log(token,narg,arg)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:23-wht@n4hgf---- 2.80 --- */
/*:04-10-1990-17:34-wht@n4hgf-provisions for ecu xmtr logging when rcvr dead */
/*:03-25-1990-14:11-wht@n4hgf------ x2.70 ------- */
/*:11-23-1989-03:11-wht-conditional compile for M_XENIX or M_UNIX */
/*:07-03-1989-22:57-wht------ x2.00 ----- */
/*:06-24-1989-16:53-wht-flush edits --- ecu 1.95 */

#include "ecu.h"
#include "ecuerror.h"
#include "ecucmd.h"

#if !defined(M_XENIX) || defined(M_UNIX)
extern int cmd_escape;		/* icmd escape */
#endif
extern int rcvr_pid;

char rcvr_log_file[256];	/* if rcvr_log!= 0,log filename */
int rcvr_log = 0;			/* rcvr log active if != 0 */
int rcvr_log_gen_title = 0;	/* gen log header on next open (rcvr process) */
FILE *rcvr_log_fp;			/* rcvr log file */
int rcvr_log_append = 1;
int rcvr_log_raw = 0;
int rcvr_log_flusheach = 0;

/*+-------------------------------------------------------------------------
	cmd_escape_str()
--------------------------------------------------------------------------*/
char *
cmd_escape_str()
{
#if defined(M_XENIX) || defined(M_UNIX)
	return("HOME ");
#else
char *make_char_graphic();
	return(make_char_graphic(cmd_escape,0));
#endif
}	/* end of cmd_escape_str */

/*+-----------------------------------------------------------------------
	display_ascii_names()
------------------------------------------------------------------------*/
void display_ascii_names()
{
register intval;

	for(intval = 0; intval < 32; intval++)
	{
		ff(se,"%s %3d %03o %02x ^%c | ",hex_to_ascii_name((intval)),
				(intval),(intval),(intval),(intval) |0x40);
		ff(se,"    %3d %03o %02x  %c | ",
				intval+32,intval+32,intval+32,intval+32);
		ff(se,"    %3d %03o %02x  %c | ",
				intval+64,intval+64,intval+64,intval+64);
		if(intval != 31)
		{
			ff(se,"    %3d %03o %02x  %c\r\n",
					intval+96,intval+96,intval+96,intval+96);
		}
		else
		{
			ff(se,"    %3d %03o %02x  ^?\r\n",
					intval+96,intval+96,intval+96);
		}
	}

}	/* end of display_ascii_names */

/*+-------------------------------------------------------------------------
	icmd_conversion(token,narg,arg)
--------------------------------------------------------------------------*/
void
icmd_conversion(token,narg,arg)
int token;
int narg;
char **arg;
{
register itmp;

	switch(token)
	{
	case CTxa:
	case CToa:
	case CTda:
		if(narg > 1)
		{
		int result;
		char format[4];
			sprintf(format,"%%%c",to_lower(*arg[0]));
			if(sscanf(arg[1],format,&result) == 0)
			{
				ff(se,"  invalid argument\r\n");
				return;
			}
			result &= 0xFF;
			if(result == ' ')
				ff(se," == ' ' 0x20\r\n",make_char_graphic(result,1));
			else
				ff(se," == %s\r\n",make_char_graphic(result,1));
		}
		else
		{
			ff(se,"\r\n");
			display_ascii_names();
		}
		break;
	case CTax:
		if(arg[1] == (char *)0)
		{
			ff(se,"\r\n");
			display_ascii_names();
			break;
		}
		switch(strlen(arg[1]))
		{
			case 1:
				ff(se," == 0x%02x\r\n",*arg[1]);
				break;
			case 2:
				if(*arg[1] == '^')
				{
					itmp = to_upper(*(arg[1] + 1));
					if((itmp < '@') || (itmp > '_'))
					{
						ff(se,"  not a valid control character\r\n");
						return;
					}
					itmp &= 0x1F;
					ff(se," == 0x%02x %s\r\n",itmp,make_char_graphic(itmp,1));
					break;
				}	/* else fall thru */
			case 3:
				if((itmp = ascii_name_to_hex(arg[1])) > -1)
				{
					ff(se," == 0x%02x %s\r\n",itmp,make_char_graphic(itmp,1));
					break;
				}	/* else fall thru */
			default:
				ff(se,"  invalid ... examples of valid parameters:\r\n");
				ff(se,"        ^A ETX  or  printable character\r\n");
				break;
		}
		break;
	default:
		ff(se,"  invalid command\r\n");
	}
}	/* end of icmd_conversion */

/*+-------------------------------------------------------------------------
	icmd_log(token,narg,arg)
--------------------------------------------------------------------------*/
icmd_log(token,narg,arg)
int token;
int narg;
char **arg;
{
register itmp;
register itmp2;

	switch(token)
	{
	case CTloff:
		goto LOG_OFF;
		break;
#if defined(M_XENIX) || defined(M_UNIX)
	case CTllp:
		icmd("log /dev/lp1");
#endif
		break;
	case CTlog:
		if(narg > 1)
		{
			if(minunique("off",arg[1],3))
			{
LOG_OFF:
				if(rcvr_log == 0)	/* "off",but not logging */
					goto RECORD_REPORT;
				ff(se,"\r\nlogging concluded (file %s)\r\n",rcvr_log_file);
				shmx_set_rcvr_log("",0,0,0);
				rcvr_log = 0;
				rcvr_log_file[0] = 0;
				return(0);
			}
/* turning logging on */
			itmp2 = -1;
			rcvr_log_append = 1;
			rcvr_log_raw = 0;
			for(itmp = 1; itmp < narg; itmp++)
			{
				if(*arg[itmp] == '-')
				{
					switch(arg[itmp][1])
					{
						case 's': rcvr_log_append = 0; break;
						case 'r': rcvr_log_raw = 1; break;
						case 'f': rcvr_log_flusheach = 1; break;
						default:
							ff(se,"   unrecognized switch -%c\r\n",
										arg[itmp][1]);
							log_cmd_usage();
							return(eFATAL_ALREADY);
					}
				}
				else
				{
					if(itmp2 > 0)
					{
						ff(se,"   too many positional arguments\r\n");
						log_cmd_usage();
						return(eFATAL_ALREADY);
					}
					itmp2 = itmp;
				}
			}
			if(itmp2 < 0)
			{
				ff(se,"   no log file name specified\r\n");
				log_cmd_usage();
				return(eFATAL_ALREADY);
			}
			if(arg[itmp2][0] != '/')	/* if log file not full path, ... */
			{							/* ... supply current directory */
				get_curr_dir(rcvr_log_file,
					sizeof(rcvr_log_file) - strlen(arg[itmp2]) - 2);
				strcat(rcvr_log_file,"/");
				strcat(rcvr_log_file,arg[itmp2]);
			}
			else
				strcpy(rcvr_log_file,arg[itmp2]);

			/* try to open the file if we can */
			rcvr_log_fp = fopen(rcvr_log_file,"a");
			if(rcvr_log_fp != NULL)		/* if success */
			{
				fclose(rcvr_log_fp);
				rcvr_log_fp = NULL;
				rcvr_log = 1;
				shmx_set_rcvr_log(rcvr_log_file,rcvr_log_append,rcvr_log_raw,
					rcvr_log_flusheach);
			}
			else		/* xmtr() could not open file */
			{
				ff(se,"   could not open ");
				perror(rcvr_log_file);
				ff(se,"\r\n");
				return(eFATAL_ALREADY);
			}
		}	/* end of if argument to command */

RECORD_REPORT:
		if(rcvr_log)
		{
			ff(se,"\r\n%sing received %s text to\r\n%s\r\n",
				(rcvr_log_append) ? "append" : "writ",
				(rcvr_log_raw)    ? "raw"    : "filtered",
				rcvr_log_file);
			ff(se,"use \"%slog off\" to stop logging\r\n",
					cmd_escape_str());
		}
		else
		{
			ff(se,"   not logging.\r\n");
			ff(se,"use \"%slog <filename>\" to start logging\r\n",
					cmd_escape_str());
		}
		break;
	default:
		ff(se,"  invalid command\r\n");
		return(eFATAL_ALREADY);
	}
	return(0);
}	/* end of icmd_log */

/*+-------------------------------------------------------------------------
	nlin_nlout_control(token,narg,arg)
--------------------------------------------------------------------------*/
void
nlin_nlout_control(token,narg,arg)
int token;
int narg;
char **arg;
{
	switch(token)
	{
	case CTnlin:
		if(narg != 1)
			shm->Ladd_nl_incoming = yes_or_no(arg[1]);
		ff(se,"  %sappending NL to incoming CR\r\n",
			(shm->Ladd_nl_incoming) ? "" : "not ");
		break;
	case CTnlout:
		if(narg != 1)
			shm->Ladd_nl_outgoing = yes_or_no(arg[1]);
		ff(se,"  %sappending NL to outgoing CR\r\n",
			(shm->Ladd_nl_outgoing) ? "" : "not ");
		break;
	default:
	case CTnl:
		ff(se,"  incoming: %s  outgoing: %s\r\n",
			(shm->Ladd_nl_incoming) ? "CR/LF" : "CR",
			(shm->Ladd_nl_outgoing) ? "CR/LF" : "CR");
		break;
	}

}	/* end of nlin_nlout_control */

/* end of ecuicmaux.c */
/* vi: set tabstop=4 shiftwidth=4: */
