/* CHK=0x6A03 */
/*+-----------------------------------------------------------------------
	ecuicmd.c
	wht%n4hgf@gatech.edu

  Defined functions:
	icmd(icmd_cmd)
	icmd_do_proc(narg,arg)
	search_cmd_list(cmd_list,cmd)

------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:23-wht@n4hgf---- 2.80 --- */
/*:04-17-1990-15:26-wht@n4hgf-add icmd call to icmd_history_add */
/*:03-31-1990-16:12-wht@n4hgf-search_cmd_list fix for pcmds similar to icmds */
/*:03-25-1990-14:11-wht@n4hgf------ x2.70 ------- */
/*:03-12-1990-05:15-wht@n4hgf-use Lrname */
/*:12-21-1989-21:53-wht-RTS/CTS flow control */
/*:11-23-1989-03:11-wht-conditional compile for M_XENIX or M_UNIX */
/*:11-17-1989-22:25-wht-use build_arg_array not build_str_array */
/*:10-14-1989-21:10-wht-allow user to change ecu color choices */
/*:09-26-1989-14:23-wht-add mkdir */
/*:07-03-1989-22:57-wht------ x2.00 ----- */
/*:06-24-1989-16:53-wht-flush edits --- ecu 1.95 */

#include "ecu.h"
#include "esd.h"
#include "ecufkey.h"
#include "patchlevel.h"

#define NEED_P_CMD
#include "ecucmd.h"

char *cmd_escape_str();

extern KDE keyset_table[];
extern char *makedate;
extern char curr_dir[];
extern char hello_str[];
extern char keyset_name[];
extern char rcvr_log_file[];	/* if rcvr_log!= 0,log filename */
extern int current_ttymode;
extern int errno;
extern int proc_level;
extern int proctrace;
extern int rcvr_log;			/* rcvr log active if != 0 */
extern int rcvr_log_append;	/* if true, append, else scratch */
extern int rcvr_log_raw;		/* if true, log all, else filter ctl chrs */
extern int rcvr_pid;
extern ulong colors_current;
extern FILE *plog_fp;
extern ESD  *plog_name;
extern char screen_dump_file_name[];

int protocol_log_packets = 0;
int want_bell_notify = 1;	/* want bell notify */
int cmd_escape;

/*+-------------------------------------------------------------------------
	search_cmd_list(cmd_list,cmd)
returns -1 if cmd not found or insufficient chars for match
else token value for command
proc-only commands have 'min' values 0
--------------------------------------------------------------------------*/
search_cmd_list(cmd_list,cmd)
register P_CMD *cmd_list;
register char *cmd;
{
	while(cmd_list->token != -1)
	{
		if(cmd_list->min && minunique(cmd_list->cmd,cmd,cmd_list->min))
			return(cmd_list->token);
		cmd_list++;
	}
	return(-1);

}	/* end of search_cmd_list */

/*+-------------------------------------------------------------------------
	icmd_do_proc(narg,arg)
--------------------------------------------------------------------------*/
icmd_do_proc(narg,arg)
int narg;
char **arg;
{
register erc;
ulong colors_at_entry = colors_current;

	kill_rcvr_process(SIGUSR1);
	ttymode(2);
	erc = do_proc(narg,arg);
	proc_file_reset();
	ttymode(1);
	setcolor(colors_notify);
	ff(se,"[procedure finished]");
	setcolor(colors_at_entry);
	ff(se,"\r\n");
	start_rcvr_process(0);
	return(erc);
}	/* end of icmd_do_proc */

/*+-----------------------------------------------------------------------
	icmd(cmd)
 This function implements the built in commands
 It returns non-zero if program should terminate
------------------------------------------------------------------------*/
int
icmd(icmd_cmd)
register char *icmd_cmd;
{

#define ARG_MAX_QUAN 40
char *arg[ARG_MAX_QUAN];
char cmd[128];
int itmp;
int token;
int narg = 0;
ESD *tesd;
char s80[80];
char *epoch_secs_to_str();
long atol();
long rcvr_chars;
long rcvr_chars_this_connect;
char *find_procedure();

	icmd_history_add(icmd_cmd);		/* add to history list */
	strcpy(cmd,icmd_cmd);			/* get local copy of cmd string */
	switch(cmd[0]) 
	{
		case '.':		/* exit xmtr() (generally program too) */
			ff(se,"  disconnecting from line %s\r\n",shm->Lline);
			return(1);
		case '!':
		case '$':
		case '>':
			ff(se,"\r\n");
			shell(cmd);
			return(0);
		case '-':
			ff(se,"\r\n");
			exec_cmd(&cmd[1]);
			return(0);
		case '^':
			ff(se,"\r\n");
			phrase_help();
			return(0);
		case '?':
			icmd_help(0,(char **)0);
			return(0);
#ifdef DEBUG_SHM_SCREEN
		case '@':
			{
			FILE *fp = fopen("/tmp/screen","w");
			int y,x;
			char *cptr = (char *)shm->screen;
				for(y = 0; y < 43; y++)
				{
					for(x = 0; x < 80; x++)
						fputc(*cptr++,fp);
					fputc('\n',fp);
				}
				fclose(fp);
			}
			break;
#endif
		default:
			break;
	}

/* not single character argument */
	build_arg_array(cmd,arg,ARG_MAX_QUAN,&narg);

/* handle phrases */
	if(isdigit(*arg[0]))
	{
		phrases(narg,arg);
		return(0);
	}

/* search command list */
	if((token = search_cmd_list(icmd_cmds,arg[0])) < 0)
	{
		ff(se,"\r\n");
		if(find_procedure(arg[0]))
		{
			icmd_do_proc(narg,arg);
			return(0);
		}
		ff(se,"invalid command... %s? for help\r\n",
			cmd_escape_str());
		return(0);
	}

	switch(token)
	{					/* keep alphabetized PLEASE */
		case CTrx:
		case CTry:
		case CTrz:
		case CTrk:
		case CTrs:
		receive_files_from_remote(narg,arg);
		break;

		case CTsx:
		case CTsy:
		case CTsz:
		case CTsk:
		case CTss:
		send_files_to_remote(narg,arg);
		break;

		case CTbaud:
		if(narg == 1)
			ff(se,"   baud rate is %u\r\n",shm->Lbaud);
		else
		{
			itmp = atoi(arg[1]);
			if(!lnew_baud_rate(itmp))
				ff(se,"   baud rate set to %u\r\n",shm->Lbaud);
			else
			{
				ff(se,"   invalid baud rate: %u\r\n",itmp);
				ff(se,"valid rates: 110,300,600,1200,2400,4800,9600,19200\r\n");
			}
		}
		break;

		case CTbreak:
		lbreak();
		ff(se,"   break sent\r\n");
		break;

		case CTcd:
		change_directory(arg[1],(narg == 1) ? 0 : 1);
		break;

		case CTclrx:
		lclear_xmtr_xoff();
		break;

		case CTdial:
		if(narg > 1)
		{
			ff(se,"\r\n");
			strcpy(shm->Llogical,arg[1]);
			if(!lookup_logical_telno())
				ff(se,"No such entry. Try 'dial' with no arguments\r\n");
			else
			{
				if(isdigit(shm->Llogical[0]) ||
					!find_procedure(shm->Llogical))
					hayes_dial();
				else
				{
					arg[0] = shm->Llogical;
					arg[1] = "!INTERACTIVE";
					icmd_do_proc(2,arg);
				}
			}
		}
		else
		{
			pde_list_manager();
		}
		break;

		case CTdo:
		ff(se,"\r\n");
		icmd_do_proc(narg - 1,&arg[1]);
		break;

		case CTptrace:
		if(narg > 1)
			proctrace = yes_or_no(arg[1]);
		ff(se,"  procedure trace %s", proctrace ? "on" : "off");
		if(proctrace > 1)
			ff(se," (%d)",proctrace);
		ff(se,"\r\n");
		break;

		case CTpcmd:
		itmp = strlen(arg[0]);
		if((tesd = make_esd(256)) == (ESD *)0)
		{
			ff(se,"  no memory!!?!\r\n");
			break;
		}
		strcpy(tesd->pb,icmd_cmd + itmp + 1);
		tesd->cb = strlen(tesd->pb);
		ff(se,"\r\n");
		kill_rcvr_process(SIGUSR1);
		ttymode(2);
		if(itmp = execute_esd(tesd))
		{
			proc_error(itmp);
			show_esd(tesd,"");
		}
		free_esd(tesd);
		ttymode(1);
		start_rcvr_process(0);
		break;

		case CTplog:
		fputs("  ",se);
		if(narg > 1)
		{
			if(!strcmp(arg[1],"off"))
				plog_control((char *)0);
			else
				plog_control(arg[1]);
		}
		
		if(plog_fp)
			ff(se,"procedure logging: %s\r\n",plog_name->pb);
		else
			fputs("procedure logging off\r\n",se);
		break;

		case CTduplex:
		if(narg > 1)
		{
			switch(to_lower(*arg[1]))
			{
				case 'f': shm->Lfull_duplex = 1; ff(se,"  now "); break;
				case 'h': shm->Lfull_duplex = 0; ff(se,"  now "); break;
				default: ff(se,
"\r\nfirst argument character must be F or H for full or half duplex\r\n");
					break;
			}
		}
		else		/* no argument */
			ff(se,"  currently ");

		ff(se,"%s duplex\r\n",(shm->Lfull_duplex) ? "full" : "half");
		break;

		case CTexit:
		ff(se,"  disconnecting from line %s\r\n",shm->Lline);
		return(1);

		case CTfi:
		file_insert_to_line(narg,arg);
		break;

		case CThangup:
		ff(se,"   hanging up ... ");
		hayes_ATH();
		ff(se,"done\r\n");
		break;

		case CThelp:
		icmd_help(narg,arg);
		break;

		case CTsdname:
		if(narg > 1)
		{
		char *new_file_name;

			itmp = 0;	/* do not need to free(new_file_name) */
			if(find_shell_chars(arg[1]))
			{

				if(expand_cmd_with_wildlist(arg[1],&new_file_name))
				{
					ff(se,"  %s\r\n",new_file_name);
					break;
				}
				itmp = 1;	/* need to free(new_file_name) */
			}
			else
				new_file_name = arg[1];

			screen_dump_file_name[0] = 0;
			if(*new_file_name != '/')
			{
				strcpy(screen_dump_file_name,curr_dir);
				strcat(screen_dump_file_name,"/");
			}
			strcat(screen_dump_file_name,arg[1]);
			if(itmp)
				free(new_file_name);
		}
		ff(se,"\r\nscreen dump name: %s\r\n",screen_dump_file_name);
		break;

		case CTlog:
		case CTloff:
		case CTllp:
		icmd_log(token,narg,arg);
		break;

		case CTnl:
		case CTnlin:
		case CTnlout:
		nlin_nlout_control(token,narg,arg);
		break;

		case CTparity:
		if(narg > 1)
		{
			switch(to_lower(*arg[1]))
			{
				case 'n':
					shm->Lparity = 0; break;
				case 'e':
				case 'o':
					shm->Lparity = to_lower(*arg[1]); break;
				default:
					ff(se,"   unrecognized parity: %c\r\n",*arg[1]);
					goto PARITY_DONE;
			}
			lset_parity(1);
		}
		ff(se,"   parity set to %c\r\n",
			(shm->Lparity) ? to_upper(shm->Lparity) : 'N');
PARITY_DONE: ;
		break;

		case CTpid:
		ff(se,": xmtr %d, rcvr: %d, parent: %d, group: %d\r\n",
			getpid(0),rcvr_pid,getppid(0),getpgrp(0));
		break;

		case CTpwd:
		fputs(" ",se);
		stand_out();
		ff(se," %s ",curr_dir);
		stand_end();
		fputs("\r\n",se);
		break;

		case CTredial:
		(void)hayes_redial(arg,narg);
		break;

		case CTrev:
		fputs("\r\n",se);
		fputs(hello_str,se);
		ff(se," patch level %d\n",PATCHLEVEL);
		fputs("\r\n",se);
		ff(se,"%s\r\n",makedate);
		break;

#if !defined(M_XENIX) || defined(M_UNIX)
		case CTte:
		if(narg > 1)
		{
			if(sscanf(arg[1],"%x",&cmd_escape) && (cmd_escape < 0x7F))
				ff(se,"   tty escape set to "); 
			else
			{
				ff(se,"  invalid parameter must be 01 to 7E: setting to ");
				cmd_escape = '%';
			}
		}
		else
			ff(se,"   tty escape is ");

		ff(se,"0x%02x (%s)\r\n",cmd_escape,make_char_graphic(cmd_escape,1));
		break;
#endif

		case CTtime:
		get_tod(4,cmd);
		ff(se,": %s\r\n",cmd);
		break;

		case CTts:
		ff(se,"\r\n");
		sprintf(s80,"TTYIN termio (ttymode=%d)",current_ttymode);
		disp_line_termio(TTYIN,s80);
		ff(se,"\r\n");
		disp_line_termio(shm->Liofd,"comm line termio");
		break;

		case CTtty:
		ff(se,"   %s\r\n",get_ttyname());
		break;

		case CTda:
		case CToa:
		case CTxa:
		case CTax:
		icmd_conversion(token,narg,arg);
		break;

		case CTbn:
		if(narg > 1)
		{
			want_bell_notify = yes_or_no(arg[1]);
			shmx_set_rcvr_bn(want_bell_notify);
		}
		ff(se,"  bell notify %s%s\r\n",
					want_bell_notify ? "on" : "off",
					(want_bell_notify == 2) ? ": ring on incoming char" : "");
		break;

		case CTrtscts:
		if(narg > 1)
		{
			lRTSCTS_control(yes_or_no(arg[1]));
		}
		ff(se,"  RTS %s CTS %s\r\n",
			(Ltermio.c_cflag & RTSFLOW) ? "on" : "off",
			(Ltermio.c_cflag & CTSFLOW) ? "on" : "off");
		break;

		case CTfkey:
		if(narg < 2)
			keyset_display();
		else
		{
			switch(keyset_read(arg[1]))
			{
				case  0: keyset_display(); break;
				case -1: ff(se," cannot find ~/.ecu/keys\r\n"); break;
				case -2: ff(se," not found in ~/.ecu/keys\r\n"); break;
			}
		}
		break;

		case CTstat:
		ff(se,"\r\n\r\nDate/time");
		get_tod(4,cmd);
		ff(se,": %s\r\n",cmd);
		ff(se,"Console: %s  communications line: %s  ",
			get_ttyname(),shm->Lline);
		ff(se,"%u-%c-1\r\n",shm->Lbaud,
			(shm->Lparity) ? to_upper(shm->Lparity) : 'N');
		if(shm->Lmodem_off_hook)
		{
			ff(se,"Connected to %s: %s (%s) at %s (elapsed %s)\r\n",
				shm->Lrname,
				shm->Ldescr,shm->Ltelno,
				epoch_secs_to_str(shm->Loff_hook_time,1,s80),
				get_elapsed_time(time((long *)0) - shm->Loff_hook_time));
		}
		ff(se,"Current directory: %s\r\n",curr_dir);
		ff(se,"Total chars transmitted: %ld",shm->xmit_chars);
		if(shm->Lmodem_off_hook)
			ff(se," (since CONNECT %ld)\r\n",
				shm->xmit_chars_this_connect);
		else
			fputs("\r\n",se);
		shmx_rc_report(&rcvr_chars,&rcvr_chars_this_connect);
		ff(se,"Total chars received:    %ld",rcvr_chars);
		if(shm->Lmodem_off_hook)
			ff(se," (since CONNECT %ld)\r\n",rcvr_chars_this_connect);
		else
			fputs("\r\n",se);
		if(keyset_name[0])
			ff(se,"Function key set '%s' loaded\r\n",keyset_name);
		else
			ff(se,"No function key set loaded\r\n");
		if(rcvr_log)
			ff(se,"Session log open: %s (%s mode)\r\n",
				rcvr_log_file,(rcvr_log_raw) ? "raw" : "filtered");
		else
			ff(se,"Session logging not active\r\n");
		ff(se,"Bell notify is %s\r\n",(want_bell_notify) ? "on" : "off");
		ff(se,"Duplex: %s\r\n",(shm->Lfull_duplex) ? "full" : "half");
		ff(se,"CR conversion:  incoming %s  outgoing %s\r\n",
			(shm->Ladd_nl_incoming) ? "CR/LF" : "CR",
			(shm->Ladd_nl_outgoing) ? "CR/LF" : "CR");
		fputs("\r\n",se);
		break;

		case CTxlog:
		if(narg > 1)
		{
			protocol_log_packets = yes_or_no(arg[1]);
		}
		ff(se,"  protocol packet logging %s\r\n",
			(protocol_log_packets) ? "on" : "off");
		break;

		case CTsgr:
		send_get_response(token,narg,arg);
		break;

/*
		case CTmkdir:
		if(narg < 2)
			ff(se,"  no argument\r\n");
		if(mkdir(arg[2],0755))
		{
			ff(se,"  ");
			pperror(arg[2]);
		}
		else
			ff(se,"  made directory %s\r\n",arg[2]);
		break;
*/

		case 0:
		ff(se,"   procedure command not allowed in interactive mode\r\n");
		break;

		default:
		do_proc(narg,arg);
		break;

	}
	return(0);		/* 0 == do not end program */

}	/* end of icmd */

/* end of ecuicmd.c */
/* vi: set tabstop=4 shiftwidth=4: */
