/* CHK=0x60C4 */
/*+-----------------------------------------------------------------------
	eculine.c -- line related procedures
 wht%n4hgf@gatech.edu

  Defined functions:
	lbreak()
	lclear_xmtr_xoff()
	lclose()
	lflush(flush_type)
	lget_xon_xoff(ixon,ixoff)
	lgetc_timeout(msec)
	lgetc_xmtr()
	lgets_timeout(lrwt)
	llookfor(lookfor,msecs,echo_flag)
	lnew_baud_rate(new_baud)
	lopen()
	lopen_err_text(lerr)
	lputc(lchar)
	lputc_paced(pace_msec,lchar)
	lputs(string)
	lputs_paced(pace_msec,string)
	lquiet(msecs,echo_flag)
	lrdchk_xmtr()
	lreset_ksr()
	lset_baud_rate(ioctl_flag)
	lset_parity(ioctl_flag)
	ltoggle_dtr()
	lxon_xoff(flag)
	valid_baud_rate(baud)
	process_xmtr_rcvd_char(rchar,echo)

------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:23-wht@n4hgf---- 2.80 --- */
/*:04-19-1990-03:07-wht@n4hgf-GCC run found unused vars -- rm them */
/*:04-10-1990-17:34-wht@n4hgf-provisions for ecu xmtr logging when rcvr dead */
/*:03-28-1990-04:58-wht@n4hgf-add to echo in process_xmtr_rcvd_char */
/*:03-27-1990-15:18-wht@n4hgf-overhaul for/wait/SIGCLD handling */
/*:03-26-1990-00:46-wht@n4hgf-if EACCES, suggest course of action to user */
/*:03-25-1990-14:11-wht@n4hgf------ x2.70 ------- */
/*:02-09-1990-16:05-wht-heuristic debug of RTS/CTS use */
/*:12-21-1989-21:53-wht-RTS/CTS flow control */
/*:12-04-1989-17:43-wht-lget_xon_xoff was returning wrong kind of data */
/*:11-10-1989-21:49-wht-send xmtr-rcvd chars thru ansi filter */
/*:11-04-1989-07:00-wht-keep ansi cursor position in shm */
/*:10-25-1989-13:16-wht-notify rcvr on parity change */
/*:07-03-1989-22:57-wht------ x2.00 ----- */
/*:06-30-1989-00:32-wht-ltoggle_dtr no longer lopen/lclose, but open/close */
/*:06-24-1989-16:53-wht-flush edits --- ecu 1.95 */

#include "ecu.h"
#include "ecukey.h"
#include "ecuhangup.h"
#if !defined(NO_SELECT)
#include <sys/select.h>
#endif

extern int rcvr_pid;
extern int errno;
extern int lgetc_count;
extern int vmin;
extern int interrupt;		/* SIGINT flag: see xmtr_SIGINT_handler */

char lopen_err_str[64] = "";

#define LPUTS_NAP_COUNT	20

/*+-------------------------------------------------------------------------
	process_xmtr_rcvd_char(rchar,echo) - feed xmtr-rcvd char to rcvr code

echo: 0 no echo
      1 echo literally
      2 "make printable"
--------------------------------------------------------------------------*/
void
process_xmtr_rcvd_char(rchar,echo)
uint rchar;
register int echo;
{
extern int rcvr_log;
extern FILE *rcvr_log_fp;
extern char rcvr_log_file[];	/* if rcvr_log!= 0,log filename */

	if(process_rcvd_char(rchar))
		return;

	if(echo == 1)
	{
		if(rchar == NL)
			fputc(CR,se);
		fputc(rchar,se);
		if(rchar != CR)
			plogc(rchar);
	}
	else if(echo == 2)
	{
	char *make_char_graphic();
		pputs(make_char_graphic(rchar,0));
	}

}	/* end of process_xmtr_rcvd_char */

/*+-------------------------------------------------------------------------
	lgetc_xmtr() -- xmtr version of get char from line
also called by rcvr code when lgetc_buf empty and vmin == 1
--------------------------------------------------------------------------*/
uchar
lgetc_xmtr()
{
int itmp;
extern int errno;
uchar char_rtnd;

READ_AGAIN:
	if((itmp = read(shm->Liofd,&char_rtnd,1)) < 1)
	{
		if(!itmp)
		{
			pperror("lgetc_xmtr: zero length read\n");
			hangup(HANGUP_LINE_READ_ERROR);
		}
		if(errno == EINTR)			/* if signal interrupted, ... */
		{
			if(interrupt)
				return(0);
			goto READ_AGAIN;
		}
		hangup(HANGUP_LINE_READ_ERROR);
	}
	shm->rcvd_chars++;
	shm->rcvd_chars_this_connect++;
	if(shm->Lparity)
		char_rtnd &= 0x7F;
	return(char_rtnd);

}	/* end of lgetc_xmtr */

/*+-------------------------------------------------------------------------
	lrdchk_xmtr() -- rdchk(shm->Liofd) for xmtr
--------------------------------------------------------------------------*/
int
lrdchk_xmtr()
{
	return(rdchk(shm->Liofd));
}	/* end of lrdchk_xmtr */

/*+-------------------------------------------------------------------------
	char *lgets_timeout(LRWT *) - may be called by xmtr only

to1 and to2 are unsigned long values in milliseconds (not
currently supported well under BSD4); to1 is the time to wait
for the first character, to2 the time to wait for subsequent
characters.

if raw_flag 0,     non-printables are stripped from beginning
                   and end of received characters (i.e., modem
                   response reads); NULs discarded, parity stripped
if raw_flag 1,     full raw read buffer returned

0x80 in raw_flag indicates console interrupts should be enabled.
if interrupt thus detected, the procedure returns "!Interrupted"
without reseting variable 'interrupt'

buffer is address to read chars into

bufsize is buffer max size (allowing room for terminating null)
which should be at least 2 if raw_size includes 0x80 bit,
else at least 12 characters if 0x80 omitted.

count is a int which, at return, receives the actual count read

--------------------------------------------------------------------------*/
char *
lgets_timeout(lrwt)
LRWT *lrwt;
{
/**********************/
#if defined(NO_SELECT)
/**********************/

register actual_count = 0;
register char *cptr = lrwt->buffer;
register echo_flag = lrwt->echo;
int max_count = lrwt->bufsize;
char *rtn_val;
int timeout_counter;
int qc1;
int qc2;
int raw_mode = lrwt->raw_flag & 0x0F;
int check_sigint = (lrwt->raw_flag & 0x80);
int old_ttymode = get_ttymode();	/* save original tty mode */
int delim_len;
long quantum;
long ltmp;
long nap(long);

	delim_len = (lrwt->delim) ? strlen(lrwt->delim) : 0;

	if((shm->Lbaud < 300) && lrwt->to2)
		if(lrwt->to2 < 300L) lrwt->to2 = 300L;
	else if((shm->Lbaud < 1200) && lrwt->to2)
		if(lrwt->to2 < 200L) lrwt->to2 = 100L;

/* shortest interval */
	ltmp = (lrwt->to1 < lrwt->to2) ? lrwt->to1 : lrwt->to2;

/* calculate wait quantum */
	quantum = ltmp / 10L;				/* try for ten ticks */

#if defined(M_I386)
	if(quantum < 40L)
		quantum = 40L;
#else
	if(quantum < 20L)
		quantum = 20L;
#endif
	qc1 = lrwt->to1 / quantum;
	if(!qc1) qc1 = 1L;
	qc2 = lrwt->to2 / quantum;
	if(!qc2) qc2 = 1L;

/* perform the lrtw function using nap() and rdchk()
   input: qc1 is first nap count (for first charcters) 
          qc2 is 2nd nap count (for subsequent characters) 
          quantum is the nap period in milliseconds
          cptr is char* to receive read string
          max_count is max number of characters incl null
          lrwt->raw_flag as described above

  output: lrwt->count is actual count of return result
          lrwt->buffer is return read buffer
*/
	max_count--;				/* leave room for null */

	if(check_sigint)
		ttymode(2);				/* let console interrupt long timeouts */

	timeout_counter = qc1;		/* first timeout */ 
	*cptr = 0;					/* init result string */
	while(timeout_counter--)
	{
		nap(quantum);

		if(check_sigint && interrupt)
			goto INTERRUPTED;

		while(lrdchk_xmtr())
		{
			*cptr = lgetc_xmtr();

			if(check_sigint && interrupt)
				goto INTERRUPTED;

			if(*cptr == 0)
				continue;

			process_xmtr_rcvd_char(*cptr,echo_flag);

			if(!raw_mode && (*cptr == CR))
					continue;

			*++cptr = 0;
			if(++actual_count == 1)
			{
				if(!lrwt->to2)
					break;
				timeout_counter = qc2;
			}

			if(--max_count == 0)
				goto BOTTOM;

			if(delim_len && (actual_count >= delim_len) &&
					!strncmp(lrwt->delim,cptr - delim_len,delim_len))
				goto BOTTOM;
		}
	}

/******************************/
#else /* other than NO_SELECT */
/******************************/
/* --- use select --- */
register actual_count = 0;
register char *cptr = lrwt->buffer;
register max_count = lrwt->bufsize;
register raw_mode = lrwt->raw_flag & 0x0F;
register echo_flag = lrwt->echo;
int check_sigint = (lrwt->raw_flag & 0x80);
int old_ttymode = get_ttymode();	/* save original tty mode */
int fdmask;
int delim_len;
struct timeval tval;
char *rtn_val;

	delim_len = (lrwt->delim) ? strlen(lrwt->delim) : 0;

	if((shm->Lbaud < 300) && lrwt->to2)
		if(lrwt->to2 < 300L) lrwt->to2 = 300L;
	else if((shm->Lbaud < 1200) && lrwt->to2)
		if(lrwt->to2 < 200L) lrwt->to2 = 100L;


/* perform the lrtw function

  output: lrwt->count is actual count of return result
          lrwt->buffer is return read buffer
*/
	max_count--;				/* leave room for null */

	if(check_sigint)
		ttymode(2);				/* let console interrupt long timeouts */

	*cptr = 0;					/* init result string */
	while(1)
	{
		if(check_sigint && interrupt)
			goto INTERRUPTED;

		errno = 0;
		fdmask = 1 << shm->Liofd;
		if(actual_count)
		{
			tval.tv_sec = lrwt->to2 / 1000L;
			tval.tv_usec = (lrwt->to2 % 1000L) * 1000L;
		}
		else
		{
			tval.tv_sec = lrwt->to1 / 1000L;
			tval.tv_usec = (lrwt->to1 % 1000L) * 1000L;
		}
		if(select(32,&fdmask,(int *)0,(int *)0,&tval) != 1)
		{
			if(errno == EINTR)
				continue;
			break;
		}

		while(rdchk(shm->Liofd))
		{
			*cptr = lgetc_xmtr();

			if(check_sigint && interrupt)
				goto INTERRUPTED;

			if(*cptr == 0)
				continue;

			process_xmtr_rcvd_char(*cptr,!!echo_flag);

			if(!raw_mode && (*cptr == CR))
					continue;

			*++cptr = 0;
			actual_count++;

			if(--max_count == 0)
				goto BOTTOM;

			if(delim_len && (actual_count >= delim_len) &&
					!strncmp(lrwt->delim,cptr - delim_len,delim_len))
				goto BOTTOM;
		}
		if(!lrwt->to2)
			break;
	}

#endif	/* NO_SELECT */

/********* common post processing for select() / no select() ************/
BOTTOM:
	if(check_sigint)
		ttymode(old_ttymode);
	if(raw_mode)
	{
		lrwt->count = actual_count;
		return(lrwt->buffer);
	}
	cptr = lrwt->buffer;
	while(((*cptr > 0) && (*cptr < SPACE)) || (*cptr >= DEL))
		cptr++;
	rtn_val = cptr;
	actual_count = 0;
	while(((*cptr &= 0x7F) >= SPACE) && (*cptr < DEL))
	{
		cptr++;
		actual_count++;
	}
	*cptr = 0;
	strcpy(lrwt->buffer,rtn_val);
	lrwt->count = actual_count;
	return(lrwt->buffer);

INTERRUPTED:
	ttymode(old_ttymode);
	strcpy(lrwt->buffer,"!Interrupted");
	lrwt->count = strlen(lrwt->buffer);
	return((char *)0);

}	/* end of lgets_timeout */

/*+-------------------------------------------------------------------------
	lgetc_timeout(msec) - may be called by xmtr only

 reads one character from line unless msec passes with no receipt.
 return char if received, else -1 if timeout
--------------------------------------------------------------------------*/
int
lgetc_timeout(msec)
long msec;
{
uchar rtn_char;
#if defined(NO_SELECT)
long nap();
long count;

AGAIN:
	count = msec;
	while(!lrdchk_xmtr())
	{
		if(interrupt)
			return(-1);
		if((count -= nap(20L)) <= 0)
			return(-1);
	}

#else

int fdmask;
struct timeval tval;

AGAIN:
	tval.tv_sec = msec / 1000L;
	tval.tv_usec = (msec % 1000L) * 1000L;
	fdmask = 1 << shm->Liofd;
	if(select(32,&fdmask,(int *)0,(int *)0,&tval) < 1)
		return(-1);
	if(!lrdchk_xmtr())
		return(-1);
	if(interrupt)
		return(-1);
#endif

	rtn_char = lgetc_xmtr();
	if(!rtn_char)
		goto AGAIN;
	return(rtn_char);

}	/* end of lgetc_timeout */

/*+-------------------------------------------------------------------------
	llookfor(lookfor,msecs,echo_flag)
return 1 if successful, else 0 if no match
echo_flag: 0 no echo
           1 echo literally
           2 "make printable"
--------------------------------------------------------------------------*/
int
llookfor(lookfor,msecs,echo_flag)
char *lookfor;
ulong msecs;
int echo_flag;
{
register lookfor_len = strlen(lookfor);
register lchar;
char *lastfew = (char *)malloc(lookfor_len);
int success_flag = 0;
int old_ttymode = get_ttymode();

	if(!lastfew)
	{
		pputs("memory exhausted\n");
		return(0);
	}

	ttymode(2);

	memset(lastfew,0,lookfor_len);
	while((lchar = lgetc_timeout(msecs)) >= 0)
	{
		if(!lchar)		/* skip nulls */
			continue;
		process_xmtr_rcvd_char(lchar,echo_flag);
		memcpy(lastfew,lastfew + 1,lookfor_len - 1);
		*(lastfew + lookfor_len - 1) = lchar;
		if(!strncmp(lastfew,lookfor,lookfor_len))
		{
			success_flag = 1;
			break;
		}
	}
	free(lastfew);
	ttymode(old_ttymode);
	return(success_flag);
}	/* end of llookfor */

/*+-------------------------------------------------------------------------
	lquiet(msecs,echo_flag)
--------------------------------------------------------------------------*/
void
lquiet(msecs,echo_flag)
ulong msecs;
int echo_flag;
{
register lchar;
int old_ttymode = get_ttymode();

	ttymode(2);
	while((lchar = lgetc_timeout(msecs)) >= 0)
	{
		if(interrupt)	/* if interrupt, return */
			break;
		if(!lchar)		/* skip nulls */
			continue;
		process_xmtr_rcvd_char(lchar,!!echo_flag);
	}
	ttymode(old_ttymode);

}	/* end of lquiet */

/*+-------------------------------------------------------------------------
	lflush(flush_type) -- flush line driver input &/or output buffers

0 == input buffer
1 == output buffer
2 == both buffers
--------------------------------------------------------------------------*/
void
lflush(flush_type)
int flush_type;
{
	lgetc_count = 0;
	switch(flush_type)
	{
		case 0:
			ioctl(TTYIN,TCFLSH,(char *)0); break;
		case 1:
			ioctl(TTYIN,TCFLSH,(char *)1); break;
		case 2:
			ioctl(TTYIN,TCFLSH,(char *)2); break;
	}
}	/* end of lflush */

/*+-------------------------------------------------------------------------
	lreset_ksr()

  This procedure restores the termio for the
  comm line to the values in Ltermio
--------------------------------------------------------------------------*/
void
lreset_ksr()
{
	ioctl(shm->Liofd,TCSETA,(char *)&Ltermio);

}	/* end of lreset_ksr */

/*+-----------------------------------------------------------------------
	lputc(lchar) -- write lchar to comm line
------------------------------------------------------------------------*/
void
lputc(lchar)
char lchar;
{
	while(write(shm->Liofd,&lchar,1) < 0)
	{
		if(errno == EINTR)
			continue;
		pperror("lputc write error");
		hangup(HANGUP_XMTR_WRITE_ERROR);
	}
	shm->xmit_chars++;
	shm->xmit_chars_this_connect++;
}	/* end of lputc */

/*+-----------------------------------------------------------------------
	lputc_paced(pace_msec,lchar) -- write lchar to comm line
  with time between each character 
------------------------------------------------------------------------*/
void
lputc_paced(pace_msec,lchar)
register pace_msec;
char lchar;
{

	lputc(lchar);	
	nap((long)(pace_msec ? pace_msec : LPUTS_NAP_COUNT));

}	/* end of lputc_paced */

/*+-----------------------------------------------------------------------
	lputs(string) -- write string to comm line
------------------------------------------------------------------------*/
void
lputs(string)
register char *string;
{
	while(*string)
		lputc(*string++);
}

/*+-----------------------------------------------------------------------
	lputs_paced(pace_msec,string) -- write string to comm line
  with time between each character 
------------------------------------------------------------------------*/
void
lputs_paced(pace_msec,string)
register pace_msec;
register char *string;
{
	while(*string)
		lputc_paced(pace_msec,*string++);

}	/* end of lputs_paced */

/*+-------------------------------------------------------------------------
	valid_baud_rate(baud) -- returns (positive) baud rate selector
or -1 if invalid baud rate
--------------------------------------------------------------------------*/
valid_baud_rate(baud)
uint baud;
{
	switch(baud)
	{
		case 110: return(B110);
		case 300: return(B300);
		case 600: return(B600);
		case 1200: return(B1200);
		case 2400: return(B2400);
		case 4800: return(B4800);
		case 9600: return(B9600);
		case 19200: return(EXTA);
		case 38400: return(EXTB);
		default: return(-1);
	}

}	/* end of valid_baud_rate */

/*+-----------------------------------------------------------------------
	lset_baud_rate(ioctl_flag)

  If 'ioctl_flag' is set, then perform ioctl call
  is executed after setting baud rate
------------------------------------------------------------------------*/
lset_baud_rate(ioctl_flag)
int ioctl_flag;
{
int baud_selector = valid_baud_rate(shm->Lbaud);

	if(baud_selector == -1)
		baud_selector = valid_baud_rate(shm->Lbaud = DEFAULT_BAUD_RATE);

	Ltermio.c_cflag &= ~CBAUD;
	Ltermio.c_cflag |= baud_selector;

#if defined(HO_HUM)
	if(ioctl_flag)
#endif
		 ioctl(shm->Liofd,TCSETA,(char *)&Ltermio);
	return(0);

}	/* end of lset_baud_rate */

/*+-------------------------------------------------------------------------
	lRTSCTS_control(flag)
--------------------------------------------------------------------------*/
void
lRTSCTS_control(flag)
int flag;
{
	switch(flag)
	{
		case 0:
			Ltermio.c_iflag |= (IXOFF);
			Ltermio.c_cflag &= ~(RTSFLOW | CTSFLOW);
			break;

		case 1:
			Ltermio.c_iflag &= ~(IXON | IXOFF | IXANY);
			Ltermio.c_cflag |= (RTSFLOW | CTSFLOW);
			break;

		case 2:
			Ltermio.c_iflag &= ~(IXON | IXOFF | IXANY);
			Ltermio.c_cflag |= RTSFLOW;
			Ltermio.c_cflag &= ~CTSFLOW;
			break;

		case 3:
			Ltermio.c_iflag &= ~(IXON | IXOFF | IXANY);
			Ltermio.c_cflag |= CTSFLOW;
			Ltermio.c_cflag &= ~RTSFLOW;
			break;
	}
	ioctl(shm->Liofd,TCSETA,(char *)&Ltermio);

}	/* end of lRTSCTS_control */

/*+-------------------------------------------------------------------------
	lnew_baud_rate(new_baud)
--------------------------------------------------------------------------*/
int
lnew_baud_rate(new_baud)
uint new_baud;
{
	if(valid_baud_rate(new_baud) < 0)
		return(-1);
	if(shm->Lbaud != new_baud)
		shm->Lmodem_already_init = 0;
	shm->Lbaud = new_baud;
	lset_baud_rate(1);
	return(0);
}	/* end of lnew_baud_rate */

/*+-----------------------------------------------------------------------
	lset_parity(ioctl_flag)

  If 'ioctl_flag' is set, then perform ioctl call
  is executed after setting parity
------------------------------------------------------------------------*/
void
lset_parity(ioctl_flag)
int ioctl_flag;
{
	Ltermio.c_cflag &= ~(CS8 | PARENB | PARODD);
	switch(to_lower(shm->Lparity))
	{
		case 'e':
			Ltermio.c_cflag |= CS7 | PARENB;
			Ltermio.c_iflag |= ISTRIP;
			break;
		case 'o':
			Ltermio.c_cflag |= CS7 | PARENB | PARODD;
			Ltermio.c_iflag |= ISTRIP;
			break;
		default:
			ff(se,"invalid parity: %c ... defaulting to no parity\r\n");
		case 0:
		case 'n':
			Ltermio.c_cflag |= CS8;
			Ltermio.c_iflag &= ~(ISTRIP);
			shm->Lparity = 0;
			break;
	}			

#if defined(HO_HUM)
	if(ioctl_flag)
#endif
	{
		ioctl(shm->Liofd,TCSETA,(char *)&Ltermio);
	}

}	/* end of lset_parity */

/*+-------------------------------------------------------------------------
	lclear_xmtr_xoff()
--------------------------------------------------------------------------*/
void
lclear_xmtr_xoff()
{
	ioctl(shm->Liofd,TCXONC,(char *)1); /* restart xmtr output */
}	/* end of lclear_xmtr_xoff */

/*+-------------------------------------------------------------------------
	lbreak()
--------------------------------------------------------------------------*/
void
lbreak()
{
	ioctl(shm->Liofd,TCSBRK,(char *)0);
}	/* end of lbreak */

/*+----------------------------------------------------------------------
	lopen()
returns negative LOPEN_ codes if failure else positive pid using line
else 0 if successful open
------------------------------------------------------------------------*/
int
lopen()
{
register itmp = strlen(shm->Lline);
struct stat ttystat;

	lopen_err_str[0] = 0;
	if(shm->Liofd >= 0)
		return(LOPEN_ALREADY);
	if(isupper(shm->Lline[itmp - 1]))
		shm->Lline[itmp - 1] = tolower(shm->Lline[itmp - 1]);
	if(!strcmp(shm->Lline,"/dev/tty"))
		return(LOPEN_INVALID);
	if(itmp = lock_tty())		/* get lock file */
		return(itmp);
	if(stat(shm->Lline,&ttystat) < 0)
		return(LOPEN_NODEV);
	shm->Liofd = open(shm->Lline,O_RDWR,0777);
	if(shm->Liofd < 0)
	{
		if(errno == EACCES)
			sprintf(lopen_err_str,"open error - try chmod +rw %s",shm->Lline);
		return(LOPEN_OPNFAIL);
	}
	else
	{
		ioctl(shm->Liofd,TCGETA,(char *) &Ltermio);
		Ltermio.c_iflag = (IGNPAR | IGNBRK | IXOFF );
		Ltermio.c_oflag = 0;
		Ltermio.c_cflag |= (CLOCAL | CREAD | HUPCL);
		Ltermio.c_lflag = 0;

		Ltermio.c_cc[VMIN]   = 1;
		Ltermio.c_cc[VTIME]  = 1;
		lset_baud_rate(0);		/* do not perform ioctl */
		lset_parity(1);			/* do perform ioctl */
	}

	lopen_err_str[0] = 0;
	return(0);

}	/* end of lopen */

/*+-----------------------------------------------------------------------
	lclose()
------------------------------------------------------------------------*/
void
lclose()
{
	if(shm->Liofd < 0)
		return;
	unlock_tty();	/* kill lock file (writes to line; must go before close) */
	close(shm->Liofd);
	shm->Liofd = -1;

}	/* end of lclose */

/*+-------------------------------------------------------------------------
	ltoggle_dtr()
--------------------------------------------------------------------------*/
void
ltoggle_dtr()
{
	close(shm->Liofd);
	nap(300L);
	shm->Liofd = open(shm->Lline,O_RDWR,0777);
	ioctl(shm->Liofd,TCSETA,(char *)&Ltermio);
	nap(600L);
}	/* end of ltoggle_dtr */

/*+-------------------------------------------------------------------------
	lxon_xoff(flag)
IXON specifies whether or not we respond to xon/xoff characters
IXOFF specifies whether or not we generate XON/XOFF characters
--------------------------------------------------------------------------*/
void
lxon_xoff(flag)
int flag;
{
	if(flag & IXON)
		Ltermio.c_iflag |= IXON;
	else
		Ltermio.c_iflag &= ~IXON;

	if(flag & IXOFF)
		Ltermio.c_iflag |= IXOFF;
	else
		Ltermio.c_iflag &= ~IXOFF;

	ioctl(shm->Liofd,TCSETA,(char *)&Ltermio);

}	/* end of lxon_xoff */

/*+-------------------------------------------------------------------------
	lget_xon_xoff(ixon,ixoff)
--------------------------------------------------------------------------*/
void
lget_xon_xoff(ixon,ixoff)
int *ixon;
int *ixoff;
{
	*ixon  = Ltermio.c_iflag & IXON;
	*ixoff = Ltermio.c_iflag & IXOFF;
}	/* end of lget_xon_xoff */

/*+-------------------------------------------------------------------------
	lopen_err_text(lerr)
--------------------------------------------------------------------------*/
char *
lopen_err_text(lerr)
int lerr;
{
static char lerr_s40[40];

	if(lopen_err_str[0])
		return(lopen_err_str);

	switch(lerr)
	{
		case LOPEN_INVALID: return("invalid line name");
		case LOPEN_UNKPID: return("unknown pid is using line");
		case LOPEN_LCKERR: return("error creating lock file");
		case LOPEN_NODEV: return("line does not exist");
		case LOPEN_ALREADY: return("line already open");
		case LOPEN_OPNFAIL: return("line open error (chmod +rw ?)");
		case LOPEN_ENABLED: return("line enabled for incoming login");
		case LOPEN_ENABLED_IN_USE: return("line in use by incoming login");
		case LOPEN_DIALOUT_IN_USE: return("line in use by another dial out");
	}
	if(lerr > 0)
		sprintf(lerr_s40,"pid %d using line",lerr);
	else
		sprintf(lerr_s40,"unknown line error %d",lerr);
	return(lerr_s40);
}	/* end of lopen_err_text */

/* end of eculine.c */

/* vi: set tabstop=4 shiftwidth=4: */
