/* CHK=0xEE4D */
/*+-------------------------------------------------------------------------
	ecutcap.c -- termcap stuff
	wht%n4hgf@gatech.edu

  Defined functions:
	clear_screen()
	curbotleft()
	curleft(count)
	curmove(y,x)
	curright(count)
	del_line()
	draw_box(y,x,height,width,title,title_x)
	draw_box_primitive(y,x,height,width)
	eeod()
	eeol()
	horiz_rule(count)
	ins_line()
	putc_stderr(character)
	read_termcap()
	stand_end()
	stand_out()
	tcap_clear_area_char(count,clrch)
	tcap_gets(buf,bufsize,delim,wait_for_key)
	vbell()
	vertical_rule(y,x,count)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:23-wht@n4hgf---- 2.80 --- */
/*:04-15-1990-17:09-wht@n4hgf-add curleft and curright */
/*:03-27-1990-15:18-wht@n4hgf-overhaul for/wait/SIGCLD handling */
/*:03-25-1990-14:11-wht@n4hgf------ x2.70 ------- */
/*:11-19-1989-16:59-wht-add eeod */
/*:07-03-1989-22:57-wht------ x2.00 ----- */
/*:06-24-1989-16:53-wht-flush edits --- ecu 1.95 */

#include "ecu.h"
#include "ecukey.h"
#include "pc_scr.h"

char *tcap_eeol = "";
char *tcap_eeod = "";
char *tcap_standout = "";
char *tcap_standend = "";
char *tcap_vbell = "";
char *tcap_move = "";
char *tcap_clear = "";
char *tcap_delline = "";
char *tcap_insline = "";
char *tcap_curleft = "";
char *tcap_curright = "";
char tcap_strbuf[256];
int tcap_standout_width;
int tcap_LINES;
int tcap_COLS;

char *tgetstr();
char *tgoto();
char *getenv();

void curmove();
void stand_out();
void stand_end();

/*+-------------------------------------------------------------------------
	read_termcap() - get termcap variables
--------------------------------------------------------------------------*/
void
read_termcap()
{
char termbuf[1024];
char *strbuf = tcap_strbuf;

	if(tgetent(termbuf,getenv("TERM")) > 0)
	{
		tcap_standout = tgetstr("so",&strbuf);
		tcap_standend = tgetstr("se",&strbuf);
		tcap_vbell    = tgetstr("vb",&strbuf);
		tcap_eeol     = tgetstr("ce",&strbuf);
		tcap_eeod     = tgetstr("cd",&strbuf);
		tcap_move     = tgetstr("cm",&strbuf);
		tcap_clear    = tgetstr("cl",&strbuf);
		tcap_delline  = tgetstr("dl",&strbuf);
		tcap_insline  = tgetstr("al",&strbuf);
		tcap_curleft  = tgetstr("kl",&strbuf);
		tcap_curright = tgetstr("kr",&strbuf);
		if((tcap_standout_width = tgetnum("sg")) < 0)
			tcap_standout_width = 0;
		tcap_LINES = tgetnum("li");
		tcap_COLS = tgetnum("co");
	}

}	/* end of read_termcap */

/*+-------------------------------------------------------------------------
	horiz_rule(count) - horizontal rule starting at current position
--------------------------------------------------------------------------*/
void
horiz_rule(count)
register count;
{
	while(count--)
		fputc(sHR,se);
}	/* end of horiz_rule */

/*+-------------------------------------------------------------------------
	vertical_rule(y,x,count) - vertical rule starting at y,x
--------------------------------------------------------------------------*/
void
vertical_rule(y,x,count)
int y;
register x;
register count;
{

	while(count--)
	{
		curmove(y++,x);
		fputc(sVR,se);
	}
		
}	/* end of vertical_rule */

/*+-------------------------------------------------------------------------
	draw_box_primitive(y,x,height,width) - ruled box
--------------------------------------------------------------------------*/
void
draw_box_primitive(y,x,height,width)
register y;
int x;
int height;
int width;
{
register i;

	curmove(y,x);
	fputc(sTL,se);
	if((i = width - 2) > 0)
		horiz_rule(i);
	fputc(sTR,se);
	if((i = height - 2) > 0)
	{
		vertical_rule(y + 1,x + width - 1,i);
		vertical_rule(y + 1,x,i);
	}
	curmove(y + height - 1,x);
	fputc(sBL,se);
	if((i = width - 2) > 0)
		horiz_rule(i);
	fputc(sBR,se);

}	/* end of draw_box_primitive */

/*+-------------------------------------------------------------------------
	draw_box(y,x,height,width,title,title_x)
--------------------------------------------------------------------------*/
void
draw_box(y,x,height,width,title,title_x)
char *title;
{
register stand = (title_x < 0);

	if(stand)
		title_x = -title_x;

	draw_box_primitive(y,x,height,width);
	curmove(y,x + title_x);
	fputc('[',se);
	if(stand)
		stand_out();
	fprintf(so," %s ",title);
	if(stand)
		stand_end();
	fputc(']',se);

}	/* end of draw_box */

/*+-------------------------------------------------------------------------
	putc_stderr(character) - utility rotuine for tputs
--------------------------------------------------------------------------*/
void
putc_stderr(character)
int character;
{
	fputc(character,se);
}	/* end of putc_stderr */

/*+-------------------------------------------------------------------------
	curmove(y,x)
--------------------------------------------------------------------------*/
void
curmove(y,x)
uint y;
uint x;
{
	if(y >= tcap_LINES)
		y = tcap_LINES - 1;
	if(x >= tcap_COLS)
		x = tcap_COLS - 1;
	tputs(tgoto(tcap_move,x,y),1,putc_stderr);
}	/* end of curmove */

/*+-------------------------------------------------------------------------
	curleft(count) - move cursor left
--------------------------------------------------------------------------*/
void
curleft(count)
register count;
{
	while(count--)
		fputs(tcap_curleft,se);
}	/* end of curleft */

/*+-------------------------------------------------------------------------
	curright(count) - move cursor right
--------------------------------------------------------------------------*/
void
curright(count)
register count;
{
	while(count--)
		fputs(tcap_curright,se);
}	/* end of curright */

/*+-------------------------------------------------------------------------
	curbotleft()
--------------------------------------------------------------------------*/
void
curbotleft()
{
	curmove(tcap_LINES - 1,0);
}	/* end of curbotleft */

/*+-------------------------------------------------------------------------
	ins_line()
--------------------------------------------------------------------------*/
void
ins_line()
{
	if(*tcap_insline)
		tputs(tcap_insline,1,putc_stderr);
}	/* end of ins_line */

/*+-------------------------------------------------------------------------
	del_line()
--------------------------------------------------------------------------*/
void
del_line()
{
	if(*tcap_delline)
		tputs(tcap_delline,1,putc_stderr);
}	/* end of del_line */

/*+-------------------------------------------------------------------------
	vbell() - output visual bell
--------------------------------------------------------------------------*/
void
vbell()
{
	if(*tcap_vbell)
		tputs(tcap_vbell,1,putc_stderr);
}	/* end of vbell */

/*+-------------------------------------------------------------------------
	clear_screen()
--------------------------------------------------------------------------*/
void
clear_screen()
{
	if(*tcap_clear)
		tputs(tcap_clear,1,putc_stderr);
}	/* end of clear_screen */

/*+-------------------------------------------------------------------------
	eeol() - erase to end of line
--------------------------------------------------------------------------*/
void
eeol()
{
	if(*tcap_eeol)
		tputs(tcap_eeol,1,putc_stderr);
}	/* end of eeol */

/*+-------------------------------------------------------------------------
	eeod() - erase to end of display
--------------------------------------------------------------------------*/
void
eeod()
{
	if(*tcap_eeod)
		tputs(tcap_eeod,1,putc_stderr);
}	/* end of eeod */

/*+-------------------------------------------------------------------------
	stand_out()
--------------------------------------------------------------------------*/
void
stand_out()
{
	if(*tcap_standout) /*  && (tcap_standout_width == 0)) */
		tputs(tcap_standout,1,putc_stderr);
}	/* end of stand_out */

/*+-------------------------------------------------------------------------
	stand_end()
--------------------------------------------------------------------------*/
void
stand_end()
{
	if(*tcap_standend) /* && (tcap_standout_width == 0)) */
		tputs(tcap_standend,1,putc_stderr);
}	/* end of stand_end */

/*+-------------------------------------------------------------------------
	tcap_clear_area_char(count,clrch)
--------------------------------------------------------------------------*/
void
tcap_clear_area_char(count,clrch)
register count;
register clrch;
{
register itmp = count;
	while(itmp--)
		fputc(clrch,se);
	itmp = count;
	while(itmp--)
		fputs(tcap_curleft,se);

}	/* end of tcap_clear_area_char */

/*+-------------------------------------------------------------------------
 numchars = tcap_gets(buf,bufsize,&delim,wait_for_key)

This procedure reads a string and returns the number
of characters read.  -1 is returned if an abort is signaled by the
keyboard user.
--------------------------------------------------------------------------*/
int
tcap_gets(buf,bufsize,delim,wait_for_key)
char *buf;
register bufsize;	/* includes room for null..field is 1 less */
register uchar *delim;
int wait_for_key;
{
register count;
register itmp;
register char *cptr = buf;

	if(!wait_for_key)
		tcap_clear_area_char(bufsize-1,'_');

	count = 0;			/* no characters in string */
	*cptr = 0;			/* start with null string */

	while(1)
	{
		*delim = ttygetc(1);
		if((*delim < 0x20) || (*delim >= 0x7F))
		{
			switch(*delim)
			{
				case CR:
					*delim = NL;
				case NL:
					return(count);

				case BS:
					if(count)
					{
						count--;
						*--cptr = 0;
						fputc(0x08,se);
						fputc('_',se);
						fputs(tcap_curleft,se);
					}
					continue;

				case ESC:
				case CTL_U:
					itmp = count;
					while(itmp--)
					{
						*--cptr = 0;
						fputc(0x08,se);
					}
					itmp = count;
					while(itmp--)
						fputc('_',se);
					itmp = count;
					while(itmp--)
						fputs(tcap_curleft,se);
					cptr = buf;
					count = 0;
					if(*delim == ESC)
						return(-1);
					continue;

				default:
					return(-1);
			}	/* end of switch(*delim) */
			/*NOTREACHED*/
		}		/* end of if read delimiter */

		if(count == bufsize-1)
			continue;
		if(wait_for_key)
		{
			tcap_clear_area_char(bufsize-1,'_');
			wait_for_key = 0;
		}
		fputc(*delim,se);
		*cptr++ = *delim;
		*cptr = 0;
		count++;
	}	/* end of while can get character */
					
}	/* end of tcap_gets */


/* end of ecutcap.c */
/* vi: set tabstop=4 shiftwidth=4: */
