/* CHK=0xFD81 */
/*+-------------------------------------------------------------------------
	ecutty.c
	wht%n4hgf@gatech.edu

  Defined functions:
	_setcolor(clrs)
	color_name_to_num(cname)
	get_ttymode()
	get_ttyname()
	ring_bell()
	setcolor(new_colors)
	setcolor_internal(ntokens,tokens)
	ttyflush(flush_type)
	ttygetc(xkey_ok)
	ttygets(str,maxsize,flags)
	ttygets_esd(tesd,flags,append_flag)
	ttyinit()
	ttymode(arg)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:23-wht@n4hgf---- 2.80 --- */
/*:04-19-1990-03:07-wht@n4hgf-GCC run found unused vars -- rm them */
/*:04-17-1990-14:48-wht@n4hgf-fix bug in ttygets test against 0x7F */
/*:04-15-1990-16:52-wht@n4hgf-expanded ttygets */
/*:03-25-1990-14:11-wht@n4hgf------ x2.70 ------- */
/*:03-05-1990-04:02-wht@n4hgf-add UNIX /dev/syscon to get_ttyname */
/*:11-23-1989-03:11-wht-conditional compile for M_XENIX or M_UNIX */
/*:10-14-1989-21:10-wht-allow user to change ecu color choices */
/*:07-03-1989-22:57-wht------ x2.00 ----- */
/*:06-24-1989-16:53-wht-flush edits --- ecu 1.95 */

#include <sys/machdep.h>
#include "ecu.h"
#include "esd.h"
#include "ecukey.h"
#include "ecuxkey.h"
#include "ecuerror.h"

#define DEFINE_TTY_DATA
#include "ecutty.h"

static char *dole_out_rd_char = (char *)0;

extern int interrupt;

int current_ttymode = 0;

struct termio tty_termio_at_entry;
struct termio tty_termio_current;

char kbdeof;			/* current input EOF */
char kbdeol2;			/* current secondary input EOL */
char kbdeol;			/* current input EOL */
char kbderase;			/* current input ERASE */
char kbdintr;			/* current input INTR */
char kbdkill;			/* current input KILL */
char kbdquit;			/* current input QUIT */
int echo_erase_char;	/* save users ECHOE bit */
int echo_kill_char;		/* save users ECHOK bit */
char kbd_is_7bit;		/* keyboard has parity */

ulong colors_current = 0x04070A00L;
ulong colors_normal = 0x04070A00L;	/* lt_green/black red/white */
ulong colors_success = 0x07000A00L;	/* lt_green/black red/white */
ulong colors_alert = 0x0E000E00L;	/* yellow */
ulong colors_error = 0x04000400L;	/* red */
ulong colors_notify = 0x08000800L;	/* gray */

int use_colors = 0;		/* set by ttyinit, but default no */

extern char screen_dump_file_name[];

/*+-------------------------------------------------------------------------
	color_name_to_num(cname)
--------------------------------------------------------------------------*/
int
color_name_to_num(cname)
char *cname;
{
register COLOR *color = colors;
register itmp;

	while(color->name)
	{
		if((itmp = strcmp(color->name,cname)) > 0)
			return(-1);
		if(!itmp)
			return(color->num);
		color++;
	}
	return(-1);

}	/* end of color_name_to_num */

/*+-------------------------------------------------------------------------
	_setcolor(clrs)
--------------------------------------------------------------------------*/
void
_setcolor(clrs)
ulong clrs;
{
	if(!use_colors)
		return;

	/* normal */
	ff(se,"\033[=%ldF\033[=%ldG",(clrs >> 8) & 0xFF,clrs & 0xFF);

	/* reverse */
	ff(se,"\033[=%ldH\033[=%ldI",(clrs >> 24) & 0xFF,(clrs >> 16) & 0xFF);

	colors_current = clrs;

}	/* end of _setcolor */

/*+-------------------------------------------------------------------------
	setcolor(new_colors)

requires termcap init to have been done
--------------------------------------------------------------------------*/
void
setcolor(new_colors)
ulong new_colors;
{
	if(!use_colors)
	{
		if(new_colors != colors_normal)
			stand_out();
		else
			stand_end();
		return;
	}

	_setcolor(new_colors);
	stand_end();

}	/* end of setcolor */

/*+-------------------------------------------------------------------------
	setcolor_internal(ntokens,tokens)

returns 0 on success, else token number in error + 1
--------------------------------------------------------------------------*/
int
setcolor_internal(ntokens,tokens)
int ntokens;
char **tokens;
{
ulong fgnd;
ulong bgnd;

	if(ntokens == 2)
		tokens[2] = "black";

	if((fgnd = (ulong)color_name_to_num(tokens[1])) > 15)
		return(2);
	if((bgnd = (ulong)color_name_to_num(tokens[2])) > 15) 
		return(3);

	if(!strcmp(tokens[0],"normal"))
	{
		colors_normal &= 0xFFFF0000L;
		colors_normal |= (fgnd << 8) | bgnd;
		setcolor(colors_normal);
	}
	else if(!strcmp(tokens[0],"reverse"))
	{
		colors_normal &= 0x0000FFFFL;
		colors_normal |= (fgnd << 24) | (bgnd << 16);
		setcolor(colors_normal);
	}
	else if(!strcmp(tokens[0],"notify"))
		colors_notify = (fgnd << 24) | (bgnd << 16) | (fgnd << 8) | bgnd;
	else if(!strcmp(tokens[0],"success"))
		colors_success = (fgnd << 24) | (bgnd << 16) | (fgnd << 8) | bgnd;
	else if(!strcmp(tokens[0],"alert"))
		colors_alert = (fgnd << 24) | (bgnd << 16) | (fgnd << 8) | bgnd;
	else if(!strcmp(tokens[0],"error"))
		colors_error = (fgnd << 24) | (bgnd << 16) | (fgnd << 8) | bgnd;
	else
		return(1);

	return(0);

}	/* end of setcolor_internal */

/*+-------------------------------------------------------------------------
	read_colors_file()
--------------------------------------------------------------------------*/
void
read_colors_file()
{
FILE *fp;
char s128[128];
#define MAX_COLOR_TOKENS 6
char *tokens[MAX_COLOR_TOKENS];
int ntokens;
char *cptr;
int itmp;

	get_home_dir(s128);
	strcat(s128,"/.ecu/colors");
	if(access(s128,4))
		return;

	fp = fopen(s128,"r");

	while(fgets(s128,sizeof(s128),fp) != NULL)
	{
		if(s128[0] == '#')			/* comment? */
			continue;
		if(itmp = strlen(s128))		/* itmp = len; if > 0 ... */
		{
			itmp--;
			s128[itmp] = 0;			/* ... strip trailing NL */
		}
		cptr = s128;				/* first call to str_token, -> buff */
		while((*cptr == 0x20) || (*cptr == TAB))
			cptr++;				/* strip leading spaces */
		if(*cptr == 0)				/* if line all blank, skip it */
			continue;

		build_str_array(s128,tokens,MAX_COLOR_TOKENS,&ntokens);
		if(ntokens < 2)
			continue;

		setcolor_internal(ntokens,tokens);

	}			/* while records left to ready */
	fclose(fp);
}	/* end of read_colors_file */

/*+-------------------------------------------------------------------------
	ring_bell()
--------------------------------------------------------------------------*/
void
ring_bell()
{
	fputc(7,se);
}	/* end of ring_bell */

/*+-------------------------------------------------------------------------
	ttyinit()
--------------------------------------------------------------------------*/
void
ttyinit()
{
int monitor_type;

	/* save initial tty state */
	ioctl(TTYIN,TCGETA,(char *)&tty_termio_at_entry);

	interrupt = 0;			/* see xmtr signal handlers */

	kbdintr =  (tty_termio_at_entry.c_cc[VINTR])
		? (tty_termio_at_entry.c_cc[VINTR]  & 0x7F) : '\377';
	kbdquit =  (tty_termio_at_entry.c_cc[VQUIT])
		? (tty_termio_at_entry.c_cc[VQUIT]  & 0x7F) : '\377';
	kbderase = (tty_termio_at_entry.c_cc[VERASE])
		? (tty_termio_at_entry.c_cc[VERASE] & 0x7F) : '\377';
	kbdkill =  (tty_termio_at_entry.c_cc[VKILL])
		? (tty_termio_at_entry.c_cc[VKILL]  & 0x7F) : '\377';
	kbdeof =   (tty_termio_at_entry.c_cc[VEOF])
		? (tty_termio_at_entry.c_cc[VEOF]   & 0x7F) : '\04';
	kbdeol2 =  (tty_termio_at_entry.c_cc[VEOL])
		? (tty_termio_at_entry.c_cc[VEOL]   & 0x7F) : '\377';
	kbdeol =   (tty_termio_at_entry.c_iflag & ICRNL)
		? '\r' : '\n';

	kbd_is_7bit = ((tty_termio_at_entry.c_cflag & PARENB) != 0);
	if(kbd_is_7bit)
	{
		ff(se,"keyboard has parity enabled\n");
		nap(2000L);
	}
	echo_erase_char = tty_termio_at_entry.c_lflag & ECHOE;
	echo_kill_char = tty_termio_at_entry.c_lflag & ECHOK;
	tty_termio_current = tty_termio_at_entry;
	current_ttymode = 0;

	get_home_dir(screen_dump_file_name);
	strcat(screen_dump_file_name,"/.ecu/screen.dump");

	read_termcap();			/* get cursor,standout/end strings */

	ioctl(TTYIN,CONS_GET,&monitor_type);
	if(use_colors = (monitor_type != MONO))
	{
		read_colors_file();
		setcolor(colors_normal);
	}

}	/* end of ttyinit */

/*+-----------------------------------------------------------------------
	ttymode(arg) -- control user console (kbd/screen)

  Where arg ==
	0 restore attributes saved at start of execution
	1 raw mode (send xon/xoff, but do not respond to it, no ISIG/SIGINT)
	2 raw mode (same as 1 but allow keyboard interrupts)
	3 attributes at start of execution, but with echo disabled and no parity

------------------------------------------------------------------------*/
void
ttymode(arg)
{
	if(arg == 0)
	{
		ioctl(TTYIN,TCSETAW,(char *)&tty_termio_at_entry);
		tty_termio_current = tty_termio_at_entry;
		current_ttymode = 0;
	}
	else if((arg == 1) || (arg == 2))
	{
		tty_termio_current = tty_termio_at_entry;

		tty_termio_current.c_cflag &= ~(PARENB | PARODD);
		tty_termio_current.c_cflag |= CS8;

		/* don't want to honor tty xon/xoff, but pass to other end */
		tty_termio_current.c_iflag &=
			~(INLCR | ICRNL | IGNCR | IXON | IUCLC | ISTRIP);
		tty_termio_current.c_iflag |= IXOFF;	/* this end will xon/xoff */

		tty_termio_current.c_oflag |= OPOST;
		tty_termio_current.c_oflag &= ~(OLCUC | ONLCR | OCRNL | ONOCR | ONLRET);

		tty_termio_current.c_lflag &= ~(ICANON | ISIG | ECHO);
		if(arg == 2)
			tty_termio_current.c_lflag |= ISIG;

		tty_termio_current.c_cc[VMIN] = 1;
		tty_termio_current.c_cc[VTIME] = 0;

		ioctl(TTYIN,TCSETAW,(char *)&tty_termio_current);
		current_ttymode = arg;
	}
	else if(arg == 3)
	{
		tty_termio_current = tty_termio_at_entry;
		tty_termio_current.c_cflag &= ~(PARENB | PARODD);
		tty_termio_current.c_cflag |= CS8;
		tty_termio_current.c_iflag &= ~(ISTRIP);
		tty_termio_current.c_lflag &= ~(ICANON | ISIG | ECHO);
		ioctl(TTYIN,TCSETAW,(char *)&tty_termio_current);
		current_ttymode = 3;
	}
}	/* end of ttymode */

/*+-------------------------------------------------------------------------
	int	get_ttymode()
--------------------------------------------------------------------------*/
int
get_ttymode()
{
	return(current_ttymode);
}	/* end of get_ttymode */

/*+-----------------------------------------------------------------------
	ttyflush(flush_type) -- flush tty driver input &/or output buffers

0 == input buffer
1 == output buffer
2 == both buffers
------------------------------------------------------------------------*/
void
ttyflush(flush_type)
int flush_type;
{
	ioctl(TTYIN,TCXONC,(char *)0); /* stop tty output */
	switch(flush_type)	/* avoid compiler warning of cast int to far ptr */
	{
		case 0:
			ioctl(TTYIN,TCFLSH,(char *)0); break;
		case 1:
			ioctl(TTYIN,TCFLSH,(char *)1); break;
		case 2:
			ioctl(TTYIN,TCFLSH,(char *)2); break;
	}
	ioctl(TTYIN,TCXONC,(char *)1); /* restart tty output */
#if defined(M_XENIX) || defined(M_UNIX)
	dole_out_rd_char = (char *)0;	/* see ttygetc() */
#endif
	return;
}	/* end of ttyflush */

/*+-------------------------------------------------------------------------
	ttygetc(xkey_ok) -- get a key from the keyboard
if XENIX, map extended keys to sign-bit-set special value
if xkey_ok is 0, disallow extended keys
--------------------------------------------------------------------------*/
uint
ttygetc(xkey_ok)
int xkey_ok;
{
uchar ctmp;
extern int errno;
register uint itmp = 0;
static uchar rd_char[10];

	if(dole_out_rd_char)		/* handle (very unlikely) FAST typist */
	{
		if(itmp = *dole_out_rd_char++)
			return(itmp);
		else
			dole_out_rd_char = (char *)0;
	}

GET_KEY:
	errno = 0;
	if(read(TTYIN,&ctmp,1) < 0)
	{
		if(errno == EINTR)
			goto GET_KEY;
		perror_errmsg("keyboard");
		hangup(HANGUP_TTYIN_READ_ERROR);
	}

	if(kbd_is_7bit)
		ctmp &= 0x7F;

	if(ctmp == ESC)	/* if escape */
	{
		itmp = 0;
		nap(60L);
		while((!isalpha(ctmp)) && (itmp < sizeof(rd_char) - 1))
		{
			if(rdchk(0) <= 0)
				break;
			read(TTYIN,&ctmp,1);
			if(kbd_is_7bit)
				ctmp &= 0x7F;
			rd_char[itmp++] = ctmp;
			nap((long)20);
		}
		rd_char[itmp] = 0;
		if(!itmp)				/* no subsequent chars, so ... */
			return(ESC);		/* return the escape */
		else if((itmp == 2) && (rd_char[0] == '['))
		{
			switch(rd_char[1] | 0x80)
			{
				case XFcur5:
					screen_dump(screen_dump_file_name);
					goto GET_KEY;
				case XFcurup: case XFcurdn: case XFcurrt: case XFcurlf:
				case XFend: case XFpgdn: case XFhome: case XFpgup: case XFins:
				case XF1: case XF2: case XF3: case XF4: case XF5: case XF6:
				case XF7: case XF8: case XF9: case XF10: case XF11: case XF12:
				case XFbktab:
					if(xkey_ok)
						return(rd_char[1] | 0x80);
					/* fall thru -- xkey not allowed */
				default:
					ring_bell();
					goto GET_KEY;
			}
			/*NOTREACHED*/
		}
		/* not func key -- we have a FAST typist */
		dole_out_rd_char = rd_char;
		return(ESC);
	}
	return(ctmp);
}	/* end if ttygetc */

/*+-----------------------------------------------------------------------
	ttygets(str,maxsize,flags)

flags & 1 - echo cr/lf terminator
flags & 2 - extended delimiter set (Home, End, PgUp, PgDn, CurUp, CurDn)
flags & 4 - redisplay/edit current string
------------------------------------------------------------------------*/
void
ttygets(str,maxsize,flags)
register char *str;
int maxsize;
int flags;
{
register inch;
register strcount = 0;
register strpos = 0;
int insert_mode = 0;

	--maxsize;		/* decrement for safety */

	if(flags & 4)
	{
		strpos = strcount = strlen(str);
		fputs(str,se);
	}

	while(1)
	{
		inch = ttygetc(1);
		if((inch == kbdintr) || (inch == ESC))
		{
			curright(strcount - strpos);
			while(strcount)
			{
				fputc(BS,se);
				fputc(SPACE,se);
				fputc(BS,se);
				strcount--;
			}
			*str       = ESC;
			*(str + 1) = 0;
			return;
		}
		else if(inch == kbdkill)
		{
			curright(strcount - strpos);
			while(strcount)
			{
				fputc(BS,se);
				fputc(SPACE,se);
				fputc(BS,se);
				strcount--;
			}
			strpos = 0;
			*str = 0;
			continue;
		}
		else if(inch == kbderase)
		{
			if(strcount)
			{
				if(strcount == strpos)
				{
					fputc(BS,se);
					fputc(SPACE,se);
					fputc(BS,se);
					strcount--,strpos--;
				}
				else
				{
					if(!strpos)
						continue;
					mem_cpy(str + strpos - 1,str + strpos,strcount - strpos);
					fputc(BS,se);
					str[--strcount] = 0;
					strpos--;
					fputs(str + strpos,se);
					fputc(' ',se);
					curleft(strcount - strpos + 1);
				}
			}
			str[strcount] = 0;
			continue;
		}
		else if(inch == XFins)
		{
			insert_mode = !insert_mode;
			continue;
		}
		else if(inch == XFcurlf)
		{
			if(strpos)
			{
				strpos--;
				curleft(1);
			}
			continue;
		}
		else if(inch == XFcurrt)
		{
			if(strpos < strcount)
			{
				strpos++;
				curright(1);
			}
			continue;
		}

		if(flags & 2)	/* extended delimiter */
		{
			switch(inch)
			{
				case XFhome:
				case XFend:
				case XFpgup:
				case XFpgdn:
				case XFcurup:
				case XFcurdn:
					curright(strcount - strpos);
					while(strcount)
					{
						fputc(BS,se);
						fputc(SPACE,se);
						fputc(BS,se);
						strcount--;
					}
					*str       = inch;
					*(str + 1) = 0;
					return;
			}
		}

		switch(inch)
		{
			case CR:
			case NL:
				str[strcount] = 0;
				curright(strcount - strpos);
				if((flags & 1))
					ff(se,"\r\n");
				return;

			case CTL_L:
			case CTL_R:
				curright(strcount - strpos);
				ff(se,"%s (insert mode %s)\r\n",make_char_graphic(inch,0),
					(insert_mode) ? "ON" : "OFF");
				eeol();
				fputs(str,se);
				curleft(strcount - strpos);
				break;

			default:
				if((inch < SPACE) || (inch >= 0x7F))
				{
					fputc('\7',se);
					break;
				}
				if(strpos == strcount)
				{
					if(strcount == maxsize)
					{
						fputc('\7',se);
						break;
					}
					str[strcount++] = inch & 0x7F;
					strpos++;
					fputc(inch,se);
				}
				else
				{
					if(insert_mode)
					{
						if(strcount == maxsize)
						{
							fputc('\7',se);
							break;
						}
						mem_cpy(str+strpos+1,str+strpos,strcount-strpos);
						str[strpos] = inch;
						strcount++;
						str[strcount] = 0;
						fputs(str + strpos++,se);
						curleft(strcount - strpos);
					}
					else
					{
						str[strpos++] = inch;
						fputc(inch,se);
					}
				}
				str[strcount] = 0;
				break;
		}
	}			/* end of while we have room left in string */

}	/* end of ttygets() */

/*+-------------------------------------------------------------------------
	ttygets_esd(tesd,flags,append_flag)
--------------------------------------------------------------------------*/
ttygets_esd(tesd,flags,append_flag)
ESD *tesd;
int flags;
int append_flag;
{
char *pb = tesd->pb;
int maxcb = tesd->maxcb;

	if(append_flag)
	{
		pb += tesd->cb;
		maxcb -= tesd->cb;
	}
	else
	{
		pb = tesd->pb;
		maxcb = tesd->maxcb;
		tesd->cb = 0;
	}

	ttygets(pb,maxcb,flags);

	if(*pb == ESC)
	{
		if(!append_flag)
			zero_esd(tesd);
		return(eProcAttn_ESCAPE);
	}

	tesd->cb = strlen(tesd->pb);
	plogs(pb);
	if(flags & 1)
		plogc(NL);
	return(0);

}	/* end of ttygets_esd */

/*+-------------------------------------------------------------------------
	char *get_ttyname() - return pointer to static string
--------------------------------------------------------------------------*/
char *
get_ttyname()
{
static char ttname[64];
#if defined(pyr)
char *ttyname();
	strcpy(ttname,ttyname(0));
	return(ttname);
#endif
#if defined(M_XENIX) || defined(M_UNIX)
register unsigned int rdev;
register char *cptr;
struct stat fst;

	fstat(0,&fst);
	rdev = (unsigned)fst.st_rdev;
	if(rdev == 0x0301)
		strcpy(ttname,"/dev/console");
#if defined(M_UNIX)
	else if(rdev == 0x0000)
		strcpy(ttname,"/dev/syscon");
#endif
	else
	{
		strcpy(ttname,"/dev/tty");
		cptr = ttname + 8;

		if(rdev < 0x000C)
		{
			*cptr++ = '0' + ((rdev + 1) / 10);
			*cptr++ = '0' + ((rdev + 1) % 10);
		}
		else if(!(rdev & ~0x58F))
		{
			*cptr++ = (rdev & 0x0008) ? '2' : '1';
			*cptr++ = ((rdev & 0x0080) ? 'A' : 'a') + (rdev & 0x0007);
		}
		else
		{
			*cptr++ = '?';
			*cptr++ = '?';
		}
		*cptr = 0;
	}

	return(ttname);
#endif	
}	/* end of get_ttyname */

/* end of ecutty.c */
/* vi: set tabstop=4 shiftwidth=4: */
