/* CHK=0xDA53 */
/*+-------------------------------------------------------------------------
	ecuungetty.c - ecu "ungetty" program
	wht%n4hgf@gatech.edu

ecuungetty /dev/ttyxx <bamboozle-str>
ecuungetty -t /dev/ttyxx <bamboozle-str>
ecuungetty -r /dev/ttyxx <bamboozle-str>

  Defined functions:
	ecu_log_event(logstr)
	errno_text(errnum)
	hangup()
	main(argc,argv,envp)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:03-27-1990-14:05-wht@n4hgf-retry kill SIGUSR2 up to five times */
/*:03-25-1990-14:12-wht@n4hgf------- ecu x2.70 -------- */
/*:11-24-1989-21:18-wht-use common utmpstat.c - UNIX upgrade */
/*:07-03-1989-22:58-wht------ x2.00 ----- */
/*:06-29-1989-17:19-wht-creation */

#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <utmp.h>
#include "../stdio_lint.h"
#include "../ecuungetty.h"
#include "../utmpstatus.h"

extern struct utmp last_utmp;

/*+-------------------------------------------------------------------------
	ecu_log_event(logstr) - dummy procedure for utmpstat.o
--------------------------------------------------------------------------*/
void
ecu_log_event(logstr)
char *logstr;
{
	;
}	/* end of ecu_log_event */

/*+-------------------------------------------------------------------------
	hangup() - "dummy" for utmpstat.c
This particular incantation will only be called if utmp is non-existent
or not readable.......
--------------------------------------------------------------------------*/
hangup()
{
	exit(UGE_NOTROOT);
}	/* end of hangup */

/*+-------------------------------------------------------------------------
	errno_text(errnum)
--------------------------------------------------------------------------*/
char *
errno_text(errnum)
int errnum;
{
static char errstr[12];
	sprintf(errstr,"E%d\n",errnum);
	return(errstr);
}	/* end of errno_text */

/*+-------------------------------------------------------------------------
	main(argc,argv,envp)
--------------------------------------------------------------------------*/
main(argc,argv,envp)
int argc;
char **argv;
char **envp;
{
int op = 'g';	/* assume "get" operation */
int status;
int itmp;
char *tty = argv[1];
char *bamboozlement = argv[2];
char *bamboozle();

	if(geteuid() != 0)
		exit(UGE_NOTROOT);

	if(*argv[1] == '-')
	{
		switch(op = *(argv[1] + 1))
		{
			case 'r':
			case 't':
				break;
			default:
				exit(UGE_BADSWITCH);
		}
		if(argc < 3)
			exit(UGE_BADARGC);
		tty = argv[2];
		bamboozlement = argv[3];
	}
	else if(argc <= 2)
		exit(UGE_BADARGC);

	if(getuid() != 0)
	{
		if(strcmp(bamboozlement,bamboozle(getppid())))
			exit(UGE_CALLER);
	}

	status = utmp_status(tty);

	switch(op)
	{
		case 'g':
			switch(status)
			{
				case US_NOTFOUND:	/* not in utmp, or getty dead */
					exit(UG_NOTENAB);
				case US_LOGIN:		/* enabled for login, idle */
					kill(last_utmp.ut_pid,SIGUSR1);
					nap(200L);
					exit(UG_ENAB);
				case US_DIALOUT:	/* enabled for login, currently dialout */
				case US_LOGGEDIN:	/* enabled for login, in use */
					exit(UG_FAIL);
			}
			break;

		case 't':
			switch(status)
			{
				case US_NOTFOUND:	/* not in utmp, or getty dead */
					exit(UGE_T_NOTFOUND);
				case US_LOGIN:		/* enabled for login, idle */
					exit(UGE_T_LOGIN);
				case US_LOGGEDIN:	/* enabled for login, in use */
					exit(UGE_T_LOGGEDIN);
				case US_DIALOUT:	/* enabled for login, currently dialout */
					exit(UG_RESTART);
			}
			break;

		case 'r':
			switch(status)
			{
				case US_NOTFOUND:	/* not in utmp, or getty dead */
				case US_LOGIN:		/* enabled for login, idle */
				case US_LOGGEDIN:	/* enabled for login, in use */
					exit(0);
				case US_DIALOUT:	/* enabled for login, currently dialout */
					itmp = 5;
					while(itmp--)
					{
						if(kill(last_utmp.ut_pid,SIGUSR2))
							break;
						nap(100L);
					}
					exit(0);
			}
			break;
	}
	exit(0);
}	/* end of main */

/* vi: set tabstop=4 shiftwidth=4: */
/* end of ecuungetty.c */
