/* CHK=0xC7DA */
/*+-----------------------------------------------------------------------
	ecuutil.c -- utility routines for extended calling unit
	wht%n4hgf@gatech.edu

  Defined functions:
	arg_token(parsestr,termchars)
	ascii_name_to_hex(str3char)
	ascii_to_hex(ascii)
	build_arg_array(cmd,arg,arg_max_quan,narg_rtn)
	build_str_array(str,arg,str_max_quan,nstr_rtn)
	disp_line_termio(fd,text)
	disp_termio(ttt,text)
	errno_text(errno)
	find_shell_chars(command)
	get_curr_dir(cdir,cdir_max)
	get_home_dir(home_dir)
	hex_to_ascii_name(char_val)
	make_char_graphic(ch,incl_3char)
	make_dirs(pathname)
	make_ecu_subdir()
	mem_cpy(dest,src,len);
	mkdir(dpath,dmode)
	mode_map(mode,mode_str)
	pad_zstr_to_len(zstr,len)
	perror_errmsg(str)
	print_cwd(curdir,buf_size)
	skip_ld_break(zstr)
	str_token(parsestr,termchars)
	yes_or_no(strarg)

------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:24-wht@n4hgf---- 2.80 --- */
/*:05-02-1990-13:25-wht@n4hgf-added double quote to shell metachar list */
/*:04-19-1990-14:56-wht@n4hgf-my GCC is uses XENIX libs/turn on mkdir for it */
/*:04-19-1990-03:07-wht@n4hgf-GCC run found unused vars -- rm them */
/*:04-16-1990-00:52-wht@n4hgf-add mem_cpy */
/*:04-02-1990-02:17-wht@n4hgf-Dialers support appears to work */
/*:03-27-1990-15:18-wht@n4hgf-overhaul for/wait/SIGCLD handling */
/*:03-25-1990-14:11-wht@n4hgf------ x2.70 ------- */
/*:02-28-1990-17:21-wht@n4hgf-chmod new ecu dir to 700 */
/*:12-21-1989-17:01-wht-add CTSFLOW/RTSFLOW to disp_termio */
/*:12-09-1989-22:54-wht-found bug in arg_token when using it elsewhere */
/*:11-23-1989-03:11-wht-conditional compile for M_XENIX or M_UNIX */
/*:11-23-1989-02:47-wht-UNIX has a mkdir system call */
/*:11-19-1989-16:10-wht-add errno_text */
/*:11-02-1989-06:59-wht-use getpwuid for home dir */
/*:10-07-1989-21:11-wht-fix arg_token+fix doc on add_token & str_token */
/*:07-29-1989-14:39-wht-add CS5,6,7, 8 to disp_termio */
/*:07-03-1989-22:57-wht------ x2.00 ----- */
/*:06-24-1989-16:53-wht-flush edits --- ecu 1.95 */

#include "ecu.h"
#include "ecuhangup.h"
#include "ecufork.h"
#include <pwd.h>

char *getenv();

extern int errno;
extern int rcvr_pid;		/* ==0 if rcvr process, else pid of rcvr */
extern char curr_dir[];

char *ascii_ctlstr =
"NULSOHSTXETXEOTENQACKBELBS HT NL VT FF CR SO SI DLEDC1DC2DC3DC4NAKSYNETBCANEM SUBESCFS GS RS US ";

/*+-------------------------------------------------------------------------
	mem_cpy(dest,src,len) - memcpy() with non-destructive overlapping copy
--------------------------------------------------------------------------*/
void
mem_cpy(dest,src,len)
register char *dest;
register char *src;
register len;
{
	if(dest > src)
	{
		dest += len;
		src += len;
		while(len--)
			*--dest = *--src;
	}
	else
	{
		while(len--)
			*dest++ = *src++;
	}
}	/* end of mem_cpy */

/*+-------------------------------------------------------------------------
	errno_text(errno)
--------------------------------------------------------------------------*/
char *
errno_text(errno)
int errno;
{
static char errant[16];

	switch(errno)
	{
		case 0: return("0");
		case EPERM: return("EPERM");
		case ENOENT: return("ENOENT");
		case ESRCH: return("ESRCH");
		case EINTR: return("EINTR");
		case EIO: return("EIO");
		case ENXIO: return("ENXIO");
		case E2BIG: return("E2BIG");
		case ENOEXEC: return("ENOEXEC");
		case EBADF: return("EBADF");
		case ECHILD: return("ECHILD");
		case EAGAIN: return("EAGAIN");
		case ENOMEM: return("ENOMEM");
		case EACCES: return("EACCES");
		case EFAULT: return("EFAULT");
		case ENOTBLK: return("ENOTBLK");
		case EBUSY: return("EBUSY");
		case EEXIST: return("EEXIST");
		case EXDEV: return("EXDEV");
		case ENODEV: return("ENODEV");
		case ENOTDIR: return("ENOTDIR");
		case EISDIR: return("EISDIR");
		case EINVAL: return("EINVAL");
		case ENFILE: return("ENFILE");
		case EMFILE: return("EMFILE");
		case ENOTTY: return("ENOTTY");
		case ETXTBSY: return("ETXTBSY");
		case EFBIG: return("EFBIG");
		case ENOSPC: return("ENOSPC");
		case ESPIPE: return("ESPIPE");
		case EROFS: return("EROFS");
		case EMLINK: return("EMLINK");
		case EPIPE: return("EPIPE");
		case EDOM: return("EDOM");
		case ERANGE: return("ERANGE");
		default:
			sprintf(errant,"E%04u",errno);
			return(errant);
	}

}	/* end of errno_text */

/*+-------------------------------------------------------------------------
	skip_ld_break(zstr) - skip leading spaces and tabs
--------------------------------------------------------------------------*/
char *
skip_ld_break(zstr)
register char *zstr;
{
	while(isspace(*zstr))
		zstr++;
	return(zstr);
}	/* end of skip_ld_break */

/*+-----------------------------------------------------------------------
	pad_zstr_to_len(zstr,len)

  pads with spaces to specified length, unless already longer than
  len in which case the string is truncated to 'len' characters.
------------------------------------------------------------------------*/
void
pad_zstr_to_len(zstr,len)
char *zstr;
int len;
{
register izstr;

	izstr = strlen(zstr);
	if(izstr >= len)
		zstr[len] = 0;
	else
	{
		while(izstr < len)
			zstr[izstr++] = 0x20;
		zstr[izstr] = 0;
	}
}	/* end of pad_zstr_to_len */

/*+-----------------------------------------------------------------------
	arg_token(parsestr,termchars)

Get next token from string parsestr ((char *)0 on 2nd, 3rd, etc.
calls), where tokens are nonempty strings separated by runs of chars
from termchars.  Writes nulls into parsestr to end tokens.
termchars need not remain constant from call to call.

Treats multiple occurrences of a termchar as one delimiter (does not
allow null fields).
------------------------------------------------------------------------*/
char *arg_token_static = (char *)0;
char *arg_token(parsestr,termchars)
char *parsestr;
char *termchars;
{
register int first = 1;
register char *termptr;
register char *parseptr;
char *token;

	if(parsestr == (char *)0 && arg_token_static == (char *)0)
		return((char *)0);

	if(parsestr)
		parseptr = parsestr;
	else
       parseptr = arg_token_static;

	while(*parseptr)
	{
		if(!strchr(termchars,*parseptr))
			break;
		parseptr++;
	}

	if(!*parseptr)
	{
		arg_token_static = (char *)0;
		return((char *)0);
	}

	token = parseptr;
	if(*token == '\'')
	{
		token++;
		parseptr++;
		while(*parseptr)
		{
			if(*parseptr == '\'')
			{
				arg_token_static = parseptr + 1;
				*parseptr = 0;
				return(token);
			}
			parseptr++;
		}
		arg_token_static = (char *)0;
		return(token);
	}
	while(*parseptr)
	{
		if(strchr(termchars,*parseptr))
		{
			*parseptr = 0;
			arg_token_static = parseptr + 1;
			while(*arg_token_static)
			{
				if(!strchr(termchars,*arg_token_static))
					break;
				arg_token_static++;
			}
			return(token);
		}
		parseptr++;
	}
	arg_token_static = (char *)0;
	return(token);
}	/* end of arg_token */

/*+-------------------------------------------------------------------------
	build_arg_array(cmd,arg,arg_max_quan,&narg)
--------------------------------------------------------------------------*/
void
build_arg_array(cmd,arg,arg_max_quan,narg_rtn)
char *cmd;
char **arg;
int arg_max_quan;
int *narg_rtn;
{
register itmp;
register narg;

	for(itmp = 0; itmp < arg_max_quan; itmp++)
		arg[itmp] = (char *)0;
	arg[0] = arg_token(cmd," \t\r\n");

	for(narg = 1; narg < arg_max_quan; ++narg)
	{
		if((arg[narg] = arg_token((char *)0," \t\r\n")) == (char *)0) 
			break;
	}

	*narg_rtn = narg;

}	/* end of build_arg_array */

/*+-----------------------------------------------------------------------
	str_token(parsestr,termchars)

Get next token from string parsestr ((char *)0 on 2nd, 3rd, etc.
calls), where tokens are nonempty strings separated by runs of chars
from termchars.  Writes nulls into parsestr to end tokens.
termchars need not remain constant from call to call.

Treats each occurrence of a termchar as delimiter (allows null
fields).
------------------------------------------------------------------------*/
static char *str_token_static = (char *)0;
char *str_token(parsestr,termchars)
char *parsestr;
char *termchars;
{
register char *termptr;
register char *parseptr;
char *token;

	if(parsestr == (char *)0 && str_token_static == (char *)0)
		return((char *)0);

	if(parsestr)
		parseptr = parsestr;
	else
       parseptr = str_token_static;

	while(*parseptr)
	{
		for(termptr = termchars; *termptr != 0; termptr++)
		{
			if(*parseptr == *termptr)
				goto FOUND_TERM;
		}
		if(!*termptr)
			break;
		parseptr++;
	}

	if(!*parseptr)
	{
		str_token_static = (char *)0;
		return((char *)0);
	}

FOUND_TERM:
	token = parseptr;
	while(*parseptr)
	{
		for(termptr = termchars; *termptr;)
		{
			if(*parseptr == *termptr++)
			{
				str_token_static = parseptr + 1;
				*parseptr = 0;
				return(token);
			}
		}
		parseptr++;
	}
	str_token_static = (char *)0;
	return(token);
}	/* end of str_token */

/*+-------------------------------------------------------------------------
	build_str_array(str,arg,str_max_quan,&narg)
--------------------------------------------------------------------------*/
void
build_str_array(str,arg,str_max_quan,nstr_rtn)
char *str;
char **arg;
int str_max_quan;
int *nstr_rtn;
{
register itmp;
register narg;


	for(itmp = 0; itmp < str_max_quan; itmp++)
		arg[itmp] = (char *)0;
	arg[0] = str_token(str," \t\r\n");

	for(narg = 1; narg < str_max_quan; ++narg)
	{
		if((arg[narg] = str_token((char *)0," \t\r\n")) == (char *)0) 
			break;
	}

	*nstr_rtn = narg;

}	/* end of build_str_array */

/*+-----------------------------------------------------------------------
	make_char_graphic(character,incl_3char) - Make all chars "printable"

  returns pointer to a static string containing printable version
  of a character.  If control char, printed as "^A", etc.
  if incl_3char set true, then space + ASCII assignment (e.g. "NUL") is
  appended to the string for non-printable graphics
------------------------------------------------------------------------*/
char *make_char_graphic(ch,incl_3char)
register char ch;
int incl_3char;
{
static char gg[16];

	ch &= 0x7F;
	if((ch >= 0x20) && (ch < 0x7F))
	{
		gg[0] = ch; gg[1] = 0;
	}
	else
	{
		gg[0] = '^'; 
		if(ch == 0x7F)
		{
			gg[1] = '?';
			if(incl_3char)
				strcpy(&gg[2]," DEL");
		}
		else
		{
			gg[1] = ch + 0x40;
			if(incl_3char)
			{
				gg[2] = 0x20;
				strncpy(&gg[3],ascii_ctlstr + (ch * 3),3);
				gg[7] = 0;
			}
			else
				gg[2] = 0;
		}
	}
	return(gg);
}	/* end of make_char_graphic */

/*+-----------------------------------------------------------------------
	disp_termio(ttt)
  display termio 'ttt' on stderr
------------------------------------------------------------------------*/
void disp_termio(ttt,text)
struct termio *ttt;
char *text;
{
register flag;
register i_cc;
register char *cptr;
int dbits;
char parity;

	pprintf("---------> %s\n",text);

	flag = ttt->c_iflag;
	pprintf("iflag: %07o IGNBRK:%d  BRKINT:%d  IGNPAR:%d  PARMRK:%d  INPCK:%d  ISTRIP:%d\n",
				flag,
				(flag & IGNBRK) ? 1 : 0,
				(flag & BRKINT) ? 1 : 0,
				(flag & IGNPAR) ? 1 : 0,
				(flag & PARMRK) ? 1 : 0,
				(flag & INPCK ) ? 1 : 0,
				(flag & ISTRIP) ? 1 : 0);
	pprintf(
"               INLCR:%d  IGNCR:%d  ICRNL:%d  IUCLC:%d  IXON:%d  IXANY:%d  IXOFF:%d\n",
				(flag & INLCR ) ? 1 : 0,
				(flag & IGNCR ) ? 1 : 0,
				(flag & ICRNL ) ? 1 : 0,
				(flag & IUCLC ) ? 1 : 0,
				(flag & IXON  ) ? 1 : 0,
				(flag & IXANY ) ? 1 : 0,
				(flag & IXOFF ) ? 1 : 0);

	flag = ttt->c_oflag;
	pprintf("oflag: %07o OPOST:%d  OLCUC:%d  ONLCR:%d  OCRNL:%d  ONOCR:%d  ONLRET:%d  OFDEL:%d\n",
				flag,
				(flag & OPOST ) ? 1 : 0,
				(flag & OLCUC ) ? 1 : 0,
				(flag & ONLCR ) ? 1 : 0,
				(flag & OCRNL ) ? 1 : 0,
				(flag & ONOCR ) ? 1 : 0,
				(flag & ONLRET) ? 1 : 0,
				(flag & OFDEL ) ? 1 : 0);

	flag = ttt->c_cflag;
	pprintf("cflag: %07o ",ttt->c_cflag);
	switch(flag & CBAUD)
	{
		case B0:	cptr = "HUP"; break;
		case B50:	cptr = "50"; break;
		case B75:	cptr = "75"; break;
		case B110:	cptr = "110"; break;
		case B134:	cptr = "134.5"; break;
		case B150:	cptr = "150"; break;
		case B200:	cptr = "200"; break;
		case B300:	cptr = "300"; break;
		case B600:	cptr = "600"; break;
		case B1200:	cptr = "1200"; break;
		case B1800:	cptr = "1800"; break;
		case B2400:	cptr = "2400"; break;
		case B4800:	cptr = "4800"; break;
		case B9600:	cptr = "9600"; break;
#if defined(B19200)
		case B19200:cptr = "19200"; break;
#endif
#if defined(B38400)
		case B38400:cptr = "38400"; break;
#endif
		default:	
			switch(flag & CBAUD)
			{
				case EXTA:	cptr = "EXTA"; break;
				case EXTB:	cptr = "EXTB"; break;
				default:	cptr = "????"; break;
			}
	}
	dbits = 5 + ((flag & CSIZE) >> 4);
	parity = (flag & PARENB) ? ((flag & PARODD) ? 'O' : 'E') : 'N';
	pprintf("%s-%d-%c-%d ",cptr,dbits,parity,(flag & CSTOPB) ? 2 : 1);
	switch(flag & CS8)
	{
		case CS8: pputs("CS8 "); break;
		case CS7: pputs("CS7 "); break;
		case CS6: pputs("CS6 "); break;
		case CS5: pputs("CS5 "); break;
	}
	pprintf("CREAD:%d  HUPCL:%d  CLOCAL:%d",
				(flag & CREAD ) ? 1 : 0,
				(flag & HUPCL ) ? 1 : 0,
				(flag & CLOCAL) ? 1 : 0);
#ifdef RTSFLOW
	pprintf(" RTSFLOW:%d  CTSFLOW:%d",
				(flag & RTSFLOW ) ? 1 : 0,
				(flag & CTSFLOW ) ? 1 : 0);
#endif
	pprintf("\n");

	flag = ttt->c_lflag;
	pprintf("lflag: %07o ISIG:%d  ICANON:%d  XCASE:%d  ECHO:%d  ECHOE:%d\n",
				flag,
				(flag & ISIG  ) ? 1 : 0,
				(flag & ICANON) ? 1 : 0,
				(flag & XCASE ) ? 1 : 0,
				(flag & ECHO  ) ? 1 : 0,
				(flag & ECHOE ) ? 1 : 0);
	pprintf("            ECHOK:%d  ECHONL:%d  NOFLSH:%d",
				(flag & ECHOK ) ? 1 : 0,
				(flag & ECHONL) ? 1 : 0,
				(flag & NOFLSH) ? 1 : 0);

#if defined(XCLUDE)
	pprintf("  XCLUDE:%d",(flag & XCLUDE) ? 1 : 0);
#endif
	pputs("\n");

	pprintf("ctl chars: ");
	for(i_cc = 0; i_cc < NCC; i_cc++)
		pprintf("%02x   ",ttt->c_cc[i_cc]);
	pprintf("  (hex)\n");
	pprintf("           INTR QUIT ERAS KILL EOF  EOL  EOL2 SWTCH  VMIN-EOF VTIME-EOL\n");

}	/* end of disp_termio */

/*+-----------------------------------------------------------------------
	disp_line_termio(fd)

Get current termio structure for file descriptor fd
and display on stderr
------------------------------------------------------------------------*/
void disp_line_termio(fd,text)
int fd;		/* file descriptor */
char *text;
{
struct termio fd_termio;
char text2[128];

	ioctl(fd,TCGETA,&fd_termio);
	sprintf(text2,"fd: %d  %s",fd,text);
	disp_termio(&fd_termio,text2);

}	/* end of disp_line_termio */

/*+-----------------------------------------------------------------------
	print_cwd(curdir) print and return current working directory
------------------------------------------------------------------------*/
void print_cwd(curdir,buf_size)
char *curdir;
int buf_size;
{
	ff(se,"%s\r\n",curr_dir);
	strncpy(curdir,curr_dir,buf_size);
}	/* end of print_cwd */

/*+-----------------------------------------------------------------------
	dummy_call() bypass optimizer (call from another module for noop)
------------------------------------------------------------------------*/
void dummy_call() { ; }

/*+-----------------------------------------------------------------------
	ascii_name_to_hex(str3char)

  return value of ascii ctl char name (e.g., "NUL") 0 - 0x1F
  returns -1 if input not valid
------------------------------------------------------------------------*/
ascii_name_to_hex(str3char)
char *str3char;
{
register char *cptr = ascii_ctlstr;
register intval;

	for(intval = 0; intval < 0x20; intval++)
	{
		if(	(to_lower(*str3char) == to_lower(*cptr))				&&
			(to_lower(*(str3char + 1)) == to_lower(*(cptr + 1)))	&&
			((to_lower(*(str3char + 2)) == to_lower(*(cptr + 2)))	||
			(*(cptr + 2) == ' ')))
			return(intval);
		cptr += 3;
	}

	return(-1);

}	/* end of ascii_name_to_hex */

/*+-------------------------------------------------------------------------
	ascii_to_hex(ascii)
--------------------------------------------------------------------------*/
ascii_to_hex(ascii)
char *ascii;
{
int hexval;
	if(strlen(ascii) == 1)
		return(*ascii);
	else if(!strncmp(ascii,"0x",2))
	{
		sscanf(ascii + 2,"%x",&hexval);
		return(hexval & 0xFF);
	}
	else if(*ascii == '^')
		return(*(ascii + 1) & 0x1F);
	else
		return(ascii_name_to_hex(ascii));
}	/* end of ascii_to_hex */

/*+-------------------------------------------------------------------------
	hex_to_ascii_name(char_val)

  Returns pointer to static string containing three character ASCII
  name for control character followed by a null.
--------------------------------------------------------------------------*/
char *
hex_to_ascii_name(char_val)
char char_val;
{
static char ascii_name[4];

	if(char_val == 0x7F)
		strcpy(ascii_name,"DEL");
	else
	{
		strncpy(ascii_name,ascii_ctlstr + (char_val * 3),3);
		ascii_name[3] = 0;
	}

	return(ascii_name);

}	/* end of hex_to_ascii_name */

/*+-------------------------------------------------------------------------
	get_curr_dir(cdir,cdir_max) - get current directory into 'cdir'
--------------------------------------------------------------------------*/
get_curr_dir(cdir,cdir_max)
char *cdir;
int cdir_max;
{
#if defined(pyr)
char *getwd();
	return((getwd(cdir) == (char *)0) ? -1 : 0);
#endif

#if defined(M_XENIX) || defined(M_UNIX)
FILE *popen();
FILE *pipefp = popen("/bin/pwd","r");
	if(!pipefp)
		return(-1);
	fgets(cdir,cdir_max,pipefp);
	cdir[strlen(cdir) - 1] = 0;
	fclose(pipefp);
	return(0);
#endif

}	/* end of get_curr_dir */

/*+-----------------------------------------------------------------------
	get_home_dir(home_dir):  leave plenty of room for result!
------------------------------------------------------------------------*/
get_home_dir(home_dir)
char *home_dir;
{
static char home_directory[256] = "";
struct passwd *pwent;
struct passwd *getpwuid();

	if(home_directory[0])
	{
		strcpy(home_dir,home_directory);
		return(0);
	}

	if(!(pwent = getpwuid(getuid())))
	{
		pperror("cannot get pwent for you!!");
		hangup(HANGUP_PWENT_ERROR);
	}
	strcpy(home_directory,pwent->pw_dir);
	strcpy(home_dir,pwent->pw_dir);
	endpwent();
	return(0);

}	/* end of get_home_dir */

/*+-------------------------------------------------------------------------
	mkdir(dpath,dmode)
 Directory-creating routines from Public Domain TAR by John Gilmore
 Make a directory.  Compatible with the mkdir() system call on 4.2BSD.
--------------------------------------------------------------------------*/
#if !defined(M_UNIX) || defined(__GNUC__)
#define	TERM_SIGNAL(status)		((status) & 0x7F)
#define TERM_COREDUMP(status)	(((status) & 0x80) != 0)
#define TERM_VALUE(status)		((status) >> 8)
mkdir(dpath,dmode)
char *dpath;
int dmode;
{
	int cpid,status;
	struct stat statbuf;
	int (*original_sighdlr)();

	if(stat(dpath,&statbuf) == 0)
	{
		errno = EEXIST;		/* Stat worked,so it already exists */
		return(-1);
	}

	/* If stat fails for a reason other than non-existence,return error */
	if(errno != ENOENT)
		return(-1);

	original_sighdlr = signal(SIGCLD,SIG_DFL);
	switch(cpid = fork())
	{

	case -1:			/* Error in fork() */
		return(-1);		/* Errno is set already */

	case 0:				/* Child process */
		/*
		 * Cheap hack to set mode of new directory.  Since this
		 * child process is going away anyway,we zap its umask.
		 * FIXME,this won't suffice to set SUID,SGID,etc. on this
		 * directory.  Does anybody care?
		 */
		status = umask(0);	/* Get current umask */
		status = umask(status | (0777 & ~dmode)); /* Set for mkdir */
		execl("/bin/mkdir","mkdir",dpath,(char *)0);
		_exit(-1);		/* Can't exec /bin/mkdir */

	default:			/* Parent process */
		while((cpid != wait(&status)) && (cpid != -1))
			;	/* Wait for kid to finish */
	}

	signal(SIGCLD,original_sighdlr);

	if(TERM_SIGNAL(status) != 0 || TERM_VALUE(status) != 0)
	{
		errno = EIO;		/* We don't know why,but */
		return(-1);		/* /bin/mkdir failed */
	}

	return(0);
}	/* end of mkdir */
#endif

/*+-------------------------------------------------------------------------
	make_dirs(pathname)

  Directory-creating routines from Public Domain TAR by John Gilmore
  After a file/link/symlink/dir creation has failed, see if it's because
  some required directory was not present, and if so, create all
  required dirs.

  returns 0 if no directory made, else # levels required to get target
--------------------------------------------------------------------------*/
int
make_dirs(pathname)
register char *pathname;
{
	register char *p;			/* Points into path */
	int madeone = 0;			/* Did we do anything yet? */
	int save_errno = errno;		/* Remember caller's errno */

	if(errno != ENOENT)
		return(0);		/* Not our problem */

	for(p = strchr(pathname,'/'); p != NULL; p = strchr(p+1,'/'))
	{
		/* Avoid mkdir of empty string,if leading or double '/' */
		if(p == pathname || p[-1] == '/')
			continue;
		/* Avoid mkdir where last part of path is '.' */
		if(p[-1] == '.' && (p == pathname+1 || p[-2] == '/'))
			continue;
		*p = 0;				/* Truncate the path there */
		if( !mkdir(pathname,0777))
		{	/* Try to create it as a dir */
			madeone++;		/* Remember if we made one */
			*p = '/';
			continue;
		}
		*p = '/';
		if(errno == EEXIST)		/* Directory already exists */
			continue;
		/*
		 * Some other error in the mkdir.  We return to the caller.
		 */
		break;
	}
	errno = save_errno;		/* Restore caller's errno */
	return(madeone);		/* Tell them to retry if we made one */
}	/* end of make_dirs */

/*+-------------------------------------------------------------------------
	make_ecu_subdir()
must be called early in execution before wierd tty states set, etc.
--------------------------------------------------------------------------*/
void
make_ecu_subdir()
{
int itmp;
struct stat fst;
char s256[256];

	get_home_dir(s256);
	strcat(s256,"/.ecu");
	if((!(itmp = stat(s256,&fst))) && ((fst.st_mode & S_IFMT) != S_IFDIR))
	{
		ff(se,"~/.ecu is not a directory. Rename the file and try again.\n\n");
		exit(1);
	}
	if(itmp)		/* if stat failed, try to make the directory */
	{
		strcat(s256,"/.");
		if(make_dirs(s256))
		{
			if(stat(s256,&fst))
			{
				ff(se,"cannot make ~/.ecu subdirectory.\n");
				perror(s256);
				exit(1);
			}
		}
	}
	chmod(s256,0700);
}	/* end of make_ecu_subdir */

/*+-------------------------------------------------------------------------
	yes_or_no(strarg)
  Returns 1 if first char is 'Y' or 'y'
	or if strarg is numeric returns the numeric value
	or if strarg is alpha == "on" returns 1
  Returns 0 otherwise
--------------------------------------------------------------------------*/
yes_or_no(strarg)
char *strarg;
{
	if((strcmp(strarg,"on") == 0) || (to_lower(*strarg) == 'y'))
		return(1);
	else
		return(atoi(strarg));
}	/* end of yes_or_no */

/*+-------------------------------------------------------------------------
	find_shell_chars(command)
--------------------------------------------------------------------------*/
find_shell_chars(command)
char *command;
{
register schar;
register cchar;
register char *scptr;
static char shell_chars[] = "\"~;*?'`{}[]$";

	while(cchar = *command++)
	{
		scptr = shell_chars;
		while(schar = *scptr++)
			if(schar == cchar)
				return(1);
	}
	return(0);
}	/* end of find_shell_chars */

/*+-------------------------------------------------------------------------
	perror_errmsg(str)
--------------------------------------------------------------------------*/
void
perror_errmsg(str)
char *str;
{
extern char errmsg[];
extern char *sys_errlist[];
extern int sys_nerr;
extern int errno;

	if(errno > sys_nerr)
		sprintf(errmsg,"%s: error %d",str,errno);
	else
		sprintf(errmsg,"%s: %s",str,sys_errlist[errno]);
}	/* end of perror_errmsg */

/*+-----------------------------------------------------------------------
	mode_map(mode,mode_str)	build drwxrwxrwx string
------------------------------------------------------------------------*/
char *
mode_map(mode,mode_str)
unsigned short mode;
char *mode_str;
{
register unsigned ftype = mode & S_IFMT;
register char *rtn;
static char result[12];

	rtn = (mode_str == (char *)0) ? result : mode_str;

	/*          drwxrwxrwx */
	/*          0123456789 */
	strcpy(rtn,"----------");

	switch(ftype)
	{
		case S_IFIFO:	*rtn = 'p'; break; /* FIFO (named pipe) */
		case S_IFDIR:	*rtn = 'd'; break; /* directory */
		case S_IFCHR:	*rtn = 'c'; break; /* character special */
		case S_IFBLK:	*rtn = 'b'; break; /* block special */
		case S_IFREG:	*rtn = '-'; break; /* regular */

#if defined(pyr) | defined(BSD4)
		case S_IFLNK:	*rtn = 'l'; break; /* symbolic link */
		case S_IFSOCK:	*rtn = 's'; break; /* socket */
#endif

#if defined (M_XENIX) || defined(M_UNIX)
		case S_IFNAM:						/* name space entry */
			if(mode & S_INSEM)				/* semaphore */
			{
				*rtn = 's';
				break;
			}
			if(mode & S_INSHD)				/* shared memory */
			{
				*rtn = 'm';
				break;
			}
#endif

		default:		*rtn = '?'; break;	/* ??? */
	}

	if(mode & 000400) *(rtn + 1) = 'r';
	if(mode & 000200) *(rtn + 2) = 'w';
	if(mode & 000100) *(rtn + 3) = 'x';
	if(mode & 004000) *(rtn + 3) = 's';
	if(mode & 000040) *(rtn + 4) = 'r';
	if(mode & 000020) *(rtn + 5) = 'w';
	if(mode & 000010) *(rtn + 6) = 'x';
	if(mode & 002000) *(rtn + 6) = 's';
	if(mode & 000004) *(rtn + 7) = 'r';
	if(mode & 000002) *(rtn + 8) = 'w';
	if(mode & 000001) *(rtn + 9) = 'x';
	if(mode & 001000) *(rtn + 9) = 't';

	return(rtn);

}	/* end of mode_map */

/* end of ecuutil.c */
/* vi: set tabstop=4 shiftwidth=4: */
