/* CHK=0x920C */
/*+-------------------------------------------------------------------------
	ecuwinutil.c - curses window utilities
	wht%n4hgf@gatech.edu

  Defined functions:
	clear_area(win,y,x,len)
	clear_area_char(win,y,x,len,fillchar)
	winbox(win)
	window_create(title,title_x,tly,tlx,lines,cols)
	window_setup(win,title,title_x)
	windows_end(last_win)
	windows_end_signal()
	windows_start()
	winget_single(win,nondelim_list,delim_list)
	wingets(win,y,x,buf,bufsize,delim,edit)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:24-wht@n4hgf---- 2.80 --- */
/*:04-19-1990-03:07-wht@n4hgf-GCC run found unused vars -- rm them */
/*:04-16-1990-00:05-wht@n4hgf-smarter wingets */
/*:03-27-1990-15:18-wht@n4hgf-overhaul for/wait/SIGCLD handling */
/*:03-25-1990-14:11-wht@n4hgf------ x2.70 ------- */
/*:11-24-1989-14:41-wht-UNIX tcap-curses blinks with high bits set on addch */
/*:11-19-1989-17:17-wht-change windows_end_signal() to not goto bottom left */
/*:07-03-1989-22:57-wht------ x2.00 ----- */
/*:06-24-1989-16:53-wht-flush edits --- ecu 1.95 */

#include <curses.h>
#include "ecukey.h"
#include "ecuxkey.h"
#include "ecuhangup.h"
#include "pc_scr.h"

#if !defined(ushort)
#define ushort unsigned short
#endif
#if !defined(uchar)
#define uchar unsigned char
#endif
#if !defined(uint)
#define uint unsigned int
#endif
#if !defined(ulong)
#define ulong unsigned long
#endif

int windows_active = 0;
int ttymode_before_window_start;

/*+-------------------------------------------------------------------------
	clear_area_char(win,y,x,len,fillchar)
--------------------------------------------------------------------------*/
void
clear_area_char(win,y,x,len,fillchar)
WINDOW *win;
int y;
int x;
int len;
char fillchar;
{
	wmove(win,y,x);
	while(len-- > 0)
		waddch(win,fillchar & 0xFF);
	wmove(win,y,x);

}	/* end of clear_area_char */

/*+-------------------------------------------------------------------------
	clear_area(win,y,x,len)
--------------------------------------------------------------------------*/
void
clear_area(win,y,x,len)
WINDOW *win;
int y;
int x;
int len;
{
	clear_area_char(win,y,x,len,' ');
}	/* end of clear_area_char */

/*+-------------------------------------------------------------------------
	windows_start()
--------------------------------------------------------------------------*/
void
windows_start()
{

	ttymode_before_window_start = get_ttymode();
	ttymode(0);
	if(!initscr())
	{
		fprintf(stderr,"curses init failure ... check term type\r\n");
		hangup(HANGUP_USAGE);
	}
	savetty(); raw(); noecho(); nonl(); clear();
	windows_active = 1;

}	/* end of windows_start */

/*+-------------------------------------------------------------------------
	windows_end(last_win)
--------------------------------------------------------------------------*/
void
windows_end(last_win)
WINDOW *last_win;
{
	if(last_win)
		wrefresh(last_win);
/*
	resetty();
*/
	endwin();
	curmove(LINES - 1,0);
	ttymode(ttymode_before_window_start);
	windows_active = 0;
}	/* end of windows_end */

/*+-------------------------------------------------------------------------
	windows_end2()
--------------------------------------------------------------------------*/
void
windows_end2()
{
	endwin();
	ttymode(ttymode_before_window_start);
	windows_active = 0;
}	/* end of windows_end2 */

/*+-------------------------------------------------------------------------
	windows_end_signal() -- called by hangup()
--------------------------------------------------------------------------*/
void
windows_end_signal()
{
	windows_end2();
}	/* end of windows_end_signal */

/*+-------------------------------------------------------------------------
	winbox(win)
--------------------------------------------------------------------------*/
void
winbox(win)
WINDOW *win;
{

	box(win,sVR,sHR);
	wmove(win,0,0); waddch(win,(unsigned)sTL);
	wmove(win,win->_maxy - 1,0); waddch(win,(unsigned)sBL);
	wmove(win,win->_maxy - 1,win->_maxx - 1); waddch(win,(unsigned)sBR);
	wmove(win,0,win->_maxx - 1); waddch(win,(unsigned)sTR);

}	/* end of winbox */

/*+-------------------------------------------------------------------------
	window_setup(win)
--------------------------------------------------------------------------*/
void
window_setup(win,title,title_x)
WINDOW *win;
char *title;
{
register stand = (title_x < 0);

	if(stand)
		title_x = -title_x;

	scrollok(win,0);		/* do not scroll */
	winbox(win);
	wmove(win,0,title_x);
	if(stand)
		wstandout(win);
	waddch(win,'[');
	wprintw(win," %s ",title);
	waddch(win,']');
	if(stand)
		wstandend(win);
}	/* end of window_setup */

/*+-------------------------------------------------------------------------
	window_create(title,title_x,tly,tlx,lines,cols)
if title_x negative, make title "stand" out
--------------------------------------------------------------------------*/
WINDOW *
window_create(title,title_x,tly,tlx,lines,cols)
char *title;
{
register WINDOW *nwin = newwin(lines,cols,tly,tlx);
	
	if(nwin)
		window_setup(nwin,title,title_x);
	return(nwin);
}	/* end of window_create */

/*+-------------------------------------------------------------------------
	wingets
 numchars = wingets(win,y,x,buf,bufsize,&delim,edit)

This procedure reads a string from win and returns the number
of characters read.  -1 is returned if an abort is signaled by the
keyboard user.
--------------------------------------------------------------------------*/
int
wingets(win,y,x,buf,bufsize,delim,edit)
WINDOW *win;
int y;
register x;
register char *buf;
int bufsize;	/* includes room for null..field is 1 less */
register uchar *delim;
int edit;
{
register count = 0;
register pos = 0;
int insert_mode = 0;


	bufsize--;
	clear_area_char(win,y,x,bufsize,'_');
	if(edit)
	{
		waddstr(win,buf);
		count = pos = strlen(buf);
	}
	else
		*buf = 0;
	wmove(win,y,x + pos);

	while(1)
	{
		wrefresh(win);
		*delim = ttygetc(1);
		if((*delim < 0x20) || (*delim >= 0x7F))
		{
			switch(*delim)
			{
				case CR:
					*delim = NL;
				case NL:
					wrefresh(win);
					return(count);

				case BS:
					if(count)
					{
						if(count == pos)
						{
							*(buf + --count) = 0;
							wmove(win,y,x + count);
							waddch(win,'_');
							wmove(win,y,x + count);
							pos--;
						}
						else
						{
							if(!pos)
								continue;
							mem_cpy(buf + pos - 1,buf + pos,count - pos);
							*(buf + --count) = 0;
							wmove(win,y,x + --pos);
							waddstr(win,buf + pos);
							waddch(win,'_');
							wmove(win,y,x + pos);
						}
					}
					continue;

				case XFcurlf:
					if(pos)
						wmove(win,y,x + --pos);
					continue;

				case XFcurrt:
					if(pos < count)
						wmove(win,y,x + ++pos);
					continue;

				case XFins:
					insert_mode = !insert_mode;
					continue;

				case ESC:
					return(-1);

				case CTL_U:
					clear_area_char(win,y,x,bufsize,'_');
					count = 0;
					pos = 0;
					continue;

				default:
					wrefresh(win);
					return(-1);
			}	/* end of switch(*delim) */
			/*NOTREACHED*/
		}		/* end of if read delimiter */

		if(count == bufsize)
			continue;

		if(insert_mode && (pos != count))
		{
			waddch(win,*delim);
			waddstr(win,buf + pos);
			mem_cpy(buf + pos + 1,buf + pos,count - pos);
			*(buf + pos++) = *delim;
			*(buf + ++count) = 0;
			wmove(win,y,x + pos);
		}
		else
		{
			waddch(win,*delim);
			*(buf + pos) = *delim;
			if(pos == count)
				*(buf + ++count) = 0;
			pos++;
		}
	}	/* end of while can get character */
					
}	/* end of wingets */

/*+-------------------------------------------------------------------------
	winget_single(win,nondelim_list,delim_list)

This procedure assumes cursor is positioned, repeats reading a non-echoing
character from the keyboard until it matches a character in nondelim_list
or delim_list.  delim_list is expected to contain printable characters
and no upper-case characters.

If no match occurs, the bell is rung and the keyboard is read again.

If the input character matches a character in delim_list, the index (0-n)
of the character in delim_list is returned.  If a match occurs, an
upper-case version of the matching character is placed in the window.

If the input character matches a character in nondelim_list, the character
is returned or'ed with 0x1000

--------------------------------------------------------------------------*/
int
winget_single(win,nondelim_list,delim_list)
WINDOW *win;
register char *nondelim_list;
register char *delim_list;
{
register itmp;
register nlen = strlen(nondelim_list);
register dlen = strlen(delim_list);
register ichar;

	wrefresh(win);

	while(1)
	{
		ichar = to_lower(ttygetc(1));
		for(itmp = 0; itmp < nlen; itmp++)
		{
			if(ichar == nondelim_list[itmp])
			{
				waddch(win,to_upper(ichar));
				wrefresh(win);
				return(itmp);
			}
		}
		for(itmp = 0; itmp < dlen; itmp++)
		{
			if(ichar == delim_list[itmp])
				return(ichar | 0x1000);
		}
		ring_bell();
	}

}	/* end of winget_single */

/* end of ecuwinutil.c */
/* vi: set tabstop=4 shiftwidth=4: */
