/* CHK=0xC6CA */
/*+-------------------------------------------------------------------------
    feval.c
	Copyright 1989 Warren H. Tucker, III.  All rights reserved.

    feval_int(param,&int_returned)
    feval_str(param,&esd_to_be_plugged)

  These routines are called with param.index as follows:

         !nnnnn       nnn is name of function
          ^
          |

  Defined functions:
	feval_int(param,value)
	feval_str(param,result_esd)
	strfunc_left(param,scratch_esd,result_esd)
	strfunc_right(param,scratch_esd,result_esd)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:24-wht@n4hgf---- 2.80 --- */
/*:04-19-1990-03:07-wht@n4hgf-GCC run found unused vars -- rm them */
/*:04-11-1990-02:53-wht@n4hgf-ensure isacii, etc. return 0 or 1 */
/*:04-11-1990-02:49-wht@n4hgf-check length of arg on isacii etc */
/*:04-05-1990-14:29-wht@n4hgf-sort/renumber FI and FS tokens */
/*:04-05-1990-14:01-wht@n4hgf-add monthz, dayz, datez, timez, timezs */
/*:03-27-1990-15:18-wht@n4hgf-overhaul for/wait/SIGCLD handling */
/*:03-25-1990-14:11-wht@n4hgf------ x2.70 ------- */
/*:03-16-1990-04:36-wht@n4hgf-add %mhack */
/*:03-12-1990-04:31-wht@n4hgf-use Lrname */
/*:11-06-1989-03:24-wht-add ctype-style operators */
/*:11-05-1989-22:51-wht-strfunc_right (%right) was not working at all */
/*:11-05-1989-22:51-wht-strfunc_right (%right) was not returning 0 on success */
/*:11-05-1989-19:45-wht-add string %screen function */
/*:11-05-1989-16:48-wht-rename ansi_y to cursor_y, ansi_x to cursor_x */
/*:11-04-1989-07:11-wht-add %curx and %cury integer functions */
/*:07-03-1989-22:57-wht------ x2.00 ----- */
/*:06-24-1989-16:53-wht-flush edits --- ecu 1.95 */

#include "ecu.h"
#include "ecuerror.h"
#include "esd.h"
#include "proc.h"
#include "var.h"

extern int shm_shmid;

#define FIargc           1
#define FIbaud           2
#define FIcolors         3
#define FIconn           4
#define FIcsec           5
#define FIctoi           6
#define FIcurx           7
#define FIcury           8
#define FIfatime         9
#define FIfmode         10
#define FIfmtime        11
#define FIfsize         12
#define FIftell         13
#define FIinstr         14
#define FIisalnum       15
#define FIisalpha       16
#define FIisascii       17
#define FIischr         18
#define FIiscntrl       19
#define FIisdigit       20
#define FIisdir         21
#define FIisgraph       22
#define FIislower       23
#define FIisprint       24
#define FIispunct       25
#define FIisreg         26
#define FIisspace       27
#define FIisupper       28
#define FIisxdigit      29
#define FIlen           30
#define FIlgetc         31
#define FImatch         32
#define FImhack         33
#define FIpid           34
#define FIrchr          35
#define FIrchrc         36
#define FIshmid         37
#define FIstoi          38
#define FIxchr          39
#define FIxchrc         40

KEYTAB feval_int_tbl[] =
{
	{"argc",FIargc},
	{"baud",FIbaud},
	{"colors",FIcolors},
	{"conn",FIconn},
	{"csec",FIcsec},
	{"ctoi",FIctoi},
	{"curx",FIcurx},
	{"cury",FIcury},
	{"fatime",FIfatime},
	{"fmode",FIfmode},
	{"fmtime",FIfmtime},
	{"fsize",FIfsize},
	{"ftell",FIftell},
	{"instr",FIinstr},
	{"isalnum",FIisalnum},
	{"isalpha",FIisalpha},
	{"isalpha",FIisalpha},
	{"isascii",FIisascii},
	{"ischr",FIischr},
	{"iscntrl",FIiscntrl},
	{"isdigit",FIisdigit},
	{"isdir",FIisdir},
	{"isgraph",FIisgraph},
	{"islower",FIislower},
	{"isprint",FIisprint},
	{"ispunct",FIispunct},
	{"isreg",FIisreg},
	{"isspace",FIisspace},
	{"isupper",FIisupper},
	{"isxdigit",FIisxdigit},
	{"len",FIlen},
	{"lgetc",FIlgetc},
	{"match",FImatch},
	{"mhack",FImhack},
	{"pid",FIpid},
	{"rchr",FIrchr},
	{"rchrc",FIrchrc},
	{"shmid",FIshmid},
	{"stoi",FIstoi},
	{"xchr",FIxchr},
	{"xchrc",FIxchrc},
	{(char *)0,0}
};

#define FSargv           1
#define FScgetc          2
#define FScgets          3
#define FSchr            4
#define FSdate           5
#define FSdatez          6
#define FSday            7
#define FSdayz           8
#define FSdir            9
#define FSedate         10
#define FSenvvar        11
#define FSerrstr        12
#define FSetime         13
#define FSfmodestr      14
#define FSitos          15
#define FSleft          16
#define FSline          17
#define FSlogname       18
#define FSmid           19
#define FSmonth         20
#define FSmonthz        21
#define FSrdesc         22
#define FSright         23
#define FSrname         24
#define FSrtel          25
#define FSscreen        26
#define FStime          27
#define FStimes         28
#define FStimez         29
#define FStimezs        30
#define FStty           31

KEYTAB feval_str_tbl[] =
{
	{"argv",FSargv},
	{"cgetc",FScgetc},
	{"cgets",FScgets},
	{"chr",FSchr},
	{"date",FSdate},
	{"datez",FSdatez},
	{"day",FSday},
	{"dir",FSdir},
	{"edate",FSedate},
	{"envvar",FSenvvar},
	{"errstr",FSerrstr},
	{"etime",FSetime},
	{"fmodestr",FSfmodestr},
	{"itos",FSitos},
	{"left",FSleft},
	{"line",FSline},
	{"logname",FSlogname},
	{"mid",FSmid},
	{"month",FSmonth},
	{"rdesc",FSrdesc},
	{"right",FSright},
	{"rname",FSrname},
	{"rtelno",FSrtel},
	{"screen",FSscreen},
	{"time",FStime},
	{"times",FStimes},
	{"timez",FStimez},
	{"timezs",FStimezs},
	{"tty",FStty},
	{(char *)0,0}
};

extern char curr_dir[];
extern int proctrace;
extern PCB *pcb_stack[];
extern int proc_level;
extern char *sys_errlist[];
extern int sys_nerr;
extern struct timeb starting_timeb;

char *day_of_week_list = "SunMonTueWedThuFriSat";
char *month_name_list = "JanFebMarAprMayJunJulAugSepOctNovDec";

/*+-------------------------------------------------------------------------
    erc = feval_int(param,&int_returned);
Functions (parameter types are expressed by the usage of variables)
--------------------------------------------------------------------------*/
feval_int(param,value)
ESD *param;
long *value;
{
register erc;
register keyword_token;
int index_save;
ESD *tesd1 = (ESD *)0;
ESD *tesd2 = (ESD *)0;
ulong int1;
char s32[32];
struct timeb now_timeb;

	index_save = param->index;

	if(erc = get_alphanum_zstr(param,s32,sizeof(s32)))
	{
		erc = eInvalidFunction;
		goto RETURN;
	}

	keyword_token = keyword_lookup(feval_int_tbl,s32);
	switch(keyword_token)
	{
/* LEN($S0)         length of $S0 */
	case FIlen:
		if(!(tesd1 = make_esd(256)))
		{
			erc = eNoMemory;
			goto RETURN;
		}
		if(erc = skip_paren(param,1))
			goto RETURN;
		if(erc = gstr(param,tesd1))
			goto RETURN;
		if(erc = skip_paren(param,0))
			goto RETURN;
		*value = (long)tesd1->cb;
		break;

/* INSTR($I0,$I1)   index of first occurrence of $I1 in $I0, -1 if none */
	case FIinstr:
		if(!(tesd1 = make_esd(256)))
		{
			erc = eNoMemory;
			goto RETURN;
		}
		if(!(tesd2 = make_esd(256)))
		{
			erc = eNoMemory;
			goto RETURN;
		}
		if(erc = skip_paren(param,1))
			goto RETURN;
		if(erc = gstr(param,tesd1))
			goto RETURN;
		if(erc = skip_comma(param))
			goto RETURN;
		if(erc = gstr(param,tesd2))
			goto RETURN;
		if(erc = skip_paren(param,0))
			goto RETURN;

		*value = (long)ulindex(tesd1->pb,tesd2->pb);
		break;

	case FImatch:
		if(!(tesd1 = make_esd(256)))
		{
			erc = eNoMemory;
			goto RETURN;
		}
		if(!(tesd2 = make_esd(256)))
		{
			erc = eNoMemory;
			goto RETURN;
		}
		if(erc = skip_paren(param,1))
			goto RETURN;
		if(erc = gstr(param,tesd1))
			goto RETURN;
		if(erc = skip_comma(param))
			goto RETURN;
		if(erc = gstr(param,tesd2))
			goto RETURN;
		if(erc = skip_paren(param,0))
			goto RETURN;

		erc = regexp_operation(tesd1->pb,tesd2->pb,value);
		break;

	case FImhack:
		ftime(&now_timeb);
		*value = ((now_timeb.time - starting_timeb.time) * 1000) +
				(now_timeb.millitm - starting_timeb.millitm);
		erc = 0;
		break;

	case FIargc:
		if(!proc_level)
		{
			pputs("not executing procedure\n");
			erc = eFATAL_ALREADY;
			break;
		}
		*value = (long)pcb_stack[proc_level - 1]->argc;
		break;

	case FIcolors:
		if(erc = ifunc_colors(value))
			goto RETURN;
		break;

	case FIftell:
		if(erc = ifunc_ftell(param,value))
			goto RETURN;
		break;

	case FIfmode:
		if(erc = ifunc_fmode(param,value))
			goto RETURN;
		break;

	case FIfsize:
		if(erc = ifunc_fsize(param,value))
			goto RETURN;
		break;

	case FIfmtime:
		if(erc = ifunc_fmtime(param,value))
			goto RETURN;
		break;

	case FIfatime:
		if(erc = ifunc_fatime(param,value))
			goto RETURN;
		break;

	case FIischr:
		if(erc = ifunc_ischr(param,value))
			goto RETURN;
		break;

	case FIisdir:
		if(erc = ifunc_isdir(param,value))
			goto RETURN;
		break;

	case FIisreg:
		if(erc = ifunc_isreg(param,value))
			goto RETURN;
		break;

	case FIbaud:
		*value = (long)shm->Lbaud;
		erc = 0;
		break;

	case FIpid:
		*value = (long)getpid();
		erc = 0;
		break;

	case FIcsec:
		*value = (shm->Lmodem_off_hook) ? shm->Loff_hook_time : -1;
		erc = 0;
		break;

	case FIconn:
		*value = (long)(shm->Lmodem_off_hook) ? (long)shm->Liofd : 0;
		erc = 0;
		break;

	case FIxchr:
		*value = shm->xmit_chars;
		erc = 0;
		break;

	case FIxchrc:
		*value = shm->xmit_chars_this_connect;
		erc = 0;
		break;

	case FIrchr:
		shmx_rc_report(value,&int1);
		break;

	case FIrchrc:
		shmx_rc_report(&int1,value);
		erc = 0;
		break;

/* LGETC($I0) get char from line, waiting for $I0 msec
returns  character read or -1 if none read in time */
	case FIlgetc:
		if(erc = skip_paren(param,1))
			goto RETURN;
		if(erc = gint(param,&int1))
			goto RETURN;
		if(erc = skip_paren(param,0))
			goto RETURN;
		*value = (long)lgetc_timeout(int1);
		break;

	case FIctoi:
		if(!(tesd1 = make_esd(256)))
		{
			erc = eNoMemory;
			goto RETURN;
		}
		if(erc = skip_paren(param,1))
			goto RETURN;
		if(erc = gstr(param,tesd1))
			goto RETURN;
		if(erc = skip_paren(param,0))
			goto RETURN;
		if(tesd1->cb == 0)
			*value = -1;
		else
			*value = (long)((unsigned)0xFF & (unsigned)tesd1->pb[0]);
		break;

	case FIstoi:
		if(!(tesd1 = make_esd(256)))
		{
			erc = eNoMemory;
			goto RETURN;
		}
		if(erc = skip_paren(param,1))
			goto RETURN;
		if(erc = gstr(param,tesd1))
			goto RETURN;
		if(erc = skip_paren(param,0))
			goto RETURN;

		tesd1->index = 0;
		skip_cmd_break(tesd1);
		*value = 0;
		gint_constant(tesd1,value);
		break;

	case FIcurx:
		*value = (long)shm->cursor_x;
		break;

	case FIcury:
		*value = (long)shm->cursor_y;
		break;

	case FIshmid:
		*value = (long)shm_shmid;
		break;

	case FIisalpha:
	case FIisupper:
	case FIislower:
	case FIisdigit:
	case FIisxdigit:
	case FIisspace:
	case FIispunct:
	case FIisalnum:
	case FIisprint:
	case FIisgraph:
	case FIiscntrl:
	case FIisascii:
		if(!(tesd1 = make_esd(256)))
		{
			erc = eNoMemory;
			goto RETURN;
		}
		if(erc = skip_paren(param,1))
			goto RETURN;
		if(erc = gstr(param,tesd1))
			goto RETURN;
		if(erc = skip_paren(param,0))
			goto RETURN;
		if(!tesd1->cb)
		{
			*value = 0;
			goto RETURN;
		}
		switch(keyword_token)
		{
			case FIisalpha:
				*value = !!isalpha(*tesd1->pb);
				break;
			case FIisupper:
				*value = !!isupper(*tesd1->pb);
				break;
			case FIislower:
				*value = !!islower(*tesd1->pb);
				break;
			case FIisdigit:
				*value = !!isdigit(*tesd1->pb);
				break;
			case FIisxdigit:
				*value = !!isxdigit(*tesd1->pb);
				break;
			case FIisspace:
				*value = !!isspace(*tesd1->pb);
				break;
			case FIispunct:
				*value = !!ispunct(*tesd1->pb);
				break;
			case FIisalnum:
				*value = !!isalnum(*tesd1->pb);
				break;
			case FIisprint:
				*value = !!isprint(*tesd1->pb);
				break;
			case FIisgraph:
				*value = !!isgraph(*tesd1->pb);
				break;
			case FIiscntrl:
				*value = !!iscntrl(*tesd1->pb);
				break;
			case FIisascii:
				*value = !!isascii(*tesd1->pb);
				break;
		}
		break;

	default:
		erc = eInvalidFunction;
	}   /* end of keyword lookup erc switch statement */

RETURN:
	if(tesd1)
		free_esd(tesd1);
	if(tesd2)
		free_esd(tesd2);
	return(erc);

}   /* end of feval_int() */

/*+------------------------------------------------------------------
    strfunc_left(param,&scratch_esd,&result_esd)
-------------------------------------------------------------------*/
int
strfunc_left(param,scratch_esd,result_esd)
ESD *param;
ESD *scratch_esd;
ESD *result_esd;
{
register erc;
int itmp;
long ltmp;

	if(erc = skip_paren(param,1))
		return(erc);
	if(erc = gstr(param,scratch_esd))
		return(erc);
	if(erc = skip_comma(param))
		return(erc);
	if(erc = gint(param,&ltmp))
		return(erc);
	itmp = (int)ltmp;
	if(itmp < 0)
		return(eBadParameter);
	if(erc = skip_paren(param,0))
		return(erc);
	/* take min of param and .cb */
	itmp = (itmp < scratch_esd->cb) ? itmp : scratch_esd->cb;
	if(itmp > (result_esd->maxcb - result_esd->cb) )
		return(eBufferTooSmall);
	memcpy(&result_esd->pb[result_esd->cb],
	    scratch_esd->pb,itmp);
	result_esd->cb += itmp;
	return(erc);
}   /* end of strfunc_left() */

/*+-------------------------------------------------------------------------
    erc = strfunc_right(param,&scratch_esd,&result_esd)
--------------------------------------------------------------------------*/
int
strfunc_right(param,scratch_esd,result_esd)
ESD *param;
ESD *scratch_esd;
ESD *result_esd;
{
register erc;
int itmp;
long ltmp;

	if(erc = skip_paren(param,1))
		return(erc);
	if(erc = gstr(param,scratch_esd))
		return(erc);
	if(erc = skip_comma(param))
		return(erc);
	if(erc = gint(param,&ltmp))
		return(erc);
	itmp = (int)ltmp;
	if(itmp < 0)
		return(eBadParameter);
	if(erc = skip_paren(param,0))
		return(erc);

/* take min of param and .cb */
	itmp = (itmp < scratch_esd->cb) ? itmp : scratch_esd->cb;
	if(itmp > (result_esd->maxcb - result_esd->cb) )
		return(eBufferTooSmall);
	memcpy(&result_esd->pb[result_esd->cb],
	    &scratch_esd->pb[scratch_esd->cb - itmp],itmp);
	result_esd->cb += itmp;
	return(0);

}   /* end of strfunc_right() */

/*+-------------------------------------------------------------------------
    erc = feval_str(param,&esd_to_be_plugged);
    results are APPENDED to 'result_esd'
--------------------------------------------------------------------------*/
feval_str(param,result_esd)
ESD *param;
ESD *result_esd;
{
register erc;
register itmp;
int index_save;
int int1,int2;
char s32[32];
char *cptr;
long ltmp;
long ltmp2;
long ltmp3;
ESD *tesd1;
char *get_ttyname();
char *getenv();
char *getlogin();
char *get_elapsed_time();
char *mode_map();

	if(!(tesd1 = make_esd(128)))
		return(eNoMemory);

	index_save=param->index;

	if(erc = get_alphanum_zstr(param,s32,sizeof(s32)-1))
	{
		free_esd(tesd1);
		return(eInvalidFunction);
	}

	erc = 0;
	itmp=keyword_lookup(feval_str_tbl,s32);
	switch(itmp)
	{
/* LEFT($S0,$I0)   return leftmost $I0 characters of $S0 */
	case FSleft:
		erc = strfunc_left(param,tesd1,result_esd);
		break;

/* RIGHT($S0,$I0)   return rightmost $I0 characters of $S0 */
	case FSright:
		erc = strfunc_right(param,tesd1,result_esd);
		break;

/* MID($S0,$I0,$I1)   return middle $I1 chars of $S0 starting at $I0 */
	case FSmid:
		if(erc = skip_paren(param,1))
			break;
		if(erc = gstr(param,tesd1))
			break;
		if(erc = skip_comma(param))
			break;
		if(erc = gint(param,&ltmp))
			break;
		int1 = (int)ltmp;
		if(int1 < 0)
		{
			erc = eBadParameter;
			break;
		}
		if(erc = skip_cmd_break(param))
			break;
		if(param->pb[param->index] == ')') /* if we find a ')' instead of ... */
		{                                  /* 2nd int param, default to max */
			++param->index;
			int2 = 256;
		}
		else
		{
			if(erc = skip_comma(param))
				break;
			if(erc = gint(param,&ltmp))
				break;
			int2 = (int)ltmp;
			if(int2 < 0)
			{
				erc = eBadParameter;
				break;
			}
			if(erc = skip_paren(param,0))
				break;
		}

		if(int1 >= tesd1->cb) /* if initial index past end of string */
			break;
		itmp = tesd1->cb - int1;
		itmp = (int2 < itmp) ? int2 : itmp;
		cptr = tesd1->pb + int1;
		goto CPTR_ITMP_COMMON;

/* ARGV($I0) */
	case FSargv:
		if(!proc_level)
		{
			pputs("not executing procedure\n");
			erc = eFATAL_ALREADY;
			break;
		}
		if(erc = skip_paren(param,1))
			break;
		if(erc = gint(param,&ltmp))
			break;
		if(erc = skip_paren(param,0))
			break;
		itmp = (long)pcb_stack[proc_level - 1]->argc; /* arg count */
		if((int)ltmp > itmp - 1)
		{
			if(proctrace)
			{
				pprintf("WARNING: %%argc=%d, %%argv(%ld) null\n",
					itmp,ltmp);
			}
			break;
		}
		cptr = (pcb_stack[proc_level - 1])->argv[(int)ltmp];
		itmp = strlen(cptr);
		goto CPTR_ITMP_COMMON;

	case FSdir:
		cptr = curr_dir;
		itmp = strlen(curr_dir);
		goto CPTR_ITMP_COMMON;

	case FSetime:
		if(erc = skip_paren(param,1))
			break;
		if(erc = gint(param,&ltmp))
			break;
		if(erc = skip_paren(param,0))
			break;
		cptr = get_elapsed_time(ltmp);
		itmp = strlen(cptr);
		goto CPTR_ITMP_COMMON;

	case FSerrstr:
		if(erc = skip_paren(param,1))
			break;
		if(erc = gint(param,&ltmp))
			break;
		if(erc = skip_paren(param,0))
			break;
		if((int)ltmp >= sys_nerr)
		{
			sprintf(s32,"error %d",(int)ltmp);
			cptr = s32;
		}
		else
			cptr = sys_errlist[(int)ltmp];
		itmp = strlen(cptr);
		goto CPTR_ITMP_COMMON;

	case FSenvvar:
		if(erc = skip_paren(param,1))
			break;
		if(erc = gstr(param,tesd1))
			break;
		if(erc = skip_comma(param))
			break;
		if(!(cptr = getenv(tesd1->pb)))
			break;
		itmp = strlen(cptr);
		goto CPTR_ITMP_COMMON;

	case FSlogname:
		if(!(cptr = getlogin()))
			break;
		itmp = strlen(cptr);
		goto CPTR_ITMP_COMMON;

	case FSfmodestr:
		if(erc = skip_paren(param,1))
			break;
		if(erc = gint(param,&ltmp))
			break;
		if(erc = skip_paren(param,0))
			break;
		cptr = mode_map((int)ltmp,(char *)0);
		itmp = strlen(cptr);
		goto CPTR_ITMP_COMMON;

	case FStty:
		cptr = get_ttyname();
		itmp = strlen(cptr);
		goto CPTR_ITMP_COMMON;

	case FSrname:
		if(!shm->Lmodem_off_hook)
			break;
		cptr = shm->Lrname;
		itmp = strlen(shm->Lrname);
		goto CPTR_ITMP_COMMON;

	case FSrdesc:
		if(!shm->Lmodem_off_hook)
			break;
		cptr = shm->Ldescr;
		itmp = strlen(shm->Ldescr);
		goto CPTR_ITMP_COMMON;

	case FSrtel:
		if(!shm->Lmodem_off_hook)
			break;
		cptr = shm->Ltelno;
		itmp = strlen(shm->Ltelno);
		goto CPTR_ITMP_COMMON;

	case FSline:
		if(shm->Liofd < 0)
			break;
		cptr = shm->Lline;
		itmp = strlen(shm->Lline);
		goto CPTR_ITMP_COMMON;

	case FSmonth:
	case FSmonthz:
	    cptr = &month_name_list[(get_month(itmp == FSmonthz) - 1) * 3];
		itmp = 3;
		goto CPTR_ITMP_COMMON;

	case FSday:
	case FSdayz:
		cptr = &day_of_week_list[get_day(itmp == FSdayz) * 3];
		itmp = 3;
		goto CPTR_ITMP_COMMON;

	case FSscreen:
		if(erc = skip_paren(param,1))
			break;
		if(erc = gint(param,&ltmp))		/* y */
			break;
		if(ltmp > 42)
		{
			erc = eBadParameter;
			break;
		}
		if(erc = skip_comma(param))
			break;
		if(erc = gint(param,&ltmp2))	/* x */
			break;
		if(ltmp2 > 79)
		{
			erc = eBadParameter;
			break;
		}
		if(erc = skip_comma(param))
			break;
		if(erc = gint(param,&ltmp3))	/* len */
			break;
		if(erc = skip_paren(param,0))
			break;

		int1 = ((int)ltmp * 80) + (int)ltmp2;	/* screen position */
		itmp = (int)ltmp3;						/* length */
		int2 = sizeof(shm->screen) - int1;		/* size from y,x to end */
		if(itmp > int2)
			itmp = int2;
		cptr = ((char *)shm->screen) + int1;
		goto CPTR_ITMP_COMMON;

CPTR_ITMP_COMMON:
		if( itmp > (result_esd->maxcb - result_esd->cb))
		{
			erc = eBufferTooSmall;
			break;
		}
		memcpy(&result_esd->pb[result_esd->cb],cptr,itmp);
		result_esd->cb += itmp;
		break;

	case FSedate:
		if(erc = skip_paren(param,1))
			break;
		if(erc = gint(param,&ltmp))
			break;
		if(erc = skip_paren(param,0))
			break;
		if(19 > (result_esd->maxcb - result_esd->cb))
		{
			erc = eBufferTooSmall;
			break;
		}
		epoch_secs_to_str(ltmp,3,&result_esd->pb[result_esd->cb]);
		result_esd->cb += 19;
		break;

	case FStime:
		if(5 > (result_esd->maxcb - result_esd->cb))
		{
			erc = eBufferTooSmall;
			break;
		}
		get_tod(0,&result_esd->pb[result_esd->cb]);
		result_esd->cb += 5;
		break;

	case FStimes:
		if(8 > (result_esd->maxcb - result_esd->cb))
		{
			erc = eBufferTooSmall;
			break;
		}
		get_tod(1,&result_esd->pb[result_esd->cb]);
		result_esd->cb += 8;
		break;

	case FStimez:
		if(5 > (result_esd->maxcb - result_esd->cb))
		{
			erc = eBufferTooSmall;
			break;
		}
		get_tod(6,&result_esd->pb[result_esd->cb]);
		result_esd->cb += 5;
		break;

	case FStimezs:
		if(8 > (result_esd->maxcb - result_esd->cb))
		{
			erc = eBufferTooSmall;
			break;
		}
		get_tod(7,&result_esd->pb[result_esd->cb]);
		result_esd->cb += 8;
		break;

	case FSdate:
		if(10 > (result_esd->maxcb - result_esd->cb))
		{
			erc = eBufferTooSmall;
			break;
		}
		get_tod(5,&result_esd->pb[result_esd->cb]);
		result_esd->cb += 10;
		break;

	case FSdatez:
		if(10 > (result_esd->maxcb - result_esd->cb))
		{
			erc = eBufferTooSmall;
			break;
		}
		get_tod(8,&result_esd->pb[result_esd->cb]);
		result_esd->cb += 10;
		break;

	case FScgets:
		erc = ttygets_esd(result_esd,1,1);
		break;

	case FScgetc:
		if(result_esd->cb == result_esd->maxcb)
		{
			erc = eBufferTooSmall;
			break;
		}
		result_esd->pb[result_esd->cb] = ttygetc(0);
		result_esd->cb++;
		break;

	case FSchr:
		if(erc = skip_paren(param,1))
			break;
		if(erc = gint(param,&ltmp))
			break;
		if(!ltmp)
		{
			pputs("cannot use %chr(0)\n");
			return(eFATAL_ALREADY);
		}
		if(erc = skip_paren(param,0))
			break;
		if(result_esd->cb == result_esd->maxcb )
		{
			erc = eBufferTooSmall;
			break;
		}
		result_esd->pb[result_esd->cb] = (char)ltmp;
		result_esd->cb++;
		break;

	case FSitos:
		if(erc = skip_paren(param,1))
			break;
		if(erc = gint(param,&ltmp))
			break;
		s32[0] = 0;
		if(!skip_comma(param))
		{
			if(erc = get_numeric_zstr(param,s32 + 1,sizeof(s32) - 4))
				strcpy(s32 + 1,"1");
			if(((itmp = atoi(s32 + 1)) < 0) ||
				(itmp > (result_esd->maxcb - result_esd->cb)))
			{
				erc = eBufferTooSmall;
				break;
			}
			s32[0] = '%';
			if(ulindex(param->pb + param->index,"x") == 0)
			{
				param->index++;
				strcat(s32,"lx");
			}
			else if(ulindex(param->pb + param->index,"o") == 0)
			{
				param->index++;
				strcat(s32,"lo");
			}
			else if(ulindex(param->pb + param->index,"d") == 0)
			{
				param->index++;
				strcat(s32,"ld");
			}
			else if(erc)
				break;
			else
				strcat(s32,"ld");
		}
		if(erc = skip_paren(param,0))
			break;
		sprintf(tesd1->pb,s32[0] ? s32 : "%ld",ltmp);
		tesd1->cb = strlen(tesd1->pb);
		if(result_esd->maxcb - result_esd->cb < tesd1->cb)
		{
			erc = eBufferTooSmall;
			break;
		}
		strcpy(&result_esd->pb[result_esd->cb],tesd1->pb);
		result_esd->cb += tesd1->cb;
		break;


	default:
		erc = eInvalidFunction;
		break;
	}   /* end of keyword lookup erc switch statement */

	null_terminate_esd(result_esd);
	free_esd(tesd1);
	return(erc);

}   /* end of feval_str() */

/* vi: set tabstop=4 shiftwidth=4: */
/* end of feval.c */
