/* CHK=0x1D7E */
#if defined(WHT)
#define LOG_UNGETTY
#define LOG_HDBDIAL
#endif
/*+-------------------------------------------------------------------------
	hdbintf.c - HDB UUCP database and /etc/utmp interface routines
	wht%n4hgf@gatech.edu

  Defined functions:
	dialstr_translate(translate_list,to_translate)
	enddlent()
	enddvent()
	getdlent()
	getdlentname(name)
	getdvbaud(baud)
	getdvent()
	getdvline(line)
	hdb_dial(presult)
	hdb_dial_error_text(errcode)
	hdb_init()
	ungetty_get_line()
	ungetty_return_line()

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:24-wht@n4hgf---- 2.80 --- */
/*:04-19-1990-03:08-wht@n4hgf-GCC run found unused vars -- rm them */
/*:04-02-1990-02:17-wht@n4hgf-Dialers support appears to work */
/*:04-02-1990-02:13-wht@n4hgf-check program execute access */
/*:04-01-1990-23:30-wht@n4hgf-add Dialers support */
/*:03-27-1990-15:18-wht@n4hgf-overhaul for/wait/SIGCLD handling */
/*:03-25-1990-14:11-wht@n4hgf------ x2.70 ------- */
/*:11-24-1989-21:03-wht-move utmp_status to its own file */
/*:11-24-1989-20:30-wht-UNIX 3.2 utmp neat but more complex */
/*:11-19-1989-15:20-wht-LOG_HDBDIAL and LOG_UNGETTY for wht only */
/*:11-16-1989-20:08-wht-fix hdb dialer failure recovery */
/*:11-14-1989-12:32-wht-add hack for credit card number trailer */
/*:10-14-1989-21:10-wht-allow user to change ecu color choices */
/*:07-03-1989-22:57-wht------ x2.00 ----- */
/*:07-01-1989-13:40-wht-creation */

#include "ecu.h"
#include "esd.h"
#include "var.h"
#include "ecuhangup.h"
#include "utmpstatus.h"
#include "ecuungetty.h"
#include "dvent.h"
#include "dlent.h"
#include "dialer.h"
#include <errno.h>
#include <utmp.h>

char *arg_token();

extern int errno;
extern char kbdintr;		/* current input INTR */
extern int interrupt;
extern ulong colors_current;
extern char *sys_errlist[];
extern int proc_level;
extern int proctrace;

int there_is_hdb_on_this_machine = 0;
FILE *fpdv = NULL;
FILE *fpdl = NULL;
char *Devices_file = "/usr/lib/uucp/Devices";
char *Dialers_file = "/usr/lib/uucp/Dialers";
char *ungetty = "/usr/lib/ecu/ecuungetty";
char ungetty_ttyname[64] = "";

/*+-------------------------------------------------------------------------
	ungetty_get_line()
--------------------------------------------------------------------------*/
ungetty_get_line()
{
int itmp;
char ungetty_log[80];
int ungetty_pid;
int (*original_sighdlr)();
int wait_status;
char bamboozlement[20];
char *bamboozle();

	if(!there_is_hdb_on_this_machine)
		return(LOPEN_ENABLED);

	strcpy(bamboozlement,bamboozle(getpid()));
	if(access(ungetty,1))
	{
		pperror(ungetty);
		return(LOPEN_ENABLED);
	}
	original_sighdlr = signal(SIGCLD,SIG_DFL);
	if((ungetty_pid = fork()) == 0)
	{
		execl(ungetty,"ungetty",shm->Lline,bamboozlement,(char *)0);
		exit(-1);
	}
	while(((itmp = wait(&wait_status)) != ungetty_pid) && (itmp != -1))
		;
	signal(SIGCLD,original_sighdlr);
	switch((wait_status >> 8) & 0xFF)
	{
		case UG_NOTENAB:	/* line acquired: not enabled */
			ungetty_ttyname[0] = 0;
			break;

		case UG_ENAB:		/* line acquired: need ungetty -r when done */
#if defined(LOG_UNGETTY)
			sprintf(ungetty_log,"UNGETTY acquired %s",shm->Lline);
			ecu_log_event(ungetty_log);
#endif
			strcpy(ungetty_ttyname,shm->Lline);
			break;

		case UG_FAIL:		/* line in use */
			ungetty_ttyname[0] = 0;
			return(LOPEN_ENABLED_IN_USE);

		default:
#if defined(LOG_UNGETTY)
			sprintf(ungetty_log,"UNGETTY status 0x%04x",wait_status);
			ecu_log_event(ungetty_log);
#endif
			ungetty_ttyname[0] = 0;
			return(LOPEN_ENABLED);
	}

	return(0);

}	/* end of ungetty_get_line */

/*+-------------------------------------------------------------------------
	ungetty_return_line()
--------------------------------------------------------------------------*/
void
ungetty_return_line()
{
int ungetty_pid;
int itmp;
int (*original_sighdlr)();
int wait_status = 0xDEAD;
char ungetty_log[80];
char bamboozlement[20];
char *bamboozle();

	if(!there_is_hdb_on_this_machine)
		return;
	if(!ungetty_ttyname[0])
		return;

	strcpy(bamboozlement,bamboozle(getpid()));

	/* call ungetty to see if we need to switch to dialin */
	if(access(ungetty,1))
	{
		pperror(ungetty);
		return;
	}
	original_sighdlr = signal(SIGCLD,SIG_DFL);
	if((ungetty_pid = fork()) == 0)
	{
		execl(ungetty,"ungetty","-t",ungetty_ttyname,bamboozlement,(char *)0);
		ecu_log_event("could not exec ecuungetty -t");
		exit(-1);
	}
	while(((itmp = wait(&wait_status)) != ungetty_pid) &&
			(itmp != -1) )
		;
	signal(SIGCLD,original_sighdlr);
	switch((wait_status >> 8) & 0xFF)
	{
		case UG_RESTART:
			break;

		default:
			sprintf(ungetty_log,"UNGETTY -t %s status %04x",
				ungetty_ttyname,wait_status);
			ecu_log_event(ungetty_log);
			ungetty_ttyname[0] = 0;
			return;
	}

	strcpy(bamboozlement,bamboozle(getpid()));

	original_sighdlr = signal(SIGCLD,SIG_DFL);
	if((ungetty_pid = fork()) == 0)
	{
		execl(ungetty,"ungetty","-r",ungetty_ttyname,bamboozlement,(char *)0);
		ecu_log_event("could not exec ecuungetty -r");
		exit(-1);
	}

	while(((itmp = wait(&wait_status)) != ungetty_pid) &&
			(itmp != -1))
		;
	if(wait_status)
	{
		sprintf(ungetty_log,"UNGETTY -r %s status 0x%04x",
			ungetty_ttyname,wait_status);
	}
	else
		sprintf(ungetty_log,"UNGETTY returned %s",ungetty_ttyname);
	ecu_log_event(ungetty_log);

	ungetty_ttyname[0] = 0;

}	/* end of ungetty_return_line */

/*+-------------------------------------------------------------------------
	getdvent() - get first or next Devices entry (a la getpwent)
--------------------------------------------------------------------------*/
struct dvent *
getdvent()
{
int itmp;
char *cptr;
#define MAX_DV_TOKENS 9
char *tokens[MAX_DV_TOKENS];
static struct dvent dve;
static char dvstr[128];
char *strchr();

	if(!there_is_hdb_on_this_machine)
		return((struct dvent *)0);

	if(fpdv == NULL)
	{
		if((fpdv = fopen(Devices_file,"r")) == NULL)
		{
			pperror(Devices_file);
			return((struct dvent *)0);
		}
	}

	while(1)
	{
		if(fgets(dvstr,sizeof(dvstr),fpdv) == NULL)
			return((struct dvent *)0);

		if((dvstr[0] == '#') || (dvstr[0] == ' ') || (strlen(dvstr) == 1))
			continue;

		build_arg_array(dvstr,tokens,MAX_DV_TOKENS,&itmp);

		if(itmp > 1)
			break;
	}

	dve.type = tokens[0];
	dve.line = tokens[1];
	dve.dialer = tokens[2];
	if(!strcmp(tokens[3],"Any"))
	{
		dve.low_baud = 1;
		dve.high_baud = 100000L;
	}
	else
	{
		dve.low_baud = atoi(tokens[3]);
		if(!(cptr = strchr(tokens[3],'-')))
			dve.high_baud = dve.low_baud;
		else
			dve.high_baud = atoi(cptr + 1);
	}
	dve.dialprog = tokens[4];
	return(&dve);

}	/* end of getdvent */

/*+-------------------------------------------------------------------------
	getdvbaud(baud) - get Devices entry matching baud rate
--------------------------------------------------------------------------*/
struct dvent *
getdvbaud(baud)
uint baud;
{
struct dvent *dve;

	while(1)
	{
		if((dve = getdvent()) == (struct dvent *)0)
			return(dve);
		if((dve->low_baud <= baud) && (baud <= dve->high_baud))
			return(dve);
	}
	/*NOTREACHED*/

}	/* end of getdvbaud */

/*+-------------------------------------------------------------------------
	getdvline(line) - get Devices entry matching line
calling argument 'line's is string AFTER "/dev/tty"
--------------------------------------------------------------------------*/
struct dvent *
getdvline(line)
char *line;
{
struct dvent *dve;

	while(1)
	{
		if((dve = getdvent()) == (struct dvent *)0)
			return(dve);
		if(!strcmp(dve->line + 3,line))
			return(dve);
	}
	/*NOTREACHED*/

}	/* end of getdvline */

/*+-------------------------------------------------------------------------
	enddvent()
--------------------------------------------------------------------------*/
void
enddvent()
{
	if(fpdv != NULL)
	{
		fclose(fpdv);
		fpdv = NULL;
	}
}	/* end of enddvent */

/*+-------------------------------------------------------------------------
	dialstr_translate(translate_list,to_translate) - translate dial strings
--------------------------------------------------------------------------*/
void
dialstr_translate(translate_list,to_translate)
register char *translate_list;
char *to_translate;
{
	register char *cptr;

	while(*translate_list && *(translate_list + 1))
	{
		for(cptr=to_translate; *cptr; cptr++)
		{
			if(*translate_list == *cptr)
				*cptr = *(translate_list + 1);
		}
		translate_list += 2;
	}
}	/* end of dialstr_translate */

/*+-------------------------------------------------------------------------
	getdlent() - get first or next Dialers entry (a la getpwent)
--------------------------------------------------------------------------*/
struct dlent *
getdlent()
{
int itmp;
char *cptr;
#define MAX_DL_TOKENS 3
char *tokens[MAX_DL_TOKENS];
static struct dlent dle;
static char dlstr[128];
char *strchr();

	if(!there_is_hdb_on_this_machine)
		return((struct dlent *)0);

	if(fpdl == NULL)
	{
		if((fpdl = fopen(Dialers_file,"r")) == NULL)
		{
			pperror(Dialers_file);
			return((struct dlent *)0);
		}
	}

	while(1)
	{
		if(fgets(dlstr,sizeof(dlstr),fpdl) == NULL)
			return((struct dlent *)0);
		if(((itmp = strlen(dlstr)) > 0) && (dlstr[itmp - 1] == '\n'))
			dlstr[--itmp] = 0;
		if((dlstr[0] == '#') || (dlstr[0] == ' ') || (!itmp))
			continue;
		for(itmp = 0; itmp < MAX_DL_TOKENS; itmp++)
			tokens[itmp] = "";
		if(tokens[0] = arg_token(dlstr," \t\r\n"))
		{
			if(tokens[1] = arg_token((char *)0," \t\r\n"))
			{
			extern char *arg_token_static;
				tokens[2] = skip_ld_break(arg_token_static);
			}
		}
		break;
	}

	dle.name = tokens[0];
	dle.tlate = tokens[1];
	dle.script = tokens[2];
	return(&dle);

}	/* end of getdlent */

/*+-------------------------------------------------------------------------
	enddlent()
--------------------------------------------------------------------------*/
void
enddlent()
{
	if(fpdl != NULL)
	{
		fclose(fpdl);
		fpdl = NULL;
	}
}	/* end of enddlent */

/*+-------------------------------------------------------------------------
	getdlentname(name) - get Dialers entry by name
--------------------------------------------------------------------------*/
struct dlent *
getdlentname(name)
char *name;
{
register struct dlent *tdle;

	while(tdle = getdlent())
	{
		if(!strcmp(name,tdle->name))
			break;
	}
	return(tdle);

}	/* end of getdlentname */

/*+-------------------------------------------------------------------------
	hdb_dial_error(errcode)

also sets iv[0] to dial command status
--------------------------------------------------------------------------*/
char *
hdb_dial_error_text(errcode)
int errcode;
{

	iv[0] = 1;
	switch(errcode & 0x7F)
	{
		case RCE_INUSE:
			return("!Line in use");
		case RCE_SIG:
			iv[0] = 2;
			return("!Interrupted");
		case RCE_ARGS:
			return("!Invalid arguments");
		case RCE_PHNO:
			return("!Invalid phone number");
		case RCE_SPEED:
			return("!Bad baud rate");
		case RCE_OPEN:
			return("!Line open error");
		case RCE_IOCTL:
			return("!Ioctl error");
		case RCE_TIMOUT:
			iv[0] = 3;
			return("!Modem Error");
		case RCE_NOTONE:
			return("NO DIAL TONE");
		case RCE_BUSY:
			return("BUSY");
		case RCE_NOCARR:
			return("NO CARRIER");
		case RCE_ANSWER:
			return("NO ANSWER");
		default:
			
		case RCE_NULL:
			return("unknown error\n");
	}
}	/* end of hdb_dial_error */

/*+-------------------------------------------------------------------------
	hdb_dial(presult) - dial with uucp dialer if we can

return 0 if connected
       1 if dial failed
       2 if interrupted
       3 if modem error
       4 if use ecu hayes dialer
--------------------------------------------------------------------------*/
int
hdb_dial(presult)
char **presult;
{
int itmp;
int dial_pid;
int wait_status;
int old_ttymode = get_ttymode();
int (*original_sighdlr)();
struct dvent *dve;
struct dlent *dle = (struct dlent *)0;
char baudstr[10];
char dial_log[100];
char stripped_num[64];
char *sptr;
char *dptr;
static char stat_s20[20];
ulong colors_at_entry = colors_current;
extern char *make_char_graphic();
#ifdef WHT	/* hack for immediate need ... make a real feature later */
FILE *fp;
char credit_file[128];
#endif

	if(!there_is_hdb_on_this_machine)
		return(4);

	strcpy(dial_log,shm->Lline + 8);
	sptr = dial_log + strlen(dial_log) - 1;
	*sptr = to_upper(*sptr);
	dve = getdvline(dial_log);
	enddvent();
	if(!dve)
	{
		pputs("no Devices entry for line ... using ecu dialer\n");
		return(4);
	}

	dial_log[0] = 0;
	if(*dve->dialprog != '/')
	{
		dle = getdlentname(dve->dialprog);
		enddlent();
		if(!dle)
		{
			sprintf(dial_log,
				"UUCPDIAL Devices entry %s: '%s' not found in Dialers",
				shm->Lline + 5,dve->dialprog);
		}
	}
	else if(access(dve->dialprog,1))
	{
		sprintf(dial_log,"UUCPDIAL Devices entry %s: (%s) %s",
			shm->Lline + 5,dve->dialprog,sys_errlist[errno]);
	}

	if(dial_log[0])
	{
#if defined(LOG_HDBDIAL)
		ecu_log_event(dial_log);
#endif
		pputs(dial_log + 9);
		pputs("\ntrying ecu dialer\n");
		return(4);
	}

	sptr = shm->Ltelno;
	dptr = stripped_num;
	while(*sptr)
	{
		if(strchr("()-",*sptr))
		{
			sptr++;
			continue;
		}
		*dptr++ = *sptr++;
	}
	*dptr = 0;

#ifdef WHT	/* hack for immediate need ... make a real feature later */
	if(*(dptr - 1) == '#')
	{
		*--dptr = 0;
		get_home_dir(credit_file);
		strcat(credit_file,"/.ecu/.credit");
		chmod(credit_file,0400);	/* let's keep this one quiet */
		if(fp = fopen(credit_file,"r"))
		{
			*dptr++ = ',';
			*dptr++ = ',';
			*dptr++ = ',';
			*dptr++ = ',';
			*dptr = 0;
			fgets(dptr,30,fp);
			fclose(fp);
		}
		if(!fp || !(*dptr))
		{
			pprintf("credit card error\n");
			iv[0] = 1;
			return(1);
		}
		*(dptr + strlen(dptr) - 1) = 0; /* kill NL */
	}
#endif /* WHT */

	pprintf("Type %s to abort ... ",
		(kbdintr == 0x7F) ? "DEL" : make_char_graphic(kbdintr,0));
	ttymode(2);

	if(!dle)
	{
		if(access(dve->dialprog,1))
		{
			pperror(ungetty);
			pputs("trying ecu dialer\n");
			return(4);
		}
		sprintf(baudstr,"%u",shm->Lbaud);
		original_sighdlr = signal(SIGCLD,SIG_DFL);
		if((dial_pid = fork()) == 0)
		{
			signal(SIGINT,SIG_DFL);
			execl(dve->dialprog,"dial",shm->Lline,stripped_num,
				baudstr,(char *)0);
			exit(-1);
		}

		wait_status = (RC_FAIL | RCE_SIG) << 8;
		while(((itmp = wait(&wait_status)) != dial_pid) && (itmp != -1))
			;
		signal(SIGCLD,original_sighdlr);
		ttymode(old_ttymode);
		ttyflush(1);

		if(interrupt)
		{
			kill(dial_pid,9);
			ltoggle_dtr();
			interrupt = 0;	 /* in case of SIGINT by user */
		}
		lreset_ksr(); /* uucp dialers are nice guys, but lets use our termio */

#if defined(LOG_HDBDIAL)
		if(wait_status)
		{
#endif
			sprintf(dial_log,"UUCPDIAL %s %s 0x%04x",
				dve->dialprog,stripped_num,wait_status);
			ecu_log_event(dial_log);
#if defined(LOG_HDBDIAL)
		}
#endif

		/*
		 * if system reports interrupt, fake dial-reported status
		 */
		if(wait_status & 0xFF)
			wait_status = (RC_FAIL | RCE_SIG) << 8;

		if(!wait_status)
		{
			sprintf(stat_s20,"CONNECT %u",shm->Lbaud);
			*presult = stat_s20;	/* hayes_dial will report result code */
			return(0);
		}

		if((wait_status & 0xFF00) == 0xFF00)
		{
			pputs("uucp dial failure (see ~/.ecu/log) ... trying ecu dialer\n");
			return(4);
		}
		*presult = hdb_dial_error_text((wait_status >> 8) & 0xFF);
		setcolor(colors_error);
		pputs(*presult);
		setcolor(colors_at_entry);
		pputc('\n');
		ltoggle_dtr();
	}
	else
	{
		extern int expresp_verbosity;
		pprintf("using Dialers entry '%s'\n",dle->name);
		expresp_verbosity = (proc_level & proctrace) ? 1 : 0;
		if(execute_expresp(dle->script))
		{
			*presult = "DIALER SCRIPT FAILED";
			setcolor(colors_error);
			pputs(*presult);
			setcolor(colors_at_entry);
			pputc('\n');
			iv[0] = 1;
		}
		else
		{
			extern char last_Speed_result[];
			if(last_Speed_result[0])
				*presult = last_Speed_result;
			else
			{
				sprintf(stat_s20,"CONNECT %u",shm->Lbaud);
				*presult = stat_s20;	/* hayes_dial will report result code */
			}
			setcolor(colors_at_entry);
			iv[0] = 0;
		}
	}

	return((int)iv[0]);

}	/* end of hdb_dial */

/*+-------------------------------------------------------------------------
	hdb_init()
--------------------------------------------------------------------------*/
void
hdb_init()
{
	there_is_hdb_on_this_machine = !access(Devices_file,4);

}	/* end of hdb_init */

/* vi: set tabstop=4 shiftwidth=4: */
/* end of hdbintf.c */
