/* CHK=0x84E5 */
/*+-------------------------------------------------------------------------
	helpgen.c -- ecu command help file maker
	wht%n4hgf@gatech.edu

  Defined functions:
	build_ecudoc()
	build_ecuhelp()
	main(argc,argv,envp)
	search_cmd_list(cmd)
	show_cmds()
	test_help()
	to_lower(ch)
	to_upper(ch)
	usage()

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:24-wht@n4hgf---- 2.80 --- */
/*:03-25-1990-14:12-wht@n4hgf------- ecu x2.70 -------- */
/*:07-03-1989-22:58-wht------ x2.00 ----- */
/*:11-07-1988-13:58-wht-source control point 1.30 */
/*:10-24-1988-16:50-wht-works under BSD4 again (mostly)!  */
/*:06-14-1988-16:06-wht-program now multi-purpose */
/*:06-13-1988-13:46-wht-creation */

#include <stdio.h>
#include <ctype.h>
#include "lint_args.h"

#if defined(M_XENIX)
#if !defined(LINT_ARGS)
#define LINT_ARGS
#endif
#endif

#define DECLARE_P_CMD
#define HELPGEN
#include "../ecucmd.h"

#include "../esd.h"
#include "../stdio_lint.h"

#define PSRC	"ecuhelp.src"
#define PDAT	"ecuhelp.data"
#define PDOC	"ecuhelp.doc"
#define PTXT	"ecuhelp.txt"
#define PFITXT	"../doc/ecufi.txt"
#define PFSTXT	"../doc/ecufs.txt"

long	start_pos[TOKEN_QUAN];
int		token_line[TOKEN_QUAN];
FILE	*fpsrc;		/* help source file */
FILE	*fpdat;		/* help data file */
FILE	*fpdoc;		/* help doc file */
FILE	*fptxt;		/* help nroff file */
P_CMD	*pcmd;
int		src_line = 0;
char	buf[128];

/*+-------------------------------------------------------------------------
	usage()
--------------------------------------------------------------------------*/
usage()
{
	fprintf(stderr,"usage: helpgen [-b] [-d] [-s] [-t]\n");
	fprintf(stderr," -b build %s from %s\n",PDAT,PSRC);
	fprintf(stderr," -d build %s from %s\n",PDOC,PDAT);
	fprintf(stderr," -n build nroff %s from %s\n",PTXT,PDAT);
	fprintf(stderr," -s show list of commands\n");
	fprintf(stderr," -t test help\n");
	fprintf(stderr,"At least one switch must be issued.  They are executed\n");
	fprintf(stderr,"in the order shown on the usage line.\n");
	exit(1);
}	/* end of usage */

/*+-------------------------------------------------------------------------
	search_cmd_list(cmd)
--------------------------------------------------------------------------*/
P_CMD *
search_cmd_list(cmd)
register char	*cmd;
{
register P_CMD	*cmd_list = icmd_cmds;

	while(cmd_list->token != -1)
	{
		if(strcmp(cmd_list->cmd,cmd) == 0)
			break;
		cmd_list++;
	}
	if(cmd_list->token == -1)
		return((P_CMD *)0);
	else
		return(cmd_list);

}	/* end of search_cmd_list */

/*+-------------------------------------------------------------------------
	show_cmds()
commands with null descriptions are "undocumented"
--------------------------------------------------------------------------*/
void
show_cmds()
{
register int itmp;
register P_CMD *pcmd = icmd_cmds;
register int longest_cmd = 0;
register int longest_descr = 0;
register int cmd_count = 0;
register int nl_flag = 0;
char	s80[80];
P_CMD *longest_cmd_p = 0;
P_CMD *longest_descr_p = 0;

	while(pcmd->token != -1)
	{
		if(!*pcmd->descr)
		{
			pcmd++;
			continue;
		}
		itmp = strlen(pcmd->cmd);
		if(itmp > longest_cmd)
		{
			longest_cmd = itmp;
			longest_cmd_p = pcmd;
		}
		itmp = strlen(pcmd->descr);
		if(itmp > longest_descr)
		{
			longest_descr = itmp;
			longest_descr_p = pcmd;
		}
		pcmd++;
	}
	pcmd = icmd_cmds;
	while(pcmd->token != -1)
	{
		if((!pcmd->min) || (!*pcmd->descr))
		{
			pcmd++;
			continue;
		}
		strcpy(s80,pcmd->cmd);
		pad_zstr_to_len(s80,longest_cmd + 2);
		for(itmp = 0; itmp < pcmd->min; itmp++)
			s80[itmp] = to_upper(s80[itmp]);
		fputs(s80,stderr);

		strcpy(s80,pcmd->descr);
		pad_zstr_to_len(s80,longest_descr + 1);
		fputs(s80,stderr);

		if(nl_flag)
			fputs("\r\n",stderr);
		else
			fputs("| ",stderr);
		nl_flag = (nl_flag) ? 0 : 1;

		pcmd++;
	}
	if(nl_flag)
		fputs("\r\n",stderr);

	itmp = longest_cmd + longest_descr + 5;
	sprintf(s80,"recwidth = %d\r\n",itmp);
	fprintf(stderr,s80);
	pcmd = longest_cmd_p;
	sprintf(s80,"longest cmd: %s: %s\r\n",pcmd->cmd,pcmd->descr);
	fprintf(stderr,s80);
	pcmd = longest_descr_p;
	sprintf(s80,"longest dsc: %s: %s\r\n",pcmd->cmd,pcmd->descr);
	fprintf(stderr,s80);

}	/* end of show_cmds */

/*+-------------------------------------------------------------------------
	build_ecuhelp()
--------------------------------------------------------------------------*/
void
build_ecuhelp()
{
register int itmp;
register char *cptr;
	printf("\nBuilding %s\n",PDAT);
	for(itmp = 0; itmp < TOKEN_QUAN; itmp++)
	{
		start_pos[itmp] = 0L;
		token_line[itmp] = 0;
	}

	if((fpsrc = fopen(PSRC,"r")) == NULL)
	{
		perror(PSRC);
		exit(1);
	}

	if((fpdat = fopen(PDAT,"w")) == NULL)
	{
		perror(PDAT);
		exit(1);
	}

	fwrite((char *)start_pos,sizeof(long),	/* write null table */
			TOKEN_QUAN,fpdat);

	while(fgets(buf,sizeof(buf),fpsrc) != NULL)
	{
		src_line++;
		itmp = strlen(buf);
		buf[--itmp] = 0;						/* kill trailing nl */
		if((buf[0] == '#') || (buf[0] == '/'))	/* ignore comments */
			continue;
		if(buf[0] == '%')		/* command indication */
		{
SEARCH_CMD_LIST:
			if(!(pcmd = search_cmd_list(&buf[1])))
			{
				printf("line %d: '%s' not in command table\n",
						src_line,&buf[1]);
				while(fgets(buf,sizeof(buf),fpsrc) != NULL)
				{
					src_line++;
					itmp = strlen(buf);
					buf[--itmp] = 0;					/* kill trailing nl */
					if(buf[0] == '%')		/* command indication */
						goto SEARCH_CMD_LIST;
				}
				break;
			}
			if(start_pos[pcmd->token])
			{
				printf("line %d: '%s' already found on line %d\n",
						src_line,&buf[1],token_line[pcmd->token]);
				exit(1);
			}
			fputs("\n",fpdat);	/* terminate previous command description */
			start_pos[pcmd->token] = ftell(fpdat);
			token_line[pcmd->token] = src_line;
			fputs("   ",fpdat);
			cptr = &buf[1];	/* command text */
			itmp = 0;
			while(*cptr)		/* show cmd and min chars required */
			{
				if(itmp < pcmd->min)
					fputc(to_upper(*cptr++),fpdat);
				else
					fputc(to_lower(*cptr++),fpdat);
				itmp++;
			}
			if(*pcmd->descr)		/* if description present */
				fprintf(fpdat," : %s\n \n",pcmd->descr);
			else
				fputs("\n \n",fpdat);
			continue;
		}
		fprintf(fpdat," %s\n",buf);
	}

	fseek(fpdat,0L,0);	/* back to position table */
	fwrite((char *)start_pos,sizeof(long),	/* write actual table */
			TOKEN_QUAN,fpdat);
	fclose(fpsrc);
	fputs("\n",fpdat);	/* terminate last command */
	fclose(fpdat);
}	/* end of build_ecuhelp */

/*+-------------------------------------------------------------------------
	build_ecudoc()
--------------------------------------------------------------------------*/
void
build_ecudoc()
{
register int itmp;
register char *cptr;

	printf("\nBuilding %s\n",PDOC);
	if((fpdat = fopen(PDAT,"r")) == NULL)
	{
		perror(PDAT);
		exit(1);
	}
	if((fpdoc = fopen(PDOC,"w")) == NULL)
	{
		perror(PDOC);
		exit(1);
	}
	fprintf(fpdoc,
		"\n     ECU  Command  Help  Documentation  (PRELIMINARY)\n\n");
	fprintf(fpdoc,
		"Commands are accessed by pressing the HOME key followed by one\n");
	fprintf(fpdoc,
		"of the following commands (capitalized portions are sufficient\n");
	fprintf(fpdoc,
		"to invoke the command):\n");
	fprintf(fpdoc,"\n");
	fprintf(fpdoc,
"---------------------------------------------------------------------\n");
	fread((char *)start_pos,sizeof(long),TOKEN_QUAN,fpdat);
	pcmd = icmd_cmds;
	while(pcmd->token != -1)
	{
		if(!pcmd->token)
		{
			pcmd++;
			continue;
		}
		if(pcmd->min && !start_pos[pcmd->token])
		{
			printf("no help available for '%s'\n",pcmd->cmd);
			pcmd++;
			continue;
		}
		fseek(fpdat,start_pos[pcmd->token],0);
		while(fgets(buf,sizeof(buf),fpdat) != NULL)
		{
			itmp = strlen(buf);
			buf[--itmp] = 0;
			if(itmp == 0)
				break;
			fprintf(fpdoc,"%s\n",buf);
		}
		fprintf(fpdoc,
"---------------------------------------------------------------------\n");
		pcmd++;
	}
	fclose(fpdat);
	fclose(fpdoc);
}	/* end of build_ecudoc */

/*+-------------------------------------------------------------------------
	build_ecutxt()
--------------------------------------------------------------------------*/
void
build_ecutxt()
{
register int itmp;
register char *cptr;

	printf("\nBuilding %s\n",PTXT);
	if((fpdat = fopen(PDAT,"r")) == NULL)
	{
		perror(PDAT);
		exit(1);
	}
	if((fptxt = fopen(PTXT,"w")) == NULL)
	{
		perror(PTXT);
		exit(1);
	}

	fread((char *)start_pos,sizeof(long),TOKEN_QUAN,fpdat);
	pcmd = icmd_cmds;
	while(pcmd->token != -1)
	{
		if(!pcmd->token)
		{
			pcmd++;
			continue;
		}
		if(pcmd->min && !start_pos[pcmd->token])
		{
			printf("no help available for '%s'\n",pcmd->cmd);
			pcmd++;
			continue;
		}
		fseek(fpdat,start_pos[pcmd->token],0);
		if(fgets(buf,sizeof(buf),fpdat) != NULL)
		itmp = strlen(buf);
		buf[--itmp] = 0;
		cptr = buf;
		while(*cptr)
			if(*cptr == ' ')
				cptr++;
			else
				break;
		fprintf(fptxt,".*s 2 \"%s\"\n.(b L",cptr);
		while(fgets(buf,sizeof(buf),fpdat) != NULL)
		{
			itmp = strlen(buf);
			buf[--itmp] = 0;
			if(itmp == 0)
				break;
			fprintf(fptxt,"%s\n",buf);
		}
		fprintf(fptxt,"\n.)b\n\n");
		pcmd++;
	}
	fclose(fpdat);
	fclose(fptxt);
}	/* end of build_ecutxt */

/*+-------------------------------------------------------------------------
	test_help()
--------------------------------------------------------------------------*/
void
test_help()
{
register int itmp;
register char *cptr;

/* test code */
	printf("\nNow to test\n");
	if((fpdat = fopen(PDAT,"r")) == NULL)
	{
		perror(PDAT);
		exit(1);
	}
	fread((char *)start_pos,sizeof(long),TOKEN_QUAN,fpdat);
	while(1)
	{
		printf("\ncommand: ");
		fgets(buf,sizeof(buf),stdin);
		itmp = strlen(buf);
		buf[--itmp] = 0;
		if(itmp == 0)
			break;
		if(!(pcmd = search_cmd_list(buf)))
		{
			printf("'%s' not found in ecu cmd table\n",buf);
			continue;
		}
		if(pcmd->min && !start_pos[pcmd->token])
		{
			printf("no help available for '%s'\n",buf);
			continue;
		}
		fseek(fpdat,start_pos[pcmd->token],0);
		while(fgets(buf,sizeof(buf),fpdat) != NULL)
		{
			itmp = strlen(buf);
			buf[--itmp] = 0;
			if(itmp == 0)
				break;
			printf("%s\n",buf);
		}
	}
}	/* end of test_help */

/*+-------------------------------------------------------------------------
	main(argc,argv,envp)
--------------------------------------------------------------------------*/
main(argc,argv,envp)
int		argc;
char	**argv;
char	**envp;
{
register int itmp;
register char *cptr;
int		iargv;
int		b_flag = 0;
int		s_flag = 0;
int		t_flag = 0;
int		f_flag = 0;
int		d_flag = 0;
int		n_flag = 0;

	setbuf(stdout,NULL);
	setbuf(stderr,NULL);

	if(argc < 1)
		usage();
	for(iargv = 1; iargv < argc; iargv++)
	{
		if(argv[iargv][0] == '-')
		{
			switch(itmp = (argv[iargv][1]))
			{
				case 'b': b_flag = 1; break;
				case 's': s_flag = 1; break;
				case 't': t_flag = 1; break;
				case 'd': d_flag = 1; break;
				case 'n': n_flag = 1; break;
				default:
					usage();
					break;
			}
		}
		else
			usage();
	}
	if(!b_flag && !s_flag && !t_flag && !d_flag && !n_flag && !f_flag)
		usage();

	if(b_flag)
		build_ecuhelp();
	if(d_flag)
		build_ecudoc();
	if(n_flag)
		build_ecutxt();
	if(s_flag)
		show_cmds();
	if(t_flag)
		test_help();

	exit(0);
}	/* end of main */
/* end of helpgen.c */
/* vi: set tabstop=4 shiftwidth=4: */
