/* CHK=0x6CA5 */
/*+-------------------------------------------------------------------------
	pcmdif.c - ecu procedure if commands
	wht%n4hgf@gatech.edu

    IFI $i0 rel-op $i1 cmd
    IFS $s0 rel-op $s1 cmd

    where rel-op is "=", "==", "!=", "<>", ">", "<", ">=", "=<"

  Defined functions:
	_if_common(param,truth)
	cmd_else(param)
	cmd_ifi(param)
	cmd_ifs(param)
	get_logicop(param,op_returned)
	get_relop(param,op_returned)
	get_truth_int(param,truth)
	get_truth_str(param,truth)
	test_truth_int(int1,relop,int2)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:24-wht@n4hgf---- 2.80 --- */
/*:03-25-1990-14:11-wht@n4hgf------ x2.70 ------- */
/*:11-06-1989-02:31-wht-add logical operators */
/*:07-03-1989-22:57-wht------ x2.00 ----- */
/*:06-24-1989-16:53-wht-flush edits --- ecu 1.95 */

#include <ctype.h>
#include "ecu.h"
#include "ecuerror.h"
#include "esd.h"
#include "var.h"
#include "proc.h"
#include "relop.h"

extern PCB *pcb_stack[];
extern int proc_level;
extern int proctrace;

#define MAX_IF 40	/* damn enough */
uchar if_level = 0;
uchar truth_already[MAX_IF];

/*+-------------------------------------------------------------------------
    get_relop(param,&op_returned)
--------------------------------------------------------------------------*/
int
get_relop(param,op_returned)
ESD *param;
int *op_returned;
{
register erc;

	if(erc = skip_cmd_break(param))
		return(eInvalidRelOp);

	switch(param->pb[param->index++]) /* index decremented in default */
	{
	case '=':
		if((param->cb != param->index) && (param->pb[param->index] == '='))
			param->index++;
		*op_returned = OP_EQ;
		return(0);

	case '!':
		if(param->cb == param->index)
			return(eInvalidRelOp);
		switch(param->pb[param->index])
		{
		case '=':
			param->index++;
			*op_returned = OP_NE;
			return(0);
		default:
			return(eInvalidRelOp);
		}

	case '<':
		if(param->cb == param->index)
		{
			*op_returned = OP_LT;
			return(0);
		}
		switch(param->pb[param->index])
		{
		case '>':
			param->index++;
			*op_returned = OP_NE;
			return(0);
		case '=':
			param->index++;
			*op_returned = OP_LE;
			return(0);
		default:
			*op_returned = OP_LT;
			return(0);
		}

	case '>':
		if(param->cb == param->index)
		{
			*op_returned = OP_LT;
			return(0);
		}
		switch(param->pb[param->index])
		{
		case '=':
			param->index++;
			*op_returned = OP_GE;
			return(0);
		default:
			*op_returned = OP_GT;
			return(0);
		}
	default:
		param->index--;
	}
	return(eInvalidRelOp);
}   /* end of get_relop */

/*+-------------------------------------------------------------------------
	get_logicop(param,op_returned)
--------------------------------------------------------------------------*/
int
get_logicop(param,op_returned)
ESD *param;
int *op_returned;
{
register erc;
register char *cptr;

	if(erc = skip_cmd_break(param))
		return(eInvalidLogicOp);

	if((param->cb - param->index) < 2)
		return(eInvalidLogicOp);

	cptr = param->pb + param->index;
	erc = eInvalidLogicOp;
	if(!strncmp(cptr,"&&",2))
	{
		*op_returned = OP_AND;
		erc = 0;
	}
	else if(!strncmp(cptr,"||",2))
	{
		*op_returned = OP_OR;
		erc = 0;
	}
	if(!erc)
		param->index += 2;
	return(erc);

}	/* end of get_logicop */

/*+-------------------------------------------------------------------------
	test_truth_int(int1,relop,int2)
--------------------------------------------------------------------------*/
int
test_truth_int(int1,relop,int2)
long int1;
int relop;
long int2;
{
register truth;

	switch(relop)
	{
	case OP_EQ:
		truth = (int1 == int2);
		break;
	case OP_NE:
		truth = (int1 != int2);
		break;
	case OP_GT:
		truth = (int1 > int2);
		break;
	case OP_LT:
		truth = (int1 < int2);
		break;
	case OP_GE:
		truth = (int1 >= int2);
		break;
	case OP_LE:
		truth = (int1 <= int2);
		break;
	}
	return(truth);

}	/* end of test_truth_int */

/*+-------------------------------------------------------------------------
	get_truth_int(param,truth)
--------------------------------------------------------------------------*/
int
get_truth_int(param,truth)
ESD *param;
int *truth;
{
register erc;
long int1;
long int2;
int operator;
int truth2;

	if(erc = gint(param,&int1))
		return(erc);
	if(erc = get_relop(param,&operator))
		return(erc);
	if(erc = gint(param,&int2))
		return(erc);
	*truth = test_truth_int(int1,operator,int2);

	while(!get_logicop(param,&operator))
	{
		if(erc = get_truth_int(param,&truth2))
			return(erc);
		switch(operator)
		{
			case OP_AND:
				*truth &= truth2;
				break;

			case OP_OR:
				*truth |= truth2;
				break;
		}
	}
	return(erc);


}	/* end of get_truth_int */

/*+-------------------------------------------------------------------------
    get_truth_str(param,truth)
--------------------------------------------------------------------------*/
int
get_truth_str(param,truth)
ESD *param;
int *truth;
{
register erc;
ESD str1,str2;
char str1_str[256];
char str2_str[256];
int operator;
int strcmp_result;
int truth2;

	str1.pb = str1_str;
	str1.maxcb = sizeof(str1_str) - 1;
	str1.cb = 0;
	str1.index = 0;

	str2.pb = str2_str;
	str2.maxcb = sizeof(str2_str) - 1;
	str2.cb = 0;
	str2.index = 0;

	if(erc = gstr(param,&str1))
		return(erc);
	if(erc = get_relop(param,&operator))
		return(erc);
	if(erc = gstr(param,&str2))
		return(erc);

	null_terminate_esd(&str1);
	null_terminate_esd(&str2);

	strcmp_result = strcmp(str1.pb,str2.pb);

	switch(operator)
	{
		case OP_EQ:
			*truth = (strcmp_result == 0);
			break;
		case OP_NE:
			*truth = (strcmp_result != 0);
			break;
		case OP_GT:
			*truth = (strcmp_result > 0);
			break;
		case OP_LT:
			*truth = (strcmp_result < 0);
			break;
		case OP_GE:
			*truth = (strcmp_result >= 0);
			break;
		case OP_LE:
			*truth = (strcmp_result <= 0);
			break;
		default:
			return(eInvalidStrOp);
	}

	while(!get_logicop(param,&operator))
	{
		if(erc = get_truth_str(param,&truth2))
			return(erc);
		switch(operator)
		{
			case OP_AND:
				*truth &= truth2;
				break;

			case OP_OR:
				*truth |= truth2;
				break;
		}
	}

	return(0);

}   /* end of get_truth_str */

/*+-------------------------------------------------------------------------
	_if_common(param,truth)
--------------------------------------------------------------------------*/
int
_if_common(param,truth)
ESD *param;
int truth;
{
register erc = 0;
char s80[80];
PCB *pcb;
ESD *else_line;
int label_on_else_line;
int truth2;
int save_index;

	if(proctrace > 1)
	{
		pprintf("if condition %s",(truth) ? "TRUE: " : "FALSE\n");
		if(truth)
		{
			skip_cmd_break(param);
			pputs(param->pb + param->index);
			pputc('\n');
		}
	}

	truth_already[if_level] = truth;

/* if end of command, execute frame else conditionally execute rest of esd */
	s80[0] = 0;
	if(end_of_cmd(param))
		erc = execute_frame(truth);
	else if(truth)
		erc = execute_esd(param);
	else
		param->index = param->cb;

	if(erc)
		return(erc);

/* check for else statement */
	pcb = pcb_stack[proc_level - 1];
	if(!pcb->current->next)		/* if no next line, no "else" */
		return(0);

	else_line = pcb->current->next->text;
	else_line->index = else_line->old_index = 0;
	if(label_on_else_line = (*else_line->pb != 0x20))
	{	/* strip label */
		if(get_alphanum_zstr(else_line,s80,sizeof(s80)))
			return(eInvalidLabel);
	}
	if(get_alphanum_zstr(else_line,s80,sizeof(s80)))
		return(0);		/* not "else" */
	if(strcmp(s80,"else"))
		return(0);		/* not "else" */
	if(label_on_else_line)
	{
		else_line->old_index = 0;
		pputs("label not allowed on else statement\n");
		return(eFATAL_ALREADY);
	}

/* we have an "else" condition */
	truth = !truth;
	pcb->current = pcb->current->next;

	trace_proc_cmd(pcb);

	if(end_of_cmd(else_line))
		erc = execute_frame(truth);
	else
	{
		save_index = else_line->old_index = else_line->index;
		s80[0] = 0;
		if((*(else_line->pb + else_line->index) != '$') &&
			get_alpha_zstr(else_line,s80,sizeof(s80)))
		{
			pputs("illegal command after 'else'\n");
			return(eFATAL_ALREADY);
		}
		if(!strcmp(s80,"ifi"))
		{
			if(erc = get_truth_int(else_line,&truth2))
				return(erc);
			erc = _if_common(else_line,!truth_already[if_level] & truth2);
			truth_already[if_level] |= truth2;
		}
		else if(!strcmp(s80,"ifs"))
		{
			if(erc = get_truth_str(else_line,&truth2))
				return(erc);
			erc = _if_common(else_line,!truth_already[if_level] & truth2);
			truth_already[if_level] |= truth2;
		}
		else if(!strncmp(s80,"while",5))
		{
			pputs("'while' command not allowed as 'else' conditional\n");
			return(eFATAL_ALREADY);
		}
		else
		{
			else_line->index = save_index;
			if(truth)
				erc = execute_esd(else_line);
		}
	}

	return(erc);
}	/* end of _if_common */

/*+-------------------------------------------------------------------------
    cmd_ifi(param)
--------------------------------------------------------------------------*/
int
cmd_ifi(param)
ESD *param;
{
register erc;
int truth;

	if(if_level == MAX_IF)
	{
		pputs("if statements nested too deeply\n");
		return(eFATAL_ALREADY);
	}
	if_level++;
	truth_already[if_level] = 0;

	if(!proc_level)
		return(eNotExecutingProc);

	if(!(erc = get_truth_int(param,&truth)))
		erc = _if_common(param,truth);
	if_level--;
	return(erc);

}   /* end of cmd_ifi */

/*+-------------------------------------------------------------------------
    cmd_ifs(param)
--------------------------------------------------------------------------*/
int
cmd_ifs(param)
ESD *param;
{
register erc;
int truth;

	if(!proc_level)
		return(eNotExecutingProc);

	if(if_level == MAX_IF)
	{
		pputs("if statements nested too deeply\n");
		return(eFATAL_ALREADY);
	}
	if_level++;
	truth_already[if_level] = 0;

	if(!(erc = get_truth_str(param,&truth)))
		erc = _if_common(param,truth);
	if_level--;
	return(erc);

}   /* end of cmd_ifs */

/*+-------------------------------------------------------------------------
	cmd_else(param)
--------------------------------------------------------------------------*/
int
cmd_else(param)
ESD *param;
{
	return(eElseCommand);
}	/* end of cmd_else */

/* vi: set tabstop=4 shiftwidth=4: */
/* end of pcmdif.c */
