/* CHK=0x1B76 */
/*+-------------------------------------------------------------------------
	pcmdxfer.c - ecu xfer-related procedure commands
	wht%n4hgf@gatech.edu

  Defined functions:
	_make_bottom_label(param,default_flag,sending_flag)
	_rx_error()
	_send_common(execcmd)
	cmd_rk(param)
	cmd_rs(param)
	cmd_rx(param)
	cmd_ry(param)
	cmd_rz(param)
	cmd_sk(param)
	cmd_ss(param)
	cmd_sx(param)
	cmd_sy(param)
	cmd_sz(param)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:24-wht@n4hgf---- 2.80 --- */
/*:04-19-1990-03:08-wht@n4hgf-GCC run found unused vars -- rm them */
/*:03-27-1990-15:18-wht@n4hgf-overhaul for/wait/SIGCLD handling */
/*:03-25-1990-14:11-wht@n4hgf------ x2.70 ------- */
/*:03-12-1990-04:32-wht@n4hgf-use Lrname */
/*:07-03-1989-22:57-wht------ x2.00 ----- */
/*:06-24-1989-16:53-wht-flush edits --- ecu 1.95 */

#include "ecu.h"
#include "ecukey.h"
#include "ecuerror.h"
#include "esd.h"
#include "var.h"
#include "proc.h"

char *strchr();

extern int proctrace;
extern int last_child_wait_status;
extern char curr_dir[];
extern proc_level;
extern PCB *pcb_stack[];

static char bottom_label[80];

/*+-------------------------------------------------------------------------
	_make_bottom_label(param,default_flag)
--------------------------------------------------------------------------*/
_make_bottom_label(param,default_flag,sending_flag)
ESD *param;
int default_flag;
int sending_flag;
{
register erc;
ESD *label = (ESD *)0;

	if(default_flag)
	{
		sprintf(bottom_label,
			(sending_flag) ? "-C \"'Connected to %s'\" "
			               : "-C 'Connected to %s' ",
			(shm->Lrname[0]) ? shm->Lrname : "?");
		return(0);
	}
	if((label = make_esd(64)) == (ESD *)0)
		return(eNoMemory);
	if(erc = gstr(param,label))
	{
		free_esd(label);
		return(erc);
	}
	strcpy(bottom_label,(sending_flag) ? "-C \"'" : "-C '");
	strcat(bottom_label,label->pb);
	strcat(bottom_label,(sending_flag) ? "'\"" : "'");
	free_esd(label);
	return(0);

}	/* end of _make_bottom_label */

/*+-------------------------------------------------------------------------
	_send_common(execcmd)
for "ecu knowledgeable" protocols only
--------------------------------------------------------------------------*/
int
_send_common(execcmd)
char *execcmd;
{
int erc = 0;

	file_xfer_start();
	if(find_shell_chars(execcmd))
	{
	char *expcmd;

		if(expand_cmd_with_wildlist(execcmd,&expcmd))
		{
			pputs("No files match wildcard list\n");
			iv[0] = -1;
			return(0);
		}
		else
		{
			if(exec_cmd(expcmd))
				erc = eFATAL_ALREADY;
			free(expcmd);
		}
	}
	else
	{
		if(exec_cmd(execcmd))
			erc = eFATAL_ALREADY;
	}
	lreset_ksr();
	file_xfer_done_bell();
	if(erc)
		return(erc);
	erc = report_send_status();
	if(!erc)
		iv[0] = 0;
	if(erc == eProcAttn_Interrupt)
	{
		erc = 0;
		iv[0] = 0x100L;
	}
	return(erc);
		

}	/* end of _send_common */

/*+-------------------------------------------------------------------------
	cmd_sx(param)

sx [-ak[l]] [<label-str>] <filelist-str>
--------------------------------------------------------------------------*/
int
cmd_sx(param)
ESD *param;
{
int erc;
char execcmd[384];
char switches[8];
ESD *pathlist;

	get_switches(param,switches,sizeof(switches));

	if(erc = _make_bottom_label(param,strchr(switches,'l') == (char *)0,1))
		return(erc);

	sprintf(execcmd,"ecusz -X -. %d ",shm->Liofd);
	strcat(execcmd,bottom_label);
	if(strchr(switches,'a'))
		strcat(execcmd,"-a ");
	else
		strcat(execcmd,"-b ");
	if(strchr(switches,'k'))
		strcat(execcmd,"-k");

	if((pathlist = make_esd(256)) == (ESD *)0)
		return(eNoMemory);

	if(erc = gstr(param,pathlist))
	{
		free_esd(pathlist);
		return(erc);
	}

	strcat(execcmd,pathlist->pb);
	free_esd(pathlist);
	return(_send_common(execcmd));

}	/* end of cmd_sx */

/*+-------------------------------------------------------------------------
	cmd_sy(param)

sy [-a[l]] [<label-str>] <filelist-str>
--------------------------------------------------------------------------*/
int
cmd_sy(param)
ESD *param;
{
int erc;
char execcmd[384];
char switches[8];
ESD *pathlist;

	get_switches(param,switches,sizeof(switches));

	if(erc = _make_bottom_label(param,strchr(switches,'l') == (char *)0,1))
		return(erc);

	sprintf(execcmd,"ecusz -Y -. %d -k ",shm->Liofd);
	strcat(execcmd,bottom_label);
	if(strchr(switches,'a'))
		strcat(execcmd,"-a ");
	else
		strcat(execcmd,"-b ");

	if((pathlist = make_esd(256)) == (ESD *)0)
		return(eNoMemory);

	if(erc = gstr(param,pathlist))
	{
		free_esd(pathlist);
		return(erc);
	}

	strcat(execcmd,pathlist->pb);
	free_esd(pathlist);
	return(_send_common(execcmd));

}	/* end of cmd_sy */

/*+-------------------------------------------------------------------------
	cmd_sz(param)

sz [-anf[l]] [<label-str>] <filelist-str>
-a ascii, else binary
-n send only newer, else all files
-f full, else simple pathnames
-l non-default bottom line label on transfer string
$i0 set to:
	0:       file transfer completely successful
	-1 program did not run
--------------------------------------------------------------------------*/
int
cmd_sz(param)
ESD *param;
{
int erc;
char execcmd[384];
char switches[8];
ESD *pathlist;

	get_switches(param,switches,sizeof(switches));

	if(erc = _make_bottom_label(param,strchr(switches,'l') == (char *)0,1))
		return(erc);

	sprintf(execcmd,"ecusz -Z -. %d -y ",shm->Liofd);
	strcat(execcmd,bottom_label);
	if(strchr(switches,'a'))
		strcat(execcmd,"-a ");
	else
		strcat(execcmd,"-b ");
	if(strchr(switches,'n'))
		strcat(execcmd,"-n ");	/* overrides -y choice earlier */
	if(strchr(switches,'f'))
		strcat(execcmd,"-f ");

	if((pathlist = make_esd(256)) == (ESD *)0)
		return(eNoMemory);

	if(erc = gstr(param,pathlist))
	{
		free_esd(pathlist);
		return(erc);
	}

	strcat(execcmd,pathlist->pb);
	free_esd(pathlist);
	return(_send_common(execcmd));

}	/* end of cmd_sz */

/*+-------------------------------------------------------------------------
	cmd_sk(param)

sk [-a] <str>
--------------------------------------------------------------------------*/
int
cmd_sk(param)
ESD *param;
{
int erc;
char execcmd[384];
ESD *pathlist;
char switches[8];

	get_switches(param,switches,sizeof(switches));

	sprintf(execcmd,"ckermit -l %d -b %u -p %c%s%s -s ",
		shm->Liofd,shm->Lbaud,
		(shm->Lparity) ? shm->Lparity : 'n',
		(strchr(switches,'a')) ? "" : " -i",
		(1 /* overwrite*/ ) ? "" : " -w");

	if(strchr(switches,'b'))
		strcat(execcmd,"-a ");
	else
		strcat(execcmd,"-b ");
	if(strchr(switches,'n'))
		strcat(execcmd,"-n ");	/* overrides -y choice earlier */
	if(strchr(switches,'f'))
		strcat(execcmd,"-f ");

	if((pathlist = make_esd(256)) == (ESD *)0)
		return(eNoMemory);

	if(erc = gstr(param,pathlist))
	{
		free_esd(pathlist);
		return(erc);
	}

	file_xfer_start();
	strcat(execcmd,pathlist->pb);
	free_esd(pathlist);
	if(exec_cmd(execcmd))
		erc = eFATAL_ALREADY;
	file_xfer_done_bell();
	lreset_ksr();
	return(erc);
}	/* end of cmd_sk */

/*+-------------------------------------------------------------------------
	cmd_ss(param)
--------------------------------------------------------------------------*/
int
cmd_ss(param)
ESD *param;
{
int erc;
char execcmd[384];
ESD *pathlist;

	sprintf(execcmd,"ecusea -. %d -/ %s -s ",shm->Liofd,curr_dir);
	if((pathlist = make_esd(256)) == (ESD *)0)
		return(eNoMemory);

	if(erc = gstr(param,pathlist))
	{
		free_esd(pathlist);
		return(erc);
	}

	strcat(execcmd,pathlist->pb);
	free_esd(pathlist);
	return(_send_common(execcmd));

}	/* end of cmd_ss */

/*+-------------------------------------------------------------------------
	_rx_error()
--------------------------------------------------------------------------*/
void
_rx_error()
{
extern int last_child_wait_status;

	iv[0] = (last_child_wait_status & 0xFF)
			? 0x100L : (long)last_child_wait_status >> 8;
	if(proctrace)
		pprintf("$i00 = %ld, (%s)\n",iv[0],
			(iv[0] == 0x100L) ? "interrupted" : "program exit status");
	file_xfer_done_bell();
}	/* end of _rx_error */

/*+-------------------------------------------------------------------------
	cmd_rx(param)

rx [-b] <str>
--------------------------------------------------------------------------*/
int
cmd_rx(param)
ESD *param;
{
int erc = 0;
char execcmd[512];
char switches[8];
ESD *tesd = make_esd(256);

	if(!tesd)
		return(eNoMemory);

	get_switches(param,switches,sizeof(switches));

	if(erc = gstr(param,tesd))
	{
		free_esd(tesd);
		return(erc);
	}

	last_child_wait_status = 0;
	if(erc = _make_bottom_label(param,strchr(switches,'l') == (char *)0,0))
		return(erc);

	sprintf(execcmd,"ecurz -X -. %d -c ",shm->Liofd);
	strcat(execcmd,bottom_label);
	if(!strchr(switches,'b'))
		strcat(execcmd,"-b ");
	else
		strcat(execcmd,"-a ");
	strcat(execcmd,tesd->pb);
	free_esd(tesd);
	file_xfer_start();
	if(exec_cmd(execcmd))
		erc = eFATAL_ALREADY;
	else
		_rx_error();
	return(erc);
}	/* end of cmd_rx */

/*+-------------------------------------------------------------------------
	cmd_ry(param)

ry
--------------------------------------------------------------------------*/
int
cmd_ry(param)
ESD *param;
{
int erc = 0;
char execcmd[512];
char switches[8];

	get_switches(param,switches,sizeof(switches));

	if(erc = _make_bottom_label(param,strchr(switches,'l') == (char *)0,0))
		return(erc);

	last_child_wait_status = 0;
	sprintf(execcmd,"ecurz -Y -. %d -y ",shm->Liofd);
	strcat(execcmd,bottom_label);
	file_xfer_start();
	if(exec_cmd(execcmd))
		erc = eFATAL_ALREADY;
	else
		_rx_error();
	return(erc);
}	/* end of cmd_ry */

/*+-------------------------------------------------------------------------
	cmd_rz(param)
--------------------------------------------------------------------------*/
int
cmd_rz(param)
ESD *param;
{
int erc = 0;
char execcmd[512];
char switches[8];

	get_switches(param,switches,sizeof(switches));

	if(erc = _make_bottom_label(param,strchr(switches,'l') == (char *)0,0))
		return(erc);

	last_child_wait_status = 0;
	sprintf(execcmd,"ecurz -Z -. %d ",shm->Liofd);
	strcat(execcmd,bottom_label);
	file_xfer_start();
	if(exec_cmd(execcmd))
		erc = eFATAL_ALREADY;
	else
		_rx_error();
	return(erc);
}	/* end of cmd_rz */

/*+-------------------------------------------------------------------------
	cmd_rk(param)

rk [-a]
--------------------------------------------------------------------------*/
int
cmd_rk(param)
ESD *param;
{
int erc = 0;
char execcmd[300];
char switches[8];

	get_switches(param,switches,sizeof(switches));

	last_child_wait_status = 0;
	sprintf(execcmd,"ckermit -r -e 512 -l %d -b %d -p %c",
		shm->Liofd,shm->Lbaud,(shm->Lparity) ? shm->Lparity : 'n');
	if(strchr(switches,'a'))
		strcat(execcmd,"-i ");
	file_xfer_start();
	if(exec_cmd(execcmd))
		erc = eFATAL_ALREADY;
	file_xfer_done_bell();
	return(erc);
}	/* end of cmd_rk */

/*+-------------------------------------------------------------------------
	cmd_rs(param)

rs
--------------------------------------------------------------------------*/
int
cmd_rs(param)
ESD *param;
{
int erc = 0;
char execcmd[128];

	file_xfer_start();
	sprintf(execcmd,"ecusea -r -. %d -/ %s",shm->Liofd,curr_dir);
	if(exec_cmd(execcmd))
		erc = eFATAL_ALREADY;
	file_xfer_done_bell();
	return(erc);
}	/* end of cmd_rs */


/* vi: set tabstop=4 shiftwidth=4: */
/* end of pcmdxfer.c */
