/* CHK=0x83D2 */
/*+-------------------------------------------------------------------------
	procframe.c - execute frame of statements
	wht%n4hgf@gatech.edu

  Defined functions:
	cmd_break(param);
	cmd_continue(param);
	execute_frame(truth)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:24-wht@n4hgf---- 2.80 --- */
/*:03-25-1990-14:12-wht@n4hgf------ x2.70 ------- */
/*:07-03-1989-22:57-wht------ x2.00 ----- */
/*:06-24-1989-16:53-wht-flush edits --- ecu 1.95 */

#include <ctype.h>
#include "ecu.h"
#include "ecukey.h"
#include "ecuerror.h"
#include "esd.h"
#include "var.h"
#include "proc.h"

extern PCB *pcb_stack[PROC_STACK_MAX];
extern int proc_level;
extern int proc_interrupt;
extern int proctrace;

/*+-------------------------------------------------------------------------
	cmd_break(param)
--------------------------------------------------------------------------*/
int
cmd_break(param)
ESD *param;
{
	return(eBreakCommand);
}	/* end of cmd_break */

/*+-------------------------------------------------------------------------
	cmd_continue(param)
--------------------------------------------------------------------------*/
int
cmd_continue(param)
ESD *param;
{
	return(eContinueCommand);
}	/* end of cmd_continue */

/*+-------------------------------------------------------------------------
	execute_frame(truth)

  pcb_stack[proc_level - 1]->current points to lcb behind frame: one
  statement or { statements }

  if truth true, execute frame, else skip it
--------------------------------------------------------------------------*/
int
execute_frame(truth)
int truth;
{
register itmp;
int erc = 0;
PCB *pcb = pcb_stack[proc_level - 1];
LCB *original_lcb = pcb->current;
LCB *begin_lcb;
ESD *text;
int nest_level = 0;
int remember_break = 0;

	if(!(pcb->current = pcb->current->next))
	{
		pcb->current = original_lcb;
		return(eNoFrame);
	}

	text = pcb->current->text;
	text->old_index = text->index = 0;

	if(*text->pb != SPACE)
		return(eLabelInvalidHere);
	skip_cmd_break(text);

/* handle single statement frame */
	if(*(text->pb + text->index) != '{')
	{
		itmp = text->cb - text->index;
		if( ((itmp > 2) && !strncmp(text->pb + text->index,"if",2)))
		{
			pputs("command must appear inside {} or on same line as else\n");
			erc = eFATAL_ALREADY;
		}
		else if( ((itmp > 5) && !strncmp(text->pb + text->index,"while",5)))
		{
			pputs("command must appear inside {} within this context\n");
			erc = eFATAL_ALREADY;
		}
		else if(truth)
		{
			trace_proc_cmd(pcb);
			erc = execute_esd(text);
		}
		return(erc);
	}

/* we've got a {} frame */
	begin_lcb = pcb->current;
	pcb->current = pcb->current->next;
	while(pcb->current)
	{
		if(proc_interrupt)
			return(eCONINT);
		text = pcb->current->text;
		text->old_index = text->index = 0;
		if(*text->pb != SPACE)
			return(eLabelInvalidHere);
		skip_cmd_break(text);
		if(*(text->pb + text->index) == '}')
		{
			if(!nest_level)
			{
				text->index = text->cb;
				if(remember_break)
					return(eBreakCommand);
				return(0);
			}
			nest_level--;
		}
		else if(truth)
		{
			trace_proc_cmd(pcb);
			if(erc = execute_esd(text))
			{
				if(erc != eBreakCommand)
					return(erc);
				remember_break = 1;
				truth = 0;
			}
		}
		else if(*(text->pb + text->index) == '{')
			nest_level++;
		pcb->current = pcb->current->next;
	}
	pcb->current = begin_lcb;
	return(eNoCloseFrame);
	
}	/* end of execute_frame */

/* vi: set tabstop=4 shiftwidth=4: */
/* end of procframe.c */
