/* CHK=0xEAB0 */

#if defined(WHT)
#define LOG_UTMP
#endif
/*+-------------------------------------------------------------------------
	utmpstat.c - utmp status for XENIX/UNIX line
	wht%n4hgf@gatech.edu

  Defined functions:
	strcmpi(s1,s2)
	utmp_status(line)

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:05-07-1990-21:24-wht@n4hgf---- 2.80 --- */
/*:03-25-1990-14:12-wht@n4hgf------ x2.70 ------- */
/*:03-23-1990-14:44-wht@n4hgf-add strcmpi */
/*:11-24-1989-21:04-wht-creation */

#include "ecu.h"
#include "ecuhangup.h"
#include "utmpstatus.h"
#include "ecuungetty.h"
#include "dialer.h"
#include <errno.h>
#include <utmp.h>

extern int errno;

char *utmp_file = "/etc/utmp";
struct utmp last_utmp;

/*+-------------------------------------------------------------------------
	strcmpi(s1,s2) - case-insensitive strcmp
--------------------------------------------------------------------------*/
int
strcmpi(s1,s2)
register char *s1;
register char *s2;
{
register char c1,c2;

	for ( ; *s1 ; s1++, s2++)
	{
		if(isupper(*s1))
			c1 = tolower(*s1);
		else
			c1 = *s1;
		if(isupper(*s2))
			c2 = tolower(*s2);
		else
			c2 = *s2;
		if(c1 != c2)
			break;
	}
	return (*s1 - *s2);
}	/* end of strcmpi */

/*+-------------------------------------------------------------------------
	utmp_status(line) - check line status in utmp
'line' is "/dev/ttyxx"-style
returns US_ value and global utmp struct last_utmp;
--------------------------------------------------------------------------*/
int
utmp_status(line)
char *line;
{
register ufd;
register itmp;
register status = US_NOTFOUND;
#if defined(LOG_UTMP)
char logstr[128];
#endif

	if((ufd = open(utmp_file,O_RDONLY,755)) < 0)
	{
		perror(utmp_file);
		hangup(HANGUP_LINE_OPEN_ERROR);
	}

	while(read(ufd,(char *)&last_utmp,sizeof(last_utmp)) > 0)
	{
#if defined(M_UNIX)
	    if((!strcmp(last_utmp.ut_name,"uugetty") ||
	     !strcmp(last_utmp.ut_name,"getty")) &&
		 (ulindex(last_utmp.ut_id,line + 8) == (strlen(last_utmp.ut_id) - 2)))
		{
			if(itmp = line_locked(line))
				status = US_DIALOUT;
			else
				status = US_LOGIN;
			break;
		}
	    if((!strcmpi(last_utmp.ut_line,line + 5)) &&
			(!kill(last_utmp.ut_pid,0) || (errno != ESRCH)))
		{
			status = US_LOGGEDIN;
			break;
		}
#else
	    if((!strcmpi(last_utmp.ut_line,line + 5)) &&
			(!kill(last_utmp.ut_pid,0) || (errno != ESRCH)))
		{
			if(!strcmp(last_utmp.ut_user,"LOGIN"))
				status = US_LOGIN;
			else if(!strcmp(last_utmp.ut_user,"DIALOUT"))
				status = US_DIALOUT;
			else
				status = US_LOGGEDIN;
			break;
		}
#endif
	}

#if defined(LOG_UTMP)
	if(status == US_NOTFOUND)
		sprintf(logstr,"UTMP %s: no entry in utmp, status=%d",line,status);
	else
	{
	char *ctime();
		sprintf(logstr,"UTMP %s:%s:%s:%d:status=%d:%s", 
		    last_utmp.ut_user, last_utmp.ut_id, last_utmp.ut_line,
		    last_utmp.ut_pid,status, ctime(&last_utmp.ut_time));
		logstr[strlen(logstr) - 1] = 0;	/* kill NL from ctime() */
	}
	ecu_log_event(logstr);
#endif
	close(ufd);
	return(status);

}	/* end of utmp_status */

/* vi: set tabstop=4 shiftwidth=4: */
/* end of utmpstat.c */
