.so tmac.scheme
.TL
Differences Between the Elk Extension Language and the
.br
Language Scheme as Defined in the Revised\v'-.3m'3\v'.3m' Report
.AU
Oliver Laumann
.AB
This memo provides a short list of the differences between 
the Elk Extension Language and the Language described in
the paper "\f2Revised\v'-.3m'3\v'.3m' Report on the Algorithmic
Language Scheme\fP" by Jonathan Rees and William Clinger
(editors).
.AE
.NH
Lexical Conventions
.LP
Upper and lower case letters in symbols are \f2always\fP distinguished.
.NH
``()''
.LP
``()'' is a legal expression; it evaluates to itself.
.NH
(if \f2test\fP \f2consequent\fP \f2alternate\fP)
.LP
\f2alternate\fP can be a sequence of expressions (implicit ``begin'').
.NH
(case \f2key\fP \f2clause\fP \f2clause\fP ...)
.LP
A \f2clause\fP can be of the form (\f2datum\fP \f2expr\fP \f2expr\fP ...).
In this case, the \f2key\fP is matched against the \f2datum\fP.
.NH
let, let*, etc.
.LP
The \f2init\fP in a binding may be omitted, in this case it defaults to
the empty list.
.NH
Quasiquotation
.LP
Quasiquotation can currently not be used to construct vectors.
.NH
define
.LP
Definitions are valid anywhere within a body (that is, not only
at the beginning of a body).
The \f2expression\fP may be omitted, in this case it defaults to ``void''
(the non-printing object).
A definition of the form
.DS C
(define (\f2form\fP . \f2expr\fP) \f2body\fP)
.DE
is converted to
.DS C
(define \f2form\fP (lambda \f2expr\fP \f2body\fP)).
.DE
Example:
.DS C
(define ((f x) y) (cons x y))
((f 1) 2)   ==>   (1 . 2)
.DE
.NH
nil, t
.LP
``nil'' and ``t'' are not provided.  If needed, they can be
defined as follows:
.DS C
(define nil #f)
(define t #t)
.DE
.NH
car, cdr
.LP
The car (or cdr, respectively) of the empty list is the empty list.
.NH
Rational and complex numbers
.LP
Rational and complex numbers are not implemented.
.NH
Exactness
.LP
All numbers are inexact.
.NH
#b #o #d #x
.LP
Radix prefixes for real numbers are currently not implemented.
.NH
#e #i
.LP
Prefixes for exact and inexact constants are not implemented.
.NH
exact\(mi>inexact, inexact\(mi>exact
.LP
Not implemented.
.NH
number\(mi>string
.LP
Currently not implemented.
Use ``(format #f "~s" \f2number\fP)'' instead.
.NH
string\(mi>number
.LP
The \f2exactness\fP and the \f2radix\fP argument are currently
not supported.
.NH
char-upper-case?, char-lower-case?
.LP
These procedures are defined on \f2all\fP characters.
.NH
make-string
.LP
A character must be supplied as second argument.
.NH
#(a b c)
.LP
Vector constants may not be quoted.  A vector constant evaluates
to itself.
.NH
open-input/output-file, call-with-input/output-file,
.br
with-input/output-from/to-file, load
.LP
\f2filename\fP may be a symbol, in this case
``symbol\(mi>string'' is applied first.
.NH
close-input-port, close-output-port
.LP
Use ``close-port'' instead, this can be applied to input ports as
well as output ports.
.NH
char-ready?
.LP
Currently not implemented.
.NH
transcript-on, transcript-off
.LP
Not implemented; use the UNIX \f2script(1)\fP command instead.
