.so tmac.scheme
.RP
.ds R "R\v'-.3m'\s-13\s0\v'.3m'RS
.ds 1 "\v'.3m'\s-11\s0\v'-.3m'
.ds 2 "\v'.3m'\s-12\s0\v'-.3m'
.TL
Functionality of the
.br
Elk Extension Language Interpreter
.sp
Overview
.AU
Oliver Laumann
.AB
This memo provides a complete list of all primitive procedures
implemented by the Elk Extension Language.
Only those procedures that are not defined in the
\f2Revised\v'-.3m'\s-13\s0\v'.3m' Report on the Algorithmic
Language Scheme\fP by Jonathan Rees and William Clinger (editors)
are described in the memo.
The procedures that are mentioned in the report are only listed \-
no description is provided.
Examples are given for many procedures.
.AE
.C Lambda Expressions, Procedures
.SH
.tl '(lambda \f2formals\fP \f2body\fP)''syntax'
.LP
See \*R.
.SH
.tl '(procedure-lambda \f2procedure\fP)''procedure'
.LP
Returns a copy of the \f2lambda\fP expression which has been
evaluated to create the given procedure.
.br
Example:
.ss
(define (square x) (* x x))
(procedure-lambda square)	==>  (lambda (x) (* x x))
.se
.SH
.tl '(procedure? \f2obj\fP)''procedure'
.LP
See \*R.
.SH
.tl '(primitive? \f2obj\fP)''procedure'
.LP
Returns #t if \f2obj\fP is a primitive procedure, #f otherwise.
.SH
.tl '(compound? \f2obj\fP)''procedure'
.LP
Returns #t if \f2obj\fP is a compound procedure (a procedure that
has been created by evaluating a lambda expression), #f otherwise.
.C Local Bindings
.SH
.tl '(let \f2bindings\fP \f2body\fP)''syntax'
.tl '(let* \f2bindings\fP \f2body\fP)''syntax'
.tl '(letrec \f2bindings\fP \f2body\fP)''syntax'
.LP
See \*R.
.C Fluid Binding
.SH
.tl '(fluid-let \f2bindings\fP \f2body\fP)''syntax'
.LP
\f2bindings\fP is of the form ((\f2variable\*1\fP \f2init1\fP) ...).
The \f2init\fPs are temporarily assigned to the \f2variable\fPs
and the \f2body\fP is executed.
The variables must be bound in an enclosing scope.
When the body is exited normally or by invoking a control point,
the old values of the variables are restored.
In the latter case, when the control returns back to the body
of the fluid-let by invocation of a control point created within
the body, the bindings are changed again to the values they had
when the body exited.
.br
Examples:
.ss
((lambda (x)
  (+ x (fluid-let ((x 3)) x))) 1)	==>  4
.se
.ss
(fluid-let ((print-length 2))
  (write '(a b c d)))	==>  '(a b ...)
.se
.ss
(define (errset thunk)
  (call-with-current-continuation
    (lambda (catch)
      (fluid-let
          ((error-handler
            (lambda msg (catch #f))))
        (list (thunk))))))
.sp
(errset (lambda () (+ 1 2)))	==>  (3)
(errset (lambda () (/ 1 0)))	==>  #f
.se
.C Definitions
.SH
.tl '(define \f2variable\fP \f2expression\fP)''syntax'
.tl '(define (\f2variable\fP \f2formals\fP) \f2body\fP)''syntax'
.tl '(define (\f2variable\fP . \f2formal\fP) \f2body\fP)''syntax'
.LP
See \*R.
.br
Returns a symbol, the identifier that has been bound.
Definitions may appear anywhere within a local body (e.\|g. a lambda
body or a \f2let\fP).
If the \f2expression\fP is omitted, \f2void\fP (the non-printing
object) is used.
.br
Examples:
.ss
(define nil #f)
.se
.ss
(define ((f x) y) (cons x y))
(define (g x) ((f x) 5))
(g 'a)	==>  (a . 5)
.se
.C Assignment
.SH
.tl '(set! \f2variable\fP \f2expression\fP)''syntax'
.LP
See \*R.
.br
Returns the previous value of \f2variable\fP.
.br
Examples:
.ss
(define-macro (swap x y)
  `(set! ,x (set! ,y ,x)))
.se
.C Procedure Application
.SH
.tl '(\f2operator\fP \f2operand\*1\fP ...)''syntax'
.LP
See \*R.
\f2operator\fP can be a macro (see below).
.SH
.tl '(apply \f2arg\*1\fP ... \f2args\fP)''procedure'
.LP
See \*R.
.C Quotation, Quasiquotation
.SH
.tl '(quote \f2datum\fP)''syntax'
.tl ,'\f2datum\fP,,syntax,
.tl '\f2constant\fP''syntax'
.LP
See \*R.
.br
Vectors need not be quoted; a vector evaluates to itself.
.SH
.tl '(quasiquote \f2expression\fP)''syntax'
.tl '(unquote \f2expression\fP)''syntax'
.tl '(unquote-splicing \f2expression\fP)''syntax'
.LP
See \*R.
.C Sequencing
.SH
.tl '(begin \f2expression\*1\fP \f2expression\*2\fP ...)''syntax'
.LP
See \*R.
.SH
.tl '(begin1 \f2expression\*1\fP \f2expression\*2\fP ...)''syntax'
.LP
Identical to \f2begin\fP, except that the result of the first
\f2expression\fP is returned.
.C Conditionals
.SH
.tl '(if \f2test\fP \f2consequent\fP \f2alternate\fP)''syntax'
.tl '(if \f2test\fP \f2consequent\fP)''syntax'
.LP
See \*R.
.br
In the first form, \f2alternate\fP can be a sequence of expressions
(implicit \f2begin\fP).
.SH
.tl '(case \f2key\fP \f2clause\*1\fP \f2clause\*2\fP ...)''syntax'
.LP
See \*R.
.br
Each \f2clause\fP not beginning with \f2else\fP can be of the form
.DS
((\f2datum\*1\fP ...) \f2expression\*1\fP \f2expression\*2\fP ...)
.DE
or
.DS
(\f2datum\fP \f2expression\*1\fP \f2expression\*2\fP ...)
.DE
In the latter case, the \f2key\fP is matched against the \f2datum\fP.
.SH
.tl '(cond \f2clause\*1\fP \f2clause\*2\fP ...)''syntax'
.LP
See \*R.
.SH
.tl '(and \f2test\*1\fP ...)''syntax'
.tl '(or \f2test\*1\fP ...)''syntax'
.LP
See \*R.
.C Booleans
.SH
.tl '(not \f2obj\fP)''procedure'
.LP
See \*R.
.SH
.tl '(boolean? \f2obj\fP)''procedure'
.LP
See \*R.
.C Iteration
.SH
.tl '(let \f2variable\fP \f2bindings\fP \f2body\fP)''syntax'
.LP
``Named \f2let\fP''.  See \*R.
.SH
.tl '(map \f2procedure\fP \f2list\*1\fP \f2list\*2\fP ...)''procedure'
.tl '(for-each \f2procedure\fP \f2list\*1\fP \f2list\*2\fP ...)''procedure'
.LP
See \*R.  \f2for-each\fP returns the empty list.
.SH
.tl '(do \f2initializations\fP \f2test\fP \f2body\fP)''syntax'
.LP
See \*R.
.C Continuations
.SH
.tl '(call-with-current-continuation \f2procedure\fP)''procedure'
.LP
See \*R.
.SH
.tl '(control-point? \f2obj\fP)''procedure'
.LP
Returns #t if \f2obj\fP is a control point (a continuation),
#f otherwise.
.SH
.tl '(dynamic-wind \f2thunk\fP \f2thunk\fP \f2thunk\fP)''procedure'
.LP
\f2dynamic-wind\fP is a generalization of the \f2unwind-protect\fP
facility provided by many Lisp systems.
.br
All three arguments are procedures of no arguments.
In the normal case, all three thunks are applied in order.
The first thunk is also applied when the body (the second thunk)
is entered by the application of a control point created within
the body (by means of \f2call-with-current-continuation\fP).
Similarly, the third thunk is also applied whenever the body is
exited by invocation of a control point created outside the body.
.br
Examples:
.ss
(define-macro (unwind-protect body . unwind-forms)
  `(dynamic-wind
    (lambda () #f)
    (lambda () ,body)
    (lambda () ,@unwind-forms)))
.se
.ss
(let ((f (open-input-file "foo")))
  (dynamic-wind
    (lambda () #f)
    (lambda () \f2do something with\fP f)
    (lambda () (close-port f))))
.se
.C Delayed Evaluation
.SH
.tl '(delay \f2expression\fP)''syntax'
.tl '(force \f2promise\fP)''procedure'
.LP
See \*R.
.SH
.tl '(promise? \f2obj\fP)''procedure'
.LP
Returns #t if \f2obj\fP is a promise, an object returned by the
application of \f2delay\fP.
Otherwise #f is returned.
.C Equivalence Predicates
.SH
.tl '(eq? \f2obj\*1\fP \f2obj\*2\fP)''procedure'
.tl '(eqv? \f2obj\*1\fP \f2obj\*2\fP)''procedure'
.tl '(equal? \f2obj\*1\fP \f2obj\*2\fP)''procedure'
.LP
See \*R.
.C Pairs and Lists
.SH
.tl '(cons \f2obj\*1\fP \f2obj\*2\fP)''procedure'
.LP
See \*R.
.SH
.tl '(car \f2pair\fP)''procedure'
.tl '(cdr \f2pair\fP)''procedure'
.LP
See \*R.
.br
\f2car\fP or \f2cdr\fP applied to the empty list returns the empty list.
.SH
.tl '(cxr \f2pair\fP \f2pattern\fP)''procedure'
.LP
\f2pattern\fP is either a symbol or a string consisting of a combination
of the characters `a' and `d'.
It encodes a sequence of \f2car\fP and \f2cdr\fP operations;
each `a' denotes the application of \f2car\fP, and each `d' denotes
the application of \f2cdr\fP.
For example, \f2(cxr p "ada")\fP is equivalent to \f2(cadar p)\fP.
.SH
.tl '(caar \f2pair\fP)''procedure'
.br
   ...
.br
.tl '(cdddr \f2pair\fP)''procedure'
.LP
See \*R.
.SH
.tl '(set-car! \f2pair\fP \f2obj\fP)''procedure'
.tl '(set-cdr! \f2pair\fP \f2obj\fP)''procedure'
.LP
See \*R.
.br
Both procedures return \f2obj\fP.
.SH
.tl '(make-list \f2k\fP \f2obj\fP)''procedure'
.LP
Returns a list of length \f2k\fP initialized with \f2obj\fP.
.br
Examples:
.ss
(make-list 0 'a)	==>  ()
(make-list 2 (make-list 2 1))	==>  ((1 1) (1 1))
.se
.SH
.tl '(list \f2obj\fP ...)''procedure'
.LP
See \*R.
.SH
.tl '(length \f2list\fP)''procedure'
.LP
See \*R.
.SH
.tl '(list-ref \f2list\fP \f2k\fP)''procedure'
.LP
See \*R.
.SH
.tl '(list-tail \f2list\fP \f2k\fP)''procedure'
.LP
See \*R.
.SH
.tl '(last-pair \f2list\fP)''procedure'
.LP
See \*R.
.SH
.tl '(append \f2list\fP ...)''procedure'
.LP
See \*R.
.SH
.tl '(append! \f2list\fP ...)''procedure'
.LP
Like \f2append\fP, except that the original
arguments are modified (destructive \f2append\fP).
The cdr of each argument is changed to point to the next argument.
.br
Examples:
.ss
(define x '(a b))
(append x '(c d))	==>  (a b c d)
x	==>  (a b)
(append! x '(c d))	==>  (a b c d)
x	==>  (a b c d)
.se
.SH
.tl '(reverse \f2list\fP)''procedure'
.LP
See \*R.
.SH
.tl '(reverse! \f2list\fP)''procedure'
.LP
Destructive \f2reverse\fP.
.SH
.tl '(memq \f2obj\fP \f2list\fP)''procedure'
.tl '(memv \f2obj\fP \f2list\fP)''procedure'
.tl '(member \f2obj\fP \f2list\fP)''procedure'
.LP
See \*R.
.SH
.tl '(assq \f2obj\fP \f2alist\fP)''procedure'
.tl '(assv \f2obj\fP \f2alist\fP)''procedure'
.tl '(assoc \f2obj\fP \f2alist\fP)''procedure'
.LP
See \*R.
.SH
.tl '(null? \f2obj\fP)''procedure'
.tl '(pair? \f2obj\fP)''procedure'
.LP
See \*R.
.C Numbers
.SH
.tl '(= \f2z\*1\fP \f2z\*2\fP ...)''procedure'
.tl '(< \f2z\*1\fP \f2z\*2\fP ...)''procedure'
.tl '(> \f2z\*1\fP \f2z\*2\fP ...)''procedure'
.tl '(<= \f2z\*1\fP \f2z\*2\fP ...)''procedure'
.tl '(>= \f2z\*1\fP \f2z\*2\fP ...)''procedure'
.LP
See \*R.
.SH
.tl '(1+ \f2z\fP)''procedure'
.tl '(1- \f2z\fP)''procedure'
.LP
Returns \f2z\fP plus 1 or \f2z\fP minus 1, respectively.
.SH
.tl '(+ \f2z\*1\fP ...)''procedure'
.tl '(* \f2z\*1\fP ...)''procedure'
.LP
See \*R.
.SH
.tl '(- \f2z\*1\fP \f2z\*2\fP ...)''procedure'
.tl '(/ \f2z\*1\fP \f2z\*2\fP ...)''procedure'
.LP
See \*R.
.SH
.tl '(zero? \f2z\fP)''procedure'
.tl '(positive? \f2z\fP)''procedure'
.tl '(negative? \f2z\fP)''procedure'
.tl '(odd? \f2z\fP)''procedure'
.tl '(even? \f2z\fP)''procedure'
.tl '(exact? \f2z\fP)''procedure'
.tl '(inexact? \f2z\fP)''procedure'
.LP
See \*R.
.br
\f2exact?\fP returns always #f; \f2inexact?\fP returns always #t.
.SH
.tl '(abs \f2z\fP)''procedure'
.LP
See \*R.
.SH
.tl '(quotient \f2n\*1\fP \f2n\*2\fP)''procedure'
.tl '(remainder \f2n\*1\fP \f2n\*2\fP)''procedure'
.tl '(modulo \f2n\*1\fP \f2n\*2\fP)''procedure'
.LP
See \*R.
.SH
.tl '(gcd \f2n\*1\fP ...)''procedure'
.tl '(lcm \f2n\*1\fP ...)''procedure'
.LP
See \*R.
.SH
.tl '(floor \f2x\fP)''procedure'
.tl '(ceiling \f2x\fP)''procedure'
.tl '(truncate \f2x\fP)''procedure'
.tl '(round \f2x\fP)''procedure'
.LP
See \*R.
.SH
.tl '(sqrt \f2z\fP)''procedure'
.LP
See \*R.
.SH
.tl '(exp \f2z\fP)''procedure'
.tl '(log \f2z\fP)''procedure'
.tl '(sin \f2z\fP)''procedure'
.tl '(cos \f2z\fP)''procedure'
.tl '(tan \f2z\fP)''procedure'
.tl '(asin \f2z\fP)''procedure'
.tl '(acos \f2z\fP)''procedure'
.tl '(atan \f2z\fP)''procedure'
.tl '(atan \f2y\fP \f2x\fP)''procedure'
.LP
See \*R.
.SH
.tl '(min \f2x\*1\fP \f2x\*2\fP ...)''procedure'
.tl '(max \f2x\*1\fP \f2x\*2\fP ...)''procedure'
.LP
See \*R.
.SH
.tl '(random)''procedure'
.LP
Returns an integer pseudo-random number in the range from 0 to
2\v'-.3m'\s-131\s0\v'.3m'-1.
.SH
.tl '(srandom \f2n\fP)''procedure'
.LP
Sets the random number generator to the starting point \f2n\fP.
\f2srandom\fP returns \f2n\fP.
.SH
.tl '(number? \f2obj\fP)''procedure'
.tl '(complex? \f2obj\fP)''procedure'
.tl '(real? \f2obj\fP)''procedure'
.tl '(rational? \f2obj\fP)''procedure'
.tl '(integer? \f2obj\fP)''procedure'
.LP
See \*R.
.SH
.tl '(string\(mi>number \f2string\fP)''procedure'
.LP
See \*R (\f2exactness\fP and \f2radix\fP are currently not supported).
.C Characters
.SH
.tl '(char\(mi>integer \f2char\fP)''procedure'
.tl '(integer\(mi>char \f2n\fP)''procedure'
.LP
See \*R.
.SH
.tl '(char-upper-case? \f2char\fP)''procedure'
.tl '(char-lower-case? \f2char\fP)''procedure'
.LP
See \*R.
.SH
.tl '(char-alphabetic? \f2char\fP)''procedure'
.tl '(char-numeric? \f2char\fP)''procedure'
.tl '(char-whitespace? \f2char\fP)''procedure'
.LP
See \*R.
.SH
.tl '(char-upcase \f2char\fP)''procedure'
.tl '(char-downcase \f2char\fP)''procedure'
.LP
See \*R.
.SH
.tl '(char=? \f2char\*1\fP \f2char\*2\fP)''procedure'
.tl '(char<? \f2char\*1\fP \f2char\*2\fP)''procedure'
.tl '(char>? \f2char\*1\fP \f2char\*2\fP)''procedure'
.tl '(char<=? \f2char\*1\fP \f2char\*2\fP)''procedure'
.tl '(char>=? \f2char\*1\fP \f2char\*2\fP)''procedure'
.LP
See \*R.
.SH
.tl '(char-ci=? \f2char\*1\fP \f2char\*2\fP)''procedure'
.tl '(char-ci<? \f2char\*1\fP \f2char\*2\fP)''procedure'
.tl '(char-ci>? \f2char\*1\fP \f2char\*2\fP)''procedure'
.tl '(char-ci<=? \f2char\*1\fP \f2char\*2\fP)''procedure'
.tl '(char-ci>=? \f2char\*1\fP \f2char\*2\fP)''procedure'
.LP
See \*R.
.SH
.tl '(char? \f2obj\fP)''procedure'
.LP
See \*R.
.C Strings
.SH
.tl '(string \f2char\fP ...)''procedure'
.LP
Returns a string containing the specified characters.
.br
Examples:
.ss
(string)	==>  ""
(string #\ea #\espace #\eb)	==>  "a b"
.se
.SH
.tl '(string? \f2obj\fP)''procedure'
.LP
See \*R.
.SH
.tl '(make-string \f2k\fP \f2char\fP)''procedure'
.LP
See \*R.
.SH
.tl '(string-length \f2string\fP)''procedure'
.LP
See \*R.
.SH
.tl '(string-ref \f2string\fP \f2k\fP)''procedure'
.LP
See \*R.
.SH
.tl '(string-set! \f2string\fP \f2k\fP \f2char\fP)''procedure'
.LP
See \*R.
.br
Returns the previous value of element \f2k\fP of the given string.
.SH
.tl '(substring \f2string\fP \f2start\fP \f2end\fP)''procedure'
.LP
See \*R.
.SH
.tl '(string-copy \f2string\fP)''procedure'
.LP
See \*R.
.SH
.tl '(string-append \f2string\fP ...)''procedure'
.LP
See \*R.
.SH
.tl '(list\(mi>string \f2chars\fP)''procedure'
.tl '(string\(mi>list \f2string\fP)''procedure'
.LP
See \*R.
.SH
.tl '(string-fill! \f2string\fP \f2char\fP)''procedure'
.LP
See \*R.
.br
Returns \f2string\fP.
.SH
.tl '(substring-fill! \f2string\fP \f2start\fP \f2end\fP \f2char\fP)''procedure'
.LP
Stores \f2char\fP in every element of \f2string\fP from \f2start\fP
(inclusive) to \f2end\fP (exclusive).
Returns \f2string\fP.
.SH
.tl '(string=? \f2string\*1\fP \f2string\*2\fP)''procedure'
.tl '(string<? \f2string\*1\fP \f2string\*2\fP)''procedure'
.tl '(string>? \f2string\*1\fP \f2string\*2\fP)''procedure'
.tl '(string<=? \f2string\*1\fP \f2string\*2\fP)''procedure'
.tl '(string>=? \f2string\*1\fP \f2string\*2\fP)''procedure'
.LP
See \*R.
.SH
.tl '(string-ci=? \f2string\*1\fP \f2string\*2\fP)''procedure'
.tl '(string-ci<? \f2string\*1\fP \f2string\*2\fP)''procedure'
.tl '(string-ci>? \f2string\*1\fP \f2string\*2\fP)''procedure'
.tl '(string-ci<=? \f2string\*1\fP \f2string\*2\fP)''procedure'
.tl '(string-ci>=? \f2string\*1\fP \f2string\*2\fP)''procedure'
.LP
See \*R.
.SH
.tl '(substring? \f2string\*1\fP \f2string\*2\fP)''procedure'
.tl '(substring-ci? \f2string\*1\fP \f2string\*2\fP)''procedure'
.LP
If \f2string\*1\fP is a substring of \f2string\*2\fP, these
procedures return the starting position of the first occurrence of the
substring within \f2string\*2\fP.
Otherwise #f is returned.
\f2substring-ci?\fP is the case insensitive version of \f2substring?\fP.
.br
Examples:
.ss
(define s "Hello world")
(substring? "foo" x)	==>  #f
(substring? "hello" x)	==>  #f
(substring-ci? "hello" x)	==>  0
(substring? "!" x)	==>  11
.se
.C Vectors
.SH
.tl '(vector? \f2obj\fP)''procedure'
.LP
See \*R.
.SH
.tl '(make-vector \f2k\fP)''procedure'
.tl '(make-vector \f2k\fP \f2fill\fP)''procedure'
.LP
See \*R.
.SH
.tl '(vector \f2obj\fP ...)''procedure'
.LP
See \*R.
.SH
.tl '(vector-length \f2vector\fP)''procedure'
.LP
See \*R.
.SH
.tl '(vector-ref \f2vector\fP \f2k\fP)''procedure'
.LP
See \*R.
.SH
.tl '(vector-set! \f2vector\fP \f2k\fP \f2obj\fP)''procedure'
.LP
See \*R.
.br
Returns the previous value of element \f2k\fP of the vector.
.SH
.tl '(vector\(mi>list \f2vector\fP)''procedure'
.tl '(list\(mi>vector \f2list\fP)''procedure'
.LP
See \*R.
.SH
.tl '(vector-fill! \f2vector\fP \f2fill\fP)''procedure'
.LP
See \*R.
.br
Returns \f2vector\fP.
.SH
.tl '(vector-copy \f2vector\fP)''procedure'
.LP
Returns a copy of \f2vector\fP.
.C Symbols
.SH
.tl '(string\(mi>symbol \f2string\fP)''procedure'
.tl '(symbol\(mi>string \f2symbol\fP)''procedure'
.LP
See \*R.
.SH
.tl '(put \f2symbol\fP \f2key\fP \f2value\fP)''procedure'
.tl '(put \f2symbol\fP \f2key\fP)''procedure'
.LP
Associates \f2value\fP with \f2key\fP in the property list of the
given symbol.
\f2key\fP must be a symbol.
Returns \f2key\fP.
.br
If \f2value\fP is omitted, the property is removed from the symbol's
property list.
.SH
.tl '(get \f2symbol\fP \f2key\fP)''procedure'
.LP
Returns the value associated with \f2key\fP in the property
list of \f2symbol\fP.
\f2key\fP must be a symbol.
If no value is associated with \f2key\fP in the symbol's property
list, #f is returned.
.br
Examples:
.ss
(put 'norway 'capital "Oslo")
(put 'norway 'continent "Europe")
(get 'norway 'capital)	==>  "Oslo"
.se
.SH
.tl '(symbol-plist \f2symbol\fP)''procedure'
.LP
Returns a copy of the property list of \f2symbol\fP as an \f2alist\fP.
.br
Examples:
.ss
(put 'norway 'capital "Oslo")
(put 'norway 'continent "Europe")
(symbol-plist 'norway)
   ==>  ((capital . "Oslo") (continent . "Europe"))
(symbol-plist 'foo)	==>  ()
.se
.SH
.tl '(symbol? \f2obj\fP)''procedure'
.LP
See \*R.
.SH
.tl '(oblist)''procedure'
.LP
Returns a list of lists containing all currently interned symbols.
Each sublist represents a bucket of the interpreters internal
hash array.
.br
Examples:
.ss
(define (apropos what)
  (let ((ret ()))
    (do ((tail (oblist) (cdr tail))) ((null? tail))
      (do ((l (car tail) (cdr l))) ((null? l))
        (if (substring? what (symbol->string (car l)))
            (set! ret (cons (car l) ret)))))
    ret))
.se
.ss
.ta 7c
(apropos "let")	==>  (let* let letrec fluid-let)
(apropos "make")	==>  (make-list make-vector make-string)
(apropos "foo")	==>  ()
.se
.C Environments
.SH
.tl '(the-environment)''procedure'
.LP
Returns the current environment.
.SH
.tl '(global-environment)''procedure'
.LP
Returns the global environment (the ``root'' environment in which
all predefined procedures are bound).
.SH
.tl '(environment\(mi>list \f2environment\fP)''procedure'
.LP
Returns a list representing the specified environment.
The list is a list of \f2frames\fP, each frame is a list of bindings
(an \f2alist\fP).
The car of the list represents the most recently established environment.
The list returned by \f2environment\(mi>list\fP can contain cycles.
.br
Examples:
.ss
(let ((x 1) (y 2))
  (car (environment->list
    (the-environment))))	==>  ((y . 2) (x . 1))
.se
.ss
((lambda (foo)
   (caar (environment->list
     (the-environment)))) "abc")	==>  (foo . "abc")
.se
.ss
(eq?
  (car (last-pair (environment->list
    (the-environment))))
  (car (environment->list
    (global-environment))))	==>  #t
.se
.SH
.tl '(procedure-environment \f2procedure\fP)''procedure'
.tl '(promise-environment \f2promise\fP)''procedure'
.tl '(control-point-environment \f2control-point\fP)''procedure'
.LP
Returns the environment in which the the body of the \f2procedure\fP
is evaluated, the environment in which a value for the \f2promise\fP
is computed when \f2force\fP is applied to it, or the environment in
which the \f2control-point\fP has been created, respectively.
.SH
.tl '(environment? \f2obj\fP)''procedure'
.LP
Returns #t if \f2obj\fP is an environment, #f otherwise.
.C Ports and Files
.LP
Generally, a file name can either be a string or a symbol.
If a symbol is given, it is converted into a string by
applying \f2symbol\(mi>string\fP.
A tilde at the beginning of a file name is expanded according
to the rules employed by the C-Shell (see \f2csh\fP(1)).
.SH
.tl '(call-with-input-file \f2file\fP \f2procedure\fP)''procedure'
.tl '(call-with-output-file \f2file\fP \f2procedure\fP)''procedure'
.LP
See \*R.
.SH
.tl '(input-port? \f2obj\fP)''procedure'
.tl '(output-port? \f2obj\fP)''procedure'
.LP
See \*R.
.SH
.tl '(current-input-port)''procedure'
.tl '(current-output-port)''procedure'
.LP
See \*R.
.SH
.tl '(with-input-from-file \f2file\fP \f2thunk\fP)''procedure'
.tl '(with-output-to-file \f2file\fP \f2thunk\fP)''procedure'
.LP
See \*R.
.br
\f2file\fP can be a string as well as a symbol.
.SH
.tl '(open-input-file \f2file\fP)''procedure'
.tl '(open-output-file \f2file\fP)''procedure'
.LP
See \*R.
.br
\f2file\fP can be a string as well as a symbol.
.SH
.tl '(close-port \f2port\fP)''procedure'
.LP
See \f2close-input-port\fP and \f2close-output-port\fP in \*R.
.SH
.tl '(clear-output-port)''procedure'
.tl '(clear-output-port \f2output-port\fP)''procedure'
.LP
If the argument is omitted, it defaults to the current output port.
.br
In case of ``buffered'' output, this procedure is used to discard
all characters that have been
output to the port but have not yet been sent to the file associated
with the port.
.SH
.tl '(flush-output-port)''procedure'
.tl '(flush-output-port \f2output-port\fP)''procedure'
.LP
If the argument is omitted, it defaults to the current output port.
.br
In case of ``buffered'' output, this procedure is used to force
all characters that have been output to the port to be printed
immediately.
This may be necessary to force output that is not terminated with a newline 
to appear on the terminal.
An output port is flushed automatically when it is closed.
.SH
.tl '(clear-input-port)''procedure'
.tl '(clear-input-port \f2input-port\fP)''procedure'
.LP
If the argument is omitted, it defaults to the current input port.
.br
In case of ``buffered'' input,
this procedure discards all characters that have already been read
from the file associated with the port but have not been processed
using \f2read\fP or similar procedures.
.SH
.tl '(port-file-name \f2port\fP)''procedure'
.LP
Returns the name of the file associated with \f2port\fP if it is
a file port, #f otherwise.
.SH
.tl '(tilde-expand \f2file\fP)''procedure'
.LP
If \f2file\fP starts with a tilde, performs tilde expansion as
described above and returns the result of the expansion
(a string); returns \f2file\fP otherwise.
\f2file\fP is a string or a symbol.
.SH
.tl '(file-exists? \f2file\fP)''procedure'
.LP
Returns #t if \f2file\fP exists, #f otherwise.
\f2file\fP is a string or a symbol; tilde expansion is not performed.
.C Input
.SH
.tl '(read)''procedure'
.tl '(read \f2input-port\fP)''procedure'
.LP
See \*R.
.SH
.tl '(read-char)''procedure'
.tl '(read-char \f2input-port\fP)''procedure'
.LP
See \*R.
.SH
.tl '(read-string)''procedure'
.tl '(read-string \f2input-port\fP)''procedure'
.LP
If the argument is omitted, it defaults to the current input port.
.br
Returns the rest of the current input line as a string (not
including the terminating newline).
.SH
.tl '(unread-char \f2char\fP)''procedure'
.tl '(unread-char \f2char\fP \f2input-port\fP)''procedure'
.LP
If the second argument is omitted, it defaults to the current input port.
.br
Pushes \f2char\fP back on the stream of input characters.
It is \f2not\fP an error for \f2char\fP not to be the last character
read from the port.
The procedure returns \f2char\fP.
.br
Examples:
.ss
(define (peek-char port)
  (unread-char (read-char port) port))
.se
.SH
.tl '(eof-object? \f2obj\fP)''procedure'
.LP
See \*R.
.C Output
.LP
Printing is controlled by the variables \f2print-length\fP and
\f2print-depth\fP; these variables are bound in the global environment.
\f2print-length\fP determines the maximum length of a list or vector
that is printed; \f2print-depth\fP determines the maximum depth.
If the value of one of the variables is not an integer, a default
value is taken.
.SH
.tl '(write \f2obj\fP)''procedure'
.tl '(write \f2obj\fP \f2output-port\fP)''procedure'
.LP
See \*R.
.SH
.tl '(display \f2obj\fP)''procedure'
.tl '(display \f2obj\fP \f2output-port\fP)''procedure'
.LP
See \*R.
.SH
.tl '(write-char \f2char\fP)''procedure'
.tl '(write-char \f2char\fP \f2output-port\fP)''procedure'
.LP
See \*R.
.SH
.tl '(newline)''procedure'
.tl '(newline \f2output-port\fP)''procedure'
.LP
See \*R.
.SH
.tl '(print \f2obj\fP)''procedure'
.tl '(print \f2obj\fP \f2output-port\fP)''procedure'
.LP
If the second argument is omitted, it defaults to the current output port.
.br
Prints \f2obj\fP using \f2write\fP and then prints a newline.
\f2print\fP returns \f2void\fP.
.SH
.tl '(format \f2destination\fP \f2format-string\fP \f2obj\fP ...)''procedure'
.LP
Prints the third and the following arguments according to the
specifications in the string \f2format-string\fP.
Characters from the format string are copied to the output.
When a tilde is encountered in the format string, the tilde and
the immediately following character are replaced in the output
as follows:
.IP "~s"
is replaced by the printed representation of the next \f2obj\fP
in the sense of \f2write\fP.
.IP "~a"
is replaced by the printed representation of the next \f2obj\fP
in the sense of \f2display\fP.
.IP "~~"
is replaced by a single tilde.
.IP "~%"
is replaced by a newline.
.LP
An error is signaled if fewer \f2obj\fPs are provided than
required by the given format string.
If the format string ends in a tilde, the tilde is ignored.
.LP
If \f2destination\fP is #t, the output is sent to the current
output port; if #f is given, the output is returned as a string;
otherwise, \f2destination\fP must be an output port.
.br
Examples:
.ss
(format #f "Hello world!")	==>  "Hello world"
(format #f "~s world!" "Hello")	==>  "\e"Hello\e" world"
(format #f "~a world!" "Hello")	==>  "Hello world"
(format #f "Hello~a")	==>  "Hello!"
.se
.ss
(define (flat-size s)
  (fluid-let ((print-length 1000) (print-depth 100))
    (string-length (format #f "~a" s))))
.se
.ss
(flat-size 1.5)	==>  3
(flat-size '(a b c))	==>  7
.se
.C String Ports
.LP
String ports are similar to file ports, except that characters are
appended to a string instead of being sent to a file, or taken
from a string instead of being read from a file.
It is not necessary to close string ports.
When an string input port has reached the end of the input string,
successive read operations return end-of-file.
.SH
.tl '(open-input-string \f2string\fP)''procedure'
.LP
Returns a new string input port initialized with \f2string\fP.
.br
Examples:
.ss
(define p (open-input-string "Hello world!"))
(read-char p)	==>  #\eH
(read p)	==>  ello
(read p)	==>  world!
(read p)	==>  \f2end of file\fP
.se
.ss
(define p (open-input-string "(cons 'a 'b)"))
(eval (read p))	==>  (a . b)
.se
.SH
.tl '(open-output-string)''procedure'
.LP
Returns a new string output port.
.SH
.tl '(get-output-string \f2string-output-port\fP)''procedure'
.LP
Returns the string currently associated with the specified string
output port.
As a side-effect, the string is reset to zero length.
.br
Examples:
.ss
(define p (open-output-string))
(display '(a b c) p)
(get-output-string p)	==>  "(a b c)"
(get-output-string p)	==>  ""
.se
.ss
(define (flat-size s)
  (let ((p (open-output-string)))
    (display s p)
    (string-length (get-output-string p))))
.se
.C Loading
.LP
A file to be loaded is searched for in the a list of directories
determined by the variable \f2load-path\fP.
\f2load-path\fP is a list of symbols or strings each of which
is used in turn as a prefix for the file name passed to \f2load\fP
until opening succeeds.
\f2load-path\fP is bound in the global environment.
.SH
.tl '(load \f2file\fP)''procedure'
.tl '(load \f2file\fP \f2environment\fP)''procedure'
.LP
Loads a source file or an object file.
If the file contains source code, the expressions in the file are
read and evaluated.
If the value of the variable \f2load-noisily?\fP is true,
the result of the evaluation of each expression is printed
(\f2load-noisily\fP is bound in the global environment; its value
is #f initially).
If the file contains object code, the contents of the file is linked
together with the running interpreter and with additional libraries
that are specified by the global string variable \f2load-libraries\fP
(for instance, ``-lc -lm'').
Object files must end in the suffix ``.o''.
\f2load\fP returns \f2void\fP.
\f2file\fP can be a string as well as a symbol.
.br
If an optional \f2environment\fP is specified, the contents of the file
is evaluated in this environment instead of the current environment.
.br
Example:
.ss
(fluid-let ((load-noisily? #t))
  (load 'test.scm))
.se
.SH
.tl '(autoload \f2symbol\fP \f2file\fP)''procedure'
.LP
Binds \f2symbol\fP in the current environment (as with \f2define\fP).
When \f2symbol\fP is evaluated the first time, \f2file\fP is loaded.
The definitions loaded from the file must provide a definition
for \f2symbol\fP different from \f2autoload\fP, otherwise an error
is signaled.
.LP
If the value of the global variable \f2autoload-notify?\fP is true,
a message is printed whenever evaluation of a symbol
triggers autoloading of a file.
\f2autoload-notify?\fP is bound to #t initially.
.C Macros
.SH
.tl '(macro \f2formals\fP \f2body\fP)''syntax'
.LP
Creates a macro.
The syntax is identical to the syntax of \f2lambda\fP expressions.
When a macro is called, the actual arguments are bound to
the formal arguments of the \f2macro\fP expression \f2in the current
environment\fP (they are \f2not\fP evaluated), then the \f2body\fP is evaluated.
The result of this evaluation is considered the \f2macro expansion\fP
and is evaluated in place of the macro call.
.SH
.tl '(define-macro (\f2variable\fP \f2formals\fP) \f2body\fP)''syntax'
.tl '(define-macro (\f2variable\fP . \f2formal\fP) \f2body\fP)''syntax'
.LP
Like \f2define\fP, except that \f2macro\fP is used instead of \f2lambda\fP.
.br
Examples:
.ss
(define-macro (++ x) `(set! ,x (1+ ,x)))
(define foo 5)
foo	==>  5
(++ foo)
foo	==>  6
.se
.ss
(define-macro (while test . body)
  `(let loop ()
     (cond (,test ,@body (loop)))))
.se
.SH
.tl '(macro? \f2obj\fP)''procedure'
.LP
Returns #t if \f2obj\fP is a macro, #f otherwise.
.SH
.tl '(macro-body \f2macro\fP)''procedure'
.LP
Returns a copy of the \f2macro\fP expression which has been evaluated to
created the given macro (similar to \f2procedure-lambda\fP).
.br
Examples:
.ss
(define-macro (++ x) `(set! ,x (1+ ,x)))
.sp
(macro-body ++)
  ==>  (macro (x) (quasiquote (set! (unquote x) (1+ (unquote x)))))
.se
.SH
.tl '(macro-expand \f2list\fP)''procedure'
.LP
If the expression \f2list\fP is a macro call, the macro call
is expanded.
.br
Examples:
.ss
(define-macro (++ x) `(set! ,x (1+ ,x)))
.sp
(macro-expand '(++ foo))	==>  (set! foo (1+ foo))
.se
.sp
The following function can be used to expand \f2all\fP macro calls
in an expression, i.\|e. not only at the outermost level:
.ss
(define (expand form)
  (if (or (not (pair? form)) (null? form))
      form
      (let ((head (expand (car form)))
            (args (expand (cdr form)))
	    (result))
        (if (and (symbol? head) (bound? head))
            (begin
              (set! result (macro-expand (cons head args)))
              (if (not (equal? result form))
                  (expand result)
                  result))
              (cons head args)))))
.se
.C Error and Exception Handling
.LP
When an error occurs or when the procedure \f2error\fP is invoked,
the interpreter calls the procedure bound to the variable
\f2error-handler\fP in the global environment.
The error handler is called with an object (either the first argument
that has been passed to \f2error\fP or a symbol identifying the
primitive procedure that has caused the error), and an error
message consisting of a format string
and a list of objects suitable to be passed to \f2format\fP.
.LP
Typically, a user-defined error handler bound to \f2error-handler\fP
prints the error message and then calls a control point that has
been created outside the error handler.
If the error handler terminates normally or if \f2error-handler\fP
is not bound to a procedure, the error message is printed in a
default way, and then a \f2reset\fP is performed.
.LP
When an interrupt occurs (typically as a result of typing the
interrupt character on the keyboard), the procedure bound
to the symbol \f2interrupt-handler\fP in the global environment
is called with no arguments.
If \f2interrupt-handler\fP is not bound to a procedure or if
the procedure terminates normally, a message is printed, and
a \f2reset\fP is performed.
.br
Examples:
.ss
(set! interrupt-handler
  (lambda ()
    (newline)
    (backtrace)
    (reset)))
.se
.SH
.tl '(error \f2obj\fP \f2string\fP \f2obj\fP ...)''procedure'
.LP
Signals an error.
The arguments of \f2error\fP are passed to the \f2error-handler\fP.
.br
Examples:
.ss
(define (foo sym)
  (if (not (symbol? sym))
      (error 'foo "argument not a symbol: ~s" sym))
  ...
.se
.SH
.tl '(reset)''procedure'
.LP
Performs a reset by calling the control point that is bound
to the symbol \f2top-level-control-point\fP in the global
environment.
The control point is called with the argument #t.
If the symbol is not bound to a control point, an error message
is printed and the interpreter is terminated.
.br
Examples:
.ss
(if (call-with-current-continuation
      (lambda (x)
        (fluid-let ((top-level-control-point x))
          \f2do something\fP
          #f)))
    (print "Got a reset!"))
.se
.SH
.tl '(exit)''procedure'
.tl '(exit \f2n\fP)''procedure'
.LP
Terminates the interpreter.
The optional argument \f2n\fP indicates the exit code;
it defaults to zero.
.C Garbage Collection
.SH
.tl '(collect)''procedure'
.LP
Causes a garbage collection.
.LP
If the value of the global variable \f2garbage-collect-notify?\fP is true,
a message indicating the amount of free memory on the heap and
the size of the heap is displayed whenever a garbage collection
is performed.
\f2garbage-collect-notify?\fP is bound to #t initially.
.C Features
.SH
.tl '(feature? \f2symbol\fP)''procedure'
.LP
Returns #t if \f2symbol\fP is a feature, i.\|e. \f2provide\fP has
been called to indicate that the feature \f2symbol\fP is present;
#f otherwise.
.SH
.tl '(provide \f2symbol\fP)''procedure'
.LP
Indicates that the feature \f2symbol\fP is present.
Returns \f2void\fP.
.SH
.tl '(require \f2symbol\fP)''procedure'
.tl '(require \f2symbol\fP \f2file\fP)''procedure'
.tl '(require \f2symbol\fP \f2file\fP \f2environment\fP)''procedure'
.LP
If the feature \f2symbol\fP is not present (i.\|e.
(feature? \f2symbol\fP) evaluates to #f), 
\f2file\fP is loaded.
A message is displayed prior to loading the file if the value of the
global variable \f2autoload-notify?\fP is true.
If the feature is still not present after the file has been loaded,
an error is signaled.
If the \f2file\fP argument is omitted, it defaults to \f2symbol\fP.
If an \f2environment\fP argument is supplied, the file is loaded
into given environment.
if the \f2environment\fP argument is omitted, it defaults to the
current environment.
.C Miscellaneous
.SH
.tl '(dump \f2file\fP)''procedure'
.LP
Writes a snapshot of the running interpreter to \f2file\fP and
returns #f.
When \f2file\fP is executed, execution of the interpreter resumes such
that the call to \f2dump\fP returns #t
(i.e., \f2dump\fP actually returns twice).
\f2dump\fP closes all ports except the current input and current
output port.
.SH
.tl '(eval \f2list\fP)''procedure'
.tl '(eval \f2list\fP \f2environment\fP)''procedure'
.LP
Evaluates the expression \f2list\fP in the specified environment.
If \f2environment\fP is omitted, the expression is evaluated
in the current environment.
.br
Examples:
.ss
(let ((car 1))
  (eval 'car (global-environment)))	==>  \f2primitive\fP \f1car\fP
.se
.ss
(define x 1)
(define env
  (let ((x 2)) (the-environment)))
(eval 'x)	==>  1
(eval 'x env)	==>  2
.se
.SH
.tl '(bound? \f2symbol\fP)''procedure'
.LP
Returns #t if \f2symbol\fP is bound in the current environment,
#f otherwise.
.SH
.tl '(type \f2obj\fP)''procedure'
.LP
Returns a symbol indicating the type of \f2obj\fP.
.br
Examples:
.ss
(type 13782343423544)	==>  integer
(type 1.5e8)	==>  real
(type (lambda (x y) (cons x y)))	==>  compound
(type #\ea)	==>  character
(type '(a b c))	==>  pair
(type ())	==>  null
(type (read
  (open-input-string "")))	==>  end-of-file
.se
.SH
.tl '(void? \f2obj\fP)''procedure'
.LP
Returns true if \f2obj\fP is the non-printing object, false otherwise.
.SH
.tl '(command-line-args)''procedure'
.LP
Returns the command line arguments of the interpreter's invocation,
a list of strings.
.de PT
..
.bp
.PX
