.TH ASM41C 1 "Sep  7,1988"
.UC 4
.SH NAME
asm41c \- HP 41c assembler
.SH SYNOPSIS
.B asm41c
[
.I options
] [ name ]
.SH DESCRIPTION
.I Asm41c
assembles the named file, or the standard input if no file name is specified.
A file suitable for processing by
.B bar41c(1)
is produced.
The available flags are:
.TP
.B \-n
Specifies that all numbers in programs are to be proceded by a zero
byte just as if the program had been entered via the 41c keyboard.
The default action of the assembler is to not include zero bytes before
numbers unless they are necessary to separate two consecutive numbers.
.TP
.B \-r
Specifies that the routine names and XROM codes for the named rom are
loaded into the operator symbol table.
.TP
.B \-X
Specifies that the XROM codes for the 41CX should be loaded.
This is the same as specifing "\-r\ CXextfcn\ \-r\ CXtime"
.TP
.B \-I
Specifies that the XROM codes for the IL interface loop rom should be
loaded.
This is the same as specifing
"\-r\ ILmass\ \-r\ ILprinter\ \-r\ ILcntl"
.TP
.B \-l
Specifies that a listing is to be produced on stdout.
The default is that error messages be produced on stderr.
.TP
.B \-t
Specifies the title to be used on listings and barcode output.
This may also be specified in the input using the ".title" directive.
.TP
.B -o
Specifies the output barcode file name.
The default is "bar.out".
.bp
.SH DIRECTIVES
An input file for the assembler consists of a sequence of directives
some of which may be followed by a sequence of HP 41c assembler code.
Blank lines and comments are ignored.
Comments are as in
.B sh ,
they start with a `#' character and continue to the end of the line.
.PP
The directives are:
.TP
.B .rom \fR\fImodule
.br
The entry points for the specifed rom are loaded into the operator
symbol table.
If the file
.I module.rom
is not found in the current directory, then a search is made of
the default rom library directory "/usr/local/lib/hprom".
This directive has the same effect as the "-r" flag on the command
line.
Example:
.sp 1
\&.rom wand
.sp 1
Loads the entry points for the barcode wand.
.sp 1
Note: if a rom module entry point has the same name as a builtin operator
then the builtin operator will hide the rom entry point.
.TP
.B .xrom \fR\fIname romnum entry
.br
.I Name
is added to the operator symbol table as the rom module entry point
specified by the two numeric arguments
.I romnum
and
.I entry .
The numeric argument are the same numbers displayed by the 41c
xrom operator when the rom module is not present in a 41c rom port.
This directive may be used to define single rom module entry points when
whole rom is unneeded.
.TP
.B .title \fR\fItext
.br
All of the text following on the line, but not including any comment,
is used on subsequent pages of the listing as the listing title.
This directive has the same effect as the "-t" flag on the command line.
.TP
.B .page
.br
Causes the start of a new page in the listing.
.TP
.B .define \fR\fIname value
.br
Define a symbolic name for a register.
.I Value
is either a number between 0 and 128 denoting a register or a previously
defined symbol. (Values greater then 99 denote the internal registers
of the 41c and should only be used with care).
The simple alpha labels A-Ja-eTZYXL and the synthetic register labels
MNOPQR (R being used instead of '|-') are predefined symbolic labels
and their definitions may be redefined using this directive.
.TP
.B .prog \fR\fI[private] name
.br
Start the assembly of a program named
.I name .
the optional keyword
.I private
specifices the barcode produced sould be for a private program
(one that can not be listed by the 41c).
the following lines in the input are take to be 41c assembly instructions
that make up the program.
The program ends at the end of the input or at the first following directive.
THis means that and roms or defined symbols should be specified before the
start of the program.
See the section below for a discription of the assembler instruction format.
.TP
.B .exec \fR\fItext
.br
The following single line of 41c code is assembled and the barcode
output is set up for an executable barcode line.
The line of barcode is labeled with the text from the rest of the line.
.TP
.B .string \fR\fI[append] \"string\" text
.br
Generate a string data barcode line labeled with the text.
The optional
.I append
keyword specifies that the string data is to be append on to the alpha
buffer, otherwise the alpha buffer will be cleared when the barcode line
is read.
.TP
.B .data \fR\fI'number' text
.br
Generate a numeric data barcode line labeled with the text.
.TP
.B .bytes \fR\fI[checksum] bytes... text
Generate a barcode line consisting of the specified decimal bytes
labeled with the text.
The optional
.I checksum
keyword specifies that a checksum is to be computer and placed
at the front of the barcode line.
The directive can be used ot create arbitrary line of barcode such as the
one and two byte paper keyboard lines.
.TP
.B ".clear rom"
.br
Delete all the rom entry points from the symbol table.
.TP
.B ".clear names"
Delete all the symbolic register names and restore the predefined
symbolic names.
.SH "ASSEMBLER CODES"
All the 41c assembler code are just as they are displayed on the 41c except
that the greek sigma character has be replaced with the '~' character,
the arrow character has been relaced with the '}' charecter,
the not equal symbol has be replaced with the string '!=' and the
angle character has been replaced with the string 'ANG' in the one place
it was used so far (the advantage module).
.PP
Strings are enclosed in double quotes with the following escapes defined:
.sp 1
.nf
\&      \e-      append character
\&      \e=      not equals character
\&      \eS      Sigma (also can use '~')
\&      \eM      mu
\&      \eL      angle
\&      \eP      pi    (also can use '{')
\&      \eA      arrow (also can use '}')
\&      \eddd    decimal byte
\&      \exhh    hex byte
\&      \e"      double quote
\&      \e\e      backslash
.fi
.ad l
.PP
Number are enclosed in single quotes (this is a hack to make parsing
easier).
Short form numbers are allowed (e.g. 'E', '-E4' etc.)
.PP
The indirect indicator IND must be in all caps
.PP
Because some of the operators have both short and long forms
(RCL STO GTO LBL),
the operators RCL2, STO2, GTO3 and LBL2 have been defined to
force the long form (RCL 0 assmebles to 032 while RCL2 0 assembles
to 144 000.)
The two spare two-byte operators are included as SPARE1 and SPARE2,
and the W string operator is defined (not that these actually do anything
usefull.)
There is the operator "XROM\ dd,dd" and a special operator "BYTE\ ddd" that
just includes the value into the output.
.SH FILES
.ta 1.5i
/usr/local/lib/hprom directory where rom library is kept
\&.../builtins      file describing the builtin functions
.br
\&.../CXextfcn.rom  41CX builtin extended functions module
.br
\&.../CXtime.rom    41CX builtin time module
.br
\&.../ILcntl.rom    IL device control
.br
\&.../ILmass.rom    IL mass storage
.br
\&.../ILprinter.rom IL printer
.br
\&.../advantage.rom HP advantage module
.br
\&.../cardr.rom     card reader internal rom
.br
\&.../extfcn.rom    extended functions module
.br
\&.../games.rom     games rom
.br
\&.../math.rom      math rom
.br
\&.../ppc.rom       PPC club rom
.br
\&.../printer.rom   non-IL printer internal rom
.br
\&.../time.rom      time module
.br
\&.../wand.rom      barcode wand internal rom
.br
bar.out           default barcode output file
.SH SEE ALSO
bar41c(l)         barcode file to troff translator
.br
Auxiliary documentation:
.in +.5i
.br
.I "HP-41c Owners Manual."
.br
.I "Creating Your Own HP-41 Bar Code Manual. 82153-90019"
.br
.I "HP-41 Synthetic Programming Made Easy"
by Keith Jarett.
.in -.5i
.SH AUTHORS
Mark A. Biggar
.br
sdcrdcf!markb
.br
markb@rdcf.sm.unisys.com
EEE_OOO_FFF
echo x - doc/asm41c.l
sed -e 's/^X//' >bar41c.l <<'EEE_OOO_FFF'
.TH BAR41C 1 "Sep  7,1988"
.UC 4
.SH NAME
bar41c \- HP 41c barcode generator
.SH SYNOPSIS
.B bar41c
[
.I options
]
.SH DESCRIPTION
.I Bar41c
is a filter that takes files like those produced by asm41c(1) and generates
*roff macros to print the barcodes.
The available flags are:
.TP
.B \-w
The integer argument specifies the maximun width of a program barcode line
in narrow bar widths.
Value is used to fill out the line of program barcode to a given width.
The default value of 312 is the equivlent to 6.5 inches using the versitec
macro set.
Narrow bars are half the width of wide bars and the spaces inbetween are
the same width as the narrow bars.
As a line of barcode can not contain more then 16 bytes the maximum usefull
value for -w is 456 (16 bytes * 8 bits * 3 bars (width of wide bar + space)
+ 9 (start and stop bars (3 narrow and 1 wide))).
.TP
.B \-t
Specifies that byte tags are to be generated.
The *roff macros will be called so that each byte in a line of barcode
is labeled with its value.
.SH DIRECTIVES
An input file for bar41c consists of a sequence of directives
some of which may be followed by a sequence of line of byte values.
.PP
The directives are:
.TP
.B title /fR/fItext
The
.I text
is used as the argument for the .Tl macro.
This directive has no other effect.
.TP
.B data \fR\fI'number' text
A numeric data barcode line type 6 is produced with
.I text
as the barcode line's label.
.B string \fR\fI[append] text
A string data barcode line is produced with
.I text
as its label.
If the optional keyword
.I append
is present then an append barcode line of type 8 is produced, otherwise
a normal string barcode line type 7 is produced.
The foolowing line in the input should be a single line of byte values
consisting of the character values for the string data.
.TP
.B bytes \fR\fI[checksum] text
A barcode line is produced using the following line of byte values
and using
.I text
as the label.
If the optional keyword
.I checksum
is present a checksum is first conputed and append on the front of the
list of byte values.
Only the first 16 values (15 if a checksum is specified) are used,
others are ignored.
.TP
.B program \fR\fI[private] name
Start generating barcode line for the program
.I name.
The following lines of byte values terminated by a line starting with
the keyword "end" consist of the instructions for the program, one
instruction per line (this is used to generate the instruction numbers
in the standard barcode line labels which look like "Line 1 (1 - 3)".)
If the optional keyword
.I private
is used then program barcode line of type 2 (private) are produced,
otherwise type 1 is used.
.SH FILES
.ta 1.5i
tmac.vbar       macro file of Versitec raster printers
.br
tmac.lwbar      macro file for Apple LaserWriter
.SH SEE ALSO
asm41c(l)       Hp41c assembler
.br
Auxiliary documentation:
.in +.5i
.br
.I "Creating Your Own HP-41 Bar Code Manual. 82153-90019"
.in -.5i
.SH BUGS
There is no provision made for the sequenced data barcode types
(9, 10 & 11).
.SH AUTHORS
Mark A. Biggar
.br
sdcrdcf!markb
.br
markb@rdcf.sm.unisys.com
EEE_OOO_FFF
echo x - doc/bar41c.l
test -d rom || mkdir rom
sed -e 's/^X//' >README <<'EEE_OOO_FFF'
How to build rom files:

rom file consist of line of perl script loading the $xrom associative
array with the two bytes of assembled data coresponding to the
xrom call.  For example the file wand.rom looks like:

$xrom{"WNDDTA"}  = "166 193";
$xrom{"WNDDTX"}  = "166 194";
$xrom{"WNDLNK"}  = "166 195";
$xrom{"WNDSUB"}  = "166 196";
$xrom{"WNDSCN"}  = "166 197";
$xrom{"WNDTST"}  = "166 198";

In order to make generating these file easier a perl script is included
(called trans) that will take files that look like:

WNDDTA  27 1
WNDDTX  27 2
WNDLNK  27 3
WNDSUB  27 4
WNDSCN  27 5
WNDTST  27 6

and generate a file like the above.  The numbers are the module number
and the routine number just as displayed by the HP 41c in the XROM op
when the module is not installed.
EEE_OOO_FFF
echo x - rom/README
sed -e 's/^X//' >builtins <<'EEE_OOO_FFF'
$single{"NULL"}    = "000";
$single{"+"}       = "064";
$single{"-"}       = "065";
$single{"*"}       = "066";
$single{"/"}       = "067";
$single{"X<Y?"}    = "068";
$single{"X>Y?"}    = "069";
$single{"X<=Y?"}   = "070";
$single{"~+"}      = "071";
$single{"~-"}      = "072";
$single{"HMS+"}    = "073";
$single{"HMS-"}    = "074";
$single{"MOD"}     = "075";
$single{"%"}       = "076";
$single{"%CH"}     = "077";
$single{"P}R"}     = "078";
$single{"R}P"}     = "079";
$single{"LN"}      = "080";
$single{"X^2"}     = "081";
$single{"SQRT"}    = "082";
$single{"Y^X"}     = "083";
$single{"CHS"}     = "084";
$single{"E^X"}     = "085";
$single{"LOG"}     = "086";
$single{"10^X"}    = "087";
$single{"E^X-1"}   = "088";
$single{"SIN"}     = "089";
$single{"COS"}     = "090";
$single{"TAN"}     = "091";
$single{"ASIN"}    = "092";
$single{"ACOS"}    = "093";
$single{"ATAN"}    = "094";
$single{"}DEC"}    = "095";
$single{"1/X"}     = "096";
$single{"ABS"}     = "097";
$single{"FACT"}    = "098";
$single{"X!=0?"}   = "099";
$single{"X>0?"}    = "100";
$single{"LN1+X"}   = "101";
$single{"X<0?"}    = "102";
$single{"X=0?"}    = "103";
$single{"INT"}     = "104";
$single{"FRC"}     = "105";
$single{"D}R"}     = "106";
$single{"R}D"}     = "107";
$single{"}HMS"}    = "108";
$single{"}HR"}     = "109";
$single{"RND"}     = "110";
$single{"}OCT"}    = "111";
$single{"CL+"}     = "112";
$single{"X<>Y"}    = "113";
$single{"PI"}      = "114";
$single{"CLST"}    = "115";
$single{"R^"}      = "116";
$single{"RDN"}     = "117";
$single{"LASTX"}   = "118";
$single{"CLX"}     = "119";
$single{"X=Y?"}    = "120";
$single{"X!=Y?"}   = "121";
$single{"SIGN"}    = "122";
$single{"X<=0?"}   = "123";
$single{"MEAN"}    = "124";
$single{"SDEV"}    = "125";
$single{"AVIEW"}   = "126";
$single{"CLD"}     = "127";
$single{"DEG"}     = "128";
$single{"RAD"}     = "129";
$single{"GRAD"}    = "130";
$single{"ENTER^"}  = "131";
$single{"STOP"}    = "132";
$single{"RTN"}     = "133";
$single{"BEEP"}    = "134";
$single{"CLA"}     = "135";
$single{"ASHF"}    = "136";
$single{"PSE"}     = "137";
$single{"CLRG"}    = "138";
$single{"AOFF"}    = "139";
$single{"AON"}     = "140";
$single{"OFF"}     = "141";
$single{"PROMPT"}  = "142";
$single{"ADV"}     = "143";

$twobyte{"RCL"}     = "144";
$twobyte{"STO"}     = "145";
$twobyte{"RCL2"}    = "144";
$twobyte{"STO2"}    = "145";
$twobyte{"ST+"}     = "146";
$twobyte{"ST-"}     = "147";
$twobyte{"ST*"}     = "148";
$twobyte{"ST/"}     = "149";
$twobyte{"ISG"}     = "150";
$twobyte{"DSE"}     = "151";
$twobyte{"VIEW"}    = "152";
$twobyte{"~REG"}    = "153";
$twobyte{"ASTO"}    = "154";
$twobyte{"ARCL"}    = "155";
$twobyte{"FIX"}     = "156";
$twobyte{"SCI"}     = "157";
$twobyte{"ENG"}     = "158";
$twobyte{"TONE"}    = "159";
$twobyte{"SF"}      = "168";
$twobyte{"CF"}      = "169";
$twobyte{"FS?C"}    = "170";
$twobyte{"FC?C"}    = "171";
$twobyte{"FS?"}     = "172";
$twobyte{"FC?"}     = "173";
$twobyte{"SPARE1"}  = "175";
$twobyte{"SPARE2"}  = "176";
$twobyte{"X<>"}     = "206";
$twobyte{"LBL2"}    = "207";
EEE_OOO_FFF
echo x - rom/builtins
sed -e 's/^X//' >trans <<'EEE_OOO_FFF'
#! /usr/local/bin/perl -i

while (<>) {
	next if /^\s*#/;                # ignore comments
	/([^ ]+) *([0-9]+) *([0-9]+)/;
	print sprintf("%-16s = \"%03d %03d\";\n",
			   "\$xrom{\"$1\"}", 160+($2/4), 64*($2%4)+$3);
}
EEE_OOO_FFF
echo x - rom/trans
sed -e 's/^X//' >CXextfcn.rom <<'EEE_OOO_FFF'
$xrom{"ALENG"}   = "166 065";
$xrom{"ANUM"}    = "166 066";
$xrom{"APPCHR"}  = "166 067";
$xrom{"APPREC"}  = "166 068";
$xrom{"ARCLREC"} = "166 069";
$xrom{"AROT"}    = "166 070";
$xrom{"ATOX"}    = "166 071";
$xrom{"CLFL"}    = "166 072";
$xrom{"CLKEYS"}  = "166 073";
$xrom{"CRFLAS"}  = "166 074";
$xrom{"CRFLD"}   = "166 075";
$xrom{"DELCHR"}  = "166 076";
$xrom{"DELREC"}  = "166 077";
$xrom{"EMDIR"}   = "166 078";
$xrom{"FLSIZE"}  = "166 079";
$xrom{"GETAS"}   = "166 080";
$xrom{"GETKEY"}  = "166 081";
$xrom{"GETP"}    = "166 082";
$xrom{"GETR"}    = "166 083";
$xrom{"GETREC"}  = "166 084";
$xrom{"GETRX"}   = "166 085";
$xrom{"GETSUB"}  = "166 086";
$xrom{"GETX"}    = "166 087";
$xrom{"INSCHR"}  = "166 088";
$xrom{"INSREC"}  = "166 089";
$xrom{"PASN"}    = "166 090";
$xrom{"PCLPS"}   = "166 091";
$xrom{"POSA"}    = "166 092";
$xrom{"POSFL"}   = "166 093";
$xrom{"PSIZE"}   = "166 094";
$xrom{"PURFL"}   = "166 095";
$xrom{"RCLFLAG"} = "166 096";
$xrom{"RCLPT"}   = "166 097";
$xrom{"RCLPTA"}  = "166 098";
$xrom{"REGMOVE"} = "166 099";
$xrom{"REGSWAP"} = "166 100";
$xrom{"SAVEAS"}  = "166 101";
$xrom{"SAVEP"}   = "166 102";
$xrom{"SAVER"}   = "166 103";
$xrom{"SAVERX"}  = "166 104";
$xrom{"SAVEX"}   = "166 105";
$xrom{"SEEKPT"}  = "166 106";
$xrom{"SEEKPTA"} = "166 107";
$xrom{"SIZE?"}   = "166 108";
$xrom{"STOFLAG"} = "166 109";
$xrom{"X<>F"}    = "166 110";
$xrom{"XTOA"}    = "166 111";
$xrom{"ASROOM"}  = "166 113";
$xrom{"CLRGX"}   = "166 114";
$xrom{"ED"}      = "166 115";
$xrom{"EMDIRX"}  = "166 116";
$xrom{"EMROOM"}  = "166 117";
$xrom{"GETKEYX"} = "166 118";
$xrom{"RESZFL"}  = "166 119";
$xrom{"~REG?"} = "166 120";
$xrom{"X=NN?"}   = "166 121";
$xrom{"X!=NN?"}  = "166 122";
$xrom{"X<NN?"}   = "166 123";
$xrom{"X<=NN?"}  = "166 124";
$xrom{"X>NN?"}   = "166 125";
$xrom{"X>=NN?"}  = "166 126";
EEE_OOO_FFF
echo x - rom/CXextfcn.rom
sed -e 's/^X//' >CXtime.rom <<'EEE_OOO_FFF'
$xrom{"ADATE"}   = "166 129";
$xrom{"ALMCAT"}  = "166 130";
$xrom{"ALMNOW"}  = "166 131";
$xrom{"ATIME"}   = "166 132";
$xrom{"ATIME24"} = "166 133";
$xrom{"CLK12"}   = "166 134";
$xrom{"CLK24"}   = "166 135";
$xrom{"CLKT"}    = "166 136";
$xrom{"CLKTD"}   = "166 137";
$xrom{"CLOCK"}   = "166 138";
$xrom{"CORRECT"} = "166 139";
$xrom{"DATE"}    = "166 140";
$xrom{"DATE+"}   = "166 141";
$xrom{"DDAYS"}   = "166 142";
$xrom{"DMY"}     = "166 143";
$xrom{"DOW"}     = "166 144";
$xrom{"MDY"}     = "166 145";
$xrom{"RCLAF"}   = "166 146";
$xrom{"RCLSW"}   = "166 147";
$xrom{"RUNSW"}   = "166 148";
$xrom{"SETAF"}   = "166 149";
$xrom{"SETDATE"} = "166 150";
$xrom{"SETSW"}   = "166 151";
$xrom{"STOPSW"}  = "166 152";
$xrom{"SW"}      = "166 153";
$xrom{"T+X"}     = "166 154";
$xrom{"TIME"}    = "166 155";
$xrom{"XYZALM"}  = "166 156";
$xrom{"CLALMA"}  = "166 159";
$xrom{"CLALMX"}  = "166 160";
$xrom{"CLRALMS"} = "166 161";
$xrom{"RCLALM"}  = "166 162";
$xrom{"SWPT"}    = "166 163";
EEE_OOO_FFF
echo x - rom/CXtime.rom
sed -e 's/^X//' >ILcntl.rom <<'EEE_OOO_FFF'
$xrom{"AUTOIO"}  = "167 027";
$xrom{"FINDID"}  = "167 028";
$xrom{"INA"}     = "167 029";
$xrom{"IND"}     = "167 030";
$xrom{"INSTAT"}  = "167 031";
$xrom{"LISTEN"}  = "167 032";
$xrom{"LOCAL"}   = "167 033";
$xrom{"MANIO"}   = "167 034";
$xrom{"OUTA"}    = "167 035";
$xrom{"PWRDN"}   = "167 036";
$xrom{"PWRUP"}   = "167 037";
$xrom{"REMOTE"}  = "167 038";
$xrom{"SELECT"}  = "167 039";
$xrom{"STOPIO"}  = "167 040";
$xrom{"TRIGGER"} = "167 041";
EEE_OOO_FFF
echo x - rom/ILcntl.rom
sed -e 's/^X//' >ILmass.rom <<'EEE_OOO_FFF'
$xrom{"CREATE"}  = "167 001";
$xrom{"DIR"}     = "167 002";
$xrom{"NEWM"}    = "167 003";
$xrom{"PURGE"}   = "167 004";
$xrom{"READA"}   = "167 005";
$xrom{"READK"}   = "167 006";
$xrom{"READP"}   = "167 007";
$xrom{"READR"}   = "167 008";
$xrom{"READRX"}  = "167 009";
$xrom{"READS"}   = "167 010";
$xrom{"READSUB"} = "167 011";
$xrom{"RENAME"}  = "167 012";
$xrom{"SEC"}     = "167 013";
$xrom{"SEEKR"}   = "167 014";
$xrom{"UNSEC"}   = "167 015";
$xrom{"VERIFY"}  = "167 016";
$xrom{"WRTA"}    = "167 017";
$xrom{"WRTK"}    = "167 018";
$xrom{"WRTP"}    = "167 019";
$xrom{"WRTPV"}   = "167 020";
$xrom{"WRTR"}    = "167 021";
$xrom{"WRTRX"}   = "167 022";
$xrom{"WRTS"}    = "167 023";
$xrom{"ZERO"}    = "167 024";
EEE_OOO_FFF
echo x - rom/ILmass.rom
sed -e 's/^X//' >ILprinter.rom <<'EEE_OOO_FFF'
$xrom{"ACA"}     = "167 065";
$xrom{"ACCHR"}   = "167 066";
$xrom{"ACCOL"}   = "167 067";
$xrom{"ACSPEC"}  = "167 068";
$xrom{"ACX"}     = "167 069";
$xrom{"BLDSPEC"} = "167 070";
$xrom{"LIST"}    = "167 071";
$xrom{"PRA"}     = "167 072";
$xrom{"PRAXIS"}  = "167 073";
$xrom{"PRBUF"}   = "167 074";
$xrom{"PRFLAGS"} = "167 075";
$xrom{"PRKEYS"}  = "167 076";
$xrom{"PRP"}     = "167 077";
$xrom{"PRPLOT"}  = "167 078";
$xrom{"PRPLOTP"} = "167 079";
$xrom{"PRREG"}   = "167 080";
$xrom{"PRREGX"}  = "167 081";
$xrom{"PR~"}     = "167 082";
$xrom{"PRSTK"}   = "167 083";
$xrom{"PRX"}     = "167 084";
$xrom{"REGPLOT"} = "167 085";
$xrom{"SKPCHR"}  = "167 086";
$xrom{"SKPCOL"}  = "167 087";
$xrom{"STKPLOT"} = "167 088";
$xrom{"FMT"}     = "167 089";
EEE_OOO_FFF
echo x - rom/ILprinter.rom
sed -e 's/^X//' >advantage.rom <<'EEE_OOO_FFF'
$xrom{"BININ"}   = "165 129";
$xrom{"BINVIEW"} = "165 130";
$xrom{"OCTIN"}   = "165 131";
$xrom{"CVTVIEW"} = "165 132";
$xrom{"HEXIN"}   = "165 133";
$xrom{"HEXVIEW"} = "165 134";
$xrom{"NOT"}     = "165 135";
$xrom{"AND"}     = "165 136";
$xrom{"OR"}      = "165 137";
$xrom{"XOR"}     = "165 138";
$xrom{"ROTXY"}   = "165 139";
$xrom{"BIT?"}    = "165 140";
$xrom{"C<>C"}    = "165 142";
$xrom{"CMAXAB"}  = "165 143";
$xrom{"CNRM"}    = "165 144";
$xrom{"CSUM"}    = "165 145";
$xrom{"DIM?"}    = "165 146";
$xrom{"FNRM"}    = "165 147";
$xrom{"I+"}      = "165 148";
$xrom{"I-"}      = "165 149";
$xrom{"J+"}      = "165 150";
$xrom{"J-"}      = "165 151";
$xrom{"M*M"}     = "165 152";
$xrom{"MAT*"}    = "165 153";
$xrom{"MAT+"}    = "165 154";
$xrom{"MAT-"}    = "165 155";
$xrom{"MAT/"}    = "165 156";
$xrom{"MATDIM"}  = "165 157";
$xrom{"MAX"}     = "165 158";
$xrom{"MAXAB"}   = "165 159";
$xrom{"MDET"}    = "165 160";
$xrom{"MIN"}     = "165 161";
$xrom{"MINV"}    = "165 162";
$xrom{"MMOVE"}   = "165 163";
$xrom{"MNAME?"}  = "165 164";
$xrom{"MR"}      = "165 165";
$xrom{"MRC+"}    = "165 166";
$xrom{"MRC-"}    = "165 167";
$xrom{"MRIJ"}    = "165 168";
$xrom{"MRIJR"}   = "165 169";
$xrom{"MRR+"}    = "165 170";
$xrom{"MRR-"}    = "165 171";
$xrom{"MS"}      = "165 172";
$xrom{"MSC+"}    = "165 173";
$xrom{"MSIJ"}    = "165 174";
$xrom{"MSIJR"}   = "165 175";
$xrom{"MSR+"}    = "165 176";
$xrom{"MSWAP"}   = "165 177";
$xrom{"MSVS"}    = "165 178";
$xrom{"PIV"}     = "165 179";
$xrom{"R<>R"}    = "165 180";
$xrom{"R>R?"}    = "165 181";
$xrom{"RMAXAB"}  = "165 182";
$xrom{"RNRM"}    = "165 183";
$xrom{"RSUM"}    = "165 184";
$xrom{"SUM"}     = "165 185";
$xrom{"SUMAB"}   = "165 186";
$xrom{"TRNPS"}   = "165 187";
$xrom{"YC+C"}    = "165 188";
$xrom{"MEDIT"}   = "165 189";
$xrom{"CMEDIT"}  = "165 190";
$xrom{"MP"}      = "165 191";
$xrom{"MATRX"}   = "165 192";
$xrom{"MTR"}     = "166 001";
$xrom{"SOLVE"}   = "166 003";
$xrom{"INTEG"}   = "166 004";
$xrom{"SILOOP"}  = "166 005";
$xrom{"SIRIN"}   = "166 006";
$xrom{"Z^N"}     = "166 007";
$xrom{"MAGZ"}    = "166 008";
$xrom{"e^Z"}     = "166 009";
$xrom{"LNZ"}     = "166 010";
$xrom{"Z^1/N"}   = "166 011";
$xrom{"SINZ"}    = "166 012";
$xrom{"COSZ"}    = "166 013";
$xrom{"TANZ"}    = "166 014";
$xrom{"a^Z"}     = "166 015";
$xrom{"LOGZ"}    = "166 016";
$xrom{"Z^1/W"}   = "166 017";
$xrom{"Z^W"}     = "166 018";
$xrom{"C+"}      = "166 019";
$xrom{"C-"}      = "166 020";
$xrom{"CINV"}    = "166 021";
$xrom{"C*"}      = "166 022";
$xrom{"C/"}      = "166 023";
$xrom{"PLV"}     = "166 024";
$xrom{"RIS"}     = "166 025";
$xrom{"DIFEQ"}   = "166 026";
$xrom{"CFIT"}    = "166 027";
$xrom{"ASIG"}    = "166 028";
$xrom{"DSIG"}    = "166 029";
$xrom{"CFIT"}    = "166 030";
$xrom{"FIT"}     = "166 031";
$xrom{"Y?X"}     = "166 032";
$xrom{"SZ?"}     = "166 033";
$xrom{"VC"}      = "166 034";
$xrom{"CROSS"}   = "166 035";
$xrom{"VS"}      = "166 036";
$xrom{"VR"}      = "166 037";
$xrom{"DOT"}     = "166 038";
$xrom{"VE"}      = "166 039";
$xrom{"V-"}      = "166 040";
$xrom{"V+"}      = "166 041";
$xrom{"VXY"}     = "166 042";
$xrom{"UV"}      = "166 043";
$xrom{"VANG"}    = "166 044";
$xrom{"VD"}      = "166 045";
$xrom{"V*"}      = "166 046";
$xrom{"TR"}      = "166 047";
$xrom{"CT"}      = "166 048";
$xrom{"AIP"}     = "166 049";
$xrom{"TVM"}     = "166 051";
$xrom{"N"}       = "166 052";
$xrom{"PV"}      = "166 053";
$xrom{"PMT"}     = "166 054";
$xrom{"FV"}      = "166 055";
$xrom{"*I"}      = "166 056";
EEE_OOO_FFF
echo x - rom/advantage.rom
sed -e 's/^X//' >cardr.rom <<'EEE_OOO_FFF'
$xrom{"MRG"}     = "167 129";
$xrom{"RDTA"}    = "167 130";
$xrom{"RDTAX"}   = "167 131";
$xrom{"RSUB"}    = "167 132";
$xrom{"VER"}     = "167 133";
$xrom{"WALL"}    = "167 134";
$xrom{"WDTA"}    = "167 135";
$xrom{"WDTAX"}   = "167 136";
$xrom{"WPRV"}    = "167 137";
$xrom{"WSTS"}    = "167 138";
EEE_OOO_FFF
echo x - rom/cardr.rom
sed -e 's/^X//' >extfcn.rom <<'EEE_OOO_FFF'
$xrom{"ALENG"}   = "166 065";
$xrom{"ANUM"}    = "166 066";
$xrom{"APPCHR"}  = "166 067";
$xrom{"APPREC"}  = "166 068";
$xrom{"ARCLREC"} = "166 069";
$xrom{"AROT"}    = "166 070";
$xrom{"ATOX"}    = "166 071";
$xrom{"CLFL"}    = "166 072";
$xrom{"CLKEYS"}  = "166 073";
$xrom{"CRFLAS"}  = "166 074";
$xrom{"CRFLD"}   = "166 075";
$xrom{"DELCHR"}  = "166 076";
$xrom{"DELREC"}  = "166 077";
$xrom{"EMDIR"}   = "166 078";
$xrom{"FLSIZE"}  = "166 079";
$xrom{"GETAS"}   = "166 080";
$xrom{"GETKEY"}  = "166 081";
$xrom{"GETP"}    = "166 082";
$xrom{"GETR"}    = "166 083";
$xrom{"GETREC"}  = "166 084";
$xrom{"GETRX"}   = "166 085";
$xrom{"GETSUB"}  = "166 086";
$xrom{"GETX"}    = "166 087";
$xrom{"INSCHR"}  = "166 088";
$xrom{"INSREC"}  = "166 089";
$xrom{"PASN"}    = "166 090";
$xrom{"PCLPS"}   = "166 091";
$xrom{"POSA"}    = "166 092";
$xrom{"POSFL"}   = "166 093";
$xrom{"PSIZE"}   = "166 094";
$xrom{"PURFL"}   = "166 095";
$xrom{"RCLFLAG"} = "166 096";
$xrom{"RCLPT"}   = "166 097";
$xrom{"RCLPTA"}  = "166 098";
$xrom{"REGMOVE"} = "166 099";
$xrom{"REGSWAP"} = "166 100";
$xrom{"SAVEAS"}  = "166 101";
$xrom{"SAVEP"}   = "166 102";
$xrom{"SAVER"}   = "166 103";
$xrom{"SAVERX"}  = "166 104";
$xrom{"SAVEX"}   = "166 105";
$xrom{"SEEKPT"}  = "166 106";
$xrom{"SEEKPTA"} = "166 107";
$xrom{"SIZE?"}   = "166 108";
$xrom{"STOFLAG"} = "166 109";
$xrom{"X<>F"}    = "166 110";
$xrom{"XTOA"}    = "166 111";
EEE_OOO_FFF
echo x - rom/extfcn.rom
sed -e 's/^X//' >games.rom <<'EEE_OOO_FFF'
$xrom{"BAGELS"}  = "162 129";
$xrom{"BIOR"}    = "162 130";
$xrom{"BIOF"}    = "162 131";
$xrom{"CRAPS"}   = "162 132";
$xrom{"HANG"}    = "162 133";
$xrom{"PINBALL"} = "162 134";
$xrom{"SWAR"}    = "162 135";
$xrom{"SUBHUNT"} = "162 136";
$xrom{"BOOM"}    = "162 137";
$xrom{"INI"}     = "162 138";
$xrom{"P"}       = "162 139";
$xrom{"SIZE?"}   = "162 140";
$xrom{"RNDM"}    = "162 141";
$xrom{"RNDMW"}   = "162 142";
EEE_OOO_FFF
echo x - rom/games.rom
sed -e 's/^X//' >math.rom <<'EEE_OOO_FFF'
$xrom{"MATRIX"}  = "160 065";
$xrom{"SIMEQ"}   = "160 066";
$xrom{"VCOL"}    = "160 067";
$xrom{"VMAT"}    = "160 068";
$xrom{"PVT"}     = "160 069";
$xrom{"DET"}     = "160 070";
$xrom{"INV"}     = "160 071";
$xrom{"EDIT"}    = "160 072";
$xrom{"SOLVE"}   = "160 073";
$xrom{"SOL"}     = "160 074";
$xrom{"POLY"}    = "160 075";
$xrom{"ROOTS"}   = "160 076";
$xrom{"INTG"}    = "160 077";
$xrom{"DIFEQ"}   = "160 078";
$xrom{"FOUR"}    = "160 079";
$xrom{"C+"}      = "160 080";
$xrom{"C-"}      = "160 081";
$xrom{"C*"}      = "160 082";
$xrom{"C/"}      = "160 083";
$xrom{"MAGZ"}    = "160 084";
$xrom{"CINV"}    = "160 085";
$xrom{"Z^N"}     = "160 086";
$xrom{"Z^1/N"}   = "160 087";
$xrom{"E^Z"}     = "160 088";
$xrom{"LNZ"}     = "160 089";
$xrom{"A^Z"}     = "160 090";
$xrom{"LOGZ"}    = "160 091";
$xrom{"Z^W"}     = "160 092";
$xrom{"Z^1/W"}   = "160 093";
$xrom{"SINZ"}    = "160 094";
$xrom{"COSZ"}    = "160 095";
$xrom{"TANZ"}    = "160 096";
$xrom{"SINH"}    = "160 097";
$xrom{"COSH"}    = "160 098";
$xrom{"TANH"}    = "160 099";
$xrom{"ASINH"}   = "160 100";
$xrom{"ACOSH"}   = "160 101";
$xrom{"ATANH"}   = "160 102";
$xrom{"SSS"}     = "160 103";
$xrom{"ASA"}     = "160 104";
$xrom{"SAA"}     = "160 105";
$xrom{"SAS"}     = "160 106";
$xrom{"SSA"}     = "160 107";
$xrom{"TRANS"}   = "160 108";
$xrom{"*FN"}     = "160 109";
EEE_OOO_FFF
echo x - rom/math.rom
sed -e 's/^X//' >ppc.rom <<'EEE_OOO_FFF'
$xrom{"+K"}      = "162 131";
$xrom{"-B"}      = "162 152";
$xrom{"1K"}      = "162 130";
$xrom{"2D"}      = "162 183";
$xrom{"A?"}      = "162 138";
$xrom{"AB"}      = "162 189";
$xrom{"AD"}      = "162 146";
$xrom{"AL"}      = "162 165";
$xrom{"AM"}      = "165 053";
$xrom{"BA"}      = "165 030";
$xrom{"BC"}      = "165 043";
$xrom{"BD"}      = "165 017";
$xrom{"BE"}      = "165 034";
$xrom{"BI"}      = "162 172";
$xrom{"BL"}      = "162 170";
$xrom{"BM"}      = "165 039";
$xrom{"BR"}      = "165 040";
$xrom{"B+"}      = "165 042";
$xrom{"BV"}      = "165 007";
$xrom{"BX"}      = "165 041";
$xrom{"C?"}      = "162 144";
$xrom{"CA"}      = "165 023";
$xrom{"CB"}      = "162 178";
$xrom{"CD"}      = "162 163";
$xrom{"CJ"}      = "165 021";
$xrom{"CK"}      = "162 134";
$xrom{"CM"}      = "165 020";
$xrom{"CP"}      = "165 027";
$xrom{"CU"}      = "162 162";
$xrom{"CV"}      = "165 008";
$xrom{"CX"}      = "162 161";
$xrom{"DC"}      = "162 139";
$xrom{"DF"}      = "165 013";
$xrom{"DP"}      = "162 181";
$xrom{"DR"}      = "165 038";
$xrom{"DS"}      = "162 157";
$xrom{"DT"}      = "162 145";
$xrom{"E?"}      = "162 190";
$xrom{"EP"}      = "162 159";
$xrom{"EX"}      = "162 155";
$xrom{"F?"}      = "162 132";
$xrom{"FD"}      = "165 011";
$xrom{"FI"}      = "162 191";
$xrom{"FL"}      = "162 171";
$xrom{"FR"}      = "165 012";
$xrom{"GE"}      = "162 188";
$xrom{"GN"}      = "165 015";
$xrom{"HA"}      = "165 025";
$xrom{"HD"}      = "162 148";
$xrom{"HN"}      = "162 169";
$xrom{"HP"}      = "165 029";
$xrom{"HS"}      = "165 026";
$xrom{"IF"}      = "162 177";
$xrom{"IG"}      = "165 009";
$xrom{"IP"}      = "162 173";
$xrom{"IR"}      = "165 037";
$xrom{"JC"}      = "165 022";
$xrom{"L-"}      = "162 151";
$xrom{"LB"}      = "162 150";
$xrom{"LF"}      = "162 133";
$xrom{"LG"}      = "165 024";
$xrom{"LR"}      = "165 002";
$xrom{"M1"}      = "165 033";
$xrom{"M2"}      = "165 031";
$xrom{"M3"}      = "165 032";
$xrom{"M4"}      = "165 035";
$xrom{"M5"}      = "165 036";
$xrom{"MA"}      = "165 054";
$xrom{"MK"}      = "162 129";
$xrom{"ML"}      = "162 140";
$xrom{"MP"}      = "165 028";
$xrom{"MS"}      = "162 176";
$xrom{"MT"}      = "162 156";
$xrom{"NC"}      = "162 166";
$xrom{"NH"}      = "162 168";
$xrom{"NP"}      = "165 014";
$xrom{"NR"}      = "165 050";
$xrom{"NS"}      = "165 049";
$xrom{"OM"}      = "162 186";
$xrom{"PA"}      = "162 187";
$xrom{"PD"}      = "162 180";
$xrom{"PK"}      = "162 137";
$xrom{"PM"}      = "165 019";
$xrom{"PO"}      = "165 051";
$xrom{"PR"}      = "165 045";
$xrom{"PS"}      = "162 174";
$xrom{"QR"}      = "162 182";
$xrom{"RB"}      = "165 052";
$xrom{"RD"}      = "165 005";
$xrom{"RF"}      = "162 141";
$xrom{"RK"}      = "165 006";
$xrom{"RN"}      = "165 016";
$xrom{"RT"}      = "162 179";
$xrom{"RX"}      = "162 185";
$xrom{"S1"}      = "165 046";
$xrom{"S2"}      = "165 048";
$xrom{"S3"}      = "165 047";
$xrom{"S?"}      = "162 143";
$xrom{"SB"}      = "165 001";
$xrom{"SD"}      = "165 003";
$xrom{"SE"}      = "165 056";
$xrom{"~?"}      = "162 142";
$xrom{"~C"}      = "162 149";
$xrom{"SK"}      = "165 004";
$xrom{"SM"}      = "165 055";
$xrom{"SR"}      = "165 000";
$xrom{"SU"}      = "162 167";
$xrom{"SV"}      = "165 010";
$xrom{"SX"}      = "162 184";
$xrom{"T1"}      = "162 175";
$xrom{"TB"}      = "165 018";
$xrom{"TN"}      = "162 160";
$xrom{"UD"}      = "162 136";
$xrom{"UG"}      = "165 044";
$xrom{"VA"}      = "162 135";
$xrom{"VF"}      = "165 058";
$xrom{"VK"}      = "162 164";
$xrom{"VM"}      = "162 154";
$xrom{"VS"}      = "162 158";
$xrom{"XD"}      = "162 153";
$xrom{"XE"}      = "162 147";
$xrom{"XL"}      = "165 057";
EEE_OOO_FFF
echo x - rom/ppc.rom
sed -e 's/^X//' >printer.rom <<'EEE_OOO_FFF'
$xrom{"ACA"}     = "167 065";
$xrom{"ACCHR"}   = "167 066";
$xrom{"ACCOL"}   = "167 067";
$xrom{"ACSPEC"}  = "167 068";
$xrom{"ACX"}     = "167 069";
$xrom{"BLDSPEC"} = "167 070";
$xrom{"LIST"}    = "167 071";
$xrom{"PRA"}     = "167 072";
$xrom{"PRAXIS"}  = "167 073";
$xrom{"PRBUF"}   = "167 074";
$xrom{"PRFLAGS"} = "167 075";
$xrom{"PRKEYS"}  = "167 076";
$xrom{"PRP"}     = "167 077";
$xrom{"PRPLOT"}  = "167 078";
$xrom{"PRPLOTP"} = "167 079";
$xrom{"PRREG"}   = "167 080";
$xrom{"PRREGX"}  = "167 081";
$xrom{"PR~"}     = "167 082";
$xrom{"PRSTK"}   = "167 083";
$xrom{"PRX"}     = "167 084";
$xrom{"REGPLOT"} = "167 085";
$xrom{"SKPCHR"}  = "167 086";
$xrom{"SKPCOL"}  = "167 087";
$xrom{"STKPLOT"} = "167 088";
EEE_OOO_FFF
echo x - rom/printer.rom
sed -e 's/^X//' >time.rom <<'EEE_OOO_FFF'
$xrom{"ADATE"}   = "166 129";
$xrom{"ALMCAT"}  = "166 130";
$xrom{"ALMNOW"}  = "166 131";
$xrom{"ATIME"}   = "166 132";
$xrom{"ATIME24"} = "166 133";
$xrom{"CLK12"}   = "166 134";
$xrom{"CLK24"}   = "166 135";
$xrom{"CLKT"}    = "166 136";
$xrom{"CLKTD"}   = "166 137";
$xrom{"CLOCK"}   = "166 138";
$xrom{"CORRECT"} = "166 139";
$xrom{"DATE"}    = "166 140";
$xrom{"DATE+"}   = "166 141";
$xrom{"DDAYS"}   = "166 142";
$xrom{"DMY"}     = "166 143";
$xrom{"DOW"}     = "166 144";
$xrom{"MDY"}     = "166 145";
$xrom{"RCLAF"}   = "166 146";
$xrom{"RCLSW"}   = "166 147";
$xrom{"RUNSW"}   = "166 148";
$xrom{"SETAF"}   = "166 149";
$xrom{"SETDATE"} = "166 150";
$xrom{"SETSW"}   = "166 151";
$xrom{"STOPSW"}  = "166 152";
$xrom{"SW"}      = "166 153";
$xrom{"T+X"}     = "166 154";
$xrom{"TIME"}    = "166 155";
$xrom{"XYZALM"}  = "166 156";
EEE_OOO_FFF
echo x - rom/time.rom
sed -e 's/^X//' >wand.rom <<'EEE_OOO_FFF'
$xrom{"WNDDTA"}  = "166 193";
$xrom{"WNDDTX"}  = "166 194";
$xrom{"WNDLNK"}  = "166 195";
$xrom{"WNDSUB"}  = "166 196";
$xrom{"WNDSCN"}  = "166 197";
$xrom{"WNDTST"}  = "166 198";
EEE_OOO_FFF
echo x - rom/wand.rom
test -d test || mkdir test
sed -e 's/^X//' >README <<'EEE_OOO_FFF'
These file were used to test the correct operation of the assembler.
They should be used to check out local changes you may make to the
perl scripts.  Every operator is checked.  See the comments at the
front of each file of information on how to run each test.
EEE_OOO_FFF
echo x - test/README
sed -e 's/^X//' >numbers.test <<'EEE_OOO_FFF'
# this test sould be run once with -n and once without
# without -n the 000* should not be there
.prog number-test
'123456789'     # 000* 017 018 019 020 021
		# 022 023 024 025
'1.2'           # 000 017 026 018
'1.2e2'         # 000 017 026 018 027 018
'1.2E2'         # 000 017 026 018 027 018
'-1.2e-2'       # 000 028 017 026 018 027
		# 028 018
+               # 064
'E-10'          # 000* 027 028 017 016
+               # 064
'-E-1'          # 000* 028 027 028 017
'E12'           # 000 028 027 018
.data '1.234e-4' input value
EEE_OOO_FFF
echo x - test/numbers.test
sed -e 's/^X//' >rom.test <<'EEE_OOO_FFF'
# note that this test needs to be invoked
#   as "asm41c -X -r wand"
.xrom foo 11 13
.rom math
.prog rom-test
SWPT            # 166 163
CLOCK           # 166 138
WNDDTA          # 166 193
C+              # 160 080
foo             # 162 205
.exec call foo
foo
.clear roms     # test of clear rom tables
.exec call foo
foo             # error & null = 000
EEE_OOO_FFF
echo x - test/rom.test
sed -e 's/^X//' >single.test <<'EEE_OOO_FFF'
.prog single-test
NULL    # 000
+       # 064
-       # 065
*       # 066
/       # 067
X<Y?    # 068
X>Y?    # 069
X<=Y?   # 070
SIG+    # 071
SIG-    # 072
HMS+    # 073
HMS-    # 074
MOD     # 075
%       # 076
%CH     # 077
P->R    # 078
R->P    # 079
LN      # 080
.title test new title
.page   # sneak test of .page
.prog private single-test2
X^2     # 081
SQRT    # 082
Y^X     # 083
CHS     # 084
E^X     # 085
LOG     # 086
10^X    # 087
E^X-1   # 088
SIN     # 089
COS     # 090
TAN     # 091
ASIN    # 092
ACOS    # 093
ATAN    # 094
->DEC   # 095
1/X     # 096
ABS     # 097
FACT    # 098
X!=0?   # 099
X>0?    # 100
LN1+X   # 101
X<0?    # 102
X=0?    # 103
INT     # 104
FRC     # 105
D->R    # 106
R->D    # 107
->HMS   # 108
->HR    # 109
RND     # 110
->OCT   # 111
CLSIG   # 112
X<>Y    # 113
PI      # 114
CLST    # 115
R^      # 116
RDN     # 117
LASTX   # 118
CLX     # 119
X=Y?    # 120
X!=Y?   # 121
SIGN    # 122
X<=0?   # 123
MEAN    # 124
SDEV    # 125
AVIEW   # 126
CLD     # 127
DEG     # 128
RAD     # 129
GRAD    # 130
ENTER^  # 131
STOP    # 132
RTN     # 133
BEEP    # 134
CLA     # 135
ASHF    # 136
PSE     # 137
CLRG    # 138
AOFF    # 139
AON     # 140
OFF     # 141
PROMPT  # 142
ADV     # 143
LBL 0   # 001
LBL 7   # 008
LBL 14  # 015
STO 0   # 048
STO 7   # 055
STO 15  # 063
RCL 0   # 032
RCL 7   # 039
RCL 15  # 047
BYTE 0  # 000
BYTE 127 # 127
BYTE 377 # 121
EEE_OOO_FFF
echo x - test/single.test
sed -e 's/^X//' >string.test <<'EEE_OOO_FFF'
# test various thing with string args
.prog string-test
"abcdefg"               # 247 097 098 099 100 101
			# 102 103
"123456789012345"       # 255 049 050 051 052 053
			# 054 055 056 057 048 049
			# 050 051 052 053
""                      # 240
"12345678901234567890"  # trunc error
			# 255 049 050 051 052 053
			# 054 055 056 057 048 049
			# 050 051 052 053
"\000\001\127\255\128"  # 245 000 001 127 255 128
"\x00\x01\x7f\xff\x80"  # 245 000 001 127 255 128
"\-\=\S\M\L\P\A"        # 248 127 029 126 012 013
			# 123 125
LBL "abcde"             # 192 000 246 000 097 098
			# 099 100 101
GTO "abcde"             # 029 245 097 098 099 100
			# 101
XEQ "abcde"             # 030 245 097 098 099 100
			# 101
W "abcde"               # 031 245 097 098 099 100
			# 101
.string "abcdefg" test string
			# 097 098 099 100 101 102
			# 103
EEE_OOO_FFF
echo x - test/string.test
sed -e 's/^X//' >symbol.test <<'EEE_OOO_FFF'
# test out predefined and user define symbols
.prog symbol-test
RCL A   # 144 102
RCL B   # 144 103
RCL C   # 144 104
RCL D   # 144 105
RCL E   # 144 106
RCL F   # 144 107
RCL G   # 144 108
RCL H   # 144 109
RCL I   # 144 110
RCL J   # 144 111
RCL T   # 144 112
RCL Z   # 144 113
RCL Y   # 144 114
RCL X   # 144 115
RCL L   # 144 116
RCL M   # 144 117
RCL N   # 144 118
RCL O   # 144 119
RCL P   # 144 120
RCL Q   # 144 121
RCL R   # 144 122
RCL a   # 144 123
RCL b   # 144 124
RCL c   # 144 125
RCL d   # 144 126
RCL e   # 144 127
RCL foo # error undefined symbol & 032
.define foo 100
.define bar 127
.define gorf foo
.prog symbol-test2
RCL foo # 144 100
RCL bar # 144 127
RCL gorf # 144 100
.clear names
.prog symbol-test3
RCL e   # 144 127
RCL gorf # error undefined symbol & 032

EEE_OOO_FFF
echo x - test/symbol.test
sed -e 's/^X//' >three.test <<'EEE_OOO_FFF'
# test out tree byte op codes
.prog three-byte-test
END             # 192 000 009
GTO 100         # 208 000 100
GTO3 0          # 208 000 000
XEQ 100         # 224 000 100
XEQ 0           # 224 000 000
EEE_OOO_FFF
echo x - test/three.test
sed -e 's/^X//' >twobyte.test <<'EEE_OOO_FFF'
# test out two byte op codes
.prog   two-byte-test
RCL     100     # 144 100
STO     100     # 145 100
RCL2    0       # 144 000
STO2    0       # 145 000
ST+     0       # 146 000
ST-     0       # 147 000
ST*     0       # 148 000
ST/     0       # 149 000
ISG     0       # 150 000
DSE     0       # 151 000
VIEW    0       # 152 000
SIGREG  0       # 153 000
ASTO    0       # 154 000
ARCL    0       # 155 000
FIX     0       # 156 000
SCI     0       # 157 000
ENG     0       # 158 000
TONE    0       # 159 000
SF      0       # 168 000
CF      0       # 169 000
FS?C    0       # 170 000
FC?C    0       # 171 000
FS?     0       # 172 000
FC?     0       # 173 000
SPARE1  0       # 175 000
SPARE2  0       # 176 000
X<>     0       # 206 000
LBL2    0       # 207 000
GTO     0       # 177 000
GTO     7       # 184 000
GTO     14      # 191 000
GTO     IND 0   # 174 000
XEQ     IND 0   # 174 128
RCL     IND 0   # 144 128
XROM    11 13   # 162 205
EEE_OOO_FFF
echo x - test/twobyte.test
sed -e 's/^X//' >test.41c <<'EEE_OOO_FFF'
.prog test # comment
LBL "A"
ENTER^
LN
+
XEQ IND 000
STO M
RTN
"ABCDEFG"
"\-\002hij"
'-1.234e-8'

EEE_OOO_FFF
echo x - test/test.41c
test -d tmac || mkdir tmac
sed -e 's/^X//' >README <<'EEE_OOO_FFF'
The following macros are defined in the tmac.*bar files:

.Sb "Label"

Start a line of barcode labeled with Label (for programs this label usually
looks something like "Line 1 (1 - 3)").  This macro automatically does a
.ne directive to assure that the line of barcode, its label and tags if any
remain on the same page.

.By Value

Each byte of a line of barcode is specified by one of these macro calls.
Value should be between 0 and 256 inclusive.

.Eb [1]

End a line of bar code. if the optional argument is non-zero then each
byte of the line is taged with its value (in a tiny pointsize) just
below the bars

The following macro calls are generated but are not defined.
Suitable definitions can be supplied by the user.

.Tl "Title"

Each asm41c .title directive (or -t flag) or bar41c title directive
generates this call.

.Pg "Program Name"

A call of this type is generated before the barcode for each
asm41c .prog or bar41c program directive.

The following macro, number or string register names are used in these
macro definitions:

Ad B0 B1 B2 B3 B4 B5 B6 B7 Bb Bs Bt By Eb Fi Sb b0 b1
EEE_OOO_FFF
echo x - tmac/README
sed -e 's/^X//' >tmac.lwbar <<'EEE_OOO_FFF'
.ds b0 \(bv\(bv   \"do not remove this comment
.ds b1 \(bv\(bv\(bv\(bv   \"do not remove this comment
.de Sb
.ne 0.75i
.nf
.nr Bs \\n(.s
.nr Ad \\n(.j
.nr Fi \\n(.u
.ps 8
\\$1
.sp 0.5v
.ps 18
.cs R 1
\\*(b0\\*(b0\\c
.cs R
.rm Bt
..
.de By
.ps 6
.am Bt ..
\\\\h'|\\n(.ku-\\\\n(.ku-1.1p'\(br \\$1\\\\c
...
.nr B7 (\\$1/128%2)
.nr B6 (\\$1/64%2)
.nr B5 (\\$1/32%2)
.nr B4 (\\$1/16%2)
.nr B3 (\\$1/8%2)
.nr B2 (\\$1/4%2)
.nr B1 (\\$1/2%2)
.nr B0 (\\$1%2)
.ps 18
.cs R 1
.di Bb
\\\\*(b\\n(B7\\\\*(b\\n(B6\\\\*(b\\n(B5\\\\*(b\\n(B4\\\\c
\\\\*(b\\n(B3\\\\*(b\\n(B2\\\\*(b\\n(B1\\\\*(b\\n(B0\\\\c
.di
.Bb
.cs R
..
.de Eb
.ps 18
.cs R 1
\*(b1\*(b0
.cs R
.ie '\\$1'1' \\{\\
.ps 6
.\".sp 0.05v
.Bt
.sp 0.1v
'br \\}
.el .sp 0.2v
.ps \\n(Bs
.if \\n(Fi .fi
.ad \\n(Ad
..
EEE_OOO_FFF
echo x - tmac/tmac.lwbar
sed -e 's/^X//' >tmac.vbar <<'EEE_OOO_FFF'
.ds b0 \(bv \"do not remove this comment
.ds b1 \(bv\(bv \"do not remove this comment
.de Sb
.ne 0.75i
.nf
.nr Bs \\n(.s
.nr Ad \\n(.j
.nr Fi \\n(.u
.ps 8
\\$1
.sp 0.6v
.ps 18
.cs R 3
\\*(b0\\*(b0\\c
.cs R
.rm Bt
..
.de By
.ps 6
.am Bt ..
\\\\h'|\\n(.ku-\\\\n(.ku-1.5p'\(br \\$1\\\\c
...
.nr B7 (\\$1/128%2)
.nr B6 (\\$1/64%2)
.nr B5 (\\$1/32%2)
.nr B4 (\\$1/16%2)
.nr B3 (\\$1/8%2)
.nr B2 (\\$1/4%2)
.nr B1 (\\$1/2%2)
.nr B0 (\\$1%2)
.ps 18
.cs R 3
.di Bb
\\\\*(b\\n(B7\\\\*(b\\n(B6\\\\*(b\\n(B5\\\\*(b\\n(B4\\\\c
\\\\*(b\\n(B3\\\\*(b\\n(B2\\\\*(b\\n(B1\\\\*(b\\n(B0\\\\c
.di
.Bb
.cs R
..
.de Eb
.ps 18
.cs R 3
\*(b1\*(b0
.cs R
.ie '\\$1'1' \\{\\
.ps 6
.sp 0.05v
.Bt
.sp 0.1v
'br \\}
.el .sp 0.2v
.ps \\n(Bs
.if \\n(Fi .fi
.ad \\n(Ad
..
EEE_OOO_FFF
echo x - tmac/tmac.vbar
exit 0

--
Mark Biggar
{allegra,burdvax,cbosgd,hplabs,ihnp4,akgua,sdcsvax}!sdcrdcf!markb
markb@rdcf.sm.unisys.com


