.TH multitee 1
.SH NAME
multitee \- send multiple inputs to multiple outputs
.SH SYNTAX
multitee
[
\fB\-o\fIsz
] [
\fB\-bBv\fI
] [
\fB\-ACHUVW\fI
] [
\fIfd-fd,fd,fd...
] ...
.SH DESCRIPTION
.I multitee
sends multiple inputs to multiple outputs.
Given an argument of the form
.I fdin-fdout,fdout,fdout...
it will send all input on file descriptor
.I fdin
to each descriptor
.I fdout.
.I multitee
will never block reading or writing a descriptor
(unless it runs out of buffer space---see below).
.PP
In the
.I fd
arguments, hyphens can be replaced with colons,
and any 0 can be omitted. Other than that, each
.I fd
must be an integer in the right range.
Future extensions to
.I multitee
may abuse the syntax of these arguments,
so don't insert any extra characters.
.PP
Options
.B ACHUVW
print the authorship notice,
copyright notice,
help notice,
short usage summary,
version number,
and warranty information respectively.
.PP
.I multitee
has several flags:
.TP 12
\fB\-o\fIsz
Set maximum output buffer size to
.I sz,
default 16384.
If
.I multitee
overflows an output buffer,
it will not read any more characters until the buffer
empties down to the maximum buffer size.
The actual maximum space taken by buffers is
.I sz
for each output buffer, plus 4096 per
output
.I fd,
though it would take a lot of effort to
achieve the latter.
.TP
\fB\-b
Never block, unless out of buffer space.
This is the default.
For example, if one output stream is blocked,
.I multitee
will not sit around waiting for it to clear up before it
processes characters waiting on other streams. This helps
prevent deadlock.
.TP
\fB\-B
Don't be so paranoid about blocking.
Some systems do not support true per-process non-blocking I/O;
on such systems,
.I multitee
will refuse to run unless you specify
.B\-B.
Under
.B\-B,
.I multitee
might block when, for instance,
it writes more bytes onto a pipe than the pipe has free.
.PP
.SH "EXIT VALUE"
0 if all inputs reach end-of-file normally.
1 for usage messages.
2 if a file descriptor is closed.
3 if
.I multitee
runs out of memory.
4 in various impossible situations.
5 if you don't specify
.B\-B
on machines that don't support true per-process non-blocking I/O.
6 on an I/O error.
7 on an unrecognized error.
8 if
.I multitee
hits a secondary storage resource limit
(past quota, no more disk space, past process file limit).
.SH DIAGNOSTICS
.TP
.I fatal: no memory
.I multitee
has run out of memory.
.TP
.I fatal: out of space
.I multitee
has exhausted a secondary storage resource,
such as disk space, disk quota, or maximum file size.
.TP
.I fatal: nonblocking I/O not supported
Self-explanatory.
.TP
.I fatal: unopened fd
One of the file descriptors is not open.
Sorry, there's no way to figure out which one.
.TP
.I fatal: I/O error
.I multitee
has encountered an error in disk input or output.
.TP
.I weird
This includes various errors, none of which can possibly happen.
.SH RESTRICTIONS
.I multitee
simply can't work on machines without
.I siginterrupt(3)
as well as it can on machines with that call.
.SH BUGS
None known.
.SH MACHINES
.I multitee
has been tested
on an Astronautics ZS-2 running ZSUnix
and a Sun 4 running SunOS.
.SH VERSION
multitee version 2.0, dated March 27, 1990.
.SH AUTHOR
Copyright 1990, Daniel J. Bernstein.
.SH "SEE ALSO"
tee(1)
