      subroutine psplot(x, y, n, uno1, uno2)
      parameter(nres=300,dummy=1.0)
      integer uno1, uno2, n
      real x(n), y(n)
      logical isopen
      external pspvals
      real xmin,xmax,ymin,ymax,xr,yr
      real pxoff,pyoff,xvoff,yvoff,xnoff,ynoff,toff
      real plwid,plht,blwid,tickl,tickw,orient,size
      real xnfsize,ynfsize,tfsize,xvfsize,yvfsize
      integer iuno,ptype,nxticks,nyticks
      logical boxit,clines,tickf,tickh,tickq,fill,sxvals,syvals
      character*80 rec,xname,yname,title,dashp,xnfont,ynfont,tfont,
     +xvfont,yvfont
      character vnames(0:42)*10, descs(0:42)*60
      common/plotcom/iuno,xr,yr,rec,vnames,descs,
     +xmin,xmax,ymin,ymax,pxoff,pyoff,xvoff,yvoff,xnoff,ynoff,toff,
     +plwid,plht,blwid,tickl,tickw,orient,size,
     +xnfsize,ynfsize,tfsize,xvfsize,yvfsize,
     +ptype,nxticks,nyticks,
     +boxit,clines,tickf,tickh,tickq,fill,sxvals,syvals,
     +xname,yname,title,dashp,xnfont,ynfont,tfont,xvfont,yvfont
      save /plotcom/
C
C PSPLOT does a bit of serious work and produces the plot.
C
C X    = array of X values;
C Y    = array of Y values;
C N    = number of (X,Y) pairs;
C UNO1 = unit number of parameter file.
C UNO2 = unit number of log file.
C
C UNO1 is the unit number of the input parameter file that should
C be opened in the before calling psplot (See PSPLOT.DOC). The  
C parameters in this file are used in plotting. They override 
C defaults. If UNO1 is negative, the plotting parameters are
C assumed to be set before the call and so none of the parameters
C are modified. Otherwise, if file UNO1 is not opened the default
C plotting parameters are used.
C
C UNO2 is the unit number of an output file that will contain
C the plotting parameters in effect when the plot was made.
C This file is written in a form that can be easily edited and
C provided as input (UNO1) if things need to be adjusted in a
C subsequent run. If the file UNO2 is not open, the log file 
C is not written.
C
      if (uno1 .lt. 0) goto 3
C
C Find Max and Min to design an appropriate scale.
C
      xmin = x(1)
      xmax = xmin
      ymin = y(1)
      ymax = ymin
      do 10 i=1,n
        if (xmin .gt. x(i)) xmin = x(i)
        if (xmax .lt. x(i)) xmax = x(i)
        if (ymin .gt. y(i)) ymin = y(i)
        if (ymax .lt. y(i)) ymax = y(i)
10    continue
C
C Read the parameter file.
C
      inquire(uno1,opened=isopen)
      if (.not. isopen) goto 3
 1    read(uno1,101,err=4,end=2) rec
101   format(a)
      call psparse()
      goto 1
 2    close(uno1)
C
C Now see if you need to write the Log file.
C
3     inquire(uno2,opened=isopen)
      if (isopen) call pslog(uno2)
C
C Set Current Transformation Matrix.
C
      write(iuno,101)'mydctm setmatrix'
      if (orient.eq.90.) then
        write(iuno,101)'90 rotate 0 -8.5 inch translate'
      else
        write(iuno,102) orient,' rotate'
      endif
102   format(e16.8,a)
C
C Translate origin.
C
      write(iuno,103) pxoff,' inch ', pyoff,' inch translate 0 setgray'
103   format(e16.8,a,e16.8,a)
      write(iuno,101)'/pctm pctm currentmatrix def'
      if (ptype.eq.0) goto 4
      write(iuno,104)'/xs ',n,' array def /ys ',n,' array def'
104   format(a,i6,a,i6,a)
C
C Do some scaling and write transformed array.
C
      xr = plwid*nres/(xmax-xmin)
      yr = plht*nres/(ymax-ymin)
      do 210 i=1,n
        xcood = xr*(x(i)-xmin)
        ycood = yr*(y(i)-ymin)
        write(iuno,105)'xs ',i-1,xcood,' put ys ',i-1,ycood,' put'
210   continue
105   format(a,i6,1x,e16.8,a,i6,e16.8,a)
C
C Check plot type.
C
      if (ptype.eq.1) then
        call pslinp(n)
      else if(ptype.eq.2) then
        call pscirp(n)
      else if(ptype.eq.3) then
        call pssqarp(n)
      else if(ptype.eq.4) then
        call psdiap(n)
      else 
        call pstrip(n)
      endif
C
C Draw box or corner lines if needed.
C
 4    if (boxit) then
        call pssetlw(blwid)
        call psbox(xmax-xmin,ymax-ymin,xmin,ymin,.false.,.true.)
      else if (clines) then
         call pssetlw(blwid)
         call psline(xmin,ymin,xmax,ymin,.true.)
         call psline(xmin,ymin,xmin,ymax,.true.)
      endif
C
C Tick the lines if necessary.
C
      if (tickf.or.tickh.or.tickq) call pssetlw(tickw)
C
C Tick the two lines meeting at the lower left corner.
C
      if (tickf) call pstickl(nxticks,tickl,xmin,ymin,xmax,ymin,.true.)

      if (tickh) call pstickl(nxticks-1,0.5*tickl,xmin+0.5*(xmax-xmin)/n
     +xticks,ymin,xmax-0.5*(xmax-xmin)/nxticks,ymin,.true.)

      if (tickq) call pstickl(2*nxticks-1,0.25*tickl,xmin+0.25*(xmax-xmi
     +n)/nxticks,ymin,xmax-0.25*(xmax-xmin)/nxticks,ymin,.true.)

      if (tickf) call pstickl(nyticks,-1.0*tickl,xmin,ymin,xmin,ymax,.tr
     +ue.)

      if (tickh) call pstickl(nyticks-1,-0.5*tickl,xmin,ymin+0.5*(ymax-y
     +min)/nyticks,xmin,ymax-0.5*(ymax-ymin)/nyticks,.true.)

      if (tickh) call pstickl(2*nyticks-1,-0.25*tickl,xmin,ymin+0.25*(ym
     +ax-ymin)/nyticks,xmin,ymax-0.25*(ymax-ymin)/nyticks,.true.)

      if (.not. boxit) goto 5
C
C Tick the other two lines also.
C
      if (tickf) call pstickl(nxticks,-1.0*tickl,xmin,ymax,xmax,ymax,.tr
     +ue.)

      if (tickh) call pstickl(nxticks-1,-0.5*tickl,xmin+0.5*(xmax-xmin)/
     +nxticks,ymax,xmax-0.5*(xmax-xmin)/nxticks,ymax,.true.)

      if (tickq) call pstickl(2*nxticks-1,-0.25*tickl,xmin+0.25*(xmax-xm
     +in)/nxticks,ymax,xmax-0.25*(xmax-xmin)/nxticks,ymax,.true.)

      if (tickf) call pstickl(nyticks,1.0*tickl,xmax,ymin,xmax,ymax,.tru
     +e.)

      if (tickh) call pstickl(nyticks-1,0.5*tickl,xmax,ymin+0.5*(ymax-ym
     +in)/nyticks,xmax,ymax-0.5*(ymax-ymin)/nyticks,.true.)

      if (tickh) call pstickl(2*nyticks-1,0.25*tickl,xmax,ymin+0.25*(yma
     +x-ymin)/nyticks,xmax,ymax-0.25*(ymax-ymin)/nyticks,.true.)
C
C Show the x-values corresponding to the ticks.
C
 5    if (.not. sxvals) goto 6
      call pssetf(xvfont,xvfsize)
130   format(a,a,e16.8,a)
      do 220 i=0,nxticks
        xval = xmin + i*(xmax-xmin)/nxticks
        call pssval(xvoff,xval,xval,ymin,xval+dummy,ymin,.true.)
 220  continue
C
C Now show the y-values corresponding to the ticks.
C
 6    if (.not. syvals) goto 7
      call pssetf(yvfont,yvfsize)

      do 230 i=0,nyticks
        yval = ymin + i*(ymax-ymin)/nyticks
        call pssval(yvoff,yval,xmin,yval,xmin,yval+dummy,.true.)
 230  continue
C
C Now show x and y axes names and title of plot.
C
 7    call pssetf(xnfont,xnfsize)
      call pspcstr(xnoff,xname,(xmax+xmin)/2.,ymin,xmax,ymin,.true.)
      call pssetf(ynfont,ynfsize)
      call pspcstr(ynoff,yname,xmin,(ymax+ymin)/2.,xmin,ymax,.true.)
      call pssetf(tfont,tfsize)
      call pspcstr(toff,title,(xmax+xmin)/2.,ymin,xmax,ymin,.true.)
      return
      end
