.TH PTAGS L "February 11, 1988" "" "Local UNIX Programmer's Manual"
.UC 4
.SH NAME
ptags \- creates a tags file from Prolog sources
.SH SYNOPSIS
.B ptags
[\-w] [\-l] [\-a] [\-p] files
.br
.SH DESCRIPTION
.I Ptags
creates a tags file from one or more Prolog source files.
It will do for Prolog predicates what
.IR ctags (1)
does for C functions.
Like
.I ctags,
output is sent to a file called
.I tags
in the current directory.
Each line in
the tags file contains three tab-separated fields:
the first is the name of each predicate defined in the source files;
the second is the name of the file where the predicate is defined;
and the third is an instruction to be executed by
.IR vi (1)
or
.IR ex (1)
to find the definition within the source file.
As with
.I ctags,
the tags file is sorted alphabetically by predicate name.
.PP
.I Ptags
accepts four flags to control its operation and output:
.TP
.B \-w
suppresses warnings about multiple definitions of the same predicate
in different files.
Only the first occurrence of a definition is entered in the tags file.
.TP
.B \-l
use line numbers,
rather than regular expressions,
to locate predicate definitions in source files.
.TP
.B \-a
append the output to an existing tags file in the current directory.
This allows you to combine output from
.I ptags
with output from
.I ctags.
A single tags file can be used to access Prolog predicates and C functions
\(em useful if you are mixing Prolog and C in a single program.
.TP
.B \-p
sends the output to
.I stdout
instead of to tags file.
.PP
.SH "FILES"
.PP
\&./tags                      tags file created or appended to
.SH "SEE ALSO"
ctags(1), ex(1), vi(1)
.SH "DIAGNOSTICS"
.I Ptags
complains about a variety of things, mostly related to reaching limits
of static arrays, e.g. predicate name too long, too many symbols (predicates),
too many source files, etc.
.sp
.I Ptags
can optionally produce warnings about predicates defined
across different files - see \-w above.
Warnings are only issued once for each different file.
Note that the arity of these predicates may be different -
.I ptags
doesn't care.
.SH "BUGS"
.I Ptags
is NOT a syntax checker for Prolog code, and will quite happily swallow
syntax errors without complaint, probably filling the tags file with
garbage at the same time.
.sp
The parser is very basic and may not tolerate your style of layout
for Prolog code.
For example,
if regular expressions are used to locate definitions
(rather than line numbers), they will
expect all definitions to begin at the leftmost side of the
screen, because the regular expressions always begin with a
leading `^' character, followed immediately by the predicate
name, and some trailing context characters.
.sp
Most Prologs will allow "strange" characters in predicate names,
provided the names are enclosed in single quotes.
These names will often not be acceptable as tags, though they seem to
be accepted on the command line with the \-t option in ex or vi.
Nothing we can do about this.
Anybody care to modify ex?
.SH "AUTHORS"
.PP
Chris Tweed
.br
Bob Kemp
