/*
 * rf(1L)	A "rotary file"-like database for names, addresses, telephone
 *		numbers and related information.
 *
 *		Copyright (c) 1985
 *		by Lawrence Lippman, larry@kitty.UUCP
 *		Recognition Research Corp., Clarence, NY
 *
 *		This program may be freely used and distributed, provided that
 *		it is not used for monetary or other commercial gain, and
 *		provided that this notice remains intact.
 *
 *		Last revision: 3 Jun 1985
 */

#include <stdio.h>
#include <curses.h>
#include <signal.h>
#include <string.h>

#ifndef SYSDATA
#define	SYSDATA	"/usr/local/lib/rf_data"	/* system database file */
#endif

/*
 * Global variables
 */
	char	name[31], title[25], org[73], dept[25], adr[73];
	char	phone[2][16], telex[2][25], fax[2][16], uucp[2][25];
	char	homephone[16], homeadr[65], comments[4][73];
	char	Key[73], buf[80];
	int	Org, File, List;
	int	hits, phlines, comlines;
	int	terminate(), more(), strsearch();
	FILE	*f1, *fopen();

main(argc,argv)
	char	**argv;
{
	char	filename[65], *getenv();
	int	keywords;

/*
 * Set up defaults
 */
	File = 0;	/* Use system-wide database file, not user file */
	List = 0;	/* Display full entry, not just hit list */
	Org = 0;	/* Search by name, not organization */

/*
 * Get options and search string
 */
	(void) strcpy(Key,"");
	keywords = 0;

	while(argc-- > 1) {
		if(*argv[1] == '-')
			switch(argv[1][1]) {
				case 'l':
					List = 1;
					break;
				case 'f':
					File = 1;
					break;
				case 'o':
					Org = 1;
					break;
				default:
					usage();
			}

		else{
			if(keywords > 0)
				(void) strcat(Key," ");
			(void) strcat(Key,argv[1]);
			keywords++;
		}
		argv++;
	}

	if(strlen(Key) == 0)
		usage();

/*
 * Select and open database file
 */
	if(File)
		(void) sprintf(filename, "%s/.rf_data", getenv("HOME"));
	else
		(void) strcpy(filename, SYSDATA);

	if((f1 = fopen(filename,"r")) == NULL){
		(void) fprintf(stderr,"Cannot open data file %s\n",filename);
		exit(-1);
	}

/*
 * Catch signals and setup curses
 */
	(void) signal(SIGINT, terminate);
	(void) signal(SIGQUIT, terminate);

	initscr();
	if(List){
		idlok(stdscr,1);
		setscrreg(0,19);
		scrollok(stdscr,1);
	}

/*
 * Read file, search for records, and do it to it...
 */
	hits = 0;

	while(fgets(buf,80,f1) != NULL){
		if(buf[0] == 'N' && buf[1] == ':'){
			if(strlen(buf) == 2)
				(void) strcpy(name,"");
			else{
				(void) strxcpy(name,buf,2,30);
				if(!Org){
					if(strsearch(name,Key)){
						if(hits > 0 && !List)
							(void) more();
						(void) strcpy(org,"");
						(void) clrrecord();
						(void) rdrecord();
						(void) display();
						hits++;
					}
				}
			}
		}
		if(buf[0] == 'O' && buf[1] == ':'){
			(void) strxcpy(org,buf,2,72);
			if(Org){
				if(strsearch(org,Key)){
					if(hits > 0 && !List)
						(void) more();
					(void) clrrecord();
					(void) rdrecord();
					(void) display();
					hits++;
				}
			}
		}
	}

	(void) terminate();
	return(0);
}

/*
 * clrrecord:	clear all display strings
 */
clrrecord()
{
	int	i;

	(void) strcpy(title,"");
	(void) strcpy(dept,"");
	(void) strcpy(adr,"");
	(void) strcpy(homephone,"");
	(void) strcpy(homeadr,"");
	(void) strcpy(comments[0],"");
	for(i = 0; i <= 1; i++){
		(void) strcpy(phone[i],"");
		(void) strcpy(telex[i],"");
		(void) strcpy(fax[i],"");
		(void) strcpy(uucp[i],"");
	}
}

/*
 * rdrecord:	read the rest of a record following a "name" read
 */
rdrecord()
{
	int	phln, fxln, txln, uuln;

	phln = fxln = txln = uuln = 0;
	phlines = 0;
	comlines = 0;

	while(fgets(buf,80,f1) != NULL){

		if(buf[1] != ':' || strlen(buf) <= 1)
			return;

		switch(buf[0]) {
			case 'O':
				(void) strxcpy(org,buf,2,72);
				break;
			case 'T':
				(void) strxcpy(title,buf,2,24);
				break;
			case 'D':
				(void) strxcpy(dept,buf,2,24);
				break;
			case 'A':
				(void) strxcpy(adr,buf,2,72);
				break;
			case 'P':
				if(phln > 1)
					break;
				(void) strxcpy(phone[phln],buf,2,15);
				phln++;
				break;
			case 'X':
				if(txln > 1)
					break;
				(void) strxcpy(telex[txln],buf,2,24);
				txln++;
				break;
			case 'F':
				if(fxln > 1)
					break;
				(void) strxcpy(fax[fxln],buf,2,15);
				fxln++;
				break;
			case 'U':
				if(uuln > 1)
					break;
				(void) strxcpy(uucp[uuln],buf,2,24);
				uuln++;
				break;
			case 'H':
				(void) strxcpy(homephone,buf,2,15);
				break;
			case 'R':
				(void) strxcpy(homeadr,buf,2,64);
				break;
			case 'C':
				if(comlines > 3)
					break;
				(void) strxcpy(comments[comlines],buf,2,72);
				comlines++;
				break;
		}
		if(phln > 1 || txln > 1 || fxln > 1 || uuln > 1)
			phlines = 1;
	}
}

/*
 * display:	Display the results of a search
 */
display()
{
	int	l;

	if(List){
		if(hits == 0){
			erase();
			refresh();
			if(Org)
				mvaddstr(0, 0, org);
			else
				mvaddstr(0, 0, name);
		}else{
			if(Org)
				printw("%s", org);
			else
				printw("%s", name);
			
		}
		refresh();
		return(0);
	}

	erase();
	refresh();

	attron(A_REVERSE);
	mvaddstr(0, 0, "NAME");
	mvaddstr(0, 32, "TITLE");
	mvaddstr(0, 56, "DEPT");
	mvaddstr(3, 0, "ORGANIZATION NAME & ADDRESS");
	mvaddstr(7, 0, "TELEPHONE");
	mvaddstr(7, 16, "TELECOPIER");
	mvaddstr(7, 32, "TELEX");
	mvaddstr(7, 56, "UUCP");
	mvaddstr(10 + phlines, 0, "HOME TELEPHONE");
	mvaddstr(10 + phlines, 16, "HOME ADDRESS");
	mvaddstr(13 + phlines, 0, "COMMENTS");
	attroff(A_REVERSE);
	
	mvaddstr(1, 0, name);
	mvaddstr(1, 32, title);
	mvaddstr(1, 56, dept);
	mvaddstr(4, 0, org);
	mvaddstr(5, 0, adr);
	for(l = 0; l <= phlines; l++){
		mvaddstr(8 + l, 0, phone[l]);
		mvaddstr(8 + l, 16, fax[l]);
		mvaddstr(8 + l, 32, telex[l]);
		mvaddstr(8 + l, 56, uucp[l]);
	}
	mvaddstr(11 + phlines, 0, homephone);
	mvaddstr(11 + phlines, 16, homeadr);
	for(l = 0; l <= comlines; l++)
		mvaddstr(14 + phlines + l, 0, comments[l]);

	refresh();
	return(0);
}

/*
 * more:	Prompt for display of entries when multiple hits occur
 */
more()
{
again:
	attron(A_REVERSE);
	mvaddstr(20, 0, "MORE HITS: CONTINUE?");
	attroff(A_REVERSE);
	mvaddstr(20, 22, "[y] [n]  ");
	refresh();
	switch(getch()){
		case 'n':
		case 'N':
			(void) terminate();
		case 'y':
		case 'Y':
			return(0);
	}
	mvaddstr(20, 31, " ");
	goto again;
}

/*
 * terminate:	Exit gracefully
 */
terminate()
{
	(void) fclose(f1);

	move(LINES-1, 0);
	refresh();
	endwin();
	exit(0);
}

/*
 * strsearch:	Search for any occurence of string "t" in string "s";
 *		return 1 if a match found, otherwise return 0
 */
strsearch(s,t)
	char	s[80], t[80];
{
	register	i, n, nn;
	int	slength, tlength;
	char	st[80];

	slength = strlen(s);
	tlength = strlen(t);
	if(slength == 0 || tlength == 0)
		return(0);

	for (i = 0; i < slength; i++){
		n=0;
		for(nn = i; nn < i + tlength; nn++){
			if(nn > slength)
				return(0);
			st[n] = s[nn];
			n++;
		}
		st[n] = '\0';
		if (strcmp(st,t) == 0)
			return(1);
	}
	return(0);
}

/*
 * strxcpy:	Copy string "t" to string "s", with "offset" characters in
 *		string "t" skipped before the copy, and with a maximum of
 *		"maxs" characters [not including NULL] copied to string "s"
 */
strxcpy(s,t,offset,maxs)
	char	s[80], t[80];
	int	offset, maxs;
{
	register	n, nn;
	int	tlength;

	nn = 0;
	tlength = strlen(t);

	for(n = offset; n <= tlength + 1; n++){
		s[nn] = t[n];
		if(t[n] == '\0')
			return(0);
		nn++;
		if(nn == maxs){
			s[nn + 1] = '\0';
			return(0);
		}
	}
	return(0);
}

/*
 * usage:	Display usage error message and exit
 */
usage()
{
	(void) fprintf(stderr,"usage: rf [-f] [-l] [-o] searchstring\n");
	exit(-1);
}
