

/*
 * pty.c - Berkeley style pseudo tty driver for system V
 *
 * Copyright (c) 1987, Jens-Uwe Mager, FOCUS Computer GmbH
 * Not derived from licensed software.
 *
 * Permission is granted to freely use, copy, modify, and redistribute
 * this software, provided that no attempt is made to gain profit from it,
 * the author is not construed to be liable for any results of using the
 * software, alterations are clearly marked as such, and this notice is
 * not modified.
 */

/*
 * Modified for use on the UnixPC by:
 * Eric H. Herrin II
 * University of Kentucky Mathematical Sciences Laboratories
 * eric@ms.uky.edu, eric@ms.uky.csnet, !cbosgd!ukma!eric
 *
 * Further modified and improved by:
 * Mike "Ford" Ditto
 * Chief Executive Hacker, Omnicron Data Systems
 * kenobi!ford@crash.CTS.COM, ...!crash!kenobi!ford
 *
 * See README.3b1 for details of port and installation.
 */

/*
 * Yet more geefing done by:
 * Keith M. Gabryelski
 * Random
 * ...crash!portnoy!ag, ag@portnoy.cts.com
 *
 * See README.SCO for detials of port and installation.
 */
 /* XON, XOFF control flow added by Bob Best (bob@dhw68k.cts.com)
    6/13/88
*/

/* The UnixPC does not have any extra bits in t_state, thus
 * we provide other means of storing the state.
 */
#define MRWAIT	0x01				/* master waiting in read */
#define t_rloc	t_cc[0]				/* wchannel for MRWAIT */
#define MWWAIT	0x02				/* master waiting in write */
#define t_wloc	t_cc[1]				/* wchannel for MWWAIT */
#define MOPEN	0x04				/* master is open */
#define SOPEN	0x08				/* slave is open */

#define KERNEL		1
#define defined_io	1

#include "sys/param.h"
#include "sys/types.h"
#include "sys/sysmacros.h"
#include "sys/systm.h"
#include "sys/file.h"
#include "sys/conf.h"
#include "sys/page.h"		/* structure defs --kmg */
#include "sys/dir.h"		/* structure defs --kmg */
#include "sys/seg.h"		/* structure defs --kmg */
#include "sys/proc.h"
#include "sys/tty.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/errno.h"
/* #include "sys/termio.h"	/* included in sys/tty.h --kmg */
#include "sys/ttold.h"

#define	eprintf	printf		/* So what do ya want, whicker? */

#include "PTYCNT"		/* find out about PTYCNT and MASTERMINOR */

/* PTYCNT is the maximum number of master/slave pairs that will work */
#ifndef PTYCNT
#define PTYCNT 32
#endif

/* MASTERMINOR is the lowest minor device number for master devices */
#ifndef MASTERMINOR
#define MASTERMINOR ((minor(0xFFFF)+1)/2)
#endif

/* This macro is true if the device number passed corresponds to */
/* the MASTER side of a pty */
#ifdef MASTERBIT
#define Master(dev)	((dev)&MASTERBIT)
#else
#define Master(dev)	(minor((dev)) >= MASTERMINOR)
#endif MASTERBIT

/* The tty structures must be local to this driver, since there
 * is no conf.c
 */

struct tty pty_tty[PTYCNT];
int ptystate[PTYCNT];

int ptsproc();


ptyopen(dev, flag)
	dev_t		dev;
	int		flag;
{
	register struct tty *tp;

	dev = minor(dev);
	if (Master(dev)) {
#		ifdef DEBUG
		eprintf("pty: open(master): \n");
#		endif
		dev -= MASTERMINOR;
		tp = &pty_tty[dev];
		if (dev >= PTYCNT) {
			u.u_error = ENXIO;
			return;
		}
		/*
		* allow only one controlling process
		*/
		if (ptystate[dev] & MOPEN) {
			u.u_error = EBUSY;
			return;
		}
		ptystate[dev] |= MOPEN;
		tp->t_state |= CARR_ON;
		if (tp->t_state & WOPEN)
		{
			tp->t_state &= ~WOPEN;
			wakeup((caddr_t)&tp->t_canq);
		}
	} else {
#		ifdef DEBUG
		eprintf("pty: open(slave): \n");
#		endif
		tp = &pty_tty[dev];
		if (dev >= PTYCNT) {
			u.u_error = ENXIO;
			return;
		}
		if ((tp->t_state & (ISOPEN|WOPEN)) == 0) {
			ttinit(tp);
			tp->t_proc = ptsproc;
		}
		/*
		 * if master is still open, don't wait for carrier
		 */
		if (!(flag & FNDELAY)) {
			while ((tp->t_state & CARR_ON) == 0) {
				/* eprintf("slave_open going to sleep!"); */
				tp->t_state |= WOPEN;
				if (sleep((caddr_t)&tp->t_canq, TTIPRI|PCATCH)) {
					tp->t_state &= ~WOPEN;
					wakeup((caddr_t)&tp->t_canq);
					/* eprintf("slave open: interrupted"); */
					u.u_error = EINTR;
					return;
				}
				/* eprintf("slave_open woke up! %x %x",
					tp->t_state, u.u_procp->p_sig); */
			}
		}
		ptystate[dev] |= SOPEN;
		(*linesw[tp->t_line].l_open)(tp);
	}
}

ptyclose(dev, flag)
	dev_t		dev;
	int		flag;
{
	register struct tty *tp;

	dev = minor(dev);
	if (Master(dev)) {
#		ifdef DEBUG
		eprintf("pty: close(master): \n");
#		endif
		dev -= MASTERMINOR;
		tp = &pty_tty[dev];
		if (!(ptystate[dev] & SOPEN)) {
			tp->t_tbuf.c_size  -= tp->t_tbuf.c_count;
			tp->t_tbuf.c_count = 0;
			tp->t_rbuf.c_size  -= tp->t_rbuf.c_count;
			tp->t_rbuf.c_count = 0;
			(*linesw[tp->t_line].l_close)(tp);
		} else {
			signal(tp->t_pgrp, SIGHUP);
			ttyflush(tp, FREAD|FWRITE);
		}
		/*
		 * virtual carrier gone
		 */
		tp->t_state &= ~(CARR_ON);
		ptystate[dev] &= ~MOPEN;
	} else {
#		ifdef DEBUG
		eprintf("pty: close(slave): \n");
#		endif
		tp = &pty_tty[dev];
		if (!(ptystate[dev] & MOPEN)) {
			tp->t_tbuf.c_size  -= tp->t_tbuf.c_count;
			tp->t_tbuf.c_count = 0;
			tp->t_rbuf.c_size  -= tp->t_rbuf.c_count;
			tp->t_rbuf.c_count = 0;
			(*linesw[tp->t_line].l_close)(tp);
		}
		ptystate[dev] &= ~SOPEN;
		if (ptystate[dev] & MRWAIT)
		{
			ptystate[dev] &= ~MRWAIT;
			wakeup((caddr_t)&tp->t_rloc);
		}
		if (ptystate[dev] & MWWAIT)
		{
			ptystate[dev] &= ~MRWAIT;
			wakeup((caddr_t)&tp->t_wloc);
		}
	}
}

ptyread(dev)
	dev_t		dev;
{
	register struct tty *tp;
	register n;

	dev = minor(dev);
	if (Master(dev)) {
		int didsome=0;
#		ifdef DEBUG
		eprintf("pty: read(master): \n");
#		endif
		dev -= MASTERMINOR;
		tp = &pty_tty[dev];
		while (u.u_count) {
			ptsproc(tp, T_OUTPUT);
			if ((tp->t_state & (TTSTOP|TIMEOUT))
			    || tp->t_tbuf.c_ptr == NULL || tp->t_tbuf.c_count == 0) {
				if (didsome)
					break;
				if ((ptystate[dev] & SOPEN) == 0) {
					u.u_error = EIO;
					return;
				}
				if (u.u_fmode & FNDELAY)
					break;
#				ifdef DEBUG
				eprintf("pty: read(master): master going to sleep\n");
#				endif
				ptystate[dev] |= MRWAIT;
				sleep((caddr_t)&tp->t_rloc, TTIPRI);
#				ifdef DEBUG
				eprintf("pty: read(master): master woke up\n");
#				endif

				continue;
			}
			n = min(u.u_count, tp->t_tbuf.c_count);
			if (n) {
#				ifdef DEBUG
				eprintf("pty: read(master): got some stuff\n");
#				endif
				++didsome;
				if (copyout(tp->t_tbuf.c_ptr, u.u_base, n)) {
					u.u_error = EFAULT;
					break;
				}
				tp->t_tbuf.c_count -= n;
				tp->t_tbuf.c_ptr += n;
				u.u_base += n;
				u.u_count -= n;
			}
		}
	} else {
#		ifdef DEBUG
		eprintf("pty: read(slave): \n");
#		endif
		tp = &pty_tty[dev];
#		ifdef DEBUG
		eprintf("pty: read(slave): got some stuff\n");
#		endif
		(*linesw[tp->t_line].l_read)(tp);
	}
}

ptywrite(dev)
	dev_t		dev;
{
	register struct tty *tp;
	register n;
	int flg;

	dev = minor(dev);
	if (Master(dev)) {
#		ifdef DEBUG
		eprintf("pty: write(master): \n");
#		endif
		dev -= MASTERMINOR;
		tp = &pty_tty[dev];
		while (u.u_count) {
			if ((tp->t_state & TBLOCK)
			    || tp->t_rbuf.c_ptr == NULL) {
				if ((ptystate[dev] & SOPEN) == 0) {
					u.u_error = EIO;
					return;
				}
				if (u.u_fmode & FNDELAY)
					break;
				ptystate[dev] |= MWWAIT;
#				ifdef DEBUG
				eprintf("pty: write(master): going to sleep\n");
#				endif

				sleep((caddr_t)&tp->t_wloc, TTOPRI);

#				ifdef DEBUG
				eprintf("pty: write: waking up\n");
#				endif

				continue;
			}
			n = min(u.u_count, tp->t_rbuf.c_count);
			if (n) {
#				ifdef DEBUG
				eprintf("pty: write(master): sending some stuff\n");
#				endif
				flg=tp->t_iflag;
				if(flg & IXON) {
					char ctmp;
					copyin(u.u_base, &ctmp, 1);
					ctmp &= 0177;
					if(tp->t_state & TTSTOP) {
						if(ctmp==CSTART || flg & IXANY)
						(*tp->t_proc)(tp,T_RESUME);
					} else if(ctmp==CSTOP) {
						(*tp->t_proc)(tp,T_SUSPEND);
					}
					if(ctmp==CSTART || ctmp==CSTOP) {
						u.u_base++;
						u.u_count--;
						continue;
					}
				}
				if (copyin(u.u_base,tp->t_rbuf.c_ptr, 1)) {
					u.u_error = EFAULT;
					break;
				}
				tp->t_rbuf.c_count--;
				u.u_base++;
				u.u_count--;
			}
			(*linesw[tp->t_line].l_input)(tp);
		}
	} else {
#		ifdef DEBUG
		eprintf("pty: write(slave): \n");
#		endif
		tp = &pty_tty[dev];
#		ifdef DEBUG
		eprintf("pty: write(slave): sending some stuff\n");
#		endif
		(*linesw[tp->t_line].l_write)(tp);
	}
}

ptyioctl(dev, cmd, arg, mode)
	dev_t		dev;
	int		cmd, arg, mode;
{
	register struct tty *tp;

	dev = minor(dev);
	if (Master(dev)) {
#		ifdef DEBUG
		eprintf("pty: ioctl(master): \n");
#		endif
		dev -= MASTERMINOR;
		tp = &pty_tty[dev];
		/*
		 * sorry, but we can't fiddle with the tty struct without
		 * having done LDOPEN
		 */
		if (tp->t_state & ISOPEN) {
			if (cmd == TCSBRK && arg == 0) {	/* 0 was NULL
								   --kmg */
				signal(tp->t_pgrp, SIGINT);
				if ((tp->t_iflag & NOFLSH) == 0)
					ttyflush(tp, FREAD|FWRITE);
			} else {
				/*
				 * we must flush output to avoid hang in ttywait
				 */
				if (cmd == TCSETAW || cmd == TCSETAF ||
				   cmd == TCSBRK || cmd == TIOCSETP)
					ttyflush(tp, FWRITE);
				ttiocom(tp, cmd, arg, mode);
			}
		} else
			u.u_error = EINVAL;
	} else {
#		ifdef DEBUG
		eprintf("pty: ioctl(slave): \n");
#		endif
		tp = &pty_tty[dev];
		ttiocom(tp, cmd, arg, mode);
	}
}

ptsproc(tp, cmd)
register struct tty *tp;
{
	register struct ccblock *tbuf;
	extern ttrstrt();

	switch (cmd) {
	case T_TIME:
#		ifdef DEBUG
		eprintf("pty: ptsproc: T_TIME:\n");
#		endif
		tp->t_state &= ~TIMEOUT;
		goto start;
	case T_WFLUSH:
#		ifdef DEBUG
		eprintf("pty: ptsproc: T_WFLUSH:\n");
#		endif
		tp->t_tbuf.c_size  -= tp->t_tbuf.c_count;
		tp->t_tbuf.c_count = 0;
		/* fall through */
	case T_RESUME:
#		ifdef DEBUG
		eprintf("pty: ptsproc: T_RESUME:\n");
#		endif
		tp->t_state &= ~TTSTOP;
		/* fall through */
	case T_OUTPUT:
start:
#		ifdef DEBUG
		eprintf("pty: ptsproc: T_OUTPUT:\n");
#		endif
		if (tp->t_state & (TTSTOP|TIMEOUT))
			break;
#		ifdef DEBUG
		eprintf("pty: ptsproc: T_OUTPUT: past(TTSTOP|TIMEOUT)");
#		endif
		tbuf = &tp->t_tbuf;
		if (tbuf->c_ptr == NULL || tbuf->c_count == 0) {
#		ifdef DEBUG
		eprintf("pty: ptsproc: T_OUTPUT: tbuf empty, may break\n");
#		endif
			if (tbuf->c_ptr)
				tbuf->c_ptr -= tbuf->c_size;
			if (!(CPRES & (*linesw[tp->t_line].l_output)(tp)))
				break;
		}
		if (tbuf->c_count && (ptystate[tp-pty_tty] & MRWAIT)) {
#		ifdef DEBUG
		eprintf("pty: ptsproc: T_OUTPUT: waking up master\n");
#		endif
			ptystate[tp-pty_tty] &= ~MRWAIT;
			wakeup((caddr_t)&tp->t_rloc);
		}
#		ifdef DEBUG
		eprintf("pty: ptsproc: T_OUTPUT: leaving end\n");
#		endif
		break;
	case T_SUSPEND:
#		ifdef DEBUG
		eprintf("pty: ptsproc: T_SUSPEND:\n");
#		endif
		tp->t_state |= TTSTOP;
		break;
	case T_BLOCK:
#		ifdef DEBUG
		eprintf("pty: ptsproc: T_BLOCK:\n");
#		endif
		/*
		 * the check for ICANON appears to be neccessary
		 * to avoid a hang when overflowing input
		 */
		if ((tp->t_iflag & ICANON) == 0)
			tp->t_state |= TBLOCK;
		break;
	case T_BREAK:
#		ifdef DEBUG
		eprintf("pty: ptsproc: T_BREAK:\n");
#		endif
		tp->t_state |= TIMEOUT;
		timeout(ttrstrt, tp, HZ/4);
		break;
#ifdef T_LOG_FLUSH
	case T_LOG_FLUSH:
#endif
	case T_RFLUSH:
#		ifdef DEBUG
		eprintf("pty: ptsproc: T_RFLUSH:\n");
#		endif
		if (!(tp->t_state & TBLOCK))
			break;
		/* fall through */
	case T_UNBLOCK:
#		ifdef DEBUG
		eprintf("pty: ptsproc: T_UNBLOCK:\n");
#		endif
		tp->t_state &= ~(TTXOFF|TBLOCK);
		/* fall through */
	case T_INPUT:
#		ifdef DEBUG
		eprintf("pty: ptsproc: T_INPUT:\n");
#		endif
		if (ptystate[tp-pty_tty] & MWWAIT) {
			ptystate[tp-pty_tty] &= ~MWWAIT;
#			ifdef DEBUG
			eprintf("pty: ptsproc: T_INPUT: waking up master\n");
#			endif
			wakeup((caddr_t)&tp->t_wloc);
		}
		break;
	default:
#		ifdef DEBUG
		eprintf("pty: ptsproc: default:\n");
#		else
		;
#		endif
	}
}

/*	PTYRELEASE is a 3b1ism.
 *
 * ptyrelease()
 * {
 *	register i;
 * #	ifdef DEBUG
 *	eprintf("pty: ptyrelease:\n");
 * #	endif
 *	for ( i=0 ; i<PTYCNT ; ++i )
 *		if ( (ptystate[i] & (SOPEN|MOPEN))
 *		   || (pty_tty[i].t_state & WOPEN) )
 *		{
 *			u.u_error = EBUSY;
 *			return;
 *		}
 *	return;
 * }
 */
