; LAST EDIT: Fri Nov  4 11:16:33 1988 by Shape - New Horizons in Software Engineering (chaos!shape) 
; LAST EDIT: Thu Nov  3 14:16:48 1988 by Shape - New Horizons in Software Engineering (chaos!shape) 
; LAST EDIT: Tue Nov  1 12:46:34 1988 by Uli Pralle (coma!uli) 
;;; This file is not part of the GNU Emacs distribution (yet).

;; SHAPE commands for Emacs
;; Copyright (C) 1985, 1986 Free Software Foundation, Inc.

;; This file is part of GNU Emacs.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the GNU Emacs General Public
;; License for full details.

;; Everyone is granted permission to copy, modify and redistribute
;; GNU Emacs, but only under the conditions described in the
;; GNU Emacs General Public License.   A copy of this license is
;; supposed to have been given to you along with GNU Emacs so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

;In loaddefs.el
(defvar shape-listing-switches "-al"
  "Switches passed to ls for shape. MUST contain the 'l' option.
	CANNOT contain the 'F' option.")
(defvar shape-compare-file1 nil)

(defun shape-readin (dirname buffer)
  (save-excursion
    (set-buffer buffer)
    (let ((buffer-read-only nil))
      (widen)
      (erase-buffer)
      (setq dirname (expand-file-name dirname))
      (if (file-directory-p dirname)
	  (call-process "vl" nil buffer nil
			shape-listing-switches dirname)
	(let ((default-directory (file-name-directory dirname)))
	  (call-process shell-file-name nil buffer nil
			"-c" (concat "vl " shape-listing-switches " "
				     (file-name-nondirectory dirname)))))
      (goto-char (point-min))
      (while (not (eobp))
	(insert "  ")
	(forward-line 1))
      (goto-char (point-min)))))

(defun shape-find-buffer (dirname)
  (let ((blist (buffer-list))
	found)
    (while blist
      (save-excursion
        (set-buffer (car blist))
	(if (and (eq major-mode 'shape-mode)
		 (equal shape-directory dirname))
	    (setq found (car blist)
		  blist nil)
	  (setq blist (cdr blist)))))
    (or found
	(progn (if (string-match "/$" dirname)
		   (setq dirname (substring dirname 0 -1)))
	       (create-file-buffer (file-name-nondirectory dirname))))))

(defun shapetools(&optional dirname)
  "\"Edit\" directory DIRNAME.  Delete some files in it.
       Shape displays a list of files in DIRNAME.
       You can move around in it with the usual commands.
       You can flag files for deletion with C-d
       and then delete them by typing `x'.
       Type `h' after entering shape for more info."
  (interactive)
  (if (equal dirname nil)
      (setq dirname (shape-get-filename nil t))
    nil)
  (if (equal dirname nil)
      (setq dirname (read-file-name "Shapetools (directory): "
				    nil default-directory nil))
    nil)

  (switch-to-buffer (shape-noselect dirname)))

(defun shape-other-window (dirname)
  "\"Edit\" directory DIRNAME.  Like M-x shape but selects in another window."
  (interactive (list (read-file-name "Shapetools in other window (directory): "
				     nil default-directory nil)))
  (switch-to-buffer-other-window (shape-noselect dirname)))

(defun shape-noselect (dirname)
  "Like M-x shape but returns the shape buffer as value, does not select it."
  (or dirname (setq dirname default-directory))
  (if (string-match "./$" dirname)
      (setq dirname (substring dirname 0 -1)))
  (setq dirname (expand-file-name dirname))
  (and (not (string-match "/$" dirname))
       (file-directory-p dirname)
       (setq dirname (concat dirname "/")))
  (let ((buffer (shape-find-buffer dirname)))
    (save-excursion
      (set-buffer buffer)
      (shape-readin dirname buffer)
      (shape-move-to-filename)
      (shape-mode dirname))
    buffer))

(defun shape-revert (&optional arg noconfirm)
  (let ((opoint (point))
	(ofile (shape-get-filename t t))
	(buffer-read-only nil))
    (erase-buffer)
    (shape-readin shape-directory (current-buffer))
    (or (and ofile (re-search-forward (concat " " (regexp-quote ofile) "$")
				      nil t))
	(goto-char opoint))
    (beginning-of-line)))

(defvar shape-mode-map nil "Local keymap for shape-mode buffers.")
(if shape-mode-map
    nil
  (setq shape-mode-map (make-keymap))
  (suppress-keymap shape-mode-map)
  (define-key shape-mode-map "r" 'shape-rename-file)
  (define-key shape-mode-map "\C-d" 'shape-flag-file-deleted)
  (define-key shape-mode-map "d" 'shape-flag-file-deleted)
  (define-key shape-mode-map "l" 'shape-vlog)
  (define-key shape-mode-map "v" 'shape-view-file)
  (define-key shape-mode-map "e" 'shape-find-file)
  (define-key shape-mode-map "f" 'shape-find-file)
  (define-key shape-mode-map "o" 'shape-find-file-other-window)
  (define-key shape-mode-map "q" '(lambda () (interactive) (kill-buffer (current-buffer))))
  (define-key shape-mode-map "u" 'shape-unflag)
  (define-key shape-mode-map "x" 'shape-do-deletions)
  (define-key shape-mode-map "\177" 'shape-backup-unflag)
  (define-key shape-mode-map "?" 'shape-summary)
  (define-key shape-mode-map "c" 'shape-copy-file)
  (define-key shape-mode-map "h" 'describe-mode)
  (define-key shape-mode-map " "  'shape-next-line)
  (define-key shape-mode-map "\C-n" 'shape-next-line)
  (define-key shape-mode-map "\C-p" 'shape-previous-line)
  (define-key shape-mode-map "n" 'shape-next-line)
  (define-key shape-mode-map "p" 'shape-previous-line)
  (define-key shape-mode-map "g" 'revert-buffer)
  (define-key shape-mode-map "R" 'shape-retrv)
  (define-key shape-mode-map "O" 'shape-vadm-change-owner)
  (define-key shape-mode-map "P" 'shape-vadm-promote)
  (define-key shape-mode-map "U" 'shape-vadm-unpromote)
  (define-key shape-mode-map "M" 'shape-vadm-change-mode)
  (define-key shape-mode-map "A" 'shape-vadm-change-author)
  (define-key shape-mode-map "S" 'shape-save)
  (define-key shape-mode-map "V" 'shape-vadm)
  (define-key shape-mode-map "C" 'shape-compare)
  (define-key shape-mode-map "F" 'shape-fold)
  (define-key shape-mode-map "X" 'shape-unfold)
  (define-key shape-mode-map "W" 'shape-mail-wishes)
  (define-key shape-mode-map "B" 'shape-mail-bugs)
  (define-key shape-mode-map "E" 'shape-execute))


;; Shape mode is suitable only for specially formatted data.
(put 'shape-mode 'mode-class 'special)

(defun shape-mode (dirname)
"- M change file's mode.                  - d flag a file for Deletion.
- G change group.                        - u unflag a file (remove its D flag).
- O change owner.                        - x execute the deletions requested.
- A change author.                       - e edit file or list directory.
- P promote a saved version.             - o find file/directory other window.
- U unpromote a saved version.           - W mail wishes (B to mail a bug).
- C compare two files.                   - c copy a file.
- S save a busy version.                 - v view a file in View mode.
- F fold directory                       - g read the directory again.
- X unfold file or directory             - E execute shape
- l show logentry
Space and Rubout can be used to move down and up by lines.
\\{shape-mode-map}"
  (kill-all-local-variables)    
  (make-local-variable 'revert-buffer-function)
  (setq revert-buffer-function 'shape-revert)
  (setq major-mode 'shape-mode)
  (setq mode-name "Shape")
  (make-local-variable 'shape-directory)
  (setq shape-directory dirname)
  (setq default-directory 
	(if (file-directory-p dirname)
	    dirname (file-name-directory dirname)))
  (setq mode-line-buffer-identification '("Shape Tools: %17b"))
  (setq case-fold-search nil)
  (setq buffer-read-only t)
  (use-local-map shape-mode-map)
  (run-hooks 'shape-mode-hook))

(defun shape-repeat-over-lines (arg function)
  (beginning-of-line)
  (while (and (> arg 0) (not (eobp)))
    (setq arg (1- arg))
    (save-excursion
      (beginning-of-line)
      (and (bobp) (looking-at "  total")
	   (error "No file on this line"))
      (funcall function))
    (forward-line 1)
    (shape-move-to-filename))
  (while (and (< arg 0) (not (bobp)))
    (setq arg (1+ arg))
    (forward-line -1)
    (shape-move-to-filename)
    (save-excursion
      (beginning-of-line)
      (funcall function))))

(defun shape-flag-file-deleted (&optional arg)
  "In shape, flag the current line's file for deletion.
With arg, repeat over several lines."
  (interactive "p")
  (shape-repeat-over-lines (or arg 1)
    '(lambda ()
       (let ((buffer-read-only nil))
	 (if (looking-at "  d")
	     nil
	   (if (or (looking-at "  .......... s ")
		   (looking-at "  .......... b "))
	       (progn
		 (delete-char 1)
		 (insert "D"))
	     (message "Only saved or busy versions may be deleted")))))))

(defun shape-summary ()
  (interactive)
  ;>> this should check the key-bindings and use substitute-command-keys if non-standard
  (message
   "Commands: ACFGMOPSUX cdegoruvx \(h for more help\)"))

(defun shape-unflag (arg)
  "In shape, remove the current line's delete flag then move to next line."
  (interactive "p")
  (shape-repeat-over-lines arg
    '(lambda ()
       (let ((buffer-read-only nil))
	 (delete-char 1)
	 (insert " ")
	 (forward-char -1)))))

(defun shape-backup-unflag (arg)
  "In shape, move up a line and remove deletion flag there."
  (interactive "p")
  (shape-unflag (- arg)))

(defun shape-next-line (arg)
  "Move down ARG lines then position at filename."
  (interactive "p")
  (next-line arg)
  (shape-move-to-filename))

(defun shape-previous-line (arg)
  "Move up ARG lines then position at filename."
  (interactive "p")
  (previous-line arg)
  (shape-move-to-filename))

(defun shape-find-file ()
  "In shape, visit the file or directory named on this line."
  (interactive)
  (if (file-folded-p (shape-get-filename))
      (shapetools (substring (shape-get-filename) 0 -3))
    (if (file-AFS-p (shape-get-filename))
	(message "Can't edit a version or folded file")
      (find-file (shape-get-filename)))))

(defun shape-view-file ()
  "In shape, examine a file in view mode, returning to shape when done."
  (interactive)
  (if (file-directory-p (shape-get-filename))
      (shapetools (shape-get-filename))
    (if (file-folded-p (shape-get-filename))
	(shapetools (substring (shape-get-filename) 0 -3))
      (if (file-AFS-p (shape-get-filename))
	  (shape-vcat)
	(view-file (shape-get-filename))))))
	    
(defun shape-find-file-other-window ()
  "In shape, visit this file or directory in another window."
  (interactive)
  (if (file-folded-p (shape-get-filename))
      (shape-other-window (substring (shape-get-filename) 0 -3))
    (if (file-AFS-p (shape-get-filename))
	(message "Can't edit a version")
      (if (file-DIR-p)
	  (shape-othe-window (shape-get-filename))
	(find-file-other-window (shape-get-filename))))))

(defun shape-get-filename (&optional localp no-error-if-not-filep)
  "In shape, return name of file mentioned on this line.
Value returned normally includes the directory name.
A non-nil 1st argument means do not include it.  A non-nil 2nd argument
says return nil if no filename on this line, otherwise an error occurs."
  (let (eol)
    (save-excursion
      (end-of-line)
      (setq eol (point))
      (beginning-of-line)
      (if (re-search-forward
	   "\\(Jan\\|Feb\\|Mar\\|Apr\\|May\\|Jun\\|Jul\\|Aug\\|Sep\\|Oct\\|Nov\\|Dec\\)[ ]+[0-9]+"
	   eol t)
	  (progn (skip-chars-forward " ")
		 (skip-chars-forward "^ " eol)
		 (skip-chars-forward " " eol)
		 (skip-chars-forward "^ " eol)
		 (skip-chars-forward " " eol)
		 (let ((beg (point)))
		   (skip-chars-forward "^ \n")
		   (if localp
		       (buffer-substring beg (point))
		     ;; >> uses default-directory, could lose on cd, multiple.
		     (concat default-directory (buffer-substring beg (point))))))
	(if no-error-if-not-filep nil
	  (error "No file on this line"))))))

(defun shape-move-to-filename ()
  "In shape, move to first char of filename on this line.
Returns position (point) or nil if no filename on this line."
  (let ((eol (progn (end-of-line) (point))))
    (beginning-of-line)
    (if (re-search-forward
	 "\\(Jan\\|Feb\\|Mar\\|Apr\\|May\\|Jun\\|Jul\\|Aug\\|Sep\\|Oct\\|Nov\\|Dec\\)[ ]+[0-9]+"
	 eol t)
	(progn
	  (skip-chars-forward " ")
	  (skip-chars-forward "^ " eol)
	  (skip-chars-forward " " eol)
	  (skip-chars-forward "^ " eol)
	  (skip-chars-forward " " eol)
	  (point)))))

(defun shape-map-shape-file-lines (fn)
  "perform fn with point at the end of each non-directory line:
arguments are the short and long filename"
  (save-excursion
    (let (filename longfilename (buffer-read-only nil))
      (goto-char (point-min))
      (while (not (eobp))
	(save-excursion
	  (and (not (looking-at "  d"))
	       (not (eolp))
	       (setq filename (shape-get-filename t t)
		     longfilename (shape-get-filename nil t))
	       (progn (end-of-line)
		      (funcall fn filename longfilename))))
	(forward-line 1)))))


(defun shape-collect-file-versions (ignore fn)
  "If it looks like fn has versions, we make a list of the versions.
We may want to flag some for deletion."
    (let* ((base-versions
	    (concat (file-name-nondirectory fn) ".~"))
	   (bv-length (length base-versions))
	   (possibilities (file-name-all-completions
			   base-versions
			   (file-name-directory fn)))
	   (versions (mapcar 'backup-extract-version possibilities)))
      (if versions
	  (setq file-version-assoc-list (cons (cons fn versions)
					      file-version-assoc-list)))))

(defun shape-trample-file-versions (ignore fn)
  (let* ((start-vn (string-match "\\.~[0-9]+~$" fn))
	 base-version-list)
    (and start-vn
	 (setq base-version-list	; there was a base version to which 
	       (assoc (substring fn 0 start-vn)	; this looks like a 
		      file-version-assoc-list))	; subversion
	 (not (memq (string-to-int (substring fn (+ 2 start-vn)))
		    base-version-list))	; this one doesn't make the cut
	 (shape-flag-this-line-for-DEATH))))

(defun shape-flag-this-line-for-DEATH ()
  (beginning-of-line)
  (delete-char 1)
  (insert "D"))

(defun shape-rename-file (to-file)
  "Rename this file to TO-FILE."
  (interactive "FRename to: ")
  (setq to-file (expand-file-name to-file))
  (rename-file (shape-get-filename) to-file)
  (let ((buffer-read-only nil))
    (beginning-of-line)
    (delete-region (point) (progn (forward-line 1) (point)))
    (setq to-file (expand-file-name to-file))
    (shape-add-entry (file-name-directory to-file)
		     (file-name-nondirectory to-file))))

(defun shape-copy-file ()
  "Copy this file to TO-FILE."
  (interactive)
  (let ((from-file (shape-get-filename t)))
  (if (file-AFS-p (shape-get-filename t))
      (message "Can't copy saved files")
    (setq to-file (read-string (concat "Copy " from-file " to: ")))
    (copy-file (shape-get-filename) to-file)
    (setq to-file (expand-file-name to-file))
    (shape-add-entry (file-name-directory to-file)
		     (file-name-nondirectory to-file)))))

(defun shape-add-entry (directory filename)
  ;; If tree shape is implemented, this function will have to do
  ;; something smarter with the directory.  Currently, just check
  ;; default directory, if same, add the new entry at point.  With tree
  ;; shape, should call 'shape-current-directory' or similar.  Note
  ;; that this adds the entry 'out of order' if files sorted by time,
  ;; etc.
  (if (string-equal directory default-directory)
      (let ((buffer-read-only nil))
	(beginning-of-line)
	(if (file-AFS-p filename)
	    (call-process "vl" nil t nil
			  shape-listing-switches
			  (concat directory filename))
	  (call-process "vl" nil t nil shape-listing-switches
			"-sb" (concat directory filename)))
	(forward-line -1)
	(insert "  ")
	(shape-move-to-filename)
	(let* ((beg (point))
	       (end (progn (end-of-line) (point))))
	  (setq filename (buffer-substring beg end))
	  (delete-region beg end)
	  (insert (file-name-nondirectory filename)))
	(beginning-of-line))))

(defun shape-chgrp (group)
  "Change group of this file."
  (interactive "sChange to Group: ")
  (let ((buffer-read-only nil)
	(file (shape-get-filename)))
    (call-process "/bin/chgrp" nil nil nil group file)
    (shape-redisplay file)))

(defun shape-redisplay (file)
  "Redisplay this line."
  (beginning-of-line)
  (delete-region (point) (progn (forward-line 1) (point)))
  (if file (shape-add-entry (file-name-directory    file)
			    (file-name-nondirectory file)))
  (shape-move-to-filename))

(defun shape-do-deletions ()
  "In shape, delete the files flagged for deletion."
  (interactive)
  (let (delete-list answer)
    (save-excursion
     (goto-char 1)
     (while (re-search-forward "^D" nil t)
       (setq delete-list
	     (cons (cons (shape-get-filename t) (1- (point)))
		   delete-list))))
    (if (null delete-list)
	(message "(No deletions requested)")
      (save-window-excursion
       (switch-to-buffer " *Deletions*")
       (erase-buffer)
       (setq fill-column 70)
       (let ((l (reverse delete-list)))
	 ;; Files should be in forward order for this loop.
	 (while l
	   (if (> (current-column) 59)
	       (insert ?\n)
	     (or (bobp)
		 (indent-to (* (/ (+ (current-column) 19) 20) 20) 1)))
	   (insert (car (car l)))
	   (setq l (cdr l))))
       (goto-char (point-min))
       (setq answer (yes-or-no-p "Delete these files? ")))
      (if answer
	  (let ((l delete-list)
		failures)
	    ;; Files better be in reverse order for this loop!
	    ;; That way as changes are made in the buffer
	    ;; they do not shift the lines still to be changed.
	    (while l
	      (goto-char (cdr (car l)))
	      (let ((buffer-read-only nil))
		(condition-case ()
		    (progn (shape-delete-file (concat default-directory
						      (car (car l))))
			   (delete-region (point)
					  (progn (forward-line 1) (point))))
		  
		  (error (delete-char 1)
			 (insert " ")
			 (setq failures (cons (car (car l)) failures)))))
	      (setq l (cdr l)))
	    (if failures
		(message "Deletions failed: %s"
			 (prin1-to-string failures))))))))


(defun shape-vcat()
  "retrieve and old version and display it."
  (interactive)
  (setq vcat-buffer (create-file-buffer (shape-get-filename)))
  (call-process shape-vcat-command nil vcat-buffer nil "-q" (shape-get-filename))
  (message "Restoring %s ..." (shape-get-filename t))
  (view-buffer vcat-buffer)
  (kill-buffer vcat-buffer)
)

(defun shape-vlog()
  "Display logentry for a particular version or entire history."
  (interactive)
  (if (file-directory-p (shape-get-filename))
      (error "Directories don't have any log-entries")
    (if (file-folded-p (shape-get-filename))
	(progn 
	  (setq history-filename (substring (shape-get-filename) 0 -3))
	  (setq msg-string 
		(concat "History log for " history-filename)))
      (if (file-AFS-p (shape-get-filename))
	  (progn (setq history-filename (shape-get-filename))
		 (setq msg-string (concat "Log entry for " history-filename)))
	(setq history-filename (shape-get-filename))
	(setq msg-string (concat "History log for " history-filename))))

    (setq vlog-buffer (create-file-buffer msg-string))
    (call-process shape-vlog-command nil vlog-buffer 
		  nil history-filename)
    (message (concat "Viewing " msg-string))
    (sit-for 2)
    (setq old-view-hook view-hook view-hook '(beginning-of-buffer))
    (view-buffer vlog-buffer)
    (setq view-hook old-view-hook)
    (kill-buffer vlog-buffer))
  )

(defun shape-vadm (vadm-input)
  "Perform vadm features."
  (interactive "svadm: ")
  (let ((buffer-read-only nil)
	(file (shape-get-filename)))
    (call-process shape-vadm-command nil nil nil "-q" vadm-input file)
    (shape-redisplay file)))

(defun shape-vadm-promote()
  "Performs vadm -promote."
  (interactive)
  (let ((buffer-read-only nil)
	(file (shape-get-filename t))
	(file2 (shape-get-filename)))
    (if (not (file-AFS-p file))
	(message "Can't promote busy file or directory %s" file)
      (message "Promoting %s ..." file)
      (call-process shape-vadm-command nil nil nil "-q" "-promote" file2)
      (sit-for 1 t)
      (shape-redisplay file2)
      (message "Done."))))

(defun shape-vadm-unpromote()
  "Performs vadm -unpromote."
  (interactive)
  (let ((buffer-read-only nil)
	(file (shape-get-filename t))
	(file2 (shape-get-filename)))
    (if (not (file-AFS-p file))
	(message "Can't unpromote busy file or directory %s" file)
      (message "Unpromoting %s ..." file)
      (call-process shape-vadm-command nil nil nil "-q" "-unpromote" file2)
      (sit-for 1 t)
      (shape-redisplay file2)
      (message "Done."))))

(defun shape-vadm-change-mode()
  "Performs vadm -chmod."
  (interactive)
  (let ((buffer-read-only nil)
	(file (shape-get-filename t))
	(file2 (shape-get-filename)))
    (setq input (read-string (concat "Change mode of " file " to: ")))
    (if (file-AFS-p file2)
	(call-process shape-vadm-command nil nil nil "-q" "-chmod" input file2)
      (call-process "/bin/chmod" nil nil nil input file2))
    (shape-redisplay file2)
    (message "Done.")))

(defun shape-vadm-change-author()
  "Performs vadm -chaut."
  (interactive)
  (let ((buffer-read-only nil)
	(file (shape-get-filename t))
	(file2 (shape-get-filename)))
    (setq input (read-string (concat "Change author of " file " to: ")))
    (call-process shape-vadm-command nil nil nil "-q" "-chaut" input file2)
    (shape-redisplay file2)
    (message "Done.")))


(defun shape-vadm-change-owner()
  "Performs vadm -chown."
  (interactive)
  (let ((buffer-read-only nil)
	(file (shape-get-filename t))
	(file2 (shape-get-filename)))
    (setq input (read-string (concat "Change owner of " file " to: ")))
    (call-process shape-vadm-command nil nil nil "-q" "-chown" input file2)
    (shape-redisplay file2)
    (message "Done.")))


(defun shape-save ()
  "saves a file via the save command."
  (interactive)
  (save-excursion
  (let ((buffer-read-only nil)
	(file (shape-get-filename))
	(file2 (shape-get-filename t)))
  (if (or (file-AFS-p file) (file-DIR-p))
      (message "This file not a busy file or a directory")
    (if (y-or-n-p "Describe this document or changes? ")
	(progn 
	  (setq descfile (make-temp-name "/tmp/save"))
	  (shape-get-description descfile)
	  (message "Saving file %s" file2)
	  (call-process shape-save-command nil t nil "-f" "-q" "-t"
			descfile file)
	  (delete-file descfile)
	  (shape-insert-new-version file2))
      (message "Saving file %s" file2)
      (call-process shape-save-command nil t nil "-f" "-q" file)
      (shape-insert-new-version file2)
      (while (search-forward file2 nil t)))))))

(defun shape-submit ()
  "submit a file via the submit command."
  (interactive)
  (save-excursion
  (let ((buffer-read-only nil)
	(file (shape-get-filename))
	(file2 (shape-get-filename t)))
  (if (or (file-AFS-p file) (file-DIR-p))
      (message "This file not a busy file or a directory")
    (if (y-or-n-p "Describe this document or changes? ")
	(progn 
	  (setq descfile (make-temp-name "/tmp/save"))
	  (shape-get-description descfile)
	  (message "Submitting file %s" file2)
	  (call-process shape-submit-command nil t nil "-f" "-q" "-t"
			descfile file)
	  (delete-file descfile)
	  (revert-buffer))
      (message "Submitting file %s" file2)
      (call-process shape-submit-command nil t nil "-f" "-q" file)
      (revert-buffer)
      )))))

(defun shape-retrv()
  "retrieves a version via the rtrv command."
  (interactive)
  (save-excursion
    (let ((buffer-read-only nil))
	  (setq file (shape-get-filename t))
	  (setq file2 (substring file 0 (string-match "\\\[" file)))
      (if (not (file-AFS-p file))
	  (message "This file is not saved file")
	(if (file-exists-p file2)
	    (progn
	      (if (y-or-n-p (concat "Writable busy version of "
				    file2
				    " exists! Overwrite it?"))
		  (progn
		    (call-process shape-retrv-command nil nil "-f" "-q" file)
		    (shape-redisplay file2)))))))))

(defun shape-compare()
  "compares two versions with diff and puts output into a view buffer."
  (interactive)
  (save-excursion
;    (local-set-key "^X^@" 'shape-compare)
    (setq shape-buffer1 nil)
    (setq shape-buffer2 nil)
    (if (eq shape-compare-file1 nil)
	(progn
	  (if (or (file-DIR-p) (file-folded-p (shape-get-filename t)))
	      (message "Cant't compare directories or folded files")
	    (defvar shape-compare-file1 nil)
	    (setq shape-compare-file1 (shape-get-filename t))
	    (message "Compare %s with ? \(goto file2 and hit C again\)"
		     shape-compare-file1)
	    (shape-flag-file-compare "<")))
      
      (setq shape-compare-file2 (shape-get-filename t))
      (if (or (file-DIR-p) (file-folded-p shape-compare-file2))
	  (message "Cant't compare directories or folded files")
	(shape-flag-file-compare ">")
	(if (file-AFS-p shape-compare-file1)
	    (progn
	      (setq shape-buffer1 (create-file-buffer shape-compare-file1))
	      (call-process shape-vcat-command nil shape-buffer1 nil
			    "-q" shape-compare-file1)
	      (setq shape-compare-file1 (concat "/tmp/" shape-compare-file1)))
	  nil)

	(if (file-AFS-p shape-compare-file2)
	    (progn
	      (setq shape-buffer2 (create-file-buffer shape-compare-file2))
	      (call-process shape-vcat-command nil shape-buffer2 nil
			    "-q" shape-compare-file2)
	      (setq shape-compare-file2 (concat "/tmp/" shape-compare-file2)))
	  nil)
	(setq diff-buffer (create-file-buffer "diff"))
	(if shape-buffer1
	    (progn
	      (save-excursion
		(set-buffer shape-buffer1)
		(write-file (concat "/tmp/" shape-compare-file1))))
	  nil)
	(if shape-buffer2
	    (progn
	      (save-excursion
		(set-buffer shape-buffer2)
		(write-file (concat "/tmp/" shape-compare-file2))))
	  nil)
	(message "Comparing %s with %s" shape-compare-file1
		 shape-compare-file2)
	(sit-for 3 t)
	(call-process "diff" nil diff-buffer nil shape-compare-file1
		      shape-compare-file2)
	(view-buffer diff-buffer)
	(if (file-AFS-p shape-compare-file1)
	    (progn
	      (delete-file shape-compare-file1)
	      (kill-buffer shape-buffer1))
	  nil)
	(if (file-AFS-p shape-compare-file2)
	    (progn
	      (delete-file shape-compare-file2)
	      (kill-buffer shape-buffer2))
	  nil)
	(kill-buffer diff-buffer)
	(setq shape-compare-file1 nil)
	(shape-unflag-file-compare)
	;(local-unset-key "^X^@")
    ))))


(defun shape-fold()
  "Compresses output; files with versions are displayed with <name>[*]."
  (interactive)
  (save-excursion
    (let ((buffer-read-only nil))
      (if (y-or-n-p "Fold whole directory? ")
	  (progn
	    (message "Folding directory ...")
	    (goto-char (point-min))
	    (while (search-forward "[" nil t)
	      (setq filename (shape-get-filename t t))
	      (kill-line 1)
	      (insert "*]")
	      (newline)
	      (setq filename2 (substring
				filename 0 (string-match "\\\[" filename)))
	      (setq filename2 (concat filename2 "\\\["))
	      (delete-matching-lines filename2))
	    (message "Done."))
	(setq filename (shape-get-filename t))
	(if (file-AFS-p filename)
	    (progn
	      (setq filename2 (substring
			       filename 0 (string-match "\\\[" filename)))
	      (goto-char (point-min))
	      (search-forward (concat filename2 "["))
	      (beginning-of-line)
	      (search-forward "[" nil t)
	      (kill-line 1)
	      (insert "*]")
	      (newline)
	      (delete-matching-lines filename2)
	      (sit-for 0)
	      (message "Done."))
	  (message "No version: %s" filename))))))

(defun shape-unfold()
  "Expands folded entries."
  (interactive)
  (save-excursion
    (let ((buffer-read-only nil))
      (if (y-or-n-p "Unfold whole directory? ")
	  (progn
	    (message "Unfolding directory ...")
	    (revert-buffer)
	    (message "Done."))
	(if (equal (substring (shape-get-filename t)
			      -3 (length (shape-get-filename t))) "[*]")
	    (progn
	      (setq filename (substring (shape-get-filename t) 0 -3))
	      (message "Unfolding %s ..." filename)
	      (beginning-of-line)
	      (kill-line 1)
	      (call-process shape-vl-command nil t nil shape-listing-switches
			    "-ss" "-sp" "-sP" "-sa" "-sf" filename)
	      (shape-update-buffer)
	      (sit-for 0)
	      (message "Done."))
	  (message "File not folded."))))))
	      

(defun file-AFS-p(name)
  "decides whether a file is an AFS file or not (']' as last char)."
  (if (string-match "]" name) t nil))

(defun file-DIR-p()
  (beginning-of-line)
  (looking-at "  d"))

(defun file-folded-p(file)
  (if (equal (substring file -3 (length file)) "[*]")
      t
    nil))

(defun shape-insert-new-version (file)
  "update buffer after save command."
  (interactive)
  (let ((buffer-read-only nil))
    (while (search-forward file nil t))
    (forward-line)
    (beginning-of-line)
    (call-process shape-vl-command nil t nil shape-listing-switches "-y"
		  (concat default-directory file))
    (forward-line -1)
    (insert "  ")))


(defun shape-get-description (descfile)
  "read the description for the save command."
  (save-excursion
  (find-file descfile)
  (switch-to-buffer descfile)
  (message "To stop type CNTL-C CNTL-C")
  (local-set-key "^C^C" 'shape-finish-edit)
  (recursive-edit)
  (write-file descfile)
  (kill-buffer (current-buffer))))

(defun shape-finish-edit ()
  (interactive)
  (throw 'exit nil))
   

(defun shape-delete-file (file)
  (if (file-AFS-p file)
      (call-process shape-vadm-command nil nil nil "-delete" file)
    (delete-file file)))

(defun shape-flag-file-compare(mark)
  (let ((buffer-read-only nil))
    (save-excursion
      (beginning-of-line)
      (delete-char 1)
      (insert mark)
      (sit-for 0))))

(defun shape-unflag-file-compare()
  (let ((buffer-read-only nil))
    (save-excursion
      (beginning-of-buffer)
      (re-search-forward "^[><]")
      (beginning-of-line)
      (delete-char 1)
      (insert " ")
      (re-search-forward "^[><]")
      (beginning-of-line)
      (delete-char 1)
      (insert " "))))
      
(defun shape-update-buffer()
  "Updates buffer after unfold."
  (interactive)
  (save-excursion
  (goto-char (point-min))
  (while (re-search-forward "^-" nil t)
    (beginning-of-line)
    (insert "  "))))

(defun shape-execute()
  "sets compile command to shape -k."
  (interactive)
  (save-excursion
    (setq filename (shape-get-filename t t))
    (setq shapefile nil)
    (setq promptstring nil)
    (setq basename (substring
		     filename 0 (string-match "\\\[" filename)))
    (if (or (equal basename "Shapefile")
	    (equal basename "shapefile")
	    (equal basename "Makefile")
	    (equal basename "makefile")
	    (equal filename "Shapefile")
	    (equal filename "shapefile")
	    (equal filename "Makefile")
	    (equal filename "makefile"))
	(setq shapefile filename)
      (setq shapefile nil))
    (if (file-folded-p filename)
	(setq shapefile nil)
      nil)
    (if (equal shapefile nil)
	(setq promptstring "shape -k ")
      (if (file-AFS-p filename)
	    (setq promptstring (concat "vcat " "\""
				       filename
				       "\"" " | shape -f - "))
	(setq promptstring (concat "shape -k -f " filename " "))))
    (setq input (read-string "shape: " promptstring))
    (if (equal input nil)
	(compile promptstring)
      (compile input))))

(defvar shape-wish-address "shape-wishes@coma.UUCP" "The mail address to report a wish.")
(defvar shape-bug-address "shape-bugs@coma.UUCP" "The mail address to report a bug.")
(defvar shape-bug-description "Description:\n\nRepeat-By:\n\nFix:\n\nShape Toolkit version:\n\n" 
  "Formular to report a bug")

(defun shape-mail-bugs () 
  (interactive)
  (mail nil shape-bug-address)
  (goto-char (point-min))
  (beginning-of-next-line)
  (insert "Index: <tool>/<source> <confid>\n")
  (goto-char (point-max))
  (insert shape-bug-description "\n")
  (mail-position-on-field "Subject")
  (message (substitute-command-keys "Type \\[mail-send] to send bug report.")))

(defun shape-mail-wishes ()
  (interactive)
  (mail nil shape-wish-address)
  (mail-position-on-field "Subject")
  (message (substitute-command-keys "Type \\[mail-send] to send wish report.")))


(defun shape-execute-vl()
  "executes vl reading parameters from the minibuffer."
  (interactive)
  (setq input (read-string "vl: "))
  (shell-command (concat shape-vl-command " " input)))
  
(defun shape-execute-save()
  "executes save reading parameters from the minibuffer."
  (interactive)
  (setq input (read-string "save: "))
  (shell-command (concat shape-save-command " " input)))

(defun shape-execute-submit()
  "executes submit reading parameters from the minibuffer."
  (interactive)
  (setq input (read-string "sbmt: "))
  (shell-command (concat shape-submit-command " " input)))

(defun shape-execute-retrv()
"executes retrv reading parameters from the minibuffer."
  (interactive)
  (setq input (read-string "retrv: "))
  (shell-command (concat shape-retrv-command " " input)))
  
(defun shape-execute-vadm()
"executes vadm reading parameters from the minibuffer."
  (interactive)
  (setq input (read-string "vadm: "))
  (shell-command (concat shape-vadm-command " " input)))
 
(defun shape-execute-vcat()
"executes vcat reading parameters from the minibuffer."
  (interactive)
  (setq input (read-string "vcat: "))
  (shell-command (concat shape-vcat-command " " input)))


(defun shape-execute-vlog()
"executes vlog reading parameters from the minibuffer."
  (interactive)
  (setq input (read-string "vlog: "))
  (shell-command (concat shape-vlog-command " " input)))

