/*
 * Copyright (C) 1989, 1990 W. Koch, A. Lampen, A. Mahler, W. Obst,
 *  and U. Pralle
 * 
 * This software is published on an as-is basis. There is ABSOLUTELY NO
 * WARRANTY for any part of this software to work correctly or as described
 * in the manuals. We do not accept any liability for any kind of damage
 * caused by use of this software, such as loss of data, time, money, or 
 * effort.
 * 
 * Permission is granted to use, copy, modify, or distribute any part of
 * this software as long as this is done without asking for charge, and
 * provided that this copyright notice is retained as part of the source
 * files. You may charge a distribution fee for the physical act of
 * transferring a copy, and you may at your option offer warranty
 * protection in exchange for a fee.
 * 
 * Direct questions to: Tech. Univ. Berlin
 * 		     Wilfried Koch
 * 		     Sekr. FR 5-6 
 * 		     Franklinstr. 28/29
 * 		     D-1000 Berlin 10, West Germany
 * 
 * 		     Tel: +49-30-314-22972
 * 		     E-mail: shape@coma.uucp or shape@db0tui62.bitnet
 */
/*
 *	Shape/AFS
 *
 *	afarchive.h - Constant-Definitions for AFS-archives
 *
 *	Author:	Andreas Lampen (andy@coma.UUCP
 *				andy@db0tui62.BITNET)
 *
 *	$Header: afarchive.h[1.2] Wed Feb 22 16:27:08 1989 andy@coma published $
 *
 */

/*=========================================================================
 * Internal Installation dependent constants
 *=========================================================================*/

#define AF_NEWREVS      8	/* add. space for new revisions in rev list */
#define AF_MAXLISTS   211       /* maximum number of read-in archives */
#define AF_LISTSEG      8       /* size of segments of revlists-list */

#define AF_BPOOLSIZE   64       /* max number of files in binary pool */
                                /* should be determined dynamically */
#define MAXDBMLEN    1024       /* maximum size of dbm-datastring */
#define DBMLOSS         8       /* size of management data in dbm entry */
                                /* ( 2 * sizeof (int)) */

/*=========================================================================
 * Identification strings in archive and binary pool files
 *=========================================================================*/

#define AF_LINESIZ      1024            /* size of input line */

#define AF_SEGSTRLEN    5
#define AF_ARHEADER 	"\02ARHD"	/* Header of Archive-file */
#define AF_DATAHEADER 	"\02DATA"	/* Header of Data-file */
#define AF_BPHEADER 	"\02BPHD"	/* Header of Binary-pool-file */
#define AF_UDASEG 	"\02USEG"	/* User defined attributes Segment */

#define AF_IDSTRLEN 	2
#define AF_NAMEID	"\02I" 	/* Identification path, name, type */ 
#define AF_OWNID	"\02O" 	/* Owner */
#define AF_PRDID	"\02P" 	/* predecessor of busy version */
#define AF_LOCKID	"\02L" 	/* Lock */
#define AF_REVID	"\02R" 	/* revision specific attributes */
				/* gen, rev, state, mode */
#define AF_AUTHORID	"\02A" 	/* Author */
#define AF_DATEID	"\02T" 	/* date of last mod., last access, */
				/* creation, save */
#define AF_REPRID	"\02M" 	/* kind of represetation, size of file, size */
				/* of delta and other delta stuff (tree) */
#define AF_NOTEID	"\02N" 	/* gen, rev, size, text of note */
#define AF_UDAID	"\02U" 	/* gen, rev, size, text of udattrs */
#define AF_DATAID	"\02D" 	/* gen, rev, size, repr, text of data */

#define AF_BPFILEID     "\02B"  /* filename prefix for BP files */
#define AF_AFSFILEID    "\02A"  /* filename prefix for AFS archives */

/******* binary pool specific stuff *********/
#define AF_BPSTDENTRY   "\01##" /* dbm entry name for special information */
#define AF_BPOOLNAME    "\02Bpool"

/*======================================================================
 * other constants
 *======================================================================*/

#define AF_ARCHMODE     0444
#define AF_ARCURVERS    1
#define AF_BPCURVERS    1

#define AF_NOSTRING     "\01$"

/**** extent and type of revision list ****/

#define AF_ATTRS	0001   /* Attributes are loaded */
#define AF_DATA		0002   /* Data are loaded */
#define AF_ALLSEGS	0003   /* Attributes and Data are loaded */
#define AF_COMPLETE     0004   /* List is complete (binary pools only) */

#define AF_BPOOL        0010   /* Binary pool */
#define AF_ARCHIVE      0020   /* Line of development */

#define AF_UXWRITE      0100   /* Unix write permission on AFS subdir */

/*=========================================================================
 * Useful macros
 *=========================================================================*/

/* test if string is empty */
#define NOTMT(str) ((str && str[0]) ? str : AF_NOSTRING)

