/*
 * $Header: rcsit.c,v 1.20 87/05/04 21:50:53 src Exp $
 *---------------------------------------------------------
 * $Source: /src/tools/rcs/src/rcsit/RCS/rcsit.c,v $
 * $Revision: 1.20 $
 * $Date: 87/05/04 21:50:53 $
 * $State: Exp $
 * $Author: src $
 * $Locker: src $
 *---------------------------------------------------------
 * Michael Cooper (mcooper@usc-oberon.arpa)
 * University of Southern California,
 * University Computing Services,
 * Los Angeles, California,   90089-0251
 * (213) 743-3469
 *---------------------------------------------------------
 *
 * $Log:	rcsit.c,v $
 * Revision 1.20  87/05/04  21:50:53  src
 * added -F - flag. This option causes an additional header field for
 * inclusion of compile-time options into object-files to be inserted
 * into the C-source. The man-entry is also updated.
 * 
 * Revision 1.19  86/06/12  20:29:43  src
 * The previous version was unable to look up templates in the 
 * directory $TEMPLATES, when -t - option is absent but the
 * environment-variable TEMPLATES is set. Fixed that. If no -t - option
 * was given, but $TEMPLATES is non NULL, then -t is assumed and
 * $TEMPLATES is used as path to template-files.
 * 
 * Revision 1.18  85/11/26  17:03:32  mcooper
 * Change message telling of what header was added.
 * 
 * Revision 1.17  85/11/26  16:40:55  mcooper
 * Changed the default -t option to FALSE.
 * Added specifying directory to look for .template.*
 * files in via -tdirectory.
 * 
 * Revision 1.16  85/11/11  21:35:34  mcooper
 * Added call to access() to see if the file
 * could be read.
 * 
 * Revision 1.15  85/11/11  21:22:33  mcooper
 * Changed comment char for fortran files
 * from "*" to "c".  This is what RCS uses.
 * 
 * Revision 1.14  85/11/11  20:08:43  mcooper
 * Added descriptions for fortran (.f) files.
 * 
 * Revision 1.13  85/11/11  19:52:17  mcooper
 * Modified default header templates to not bother specifying
 * the RCS file name of the file.  co(1) worries about it.
 * 
 * Revision 1.12  85/10/27  19:10:07  mcooper
 * Fixed bug that would not use template files if a file
 * type was forced with -c, -h, etc.
 * 
 * Revision 1.11  85/10/27  18:48:27  mcooper
 * Extended template file.  You can now have template
 * files describing all the types of files that
 * rcsit "knows" about.  The file $HOME/.template.*
 * (where ``*'' is a ``.'' suffix rcsit can guess at or
 * the type of file that is specified with an override)
 * is checked for existance.  If not present, the default
 * header (built into rcsit) will be used.
 * 
 * Revision 1.10  85/10/27  16:15:53  mcooper
 * Added printing of what rcsit is doing if tflag is
 * true.  Also added new headers.
 * 
 * Revision 1.9  85/10/27  14:47:39  mcooper
 * Added new template feature.  If the file
 * .template exists in the users HOME directory,
 * then that file is used as the header file instead
 * of the defaults for each type of file.  This can
 * be disabled with the -t option in case the file
 * is say a shell script.  With the template feature
 * turned off, the auto guessing is re-inabled.
 * Also, rcsit now removes its temporary files.
 * 
 * Revision 1.8  85/09/28  14:11:45  mcooper
 * Added feature: if the environment variable RCSDIR is
 * present, rcsit will attempt to make a symbolic
 * link to the directory when the -I flag is used.
 * This is done only when -I is specified AND the 
 * directory RCS is not present.  You may disable this
 * feature with the -d option.  Note also that if RCSDIR
 * is not in the environment and the above conditions
 * are true, that a normal directory called RCS will
 * be created.
 * 
 * Revision 1.7  85/09/19  15:59:53  mcooper
 * Kludge part 2 -- If you specify a ci -l of a
 * man file, then the header is messed up. 
 * Fix: After initializing the comment string,
 * unlink the file and then run co -l.
 * 
 * Revision 1.6  85/09/19  15:39:57  mcooper
 * Now knows about ``Manual'' type files.
 * 
 * Revision 1.5  85/09/19  14:23:24  mcooper
 * Added lineprint() function to print things out
 * nicely.  Fixed bug for Manual type files.  Due
 * to the fact that RCS does not not the suffixes of
 * manuals, it therefor does not know what kind of
 * comment string to use.  Thus, I kludge by running
 * a ``rcs -c`... ' file'' to tell rcs the comment
 * string.
 * 
 * Revision 1.4  85/09/19  13:28:22  mcooper
 * Fixed bug in auto_guess.  Would not continue through function
 * when file type was ``Makefile''.
 * 
 * Revision 1.3  85/09/19  13:19:50  mcooper
 * Added ``Shell Script'' file type.
 * 
 * Revision 1.2  85/09/19  10:08:36  mcooper
 * Added code to run RCS commands (rcs & ci) on files.
 * Fixed bug that limited number of command line files specified to
 * nine.  Several other minor fixes and improvements.
 * 
 * Revision 1.1  85/09/17  11:33:33  mcooper
 * Initial revision
 * 
 */

/*
 * rcsit -- 	Prepare files for RCS.  rcsit puts the correct headings
 *		at the top of files to prepare them for RCS headings
 *		and log tracking.
 *
 * Michael Cooper	(mcooper@usc-oberon.arpa)
 * University Computing Services, USC
 *
 * 9-16-85
 */

#include <sys/file.h>
#include <stdio.h>
#include <ctype.h>
#include <strings.h>

#ifdef NULL
#undef NULL
#endif
#define NULL		'\0'
#define LENGTH		512		/* length of line */
#define TRUE		1
#define FALSE		0

#ifdef DEBUG
 int debugon = TRUE;
#else
 int debugon = FALSE;
#endif

static char 	*progname;		/* program name */
static char 	*rcsdir;

/*
 * Messages to be printed for the user.
 */
static char	*msg_name;		
static char 	*m_stdc = "Standard C",
                *m_stdcflg = "Standard C with compile flags",
		*m_include = "C Include",
		*m_fortran = "Fortran",
		*m_pascal = "Pascal",
		*m_make	= "Makefile",
		*m_shell = "Shell Script",
		*m_manual = "Manual";

/*
 * The headers to put at the beginning of the file(s).
 * Notice that the words Header and Log do not appear here
 * because RCS will put in the keyword substitutions when rcsit.c
 * is co'ed.
 */
static char	*header;
#ifdef AFSIT
static char	*h_stdc = 
	"#ifndef lint\nstatic char *AFSid = \"$%s$\";\n#endif\n/*\n * $%s$\n */\n\n";
static char	*h_stdcflg = 
	"#ifndef lint\nstatic char *AFSid = \"$%s$\";\n#ifdef CFFLGS\n\
static char *ConfFlg = CFFLGS;\n\t/* should be defined from within\
 Makefile */\n#endif\n#endif\n/*\n * $%s$\n */\n\n";
#else
static char	*h_stdc = 
	"#ifndef lint\nstatic char *RCSid = \"$%s$\";\n#endif\n/*\n * $%s$\n */\n\n";
static char	*h_stdcflg = 
	"#ifndef lint\nstatic char *RCSid = \"$%s$\";\n#ifdef CFFLGS\n\
static char *ConfFlg = CFFLGS;\n\t/* should be defined from within\
 Makefile */\n#endif\n#endif\n/*\n * $%s$\n */\n\n";
#endif
static char	*h_include = 
	"/*\n * $%s$\n *\n * $%s$\n */\n\n";
static char	*h_make =
	"#\n# $%s$\n#\n# $%s$\n#\n";
static char 	*h_manual =
	"...\n... $%s$\n... \n... $%s$\n...\n";
static char 	*h_fortran =
	"c\nc $%s$\nc\nc $%s$\nc\n";

/*
 * Template file names
 */
static char	*template_c 	= ".template.c";	/* .c template */
static char 	*template_h 	= ".template.h";	/* .h template */
static char 	*template_f 	= ".template.f";	/* .f template */
static char 	*template_p 	= ".template.p";	/* .p template */
static char 	*template_man 	= ".template.man";	/* man template */
static char	*template_make	= ".template.make";	/* make template */
static char	*template_sh	= ".template.sh";	/* sh script template */
static char	*tpath;					/* path to template */
static char	tfile[BUFSIZ];				/* template file */
static char	tbuf[BUFSIZ];				/* current tfile */

/*
 * Command line flags
 */
int	Iflag	= FALSE;			/* run ci(1) */
int	rcsflag = FALSE;			/* run rcs(1) */
int	aflag	= TRUE;				/* do auto guess */
int	dflag	= TRUE;				/* creat RCS dir. */
int	qflag	= FALSE;			/* be quiet! */
int 	cflag	= FALSE;			/* std c file */
int	fflag	= FALSE;			/* fortran file */
int     Fflag   = FALSE;                        /* insert C-Flags header */
int	pflag	= FALSE;			/* pascal file */
int	hflag	= FALSE;			/* include file */
int	sflag	= FALSE;			/* shell script */
int 	mflag	= FALSE;			/* Makefile */
int	Mflag	= FALSE;			/* manual */
int	tflag	= FALSE;			/* template flag */

main(argc, argv)
int	argc;
char 	*argv[];
{
	int x;
	char	tmp[LENGTH];
	char	*file;
	char	*flags;
	char 	tmpfile[32];
	char    *cp;
	char 	*mktemp();
	char	*gettmp();
	char	*getenv();
	FILE 	*fd, 
		*fdtmp,
		*fopen();

	progname = (cp = rindex (argv[0], '/')) ? ++cp : argv[0];
	sprintf (tmpfile, "/tmp/%sXXXXXX", progname);
	for (x = 1; x < argc; x++) {
		if (argv[x][0] != '-')
			break;
		switch (argv[x][1]) {
			case 'a':
				aflag = FALSE;
				break;
			case 'q':
				qflag = TRUE;
				break;
			case 'd':
				dflag = FALSE;
				break;
			case 'f':
				fflag = TRUE;
				break;
                        case 'F':
				Fflag = TRUE;
				cflag = TRUE; /* Only for C-files yet */
				break;
			case 'h':
				hflag = TRUE;
				break;
			case 's':
				sflag = TRUE;	
				break;
			case 'm':
				mflag = TRUE;
				break;
			case 'M':
				Mflag = TRUE;
				break;
			case 'i':
			case 'I':
				Iflag = TRUE;
				flags = &argv[x][2];
				break;
			case 'r':
			case 'R':
				rcsflag = TRUE;
				flags = &argv[x][2];
				break;
			case 't':
				tflag = TRUE;
				tpath = &argv[x][2];
				break;
			case 'c':
				cflag = TRUE;
				break;
			default:
				fatal("Unknown flag %s.",argv);
		}
	}
	argc -= (x - 1);
	argv += (x - 1);

	if((hflag && (mflag || Mflag || cflag)) ||
		(mflag && (hflag || cflag || Mflag)) ||
		(Mflag && (cflag || hflag || mflag)) ||
		(cflag && (hflag || Mflag || mflag))) {
			fatal("Only ONE of -c,-f,-m,-M,-h,-s may be specified.");
	}
	if(Iflag && rcsflag) {
			fatal("Only ONE of ``-i'' and ``-r'' may be specified.");
	}

	if(cflag || hflag || mflag || Mflag || fflag || sflag)
		aflag = FALSE;

	if((rcsdir = getenv("RCSDIR")) == NULL)
		rcsdir = "RCS";
	if(Iflag && dflag)
		checkdir();	/* Make RCS directory for ci */
	if((!tflag) && ((tpath = getenv ("TEMPLATE")) != NULL))
	        tflag = TRUE;   /* fixed by axel@coma.uucp */
	if((*tpath == NULL) && ((tpath = getenv("TEMPLATE")) == NULL))
		if((tpath = getenv("HOME")) == NULL)
			fatal("Cannot find environment variable HOME or TEMPLATE");

	/*
	 * make tmp file once.
	 */
	mktemp(tmpfile);

	while (--argc) {	/* Main loop */
		file = *++argv;
		debug(sprintf(tmp, "...file (*++argv) = %s...", file));

		if(access(file, 4) != 0)
			fatal("Cannot access %s.  No read permission OR file does not exist.",
				file);
		if((fdtmp = fopen(tmpfile, "w")) == NULL) {
			fatal("Cannot open tmpfile (%s).", tmpfile);
		}

		if(aflag)
			auto_guess(file); /* try and guess file type */
		else
			set_flags();	  /* check and set flags */

		if(tflag) {
			/*
			 * first get names of templates, then create
			 * path name to it.
			 */
			get_temp();
			sprintf(tfile, "%s/%s", tpath, tbuf);
		}
		if(access(tfile, 0) == 0 && tflag) {
			if(!qflag || debugon)
				printf("Adding %s header file to %s...",
					msg_name, file);
			copy(tfile, tmpfile, "w");
			copy(file, tmpfile, "a");
		} else {
			if(!qflag || debugon)
				printf(
				"Adding default header (%s format) to %s...",
					msg_name, file);
			/*
			 * put the Keywords into header string
			 */
#ifdef AFSIT
			sprintf(tmp, header, "__Header", "__Log");
#else
			sprintf(tmp, header, "Header", "Log");
#endif
			fputs(tmp, fdtmp);
			/*
			 * fclose'em, just in case.
			 */
			fclose(fdtmp);
			copy(file, tmpfile, "a");
		}
		unlink(file);
		copy(tmpfile, file, "w");
		unlink(tmpfile);

		if(!qflag || debugon)
			printf("done.\n");
		if (Fflag)
		  puts ("-> Don't forget to define CFFLGS on compiler\
 commandline <-\n");

		if(Iflag){
			rcs("ci", file, flags);
			if(Mflag){	/* kludge to tell rcs about manuals */
				rcs("rcs", file, "c'... '");
				/*
				 * kludge part 2 - if the user tried a ci
				 * with a -l option, then the header is
				 * messed up in the currently checked out
				 * man file.  So we have to co the file to 
				 * clean up the header.  Plus we use the
				 * -l option of co to insure file locking.
				 */
				if(checkfor("l", flags)){
					unlink(file);
					rcs("co", file, "l");
				}
			}
		}
		if(rcsflag)
			rcs("rcs", file, flags);
	}
}

/*
 * debug - print (useless) debugging info.
 */
 
debug(msg)
char *msg;
{
#ifdef DEBUG
	fprintf(stderr, msg);
	putchar ('\n');
#endif
}

/*
 * auto_guess - try and be intelligent and guess type of file
 *		by looking at the suffix or the whole name
 *		in the case of a makefile.
 */

auto_guess(file)
char	*file;
{
	char *suffix;
	char *rindex();

	suffix = rindex(file, '.')+1;
	if((strcmp(file, "makefile") == 0) || (strcmp(file, "Makefile") == 0) ||
	    (strcmp(suffix, "mk") == 0)) {	/* sys V std suffix */
		mflag = TRUE;
		sflag = FALSE;
		cflag = FALSE;
		hflag = FALSE;
		Mflag = FALSE;
		fflag = FALSE;
	}
	if((strcmp(suffix, "sh") == 0) || (strcmp(suffix, "csh") == 0)) {
		sflag = TRUE;
		cflag = FALSE;
		hflag = FALSE;
		mflag = FALSE;
		Mflag = FALSE;
		fflag = FALSE;
	}
	if(strcmp(suffix, "c") == 0){
		cflag = TRUE;
		hflag = FALSE;
		mflag = FALSE;
		Mflag = FALSE;
		sflag = FALSE;
		fflag = FALSE;
	}
	if(strcmp(suffix, "h") == 0){
		hflag = TRUE;
		cflag = FALSE;
		mflag = FALSE;
		Mflag = FALSE;
		sflag = FALSE;
		fflag = FALSE;
	}
	if(strcmp(suffix, "f") == 0){
		fflag = TRUE;
		hflag = FALSE;
		cflag = FALSE;
		mflag = FALSE;
		Mflag = FALSE;
		sflag = FALSE;
	}
	if(isdigit(*suffix) != 0) {
		Mflag = TRUE;
		hflag = FALSE;
		cflag = FALSE;
		mflag = FALSE;
		sflag = FALSE;
		fflag = FALSE;
	}
	set_flags();
	if(!qflag || debugon)
		printf("Hmm.  This file looks like a %s file.\n", msg_name);
}

/*
 * set_flags - set & check flags
 */
 
set_flags()
{
	if(cflag || hflag || mflag || Mflag || sflag || fflag) {
		if(cflag) {
			msg_name = m_stdc;
			header = h_stdc;
		}
		if (Fflag) { /* This is more than just cflag - axel@coma */
		        msg_name = m_stdcflg;
			header = h_stdcflg;
		}
		if(hflag) {
			msg_name = m_include;
			header = h_include;
		}
		if(mflag) {
			msg_name = m_make;
			header = h_make;
		}
		if(Mflag) {
			msg_name = m_manual;
			header = h_manual;
		}
		if(sflag) {
			msg_name = m_shell;
			header = h_make;
		}
		if(fflag) {
			msg_name = m_fortran;
			header = h_fortran;
		}
	} else {
		cflag = TRUE;
		set_flags();
	}
}

/*
 * copy from -> to
 */

copy(from, to, mode)
char *from;
char *to;
char *mode;
{
	FILE *fdfrom, *fdto, *fopen();
	char tmp[LENGTH];
	char s[LENGTH];

	if((fdfrom = fopen(from, "r")) == NULL) {
		fatal("Cannot open %s for reading.",from);
	}
	if((fdto = fopen(to, mode)) == NULL) {
		fatal("Cannot open %s for \"%s\".",to,mode);
	}
	while(fgets(s, sizeof(s), fdfrom) != NULL)
		fputs(s, fdto);
	fclose(fdfrom);
	fclose(fdto);
}

/*
 * Run RCS's rcsprog on file with flags.
 */

rcs(rcsprog, file, flags)
char *rcsprog;
char *file;
char *flags;
{
	char buf[LENGTH];
	char tmp[LENGTH];

	if(!checkfor("q", flags) && qflag)
		flags = "q";
	if(strcmp(flags, NULL) == 0)
		sprintf(buf, "%s %s", rcsprog, file);
	else
		sprintf(buf, "%s -%s %s", rcsprog, flags, file);
	debug(sprintf(tmp,"Running ``%s''...\n", buf));
	if(!qflag)
		lineprint(sprintf(tmp, "Start of ``%s''", buf));
	system(buf);
	if(!qflag)
		lineprint(sprintf(tmp, "End of ``%s''", buf));
}

/*
 * checkdir - make RCS directory if not present.
 */

checkdir()
{
	if(access("RCS", 0) != 0){
		if(!qflag || debugon)
			printf("Cannot find \"RCS\" directory.  Creating...\n");
		if(strcmp(rcsdir, "RCS") != 0) { 
			if(symlink(rcsdir, "RCS") != 0)
				fatal("Symbolic link of %s to RCS failed.", 
					rcsdir);
		} else {
			if(mkdir(rcsdir, 0755) != 0)
				fatal("Cannot create \"%s\" directory.", 
					rcsdir);
		}
	}
}

/*
 * checkfor(x, str) -- check for x in str.  Return 1 (TRUE) if exists.
 *			Otherwise 0 (FALSE).
 */

checkfor(x, str)
char 	*x;
char 	*str;
{
	while(*str) {
		if(strcmp(str, x) == 0)
			return(TRUE);
		*str++;
	}
	return(FALSE);
}

/*
 * lineprint - print msg in a nice line
 */

lineprint(msg)
char *msg;
{
	int len, left, right, x;

	len = strlen(msg);
	right = (75-len)/2;
	left = right;
	for(x = 0; x < right; ++x)
		putchar('-');
	printf("[ %s ]", msg);
	for(x = 0; x < left; ++x)
		putchar('-');
	putchar('\n');
}

/*
 * fatal - print error and then exit(1).
 */
fatal(format, str)
char *format;
{
	static char namefmt[100];

	sprintf(namefmt, "%s: %s\n", progname, format);
	_doprnt(namefmt, &str, stderr);
	exit(1);
}

/*
 * zap str with NULL's
 */

zap(str)
char str[];
{
	int i, x;

	i = strlen(str);
	for(x = 0; x <= i; )
		str[x++] = NULL;
}

/*
 * get template names
 */

get_temp()
{
	zap(tbuf);
	if(mflag)
		strcpy(tbuf, template_make);
	if(Mflag)
		strcpy(tbuf, template_man);
	if(hflag)
		strcpy(tbuf, template_h);
	if(cflag)
		strcpy(tbuf, template_c);
	if(sflag)
		strcpy(tbuf, template_sh);
	if(fflag)
		strcpy(tbuf, template_f);
}
