.\"
.\" Copyright (C) 1989, 1990 W. Koch, A. Lampen, A. Mahler, W. Obst,
.\"  and U. Pralle
.\" 
.\" This software is published on an as-is basis. There is ABSOLUTELY NO
.\" WARRANTY for any part of this software to work correctly or as described
.\" in the manuals. We do not accept any liability for any kind of damage
.\" caused by use of this software, such as loss of data, time, money, or 
.\" effort.
.\" 
.\" Permission is granted to use, copy, modify, or distribute any part of
.\" this software as long as this is done without asking for charge, and
.\" provided that this copyright notice is retained as part of the source
.\" files. You may charge a distribution fee for the physical act of
.\" transferring a copy, and you may at your option offer warranty
.\" protection in exchange for a fee.
.\" 
.\" Direct questions to: Tech. Univ. Berlin
.\" 		     Wilfried Koch
.\" 		     Sekr. FR 5-6 
.\" 		     Franklinstr. 28/29
.\" 		     D-1000 Berlin 10, West Germany
.\" 
.\" 		     Tel: +49-30-314-22972
.\" 		     E-mail: shape@coma.uucp or shape@db0tui62.bitnet
.\" 
.NH 1
A quick guide through shape for make users
.PP
This chapter is intended to give users of \fImake\fR a quick guide
how to change their development environment when switching
from the usage of \fImake\fR to \fIshape\fR.
.PP
The first step is quite simple. As the shape program is upward
compatible to make, you can just type \fCshape\fR instead of \fCmake\fR
and shape will fulfill your (make\-)wishes.
Ooops, we nearly forgot to mention, that you have to create a
subdirectory named \fCAFS\fR first.
You will later see why.
.PP
By using shape instead of make on your old Makefile you can
already take advantage of one important improvement of shape
with respect to make.
Shape is sensible for changes in the compile environment.
If you change for example your compiler flags or preprocessor
definitions in the Makefile, shape will recompile all 
files that are affected by this change (make does not).
For example, if \fCfoo.o\fR was generated by \fCcc -c -DBSD43\fR
and you change your Makefile in order to produce a version
suitable for System V so that \fCfoo.c\fR should be
produced by \fCcc -c -DSYSV\fR, shape recognizes this change,
in contrary to make.
.PP
You might ask yourself now "how can shape do this ?".
Before we can give you an answer to this question, you got to learn
to know a few things about the architecture of the shape toolkit.
The whole shape toolkit is implemented on top af an object base (AFS),
that is able to store software objects (regular UNIX files for instance)
as a bunch
of content data together with an arbitrary number of attributes.
These things are called \fIattributed software objects\fR (ASO).
Attributes are strings of the general form \fCname=value\fR.
AFS supports a set of standard attributes (examples are:
name, type, version number, owner, author) an unlimited number
of \fIuser defined attributes\fR.
Each application may introduce its own user defined attributes.
Additionally, AFS has a built-in version control system that
supports the storage and identification of different versions
of software objects.
For further details on the object base see afintro (3).
.PP
Shape uses the mechanism of application defined attributes for recording
the compile
environment (the name of the transformation tool and all options and
arguments) of each derived file (you may know be more familiar with
the term \fI.o file\fR).
So, shape bases its decision whether to recompile a system component
from its source not only on comparing the date of last modification
(like make does) but also on the comparison of the compile environments.
.PP
Due to that mechanism, you have to bear one inconvenience on your journey
from make to shape. Your first \fCshape\fR call causes all your files
to be recompiled. This is necessary to initialize the object base.
.PP
But you get indemnified for that with the next step.
Suppose, you want to produce different versions of your system
by setting different compiler options.
Typically, you have a local test system that is compiled and linked
with the \fC-g\fR option (for symbolic debugging) and you publish
(install) product versions of your system compiled with \fC-O\fR
(optimizing) and the symbol table stripped off.
.PP
So you might think, shape recompiles the whole stuff with \fC-O\fR
for publication and afterwards recompiles it again with \fC-g\fR
for continuing the work.
.br
Not true !
.br
All derived files generated during a system building operation driven
by shape are stored in a cache named \fIderived object pool\fR
(or \fIbinary pool\fR).
The derived object pool is able to store multiple versions of
equally named derived files and to distinguish between them by the
user defined attributes they have.
So when you want to continue your work after an installation,
you old \fC.o\fR files are still there and
shape restores them from the derived object pool instead
of regenerating them from their source files.
.PP
To prevent your disk from being flooded with derived files, each binary
pool has a limited size (a maximum number of files stored in it).
Binary pools are administered in a cache fashion.
The oldest (access date) derived files get removed if no additional
files can be inserted due to the size limit.
You can define the size of your private binary pool by setting
the environment variable \fCAFSBPSIZ\fR to an appropriate value.
A reasonable value is (\fI number of your source files * 2 \fR).
The default value is 64.
.PP
Remember, all your work with shape up to now did not require
any modification of your Makefile.
If you don't like shape up to now \- no problem, \fCrm -rf AFS\fR
removes all shape specific stuff and you can go on with make.
But if you like it, it is the right time to rename your Makefile to
\fCShapefile\fR and begin to implant shape specific things.
.NH 2
Version Control
.PP
Beside the make functionality (system building), a main feature
of the shape toolkit is the integration of a version control
system in your system building environment.
Shape does not only operate on regular UNIX files, but on
multiple versions of files.
For this reason, we constructed a version control system which is
integral part of the shape toolkit.
.PP
This version control system has a lot of similarities to existing
version control systems like RCS and SCCS, but also some
improvements with respect to these systems.
You can read more about the version control in the manual pages.
We just want to mention one important feature of the version control system
here, that is important to know while reading this chapter.
.PP
Shape's version control system has a built in version status model.
The essential thing for you to know about this is, that there are
"busy" and "saved" versions. Saved versions have
a version number that consists of a generation number and
a revision number (eg. 4.3). Saved versions cannot be modified;
once saved, they are "read only". Busy versions are the place
where development takes place. They are alterable and have no
version number yet. Each regular UNIX file is a busy version
in this sense. Saved versions come into being by making a copy
of a busy version, tagging a version number to it and entering it
(as delta) in the appropriate archive file in the AFS subdirectory.
The status model is more complex, but for further reading of this
chapter it is enough when you know the states "busy", "saved"
and "published".
Published versions play a role when you working in a multi
programmer team. You can use the publishing mechanism of the
shape toolkit then to make specific versions of the system
component you work on accessible for your co workers.
.PP
If you have worked with another version control system before
it is advisable to transform your old version histories to
shape histories.
If you have used RCS before, you can have this be done
automatically by running the program \fIrcs2afs\fR.
.NH 2
Version Selection Rules
.PP
Ok, long enough now, let's dive into shape's selection mechanism.
Make's task is it, to produce names \- so called \fItargets\fR.
It does this by traversing a system model consisting of
\fItarget \- dependent\fR lines to determine which components
go into a system and producing targets from their dependents.
Targets and dependents may correspond to equally named files in the
filesystem to which transformation tools are applied.
When doing make's job, due to the introduction of multiple versions
of files, shape has the additional task to select one
specific version of each file involved.
If does this by evaluating \fIversion selection rules\fR 
given by the
programmer in a special part of the Shapefile, called the \fIrule-section\fR.
A version selection is driven by evaluating a selection rule
with the name given as parameter.
Remember that each version (software object) has a number of
attributes attached to it. These attributes are the ground,
selection rules operate on.
So let's look at the following rule section:
.nf
.sp
\fC#% RULE-SECTION
defaultrule:
	*, attr (state, busy).
#% END-RULE-SECTION\fR
.fi
.sp
.PP
A rule section always begins with a comment symbol (\fC#\fR) followed
by a percent sign and the string \fCRULE-SECTION\fR.
It ends accordingly (see example).
Each rule begins with the rule name followed by a colon (in the above case
\fCdefaultrule:\fR).
The rest ot the selection rule is a named sequence of \fIalternatives\fR,
separated by semicolons, which form a logical OR expression. Each alternative
consists of a sequence of \fIpredicates\fR, separated by commas, which form
a logical AND expression. A selection rule succeeds if one of its
alternatives succeeds (i.e. leads to the unique identification of some
document instance).
In alternatives, the first predicate (e.g. \fC*\fR or \fC*.c\fR) is usually a
pattern against which the name of a document that has to be retrieved
is matched.  The patterns used by shape have the same structure as those
used by \fCed\fR.
.PP
The rule above consists of only one alternative, so do not worry if you
do not find the semicolon in it.
Beside that, you never need to write a rule like this, because it is
the default rule \- this rule is active if no other selection rule
is given.
It makes exactly what make does: for each name (\fC*\fR) it selects the
busy version, which is an equally named regular UNIX file.
This rule makes sense if you have no saved versions.
.PP
To go on with shape it is necessary that you begin to work with
the version control system.
You should save a releasable state of
your work by applying the command \fCsave\fR to all the files
belonging to your system.
This can probably be done by inserting an entry in your Shapefile
that looks like
.sp
.nf
\fCsave:
	save $(COMPONENTS)\fR
.fi
.sp
and calling \fCshape save\fR.
.PP
If you have done this, a rule like
.nf
.sp
\fC#% RULE-SECTION
releaserule:
	*, attr (state, saved), attrmax (version).
#% END-RULE-SECTION\fR
.fi
.sp
makes sense for building releases.
This rule selects for each component (\fC*\fR) of your system the
latest (\fCattrmax (version)\fR) saved (\fCattr (state, saved)\fR)
version.
If for example there is \fIno\fR saved version of your component
\fCfoo.c\fR shape complains about that with the message:
\fCshape - selection rule for foo.c failed\fR
.PP
You can activate a selection rule by inserting its name in the 
target - dependents line of your system at the place of the first dependent.
For example
.nf
.sp
\fCall: releaserule program1 program2\fR
or
\fCprogram: releaserule $(COMPONENTS)\fR
.sp
.fi
can be meaningful rule activations.
Once set active, a selection rule is inherited to all subtargets.
But be careful, the active selection rule can also be redefined
with each single target \-
a mechanism that takes a little while to get used to.
So if you are wondering about shape selecting other versions than 
you expect, check your Shapefile for constructions like:
.nf
.sp
\fCall: releaserule program1 program2\fR

\fCprogram1: defaultrule $(COMPONENTS1)\fR

\fCprogram2: $(COMPONENTS2)\fR
.fi
.sp
In this case \fCshape all\fR builds
\fCprogram1\fR with \fCdefaultrule\fR (explicitly set) and program2
with \fCreleaserule\fR (inherited).
.PP
We cannot deal with all issues of the selection rule mechanism here;
you find more details on this in the chapter (.\" to be inserted),
but just to illustrate the power of expression sleeping in the
selection rule mechanism,
an example for a very complex selection rule.
It does not make very much sense but it shall give you an idea
of what you can express with the selection rule mechanism.
.nf
.sp
\fC#% RULE-SECTION
\&...
complexrule:
	# to be inserted
#% END-RULE-SECTION\fR
.fi
.sp
.NH 2
Variants
.PP
After having played around with selection rules
you may turn your attention to another very powerful mechanism of shape.
Remember the example from the beginning of this chapter that dealt with
different \fIvariants\fR (<- thats the buzzword) of your system for
different machines.
Let's introduce another section in the Shapefile.
The first and the last line of this section look very much like
these of the rule section with the word \fCVARIANT\fR instead of \fCRULE\fR.
.nf
.sp
\fC#% VARIANT-SECTION
vclass system ::= (bsd, sysV, sunos)

bsd:
	vpath = bsd
	vflags = -DBSD
sysV:
	vpath = sysV
	vflags = -DSYSV
sunos:
	vpath = sunos
	vflags = -DSUNOS
#% END-VARIANT-SECTION\fR
.fi
.sp
.PP
This helps you to deal on one hand with variants that are generated from
\fIone\fR source file by setting different precompiler switched and
on the other hand with variants that are stored in equally named 
files in different directories.
With a variant activated, shape passes variant dependent options
(vflags: eg. \fC-DBSD\fR) to each transformation tool (eg. \fCcc\fR)
it triggers and it extends its search path for files by the
directories named in vpath (eg. the subdirectory \fCbsd\fR).
.PP
The activation mechanism for variants is quite similar to the rule
activation.
Just after the rule name you can optionally insert the variant name
preceded by a plus sign (\fC+\fR).
The following example of a target \- dependent line is suitable for
calling shape with \fCshape SYSTEM=sunos all\fR. This builds the
bsd variant of your system.
.nf
.sp
\fCall: releaserule +$(SYSTEM) program1 program2\fR
.fi
.sp
It is advisable to have a line like
.nf
.sp
\fCSYSTEM=bsd\fR
.fi
.sp
somewhere in your Shapefile in order to define a default for your
SYSTEM macro. You can then omit the definition for SYSTEM on the command
line when you want to build the default variant.
One little hint in respect to activation of variants. When shape
comes up with an error message like
\fCshape - don't know how to shape +blabla\fR, this points to a
spelling mismatch between variant naming and variant activation.
.PP
In a variant definition, not only vpath and vflags can be defined.
You can also redefine any macro.
Another typical application is
.nf
.sp
\fC#% VARIANT-SECTION
vclass quality ::= (test, test_profiling, final)

test:
	CFLAGS = -g
	LDFLAGS = -g
test_profiling:
	CFLAGS = -pg -g
	LDFLAGS = -pg -g
final:
	CFLAGS = -O
	LDFLAGS = -s
#% END-VARIANT-SECTION\fR
.fi
.sp
that helps to manage the other case of system variants that occured in an
example earlier in tis paper.
With the targets
.nf
.sp
\fCall: +$(SYSTEM) +test_profiling program1 program2

install: +$(SYSTEM) +final program1 program2\fR
.fi
.sp
you can build test systems by calling \fCshape all\fR and release systems
by calling \fCshape install\fR.
.PP
A few words about vclasses (to be inserted).
The name of the vclass is not explicitely used up to now, but there should
be a equally named Macro (in capital letters) that holds the actual
variant setting.
For this macro (as said earlier), a default should be defined.
.NH 2
Building up a target raster.
.PP
Rule and variant activations should only be inserted in top level targets.
See the sample Shapefiles for examples.
.NH 2
Macros
.PP
Macro naming conventions -> shape manual.
.NH 2
Transformation Rules
.PP
Like in make, you can write implicit transformation rules.
Shape understands make syntax, but we recommend to use
the extended syntax described in the following paragraph
.PP
Transformation rules consist of a \fItransformation specification\fR,
and a \fItransformation script\fR. Transformation specifications
describe prerequisites and resulting objects of a transformation. The
transformation script is a template for a shell script that
will be passed to a shell process upon activation of a transformation.
The syntax for these rules is an extension of Make's default rule
specification formalism. \*(sh's rule by which the transformation
of a C source object into a '\fC\&.\&o\fR' (speak: dot-o) derived object
is defined looks like:
.sp
.nf
\fC\s-1%.o: %.c : +(CC) +(CFLAGS)
	$(CC) -c $(CFLAGS) %.c\s+1\fR
.fi
.LP
.PP
A request to produce a derived object \- let's say \fCxyzzy.o\fR \- would,
by evaluating the transformation rule, result in the compilation
of \fCxyzzy.c\fR with the compiler defined by 
the macro \fCCC\fR, with compile switches defined by \fCCFLAGS\fR.
The percent sign is consistently substituted by the name \(lqxyzzy\(rq
throughout the transformation rule. The notation \fI$(NAME)\fR substitutes
the value of the macro \fINAME\fR, while \fI+(NAME)\fR substitutes the
entire macro definition (i.e. \fINAME=value\fR).

