%{
/*	This file is part of the software similarity tester SIM.
	Written by Dick Grune, Vrije Universiteit, Amsterdam.
*/

/*
	C language front end for the similarity tester.
*/

/* Language-dependent Code */
#include	"idf.h"

static struct idf ppcmd[] =	{
	"define",	'D',
	"else",		'E',
	"endif",	'Z',
	"if",		'F',
	"ifdef",	'Y',
	"ifndef",	'N',
	"include",	'I',
	"line",		'L',
	"undef",	'U'
};

static struct idf reserved[] =	{
	"auto",		'a',
	"break",	'b',
	"case",		'k',
	"char",		'c',
	"continue",	'z',
	"default",	'_',
	"do",		'd',
	"double",	'm',
	"else",		'e',
	"enum",		'n',
	"extern",	'q',
	"float",	'y',
	"for",		'f',
	"goto",		'g',
	"if",		'i',
	"int",		'j',
	"long",		'l',
	"register",	'\0',		/* ignore */
	"return",	'r',
	"short",	'h',
	"sizeof",	'o',
	"static",	'p',
	"struct",	's',
	"switch",	'x',
	"typedef",	't',
	"union",	'u',
	"unsigned",	'v',
	"void",		'\0',		/* ignore */
	"while",	'w'
};

static int
is_trailer(ch)	{
	return ch == ')' || ch == '}' || ch == ';';
}

int
MayBeStartOfRun(ch)	{
	return !is_trailer(ch);
}

unsigned int
CheckRun(str, size) char *str; unsigned int size;	{
	/*	Checks the run starting at str with length size for
		acceptability in the language.  Cuts from the end if
		necessary and returns the accepted length (which may
		be zero).
	*/
	extern char options[];
	
	if (options['f'])	{	/* function-like forms only */
		unsigned int pos;
		unsigned int lb_pos = 0;/* latest balancing position */
		int braces = 0;
		int parens = 0;
		int brackets = 0;
		
		for (pos = 0; pos < size; pos++)	{
			switch (str[pos])	{
			case '{': braces++; break;
			case '}': braces--; break;
			case '(': parens++; break;
			case ')': parens--; break;
			case '[': brackets++; break;
			case ']': brackets--; break;
			}
			if (	/* this was one closer too many */
				braces < 0 || parens < 0 || brackets < 0
			)	{
				break;
			}
			if (	/* it happens to balance here */
				braces == 0 && parens == 0 && brackets == 0
			)	{
				lb_pos = pos + 1;
			}
		}
		size = lb_pos;			/* cut to size */
	}
	else	{
		while (	/* there is trailing garbage */
			size != 0 &&
			(str[size - 1] == '@' || is_trailer(str[size - 1]))
		)	{
			/* remove it */
			size--;
		}
	}
	return size;
}

/* Language-INdependent Code */
#include	"stream.h"

yystart()	{
	BEGIN INITIAL;
}

static int
yywrap()	{
	return 1;
}

%}

%Start	Comment

AnyQuoted	(\\.)
StrChar		([^"\n\\]|{AnyQuoted})
ChrChar		([^'\\]|{AnyQuoted})
ComChar		([^*\n]|(\*[^/]))
Idf		([A-Za-z][A-Za-z0-9_]*)

%%

\"{StrChar}*\"	{	/* strings */
		cput('"');
		count();
	}

\'{ChrChar}\'	{	/* characters */
		cput('\'');
		count();
	}

"/*"	{	/*	We cannot have one pattern for a comment
			(although one can be written), since the matched
			string would overflow lex-internal buffers like
			yysbuf and yytext. So we have to break up the string
			into lines and keep track of where we are in a start
			condition <Comment>.
		*/
		BEGIN Comment;
		count();
	}

<Comment>{ComChar}*	{	/* comment up to \n or end-of-comment */
		count();
	}

<Comment>"*/"	{		/* end-of-comment */
		BEGIN INITIAL;
		count();
	}

#[ \t]*include.*	{	/* skip #include line */
		count();
	}

#[ \t]*{Idf}	{	/* a preprocessor line */
		char *n = yytext+1;
		int ch;

		/* skip layout in front of preprocessor identifier */
		while (*n == ' ' || *n == '\t')
			n++;
		ch = idf2char(n, ppcmd);
		if (ch < 0)	{
			cput('#');
		}
		else
			cput(ch);
		count();
	}

{Idf}	{
		int ch = idf2char(yytext, reserved);

		if (ch < 0)
			cput('@');
		else
		if (ch > 0)
			cput(ch);
		count();
	}

[ \t]	{	/* layout */
		count();
	}

\n	{	/* count newlines */
		count();
		c_eol();
	}

.	{	/* copy other text */
		cput(yytext[0]);
		count();
	}

%%
