/*
 * File: lmisc.c
 *  Contents: create, keywd, limit, llist
 */

#include "../h/rt.h"
#include "../h/keyword.h"
#include "../h/version.h"

/*
 * The following code is operating-system dependent. Include files that
 *  are system-dependent.
 */

#if PORT
   /* probably needs something */
#endif					/* PORT */

#if AMIGA
#include "time.h"
#endif					/* AMIGA */

#if ATARI_ST
struct tms {
   unsigned int tms_utime;
   int tm_year;
   int tm_mon;
   int tm_mday;
   int tm_hour;
   int tm_min;
   int tm_sec;
   };
#endif					/* ATARI_ST */

#if MACINTOSH
#if MPW
#include <types.h>
#include SysTime
#include <OSUtils.h>
#include <Events.h>
#endif					/* MPW */
#endif					/* MACINTOSH */

#if MSDOS
#include <time.h>
#if MICROSOFT
#include <sys/types.h>
#endif					/* MICROSOFT */
#endif					/* MSDOS */

#if UNIX
#include <sys/types.h>
#include <sys/times.h>
#include SysTime
#endif					/* UNIX */

#if VMS
#include <types.h>
#include <time.h>
struct tms {
    time_t    tms_utime;	/* user time */
    time_t    tms_stime;	/* system time */
    time_t    tms_cutime;	/* user time, children */
    time_t    tms_cstime;	/* system time, children */
};
#endif					/* VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

/*
 * create - return an entry block for a co-expression.
 */

OpBlock(create,1,"create",0)

create(entryp, cargp)
word *entryp;
register struct descrip *cargp;
   {
#ifndef NoCoexpr
   register struct b_coexpr *sblkp;
   register struct b_refresh *rblkp;
   register struct descrip *dp, *ndp, *dsp;
   register word *newsp;
   int na, nl, i;
   struct b_proc *cproc;
   extern struct b_coexpr *alccoexp();
   extern struct b_refresh *alcrefresh();

   /*
    * Get a new co-expression stack and initialize.
    */
   if ((sblkp = alccoexp()) == NULL) 
      RunErr(0, NULL);

   /*
    * Icon stack starts at word after co-expression stack block.  C stack
    *  starts at end of stack region on machines with down-growing C stacks
    *  and somewhere in the middle of the region.
    *
    * The C stack is aligned on a doubleword boundary.	For upgrowing
    *  stacks, the C stack starts in the middle of the stack portion
    *  of the static block.  For downgrowing stacks, the C stack starts
    *  at the end of the static block.
    */
   newsp = (word *)((char *)sblkp + sizeof(struct b_coexpr));
#ifdef UpStack
   sblkp->cstate[0] =
      ((word)((char *)sblkp + (stksize - sizeof(*sblkp))/2)
       &~(WordSize*2-1));
#else					/* UpStack */
   sblkp->cstate[0] =
	((word)((char *)sblkp + stksize - WordSize)&~(WordSize*2-1));
#endif					/* UpStack */
   sblkp->es_argp = (struct descrip *)newsp;
   /*
    * Calculate number of arguments and number of local variables.
    *  na is nargs + 1 to include Arg0.
    */
   na = pfp->pf_nargs + 1;
   cproc = (struct b_proc *)BlkLoc(argp[0]);
   nl = cproc->ndynam;

   /*
    * Get a refresh block for the new co-expression.
    */
   if (blkreq((word)sizeof(struct b_refresh) +
         (na + nl) * sizeof(struct descrip)) == Error) 
      RunErr(0, NULL);
   rblkp = alcrefresh(entryp, na, nl);
   sblkp->freshblk.dword = D_Refresh;
   BlkLoc(sblkp->freshblk) = (union block *) rblkp;

   /*
    * Copy current procedure frame marker into refresh block.
    */
   rblkp->pfmkr = *pfp;
   rblkp->pfmkr.pf_pfp = 0;

   /*
    * Copy arguments into refresh block and onto new stack.
    */
   dp = &argp[0];
   ndp = &rblkp->elems[0];
   dsp = (struct descrip *)newsp;
   for (i = 1; i <= na; i++) {
      *dsp++ = *dp;
      *ndp++ = *dp++;
      }

   /*
    * Copy procedure frame to new stack and point dsp to word after frame.
    */
   *((struct pf_marker *)dsp) = *pfp;
   sblkp->es_pfp = (struct pf_marker *)dsp;
   sblkp->es_pfp->pf_pfp = 0;
   dsp = (struct descrip *)((word *)dsp + Vwsizeof(*pfp));
   sblkp->es_ipc.opnd = entryp;
   sblkp->es_gfp = 0;
   sblkp->es_efp = 0;
   sblkp->es_ilevel = 0;
   sblkp->tvalloc = NULL;

   /*
    * Copy locals to new stack and refresh block.
    */
   dp = &(pfp->pf_locals)[0];
   for (i = 1; i <= nl; i++) {
      *dsp++ = *dp;
      *ndp++ = *dp++;
      }
   /*
    * Push two null descriptors on the stack.
    */
   *dsp++ = nulldesc;
   *dsp++ = nulldesc;

   sblkp->es_sp = (word *)dsp - 1;

   /*
    * Return the new co-expression.
    */
   Arg0.dword = D_Coexpr;
   BlkLoc(Arg0) = (union block *) sblkp;
   Return;
#else					/* NoCoexpr */
   RunErr(-401, NULL);
#endif					/* NoCoexpr */
   }

static char *day[] = {
   "Sunday", "Monday", "Tuesday", "Wednesday",
   "Thursday", "Friday", "Saturday"
   };

static char *month[] = {
   "January", "February", "March", "April", "May", "June",
   "July", "August", "September", "October", "November", "December"
   };

/*
 * keywd - process keyword.
 */

char *feattab[] = {
#if AMIGA
   "Amiga",
#endif					/* AMIGA */
#if ATARI_ST
   "Atari ST",
#endif					/* ATARI_ST */
#if MACINTOSH
   "Macintosh",
#endif					/* MACINTOSH */
#if MSDOS
   "MS-DOS",
#ifdef DosFncs
   "MS-DOS extensions",
#endif					/* DosFncs */
#endif					/* MSDOS */
#if MVS
   "MVS",
#endif					/* MVS */
#if UNIX
   "UNIX",
#ifdef UnixFncs
   "UNIX extensions",
#endif					/* UnixFncs */
#endif					/* UNIX */
#if VM
   "VM",
#endif					/* VM */
#if VMS
   "VMS",
#endif					/* VMS */
#ifndef NoCoexpr
   "co-expressions",
#endif					/* NoCoexpr */
#ifndef NoOver
   "overflow checking",
#endif					/* NoOver */
#ifndef NoHeader
   "direct execution",
#endif					/* NoHeader */
#ifndef NoEnvVars
   "environment variables",
#endif					/* NoEnvVars */
#ifndef NoTraceBack
   "error traceback",
#endif					/* NoTraceBack */
#ifdef ExecImages
   "executable images",
#endif					/* ExecImages */
#ifndef NoStrInvoke
   "string invocation",
#endif					/* NoTraceBack */
#ifdef FixedRegions
   "fixed regions",
#else					/* FixedRegions */
   "expandable regions",
#endif					/* FixedRegions */
   ""
   };

LibDcl(keywd,0,"&keywd")
   {
   register int hour;
   register word i;
   register char *merid;
   char **p;
   char sbuf[MaxCvtLen];
   extern int coll_stat, coll_str, coll_blk, coll_tot;
   extern int findline();
   extern char *findfile();

/*
 * The following code is operating-system dependent. Define structure if
 *  necessary for handling time.
 */

#if PORT
   struct tm *tbuf, *localtime();
#endif					/* PORT */

#if AMIGA || MACINTOSH || MSDOS
   struct tm *tbuf, *localtime();
#endif					/* AMIGA || MACINTOSH || MSDOS */

#if ATARI_ST
   struct tms *tbuf;
#endif					/* ATARI_ST */

#if UNIX || VMS
   struct tm *tbuf, *localtime();
   struct tms tp;
#endif					/* UNIX || VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

   long time();
   long clock, runtim;
   char *alcstr();
#ifndef NoCoexpr
   struct b_coexpr *topact();
#endif					/* NoCoexpr */

   /*
    * This is just plug and chug code.	For whatever keyword is desired,
    *  the appropriate value is dug out of the system and made into
    *  a suitable Icon value.
    *
    * A few special cases are worth noting:
    *  &pos, &random, &trace - built-in trapped variables are returned
    */
   switch ((int)IntVal(Arg0)) {
      case K_ASCII:
         Arg0.dword = D_Cset;
         BlkLoc(Arg0) = (union block *) &k_ascii;
         break;
      case K_CLOCK:
         if (strreq((word)8) == Error) 
            RunErr(0, NULL);
/*
 * The following code is operating-system dependent.  Get time.
 */

#if PORT
   time(&clock);
   tbuf = localtime(&clock);
#endif					/* PORT */

#if ATARI_ST
   gettime(&tbuf);
#endif					/* ATARI_ST */

#if AMIGA || MACINTOSH || MSDOS || UNIX || VMS
   time(&clock);
   tbuf = localtime(&clock);
#endif					/* AMIGA || MACINTOSH || MSDOS ... */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

         sprintf(sbuf,"%02d:%02d:%02d",tbuf->tm_hour,tbuf->tm_min,
            tbuf->tm_sec);
         StrLen(Arg0) = 8;
         StrLoc(Arg0) = alcstr(sbuf,(word)8);
         break;
      case K_COLLECTIONS:
         MkIntT(coll_tot, &Arg0);
         Suspend;
         MkIntT(coll_stat, &Arg0);
         Suspend;
         MkIntT(coll_str, &Arg0);
         Suspend;
         MkIntT(coll_blk, &Arg0);
         Return;
      case K_CSET:
         Arg0.dword = D_Cset;
         BlkLoc(Arg0) = (union block *) &k_cset;
         break;
      case K_CURRENT:
         Arg0 = k_current;
         break;
      case K_DATE:
         if (strreq((word)10) == Error) 
            RunErr(0, NULL);

/*
 * The following code is operating-system dependent.  Get time.
 */

#if PORT
   time(&clock);
   tbuf = localtime(&clock);
#endif					/* PORT */

#if ATARI_ST
   gettime(&tbuf);
#endif					/* ATARI_ST */

#if AMIGA || MACINTOSH || MSDOS || UNIX || VMS
   time(&clock);
   tbuf = localtime(&clock);
#endif					/* AMIGA || MACINTOSH || MSDOS ... */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

         sprintf(sbuf, "%04d/%02d/%02d",
                      (tbuf->tm_year)+1900,tbuf->tm_mon+1,tbuf->tm_mday);
         StrLen(Arg0) = 10;
         StrLoc(Arg0) = alcstr(sbuf,(word)10);
         break;
      case K_DATELINE:
/*
 * The following code is operating-system dependent.  Get time.
 */

#if PORT
   time(&clock);
   tbuf = localtime(&clock);
#endif					/* PORT */

#if ATARI_ST
   gettime(&tbuf);
#endif					/* ATARI_ST */

#if AMIGA || MACINTOSH || MSDOS || UNIX || VMS
   time(&clock);
   tbuf = localtime(&clock);
#endif					/* AMIGA || MACINTOSH || MSDOS ... */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

         if ((hour = tbuf->tm_hour) >= 12) {
            merid = "pm";
            if (hour > 12)
               hour -= 12;
            }
         else {
            merid = "am";
            if (hour < 1)
               hour += 12;
            }
         sprintf(sbuf, "%s, %s %d, %d  %d:%02d %s",
#if ATARI_ST
         /* This is totally bogus and needs replacing, if only to
          *  restore order for non-Atari implementations.
          */
            day[1],
#else					/* ATARI_ST */
            day[tbuf->tm_wday],
#endif					/* ATARI_ST */
            month[tbuf->tm_mon], tbuf->tm_mday,
            1900 + tbuf->tm_year, hour, tbuf->tm_min, merid);
         if (strreq(i = strlen(sbuf)) == Error) 
            RunErr(0, NULL);
         StrLen(Arg0) = i;
         StrLoc(Arg0) = alcstr(sbuf, i);
         break;
      case K_DIGITS:
         Arg0.dword = D_Cset;
         BlkLoc(Arg0) = (union block *)&k_digits;
         break;
      case K_ERROR:
         Arg0.dword = D_Tvkywd;
         BlkLoc(Arg0) = (union block *)&tvky_err;
         break;
      case K_ERRORNUMBER:
         if (k_errornumber == 0)
            Fail;
         MkIntT((k_errornumber > 0 ? k_errornumber : -k_errornumber), &Arg0);
         break;
      case K_ERRORTEXT:
         if (k_errornumber == 0)
            Fail;
         StrLoc(Arg0) = k_errortext;
         StrLen(Arg0) = strlen(k_errortext);
         break;
      case K_ERRORVALUE:
         if (k_errornumber <= 0)
            Fail;
         Arg0 = k_errorvalue;
         break;
      case K_ERROUT:
         Arg0.dword = D_File;
         BlkLoc(Arg0) = (union block *)&k_errout;
         break;
      case K_FEATURES:
         p = feattab;
         for(;;) {
            StrLen(Arg0) = strlen(*p);
            if (StrLen(Arg0) == 0)
               Fail;
            StrLoc(Arg0) = *p;
            Suspend;
            p++;
            }
      case K_FILE:
         StrLoc(Arg0) = findfile(ipc.opnd);
         StrLen(Arg0) = strlen(StrLoc(Arg0));
         break;
      case K_HOST:
         iconhost(sbuf);
         if (strreq(i = strlen(sbuf)) == Error) 
            RunErr(0, NULL);
         StrLen(Arg0) = i;
         StrLoc(Arg0) = alcstr(sbuf, i);
         break;
      case K_INPUT:
         Arg0.dword = D_File;
         BlkLoc(Arg0) = (union block *)&k_input;
         break;
      case K_LCASE:
         Arg0.dword = D_Cset;
         BlkLoc(Arg0) = (union block *)&k_lcase;
         break;
      case K_LEVEL:
         MkIntT(k_level, &Arg0);
         break;
      case K_LINE:
         MkIntT(findline(ipc.opnd), &Arg0);
         break;
      case K_MAIN:
         Arg0 = k_main;
         break;
      case K_OUTPUT:
         Arg0.dword = D_File;
         BlkLoc(Arg0) = (union block *)&k_output;
         break;
      case K_POS:
         Arg0.dword = D_Tvkywd;
         BlkLoc(Arg0) = (union block *) &tvky_pos;
         break;
      case K_RANDOM:
         Arg0.dword = D_Tvkywd;
         BlkLoc(Arg0) = (union block *) &tvky_ran;
         break;
      case K_REGIONS:
         MkIntT((word)statend - (word)statbase, &Arg0);
         Suspend;
         MkIntT((word)strend - (word)strbase, &Arg0);
         Suspend;
         MkIntT((word)blkend - (word)blkbase, &Arg0);
         Return;
      case K_SOURCE:
#ifdef NoCoexpr
         Arg(0) = k_main;
#else					/* NoCoexpr */
 	 Arg0.dword = D_Coexpr;
 	 BlkLoc(Arg0) =
            (union block *)topact((struct b_coexpr *)BlkLoc(k_current));
#endif					/* NoCoexpr */

         break;
      case K_STORAGE:
         MkIntT((word)statfree - (word)statbase, &Arg0);
         Suspend;
         MkIntT((word)strfree - (word)strbase, &Arg0);
         Suspend;
         MkIntT((word)blkfree - (word)blkbase, &Arg0);
         Return;
      case K_SUBJECT:
         Arg0.dword = D_Tvkywd;
         BlkLoc(Arg0) = (union block *) &tvky_sub;
         break;
      case K_TIME:

/*
 * The following code is operating-system dependent. Get time.
 */

#if PORT
   /* needs something */
#endif					/* PORT */

#if AMIGA || MSDOS
         runtim = 1000 * (time(NULL) - starttime);
#endif					/* AMIGA || MSDOS */

#if ATARI_ST
         gettime(&tbuf);
         runtim = tbuf->tms_utime - starttime;
#endif					/* ATARI_ST */

#if MACINTOSH
#if MPW
         runtim = 1000 * ((extended)(TickCount() - starttime) / (extended)Hz);
#endif					/* MPW */
#endif					/* MACINTOSH */

#if AMIGA || MSDOS
         runtim = 1000 * (time(NULL) - starttime);
#endif					/* AMIGA || MSDOS */

#if UNIX || VMS
         times(&tp);
         runtim = 1000 * ((tp.tms_utime - starttime) / (double)Hz);
#endif					/* UNIX || VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

         MkIntT(runtim, &Arg0);
         break;
      case K_TRACE:
         Arg0.dword = D_Tvkywd;
         BlkLoc(Arg0) = (union block *)&tvky_trc;
         break;
      case K_UCASE:
         Arg0.dword = D_Cset;
         BlkLoc(Arg0) = (union block *)&k_ucase;
         break;
      case K_VERSION:
         if (strreq(i = strlen(VERSION)) == Error) 
            RunErr(0, NULL);
         StrLen(Arg0) = i;
         StrLoc(Arg0) = VERSION;
         break;
      default:
         syserr("keyword: unknown keyword type.");
      }
   Return;
   }


/*
 * limit - explicit limitation initialization.
 */

LibDcl(limit,2,"\\")
   {
   long l;

   /*
    * The limit is both passed and returned in Arg0.  The limit must
    *  be an integer.  If the limit is 0, the expression being evaluated
    *  fails.  If the limit is < 0, it is an error.  Note that the
    *  result produced by limit is ultimately picked up by the lsusp
    *  function.
    */
   if (DeRef(Arg0) == Error) 
      RunErr(0, NULL);

   switch (cvint(&Arg0, &l)) {

      case T_Integer:
         MkIntT(l, &Arg0);
         break;

      default:
         RunErr(101, &Arg0);
      }

   if (l < 0) 
      RunErr(205, &Arg0);
   if (l == 0)
      Fail;
   Return;
   }


/*
 * [ ... ] - create an explicitly specified list.
 */

LibDcl(llist,-1,"[...]")
   {
   register word i;
   register struct b_list *hp;
   register struct b_lelem *bp;
   extern struct b_list *alclist();
   extern struct b_lelem *alclstb();
   word nslots;

   /*
    * Round the number of elements in the list (as indicated by nargs)
    *  up to MinListSlots and ensure space for the list.
    */
   nslots = nargs;
   if (nslots < MinListSlots)
      nslots = MinListSlots;

   if (blkreq((word)sizeof(struct b_list) + sizeof(struct b_lelem) +
         nslots * sizeof(struct descrip)) == Error) 
      RunErr(0, NULL);

   /*
    * Allocate the list and a list block.
    */
   hp = alclist((word)nargs);
   bp = alclstb(nslots, (word)0, (word)nargs);

   /*
    * Make the list block just allocated into the first and last blocks
    *  for the list.
    */
   hp->listhead.dword = hp->listtail.dword = D_Lelem;
   BlkLoc(hp->listhead) = BlkLoc(hp->listtail) = (union block *) bp;
   /*
    * Dereference each argument in turn and assign it to a list element.
    */
   for (i = 1; i <= nargs; i++) {
      if (DeRef(Arg(i)) == Error) 
         RunErr(0, NULL);
      bp->lslots[i-1] = Arg(i);
      }
   /*
    * Point Arg0 at the new list and return it.
    */
   ArgType(0) = D_List;
   Arg(0).vword.bptr = (union block *)hp;
   Return;
   }
