.TH SUN2PS L "" "" "User Contributed Software"
.SH NAME
sun2ps \- a Sun rasterfile to PostScript image filter using run length 
encoding
.SH SYNOPSIS
.B sun2ps
[
.B \-s sizex sizey
] [ 
.B \-t transx transy
] [
.B \-r rot
] [
.B \-i
] [
.B \-a
] [
.B \-l
] [ rasfile ]
.SH OPTIONS
.TP
.B \-s
Set the x and y size of the image to
.I sizex
and
.I sizey
inches, respectively. Defaults to 7.5 x 10 inches.
.TP
.B \-t
Translate the origin of the x-y axis to 
.I transx
and
.I transy
inches, respectively. Defaults to (0.5, 0.5) inches.
.TP
.B \-r 
Rotate the image by
.I rot
degrees. Note: rotating the image
.I does not
translate the x-y origin. When using this option, be sure to translate
the origin accordingly. Default is 0 degrees. See the -l option.
.TP
.B \-i
Invert the image. This implies a white image on a black background.
Default is a black image on a white background.
.TP
.B \-l
Landscape mode. This option will override any previous size, translation
and rotation settings. Its is intended to provide a quick, landscape
image. Default is off (or portrait mode).
.TP
.B \-a
Automatically resize the image to conform to its aspect ratio. Normally,
.I sun2ps
will scale the image to the full page size (or 
.I sizex
and
.I sizey,
if given). Default is off.
.SH DESCRIPTION
.sp 1
.TP
.B Overview
.br
.sp 1
.I Sun2ps
translates images stored in Sun Rasterfile format (see
.I rasterfile(5)
for info on file format) to standard PostScript code. Run length
encoding/decoding was also incorporated into 
.I sun2ps
to decrease the size of the text file produced (full sized screendumps
produce a file in excess of 300K, without compression). The specifics
of the encoding algorithm are described below. Using this compression,
full screen dumps (using
.I screendump(1))
can be compressed from 20% to 30% of the original. Unfortunately,
due to the nature of the text file, binary image data is doubled
since each byte is stored as two hex digits. 
.br
.TP
.B Run Length Encoding
.br
.sp 1
The run length encoding strategy used in
.I sun2ps
is a byte-level scheme that allows runs up to 128. A maximum run length
of 128 was chosen so as to conform with images produced by the Degas Elite
system for the Atari ST. Our intent is to allow for the display of Degas
images on NeWS windows. Using this scheme, image data is sent in two types
of packets, or buffers. Along with each buffer is a count, which identifies
the size of the packet and its type. Copy packets, or "runs" in their most
conventional sense, consist of a single byte with its count specifying
the number of times this byte is to be repeated. Literal packets consist of
a buffer of non-repeated bytes with its count specifying the size of the
packet. A buffer count of 
.I N
indicates:
.br
.sp 1
.I N = 0..127 
: Literal mode. Send out the next
.I N
+ 1 bytes as is.
.br
.I N = -1..-127
: Copy mode. Copy the next byte 
.I -N
+ 1 times.
.br
.I N = -128
: Ignore.
.br
.TP
.B Input, Output & Stderr
.sp 1
.I Sun2ps
sends its PostScript output to stdout. An input rasterfile may be specified
on the command line, or if excluded, stdin is used. 
.I Sun2ps
can be compiled with diagnostic information being reported to stderr.
Diagnostics include compression ratio, average run length, and number of
runs over the maximum. See the Makefile for setting the appropriate
flags for the diagnostic output.
.SH EXAMPLES
.br
.sp 1
Standard usage is:
.br
.na
.sp 1
MyPrompt% screendump | sun2ps -l | lpr   OR
.br
.sp 1
MyPrompt% screendump | sun2ps -l > screen.ps   OR
.br
.sp 1
MyPrompt% sun2ps -a /u/big/jerk/image.ras > image.ps
.br
.ad
.SH AUTHORS
.na
Glenn Boysko    (boysko@mandrill.ces.cwru.edu)
                ({decvax, sun}!mandrill!boysko)
.ad
.br
.sp 1
With references to prior work ("postimage") by:
.na
.br
.sp 1
J.R. Bammi      (bammi@mandrill.ces.cwru.edu)
                ({decvax, sun}!mandrill!bammi)
.ad
.SH SEE ALSO
screendump(1), rasterfile(5), rasfilter8to1(1).
.SH BUGS
No Bugs! Just features!
.SH DATE
3/25/88
